/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.contribution;

import static org.fortiss.af3.exploration.alg.port.DSEPortingUtils.createSystemParameterContainer;
import static org.fortiss.af3.exploration.alg.util.DesignSpaceExplorationModelElementFactory.createDefaultDesignSpaceExploration;
import static org.fortiss.af3.exploration.util.DesignSpaceExplorationModelElementFactory.createExplorationSpecification;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashSet;
import java.util.Optional;
import java.util.Set;

import org.eclipse.core.runtime.IProgressMonitor;
import org.fortiss.af3.component.model.ComponentArchitecture;
import org.fortiss.af3.exploration.alg.dse.CompositeExplorationSolution;
import org.fortiss.af3.exploration.alg.dse.Explorer;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.solution.StrictTTSchedule;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemModelAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemParameterContainer;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.af3.AF3SystemModelAdapter;
import org.fortiss.af3.exploration.alg.port.plot.XYPlotter;
import org.fortiss.af3.exploration.alg.service.IExplorationEncoding;
import org.fortiss.af3.exploration.backend.DseInputContainer;
import org.fortiss.af3.exploration.backend.IDseBackend;
import org.fortiss.af3.exploration.backend.IDseInputContainer;
import org.fortiss.af3.exploration.backend.RuleSetContainer;
import org.fortiss.af3.exploration.model.ExplorationSolution;
import org.fortiss.af3.exploration.model.ExplorationSpecification;
import org.fortiss.af3.exploration.model.IExplorationFeature;
import org.fortiss.af3.exploration.model.solutions.ExplorationSolutionSet;
import org.fortiss.af3.exploration.moea.model.DseSpecification;
import org.fortiss.af3.exploration.moea.model.feature.FeatureFactory;
import org.fortiss.af3.exploration.projectmodel.RuleSet;
import org.fortiss.af3.platform.model.PlatformArchitecture;
import org.fortiss.af3.task.model.TaskArchitecture;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;
import org.fortiss.tooling.kernel.service.IPersistencyService;
import org.jfree.ui.RefineryUtilities;

/**
 * Contributor class to allow launching this DSE backend from the DSE perspective.
 * 
 * @author diewald
 */
public class Opt4JDseBackend implements IDseBackend {

	/** Specifies the required input models from the DSE. */
	private final Collection<Class<?>> requiredInputModels =
			Arrays.asList(ComponentArchitecture.class, TaskArchitecture.class,
					PlatformArchitecture.class, RuleSetContainer.class);

	// TODO: Remove after migration to AF3 visiualization
	/** Last {@link XYPlotter} (in order to close it before creating a new one. */
	protected XYPlotter lastPlotter;

	/** {@inheritDoc} */
	@Override
	public Collection<Class<?>> getRequiredInputSet() {
		return requiredInputModels;
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public Optional<ExplorationSolution> executeDSE(RuleSetContainer expSpec,
			Set<Class<? extends IModelElement>> solutionModels, IDseInputContainer inputContainer,
			Collection<IExplorationFeature> explorationModules, IProgressMonitor monitor,
			int timeoutMS) throws Exception {
		ExplorationSolutionSet explorationSolutionSet;
		SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter;
		// TODO: Find a cleaner way to extract the model context.
		ComponentArchitecture refModel = inputContainer.getInputModel(ComponentArchitecture.class);
		ITopLevelElement context =
				IPersistencyService.getInstance().getTopLevelElementFor(refModel);

		ExplorationSpecification spec = createExplorationSpecification();
		for(RuleSet ruleSet : expSpec) {
			context.runAsNonDirtyingCommand(
					() -> spec.getTargets().addAll(ruleSet.getExplorationTargets()));
		}

		// Input to DSE: {@link SystemModelAdapter} and {@link DesignSpaceExploration}.
		// SystemModelAdapter systemModelAdapter;
		ComponentArchitecture componentArchitecture =
				inputContainer.getInputModel(ComponentArchitecture.class);
		PlatformArchitecture platformArchitecture =
				inputContainer.getInputModel(PlatformArchitecture.class);
		TaskArchitecture taskArchitecture = inputContainer.getInputModel(TaskArchitecture.class);
		SystemParameterContainer paramContainer =
				createSystemParameterContainer(platformArchitecture, taskArchitecture, spec);
		systemModelAdapter = new AF3SystemModelAdapter(componentArchitecture, platformArchitecture,
				paramContainer);
		DseSpecification dseSpec = createDefaultDesignSpaceExploration(spec, context);

		// Set the requested solutions.
		Set<Class<?>> solutionTypes = new HashSet<>();
		solutionTypes.add(StrictTTSchedule.class);

		// Features enabled by default.
		Collection<IExplorationFeature> dseFeatures = new ArrayList<>();
		dseFeatures.add(FeatureFactory.eINSTANCE.createTaskMapping());
		dseFeatures.add(FeatureFactory.eINSTANCE.createTaskInstantiation());
		dseSpec.getFeatures().addAll(dseFeatures);

		try {
			Explorer explorer = new Explorer();
			Set<Class<? extends IExplorationEncoding>> solPhenotypeTypes = new HashSet<>();
			for(Class<?> solType : solutionTypes) {
				if(!IExplorationEncoding.class.isAssignableFrom(solType)) {
					throw new Exception(
							"The demanded solution is not available in the sleceted DSE backend.");
				}
				solPhenotypeTypes.add((Class<? extends IExplorationEncoding>)solType);
			}
			CompositeExplorationSolution<?> expResult =
					explorer.explore(dseSpec, systemModelAdapter, solPhenotypeTypes, monitor, true);
			explorationSolutionSet = expResult.convertToExplorationSolution();
			explorationSolutionSet.setInputModelAdapter(systemModelAdapter);

			explorationSolutionSet.setTargetSpecification(spec);
		} catch(Exception ex) {
			// TODO: readd code to present the schedule which failed... Use the LoggingService from
			// the kernel.
			// ExplorationUiUtil.showExceptionErrorAsync(
			// "Design Space Exploration: Error", ex.getMessage(), ex);
			// if(ex instanceof ScheduleRuntimeException) {
			// StrictTTSchedule<?, ?> faultySchedule =
			// ((ScheduleRuntimeException)ex).getSchedule();
			// SchedulePlotter sp = new SchedulePlotter(faultySchedule);
			// sp.createPlot();
			// }
			throw ex;
		}
		return Optional.of(explorationSolutionSet);
	}

	/**
	 * Creates and launches the Dialog that displays the results from the exploration.
	 * Pareto-optimal points are displayed in a 2D-graph, while schedules can be inspected for
	 * each solution.
	 */
	@Override
	public void plotResult(ExplorationSolutionSet solutionSet) {
		if(solutionSet != null) {
			// Keep only one DSE result window
			if(lastPlotter != null) {
				lastPlotter.dispose();
			}

			// draw figure
			// FIXME: Unchecked cast to ExplorationSolutionSet
			XYPlotter plotter = new XYPlotter(solutionSet.getTargetSpecification(), solutionSet,
					"Design Space Exploration Result (generations: " + "N/A" + ")");
			plotter.createPlot();
			plotter.pack();
			RefineryUtilities.centerFrameOnScreen(plotter);
			plotter.setVisible(true);

			lastPlotter = plotter;
		}
	}

	/** Defines the required input parameters of this DSE backend. */
	protected class Opt4JDSEInputParameters extends DseInputContainer {
		// Parameter container wrapping a standard java map.
	}

	/** {@inheritDoc} */
	@Override
	public String getName() {
		return "MOEA-based Exploration";
	}

	/** {@inheritDoc} */
	@Override
	public String toString() {
		return getName();
	}
}
