/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse;

import static org.fortiss.af3.exploration.alg.util.DesignSpaceExplorationModelElementFactory.createSingleExplorationSolutionMap;

import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.InstantiatedTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.solution.IExplorationSolution;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.solution.IScheduleSolution;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.solution.StrictTTSchedule;
import org.fortiss.af3.exploration.model.ExplorationTarget;
import org.fortiss.af3.exploration.model.solutions.ExplorationResult;
import org.fortiss.af3.exploration.model.solutions.ExplorationSolutionSet;
import org.fortiss.af3.exploration.model.solutions.SingleExplorationSolutionMap;
import org.fortiss.af3.exploration.model.solutions.SolutionsFactory;
import org.opt4j.core.Value;

/**
 * Result of a DSE run, i.e. the {@link IExplorationSolution}s and {@link SolutionQuantification}s
 * for all pareto-optimal {@link IExplorationSolution}s.
 * 
 * @author barner
 */
public class CompositeExplorationSolution<P extends Phenotype> {

	/**
	 * Map to store the {@link SolutionQuantification} for all pareto-optimal
	 * {@link IExplorationSolution}s of an exploration run.
	 */
	private Map<IExplorationSolution<?>, SolutionQuantification> explorationResultMap =
			new HashMap<IExplorationSolution<?>, SolutionQuantification>();

	/** Number of completed iterations that produced this {@link CompositeExplorationSolution}. */
	private int completedIterations;

	/**
	 * Adds the {@link SolutionQuantification} for a given {@link IExplorationSolution}.
	 */
	public void addExplorationSolution(IExplorationSolution<?> solution,
			SolutionQuantification result) {
		explorationResultMap.put(solution, result);
	}

	/**
	 * Deletes a {@link IExplorationSolution} and its according {@link SolutionQuantification} from
	 * the map containing the results.
	 * This method is useful for removing, e.g., infeasible results.
	 */
	public void deleteSolutionQuantification(IExplorationSolution<?> solution) {
		explorationResultMap.remove(solution);
	}

	/**
	 * Returns the Map of exploration solutions in the form {@link IExplorationSolution},
	 * {@link SolutionQuantification}. This method is intended to provide intermediate or final
	 * solution to the user.
	 */
	public Map<IExplorationSolution<?>, SolutionQuantification> getSolutionQuantification() {
		return explorationResultMap;
	}

	/** {@inheritDoc} */
	@Override
	public String toString() {
		String rval = "After " + completedIterations +
				" completed iterations, the exploration result contains " +
				explorationResultMap.size() + " individual(s) (= archive size).\n\n";

		for(IExplorationSolution<?> solution : explorationResultMap.keySet()) {
			IScheduleSolution<?> scheduleSolution = solution.getSubSolution(StrictTTSchedule.class);
			InstantiatedTaskMappingEncoding<?> mappingSolution =
					solution.getSubSolution(InstantiatedTaskMappingEncoding.class);
			rval += "------------------------------------------------\n";
			rval += "Encoding (genotype):\n" + mappingSolution + "\n";
			rval += "Solution (phenotype):\n" + scheduleSolution + "\n";
		}

		return rval;
	}

	/** Creates a copy of the ExplorationResult object this operator is applied to. */
	public CompositeExplorationSolution<P> copy() {
		CompositeExplorationSolution<P> duplicatedExplorationResult =
				new CompositeExplorationSolution<P>();

		for(IExplorationSolution<?> solution : explorationResultMap.keySet()) {
			duplicatedExplorationResult.addExplorationSolution(solution,
					explorationResultMap.get(solution));
		}

		duplicatedExplorationResult.setCompletedIterations(completedIterations);

		return duplicatedExplorationResult;
	}

	/** Records the number of completed iterations. */
	public void setCompletedIterations(int completedIterations) {
		this.completedIterations = completedIterations;
	}

	/** Returns the number of completed iterations. */
	public int getCompletedIterations() {
		return completedIterations;
	}

	/**
	 * Returns {@code this} {@link CompositeExplorationSolution} as a {@link ExplorationSolutionSet}
	 * .
	 */
	@SuppressWarnings("unchecked")
	public <T> ExplorationSolutionSet convertToExplorationSolution() throws Exception {
		ExplorationSolutionSet expSol = SolutionsFactory.eINSTANCE.createExplorationSolutionSet();
		for(Entry<IExplorationSolution<?>, SolutionQuantification> singleSolutionEntry : explorationResultMap
				.entrySet()) {
			SingleExplorationSolutionMap expSolutionMap = createSingleExplorationSolutionMap();
			IExplorationSolution<?> singleSolution = singleSolutionEntry.getKey();

			// Be transparent, provide all Phenotypes
			for(Class<? extends Phenotype> subSolutionType : singleSolution
					.getAllSubSolutionClasses()) {
				// TODO: Dirty cast, remove.
				expSolutionMap.putSolutionModel((Class<T>)subSolutionType,
						(T)singleSolution.getSubSolution(subSolutionType));
			}

			for(Entry<ExplorationTarget<?>, Value<?>> solEntry : singleSolutionEntry.getValue()
					.entrySet()) {
				if(expSolutionMap.containsKey(solEntry.getKey())) {
					throw new Exception(
							"Could not convert to the ExplorationResult to an ExplorationSolutionSet. The exist multiple results for the ExplorationTarget " +
									solEntry.getKey().getName() +
									" where there should be only one.");
				}
				// TODO: this is unsafe, the SolutionQuantification class should be either
				// eliminated, or adapted such that we obtain the type safety that is also aimed
				// at in the EMM.
				@SuppressWarnings("rawtypes") ExplorationResult res =
						SolutionsFactory.eINSTANCE.createExplorationResult();
				res.setResult(solEntry.getValue().getDouble());
				expSolutionMap.put((ExplorationTarget<?>)solEntry.getKey(), res);
			}
			expSol.getSolutions().add(expSolutionMap);
		}

		return expSol;
	}
}
