/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.create.taskmapping.AbstractTaskMappingCreator;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.create.taskmapping.TaskMappingCreator;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskgraph.InstantiatedTaskGraphEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.TaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.TaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.abstractmapping.AbstractTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.InstantiatedTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.InstantiatedTaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.graph.DecoderDependencyGraph;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.operator.crossover.CrossoverRandManyAllocation;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.operator.crossover.CrossoverSingleAllocation;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.operator.mutate.taskmapping.MutateAllocation;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.operator.mutate.taskmapping.MutateRedundancyDecrement;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.operator.mutate.taskmapping.MutateRedundancyIncrement;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem.EvaluatorWrapper;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem.StrictTTDecoder;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem.comm.MessageDecoder;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem.instantiatemapping.AbstractTaskMappingDecoder;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem.instantiatetaskgraph.InstantiatedTaskMappingDecoder;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem.instantiatetaskgraph.InstantiatedTaskMappingDecoderAcyclic;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IDeployableComponentAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemModelAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.mapping.comm.IMessageRouter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.mapping.comm.ShortestPathRouter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.sched.proc.EDFSchedulerBase;
import org.fortiss.af3.exploration.alg.service.internal.ExplorationEvaluationService;
import org.fortiss.af3.exploration.model.ExplorationSpecification;
import org.fortiss.af3.exploration.model.ExplorationTarget;
import org.fortiss.af3.exploration.model.time.DeadlineConstraint;
import org.fortiss.af3.exploration.model.time.PeriodConstraint;
import org.fortiss.af3.exploration.moea.model.DseSpecification;
import org.fortiss.af3.exploration.moea.model.predefined.ComponentMultiAllocationConstraint;
import org.fortiss.af3.exploration.moea.model.predefined.ComponentMultiDislocationConstraint;
import org.fortiss.af3.exploration.moea.model.predefined.FailureMinObjective;
import org.fortiss.af3.exploration.moea.model.predefined.SafetyIntegrityLevelConstraint;
import org.fortiss.af3.safety.model.SIL;

import com.google.inject.TypeLiteral;

/**
 * This abstract factory is responsible for creating the EvaluatorWrappers that are required to hand
 * over objectives and constraints to the optimization. Here, a differentiation is made for the
 * objectives and constraints. The implementation of the specific create functions must be made
 * within factories that are specific for a execution model (like fail silent, fault detection
 * etc.).
 * 
 * @author diewald
 */
public abstract class TaskMappingFactory<S extends InstantiatedTaskMappingEntry, T extends InstantiatedTaskMappingEncoding<S>> {

	/** Error message for non-implemented operators etc. related to schedules */
	private String SCHED_NOT_IMPL_MSG =
			"Error: This operation is not implemented for the execution model defined in the design space exploration. Please implement the operation for this execution model, or avoid calling this function if it is not necessary.";

	/** Error message for non-implemented operators etc. related to optimization goals */
	private String GOAL_NOT_IMPL_MSG =
			"Error: This optimization goal is not implemented. Please implement the goal, or remove it from your design space exploration model.";

	/** The instance (singleton) of the EvaluatorWrapper factory. */
	protected static TaskMappingFactory<?, ?> instance;

	/** Returns the singleton implementing this abstract factory. */
	public static TaskMappingFactory<?, ?> getInstance() {
		if(instance == null) {
			throw new UnsupportedOperationException(
					"TaskMapping factory in not yet initialized. A factory of a concrete TaskMapping must be created first.");
		}
		return instance;
	}

	/* Construct Creators for the task mapping. */

	/** Creates and returns a {@link TypeLiteral} of the specific encoding of the task mapping. */
	public abstract TypeLiteral<T> createGenotypeTypeLiteral();

	/**
	 * Creates an abstract TaskMappingCreator which maps components that needs to be instatiated
	 * first before they can be deployed onto a platform.
	 * 
	 * @param dse
	 *            design space exploration model defined in the GUI
	 * @param systemModelAdapter
	 *            the system model deduced from the logical and technical architecture
	 * @param execDepGraph
	 *            Dependency graph defining the execution order.
	 * @return instance of an {@link AbstractTaskMappingCreator}.
	 */
	public AbstractTaskMappingCreator createAbstractTaskMappingCreator(DseSpecification dse,
			SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter,
			DecoderDependencyGraph execDepGraph) {
		return new AbstractTaskMappingCreator(dse, systemModelAdapter, execDepGraph);
	}

	/**
	 * Creates a TaskMappingCreator that matches the given execution model.
	 * 
	 * @param dse
	 *            design space exploration model defined in the GUI
	 * @param systemModelAdapter
	 *            the system model deduced from the logical and technical architecture
	 * @param execDepGraph
	 *            Dependency graph defining the execution order.
	 * @return instance of a {@link TaskMappingCreator}.
	 */
	// TODO: document execDepGraph param.
	public TaskMappingCreator<S, T> createTaskMappingCreator(DseSpecification dse,
			SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter,
			DecoderDependencyGraph execDepGraph) {
		throw new UnsupportedOperationException(SCHED_NOT_IMPL_MSG);
	}

	/**
	 * Creates a TaskMappingCreator that uses an {@link AbstractTaskMappingEncoding} to generate a
	 * Tank Mapping that consists only of tasks instances.
	 * 
	 * @param dse
	 *            design space exploration model defined in the GUI
	 * @param systemModelAdapter
	 *            the system model deduced from the logical and technical architecture
	 * @param abstractEnc
	 *            encoding that may contain abstract task which need to be instantiated.
	 * @return instance of a {@link TaskMappingCreator}.
	 */
	// TODO: update doc
	public TaskMappingCreator<S, T> createTaskMappingCreator(DseSpecification dse,
			SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter,
			AbstractTaskMappingEncoding abstractEnc) {
		throw new UnsupportedOperationException(SCHED_NOT_IMPL_MSG);
	}

	/* Construct execution model specific operators. */

	/**
	 * Creates an execution model specific {@link MutateAllocation} operator.
	 * 
	 * @param systemModelAdapter
	 *            the system model deduced from the logical and technical architecture
	 * @return Mutation Operator for Allocations.
	 */
	public MutateAllocation<? extends TaskMappingEntry<?, ?>, ? extends TaskMappingEncoding<?, ?>>
			createMutateAllocationOperator(
					SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter) {
		throw new UnsupportedOperationException(SCHED_NOT_IMPL_MSG);
	}

	/**
	 * Creates an execution model specific {@link MutateRedundancyDecrement} operator.
	 * 
	 * @param systemModelAdapter
	 *            the system model deduced from the logical and technical architecture
	 * @return Mutation operator that decreases the redundant instances of tasks.
	 */
	public MutateRedundancyDecrement<?, ?> createMutateRedundancyDecrementOperator(
			SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter) {
		throw new UnsupportedOperationException(SCHED_NOT_IMPL_MSG);
	}

	/**
	 * Creates an execution model specific {@link MutateRedundancyIncrement} operator.
	 * 
	 * @param systemModelAdapter
	 *            the system model deduced from the logical and technical architecture
	 * @return Mutation operator that increases the redundant instances of tasks.
	 */
	public MutateRedundancyIncrement<?, ?> createMutateRedundancyIncrementOperator(
			SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter) {
		throw new UnsupportedOperationException(SCHED_NOT_IMPL_MSG);
	}

	/**
	 * Creates an execution model specific {@link CrossoverRandManyAllocation} operator
	 * 
	 * @param systemModelAdapter
	 *            The system model deduced from the logical and technical architecture.
	 * @param rate
	 *            The rate or probability of such that the the operator is applied.
	 * @return Crossover operator that swaps randomly many allocations from one resource with
	 *         another.
	 */
	public CrossoverRandManyAllocation<?, ?> createCrossoverRandManyAllocationOperator(
			SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter, double rate) {
		throw new UnsupportedOperationException(SCHED_NOT_IMPL_MSG);
	}

	/**
	 * Creates an execution model specific {@link CrossoverRandManyAllocation} operator
	 * 
	 * @param systemModelAdapter
	 *            The system model deduced from the logical and technical architecture.
	 * @param rate
	 *            The rate or probability of such that the the operator is applied.
	 * @return Crossover operator that swaps a single allocation from one resource with another.
	 */
	public CrossoverSingleAllocation<?, ?> createCrossoverSingleAllocationOperator(
			SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter, double rate) {
		throw new UnsupportedOperationException(SCHED_NOT_IMPL_MSG);
	}

	/* Construct Decoders for the genotypes, aka task mappings. */

	// TODO: correct the type parameters --> Instantiated XY
	/**
	 * Decoder to instantiate an {@link AbstractTaskMappingEncoding} into an
	 * {@link InstantiatedTaskMappingEncoding}.
	 * 
	 * @param systemModelAdapter
	 *            The system model deduced from the logical and technical architecture.
	 * @param explorationDefinition
	 *            The model of the DSE; This information can be useful for the decoders to optimize
	 *            the decoding strategy
	 * @return instance of an {@link AbstractTaskMappingDecoder}.
	 */
	public AbstractTaskMappingDecoder<?, ?> createAbstractTaskMappingDecoder(
			SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter,
			DseSpecification explorationDefinition) {
		throw new UnsupportedOperationException(
				"The Decoder to instantiate an \"abstract\" Task Mapping has not been defined for current Execution Model.");
	}

	/**
	 * Creates and returns a {@link InstantiatedTaskMappingDecoder}.
	 * 
	 * @param systemModelAdapter
	 *            The system model deduced from the logical and technical architecture.
	 * @return An instance of a {@link InstantiatedTaskMappingDecoder}.
	 */
	// TODO: remove these suppress warnings by propagating the generic C.
	@SuppressWarnings({"rawtypes", "unchecked"})
	public InstantiatedTaskMappingDecoder<S, InstantiatedTaskGraphEncoding>
			createInstantiatedTaskMappingDecoder(
					SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter) {
		return new InstantiatedTaskMappingDecoder(systemModelAdapter);
	}

	/**
	 * Creates and returns a {@link InstantiatedTaskMappingDecoderAcyclic}.
	 * 
	 * @param systemModelAdapter
	 *            The system model deduced from the logical and technical architecture.
	 * @return An instance on an {@link InstantiatedTaskMappingDecoderAcyclic}.
	 */
	public InstantiatedTaskMappingDecoderAcyclic<S, ?> createAbstractToAcyclicTaskGraphDecoder(
			SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter) {
		throw new UnsupportedOperationException(
				"The Decoder to instantiate an \"abstract\" and acyclic Task Graph has not been defined for current Execution Model.");
	}

	/**
	 * Creates and returns a {@link MessageDecoder}.
	 * 
	 * @param systemModelAdapter
	 *            The system model deduced from the logical and technical architecture.
	 * @return An instance of a {@link MessageDecoder}.
	 */
	public MessageDecoder<S, T>
			createMessageDecoder(SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter) {
		throw new UnsupportedOperationException(
				"The Decoder which generates messages from task mappings has not been defined for current Execution Model.");
	}

	/**
	 * Creates a scheduling strategy-"independent" decoder. Here, the basic time architecture is to
	 * be defined, e.g. time triggered scheduling.
	 * 
	 * @param systemModelAdapter
	 *            the system model deduced from the logical and technical architecture.
	 * @param explorationDefinition
	 *            The model of the DSE; This information can be useful for the decoders to optimize
	 *            the decoding strategy.
	 * @return instance of a {@link StrictTTDecoder}
	 */
	public StrictTTDecoder<S, T> createStrictTTDecoder(
			SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter,
			DseSpecification explorationDefinition) {
		throw new UnsupportedOperationException(SCHED_NOT_IMPL_MSG);
	}

	/**
	 * Creates a scheduling strategy-"independent" decoder for task mapping encodings. It provides a
	 * {@link ShortestPathRouter} to find the shortest communication path between two communicating
	 * instantiations of an {@link IDeployableComponentAdapter} for a given platform.
	 */
	public IMessageRouter
			createShortestPathRouter(SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter) {
		return new ShortestPathRouter(systemModelAdapter);
	}

	/**
	 * Creates an basic EDF scheduler that is used to decode task mappings into schedules for the
	 * given execution model
	 * 
	 * @param systemModelAdapter
	 *            the system model deduced from the logical and technical architecture
	 * @param expSpec
	 *            DSE goal specification used within the exploration, useful for scheduling
	 * @return instance of the EDF schedule decoder.
	 */
	public EDFSchedulerBase<?, ?> createEDFScheduler(
			SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter,
			ExplorationSpecification expSpec) {
		throw new UnsupportedOperationException(SCHED_NOT_IMPL_MSG);
	}

	/* Construct Evaluators for Optimization Goals. */

	/**
	 * Decides, which create method has to be called for a specific goal.
	 * 
	 * @throws Exception
	 */
	@SuppressWarnings("unchecked")
	public <RT, ET extends ExplorationTarget<RT>> EvaluatorWrapper<?, RT, ET>
			createEvaluatorWrapper(ExplorationTarget<?> goal, int priority,
					SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter) throws Exception {
		ExplorationEvaluationService<?> evalService = ExplorationEvaluationService.getInstance();
		EvaluatorWrapper<?, RT, ET> evaluatorWrapper = (EvaluatorWrapper<?, RT, ET>)evalService
				.createEvalWrapperFor(systemModelAdapter, goal);

		// TODO: remove the block below when all evaluators are migrated to the service interface.
		if(evaluatorWrapper == null) {
			/* First, the wrappers for the objectives are created */
			if(goal instanceof FailureMinObjective) {
				evaluatorWrapper =
						(EvaluatorWrapper<?, RT, ET>)createFailureMinObjectiveWrapper(goal,
								priority, systemModelAdapter);
			}

			/* Then, the constraints are handled */
			else if(goal instanceof DeadlineConstraint) {
				evaluatorWrapper =
						(EvaluatorWrapper<?, RT, ET>)createDeadlineConstraintWrapper(goal, priority,
								systemModelAdapter);
			} else if(goal instanceof PeriodConstraint) {
				evaluatorWrapper = (EvaluatorWrapper<?, RT, ET>)createPeriodConstraintWrapper(goal,
						priority, systemModelAdapter);
			} else if(goal instanceof ComponentMultiAllocationConstraint) {
				evaluatorWrapper =
						(EvaluatorWrapper<?, RT, ET>)createComponentMultiAllocationConstraintWrapper(
								goal, priority, systemModelAdapter);
			} else if(goal instanceof ComponentMultiDislocationConstraint) {
				evaluatorWrapper =
						(EvaluatorWrapper<?, RT, ET>)createComponentMultiDislocationConstraintWrapper(
								goal, priority, systemModelAdapter);
			} else if(goal instanceof SafetyIntegrityLevelConstraint) {
				evaluatorWrapper =
						(EvaluatorWrapper<?, RT, ET>)createSafetyIntegrityLevelConstraintWrapper(
								goal, priority, systemModelAdapter);
			}
		}

		if(evaluatorWrapper == null) {
			throw new InstantiationException("Error: The optimization goal " + goal.getName() +
					" is unknown. Please implement it or remove it from your design space exploration model");
		}
		return evaluatorWrapper;
	}

	/**
	 * Creates the Evaluator Wrapper for the reliability maximization objective.
	 * 
	 * @param optGoal
	 *            the optimization goal for which the wrapper shall be created
	 * @param priority
	 *            Gives a priority to the solver performing the evaluation for each target. It
	 *            defines the order of evaluation.
	 * @param systemModelAdapter
	 *            the system model deduced from the logical and technical architecture
	 * @return instance of an failure minimization objective wrapper.
	 * @throws Exception
	 *             if the input model validation fails.
	 */
	protected <RT, ET extends ExplorationTarget<RT>>
			EvaluatorWrapper<?, Double, FailureMinObjective>
			createFailureMinObjectiveWrapper(ExplorationTarget<?> optGoal, int priority,
					SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter) throws Exception {
		throw new UnsupportedOperationException(GOAL_NOT_IMPL_MSG);
	}

	/**
	 * Creates the Evaluator Wrapper for the deadline constraints.
	 * 
	 * @param optGoal
	 *            the optimization goal for which the wrapper shall be created
	 * @param priority
	 *            Gives a priority to the solver performing the evaluation for each target. It
	 *            defines the order of evaluation.
	 * @param systemModelAdapter
	 *            the system model deduced from the logical and technical architecture
	 * @return instance of a deadline evaluator wrapper.
	 * @throws Exception
	 *             if the input model validation fails.
	 */
	protected <RT, ET extends ExplorationTarget<RT>> EvaluatorWrapper<?, Double, DeadlineConstraint>
			createDeadlineConstraintWrapper(ExplorationTarget<?> optGoal, int priority,
					SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter) throws Exception {
		throw new UnsupportedOperationException(GOAL_NOT_IMPL_MSG);
	}

	/**
	 * Creates the Evaluator Wrapper for the (hidden/implicit) period constraints.
	 * 
	 * @param optGoal
	 *            the optimization goal for which the wrapper shall be created
	 * @param priority
	 *            Gives a priority to the solver performing the evaluation for each target. It
	 *            defines the order of evaluation.
	 * @param systemModelAdapter
	 *            the system model deduced from the logical and technical architecture
	 * @return instance of a periodicity evaluator wrapper.
	 * @throws Exception
	 *             if the input model validation fails.
	 */
	protected <RT, ET extends ExplorationTarget<RT>> EvaluatorWrapper<?, Double, PeriodConstraint>
			createPeriodConstraintWrapper(ExplorationTarget<?> optGoal, int priority,
					SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter) throws Exception {
		throw new UnsupportedOperationException(GOAL_NOT_IMPL_MSG);
	}

	/**
	 * Creates the Evaluator Wrapper for the fixed deployment constraints.
	 * 
	 * @param optGoal
	 *            the optimization goal for which the wrapper shall be created
	 * @param priority
	 *            Gives a priority to the solver performing the evaluation for each target. It
	 *            defines the order of evaluation.
	 * @param systemModelAdapter
	 *            the system model deduced from the logical and technical architecture
	 * @return instance of a allocation evaluator wrapper.
	 * @throws Exception
	 *             if the input model validation fails.
	 */
	protected <RT, ET extends ExplorationTarget<RT>>
			EvaluatorWrapper<?, Boolean, ComponentMultiAllocationConstraint>
			createComponentMultiAllocationConstraintWrapper(ExplorationTarget<?> optGoal,
					int priority, SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter)
					throws Exception {
		throw new UnsupportedOperationException(GOAL_NOT_IMPL_MSG);
	}

	/**
	 * Creates the Evaluator Wrapper for the exclude (forbidden) deployment constraints.
	 * 
	 * @param optGoal
	 *            the optimization goal for which the wrapper shall be created
	 * @param priority
	 *            Gives a priority to the solver performing the evaluation for each target. It
	 *            defines the order of evaluation.
	 * @param systemModelAdapter
	 *            the system model deduced from the logical and technical architecture
	 * @return instance of a dislocation evaluator wrapper.
	 * @throws Exception
	 *             if the input model validation fails.
	 */
	protected <RT, ET extends ExplorationTarget<RT>>
			EvaluatorWrapper<?, Boolean, ComponentMultiDislocationConstraint>
			createComponentMultiDislocationConstraintWrapper(ExplorationTarget<?> optGoal,
					int priority, SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter)
					throws Exception {
		throw new UnsupportedOperationException(GOAL_NOT_IMPL_MSG);
	}

	/**
	 * Creates the Evaluator Wrapper for {@link SafetyIntegrityLevelConstraint}s.
	 * 
	 * @param optGoal
	 *            the optimization goal for which the wrapper shall be created
	 * @param priority
	 *            Gives a priority to the solver performing the evaluation for each target. It
	 *            defines the order of evaluation.
	 * @param systemModelAdapter
	 *            the system model deduced from the logical and technical architecture
	 * @return instance of a SIL evaluator wrapper.
	 * @throws Exception
	 *             if the input model validation fails.
	 */
	protected <RT, ET extends ExplorationTarget<RT>>
			EvaluatorWrapper<?, SIL, SafetyIntegrityLevelConstraint>
			createSafetyIntegrityLevelConstraintWrapper(ExplorationTarget<?> optGoal, int priority,
					SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter) throws Exception {
		throw new UnsupportedOperationException(GOAL_NOT_IMPL_MSG);
	}
}
