/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend;

import java.util.Collection;
import java.util.Map;
import java.util.Set;

import javax.activation.UnsupportedDataTypeException;

import org.eclipse.core.runtime.IProgressMonitor;
import org.fortiss.af3.exploration.alg.dse.DSEFactory;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.ComposablePhenotype;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem.EvaluatorWrapper;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.solution.IExplorationSolution;
import org.fortiss.af3.exploration.alg.dse.evaluator.constraint.IConstraintEvaluator;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemModelAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.random.RandomExploration;
import org.fortiss.af3.exploration.alg.dse.sysmodel.random.RandomExploration.RandomnessType;
import org.fortiss.af3.exploration.alg.service.IExplorationEncoding;
import org.fortiss.af3.exploration.model.ExplorationConstraint;
import org.fortiss.af3.exploration.model.ExplorationSpecification;
import org.fortiss.af3.exploration.model.ExplorationTarget;
import org.fortiss.af3.exploration.moea.model.DseSpecification;
import org.fortiss.af3.exploration.moea.model.parameters.IOptimizer;
import org.opt4j.core.Value;
import org.opt4j.core.problem.Evaluator;

/**
 * Base class for general optimization backends.
 * 
 * @author barner
 */
public abstract class ExplorerBackendBase<O extends IOptimizer> implements IExplorerBackend<O> {

	/** DSE specification (optimizer parameters, goal specification, input model, ...) */
	protected DseSpecification dse;

	/** Contains all ExplorationTargets of the exploration. */
	protected ExplorationSpecification expSpec;

	/**
	 * Map that relates the {@link ExplorationTarget}s from input specification with the
	 * {@link ExplorationTarget}s that are actually used in the DSE.
	 */
	protected Map<ExplorationTarget<?>, ExplorationTarget<?>> targetAssocMap;

	/** Representation of system model (=input) that is evaluated during this DSE run */
	protected SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter;

	/** Set of requested {@link ComposablePhenotype}s that form the solution set. */
	protected Set<Class<? extends IExplorationEncoding>> requestedSolutions;

	/**
	 * {@link IProgressMonitor} used to visualize the progress of the exploration process (may be
	 * {@code null}).
	 */
	protected IProgressMonitor progressMonitor;

	/**
	 * Contains the EvaluatorWrapper factory. It is used to instantiate the objectives and
	 * constraints that are used during the exploration for specific execution models.
	 */
	protected DSEFactory explorationFactory;

	/**
	 * Constructor that initializes the DSE backend with the given optimizer independent parameters
	 * and the goal specification
	 */
	protected ExplorerBackendBase(DseSpecification dse, ExplorationSpecification expSpec,
			Map<ExplorationTarget<?>, ExplorationTarget<?>> targetAssocMap,
			SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter,
			Set<Class<? extends IExplorationEncoding>> requestedSolutions,
			IProgressMonitor progressMonitor) {
		this.dse = dse;
		this.expSpec = expSpec;
		this.targetAssocMap = targetAssocMap;
		this.systemModelAdapter = systemModelAdapter;
		this.requestedSolutions = requestedSolutions;
		this.progressMonitor = progressMonitor;

		// TODO: Initialization of the TaskMappingFactory is currently done here to allow an
		// instantiation of the Evaluators by the goals. Those are generified, and hence, the
		// instance must exist already.
		try {
			DSEFactory.getInstance().createTaskMappingFactory(
					dse.getParameters().getExecutionModel());
		} catch(UnsupportedDataTypeException e) {
			// See todo above
			e.printStackTrace();
		}

		// (Re-)Initialize the Random number generator
		initCentralRandomGenerator();
	}

	/**
	 * (Re-)Initializes the central random number generator of the DSE. It must be re-initialized on
	 * successive runs to obtain deterministic behavior.
	 */
	private void initCentralRandomGenerator() {
		if(!RandomExploration.getInstance().isInitialized()) {
			try {
				RandomExploration.getInstance().init(RandomnessType.realRandom, 30);
			} catch(Exception e) {
				// This is the correct place to initialize the central random generator.
				assert (false) : "The centralized random number generator of the DSE has been initialized in another location, where it should not be initialized.";
			}
		} else {
			try {
				RandomExploration.getInstance().reInit();
			} catch(Exception e) {
				// This is the correct place to re-initialize the central random generator, it has
				// been initialized in this place before.
				assert (false) : "The centralized random number generator of the DSE has not been initialized.";
			}
		}
	}

	/**
	 * Performs the initialization of the optimizer backend. Must be called from constructor in
	 * concrete derived classes.
	 */
	protected abstract void init(O optimizer);

	/**
	 * Feasibility check for the results from the exploration. It validates that each constraint is
	 * fulfilled by applying the validate function of the {@link IConstraintEvaluator}s onto each
	 * possible solution.
	 */
	// protected CompositeExplorationSolution<?> validate(
	// CompositeExplorationSolution<?> explorationResult,
	// Collection<EvaluatorWrapper<?, ?, ?>> evalWrappers) {
	// CompositeExplorationSolution<?> feasibleExplorationResult = explorationResult.copy();
	//
	// // Iterate over all (feasible + infeasible) solutions
	// for(Entry<IExplorationSolution<?>, SolutionQuantification> solution : explorationResult
	// .getSolutionQuantification().entrySet()) {
	// boolean isFeasible =
	// validateSolution(solution.getKey(), solution.getValue(), evalWrappers);
	//
	// // Remove the current solution if the feasibility check failed.
	// if(!isFeasible) {
	// feasibleExplorationResult.deleteSolutionQuantification(solution.getKey());
	// }
	// }
	//
	// return feasibleExplorationResult;
	// }

	/**  */
	protected <M extends Phenotype> boolean validateSolution(IExplorationSolution<?> solution,
			ExplorationConstraint<?> target, Value<?> value,
			Collection<EvaluatorWrapper<?, ?, ?>> evalWrappers) {
		EvaluatorWrapper<?, ?, ?> evalWrapper = getEvaluatorOf(target, evalWrappers);

		assert (evalWrapper.getEvaluator() instanceof IConstraintEvaluator) : "The evaluator for the exploration constraint " +
				target.getName() + " has an evaluator which is not an IConstraintEvaluator.";

		@SuppressWarnings("unchecked") Class<M> phenoClass =
				(Class<M>)evalWrapper.getTargetPhenotypeType();
		M targetSubSolution = solution.getSubSolution(phenoClass);
		return ((IConstraintEvaluator<?, ?, ?>)evalWrapper.getEvaluator()).validate(target,
				targetSubSolution, phenoClass, value);
	}

	/**
	 * From a given set of {@link Evaluator}s that are {@link EvaluatorWrapper}s, return the first
	 * found evaluator that is associated with the given {@link ExplorationTarget}.
	 */
	private <E extends EvaluatorWrapper<?, ?, ?>> E getEvaluatorOf(ExplorationTarget<?> expTarget,
			Collection<E> evaluators) {
		for(E evaluator : evaluators) {
			if(expTarget.getClass().isAssignableFrom(evaluator.getTargetType())) {
				return evaluator;
			}
		}
		return null;
	}
}
