/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j;

import static org.fortiss.tooling.kernel.utils.EcoreUtils.pickFirstInstanceOf;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.pickInstanceOf;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.eclipse.core.runtime.IProgressMonitor;
import org.fortiss.af3.exploration.alg.dse.CompositeExplorationSolution;
import org.fortiss.af3.exploration.alg.dse.backend.ExplorerBackendBase;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.mating.MatingCrossoverMutateRepairModule;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.operator.copy.safetyfunctionarch.CopySafetyFunctionArchModule;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.operator.copy.taskmapping.CopyTaskMappingModule;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.operator.crossover.CrossoverAbstractTaskMappingModule;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.operator.crossover.NopCrossoverInstantiatedTaskMappingModule;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.operator.crossover.NopCrossoverSFArchModule;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.operator.mutate.partitionmapping.PartitionMappingModule;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.operator.mutate.safetyarch.MutateSafetyFunctionArchitectureModule;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.operator.mutate.taskinstantiation.MutateTaskInstanceModule;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.operator.mutate.taskmapping.MutateTaskMappingModule;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem.ArchitectureExplorationProblemModule;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.solution.ArchitectureSolution;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.solution.IExplorationSolution;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemModelAdapter;
import org.fortiss.af3.exploration.alg.service.IExplorationEncoding;
import org.fortiss.af3.exploration.model.ExplorationConstraint;
import org.fortiss.af3.exploration.model.ExplorationSpecification;
import org.fortiss.af3.exploration.model.ExplorationTarget;
import org.fortiss.af3.exploration.moea.model.DseSpecification;
import org.fortiss.af3.exploration.moea.model.feature.SafetyExploration;
import org.fortiss.af3.exploration.moea.model.parameters.EAOptimizer;
import org.opt4j.core.Individual;
import org.opt4j.core.Objective;
import org.opt4j.core.Objectives;
import org.opt4j.core.Value;
import org.opt4j.core.common.archive.DefaultArchive;
import org.opt4j.core.common.completer.IndividualCompleterModule;
import org.opt4j.core.config.Task;
import org.opt4j.core.config.TaskStateListener;
import org.opt4j.core.optimizer.Archive;
import org.opt4j.core.optimizer.Control;
import org.opt4j.core.start.Opt4JTask;
import org.opt4j.optimizers.ea.EvolutionaryAlgorithmModule;
import org.opt4j.optimizers.ea.Spea2Module;

import com.google.inject.Module;

/**
 * Opt4J optimization backend that defines the DSE task for this particular optimization framework.
 * Here, the problem definitiona, the variation modules, the optimizer, and the solution format are
 * defined. Additionally, the optimization task is started here and the results are returned.
 * 
 * @author huang
 */
public class Opt4JExplorerBackend extends ExplorerBackendBase<EAOptimizer> {

	/** Opt4J optimization process */
	private Opt4JTask opt4JTask;

	/** EA parameters. */
	private EAOptimizer opt4JParameters;

	/**
	 * Contains the problem module that includes the evaluators associated with the goals of the
	 * DSE.
	 */
	private ArchitectureExplorationProblemModule archExplorationProblemModule;

	/**
	 * Constructor that initializes the Opt4J backend, and sets optimizer independent parameters
	 * and the goal specification
	 * 
	 * @throws Exception
	 */
	public Opt4JExplorerBackend(DseSpecification dse, ExplorationSpecification expSpec,
			Map<ExplorationTarget<?>, ExplorationTarget<?>> targetAssocMap,
			SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter,
			Set<Class<? extends IExplorationEncoding>> requestedSolutions,
			IProgressMonitor progressMonitor) throws Exception {
		// Optimizer independent configuration
		super(dse, expSpec, targetAssocMap, systemModelAdapter, requestedSolutions,
				progressMonitor);

		if(dse.getOptimizer() instanceof EAOptimizer) {
			init((EAOptimizer)dse.getOptimizer());
		} else {
			throw new Exception(
					"Opt4JExplorerBackend cannot be configured with optimizer dependent configuration of type " +
							dse.getClass());
		}
	}

	/** {@inheritDoc} */
	@Override
	protected void init(EAOptimizer opt4JParameters) {
		this.opt4JParameters = opt4JParameters;

		System.out.println("Initializing EA optimizer (generations: " + "" +
				opt4JParameters.getIterations() + ", alpha: " + opt4JParameters.getAlpha() + ").");

		// Instantiate modules that describe the optimization problem
		Collection<Module> modules = new ArrayList<Module>();

		// Evolutionary algorithm module
		EvolutionaryAlgorithmModule evolutionaryAlgorithmModule = new EvolutionaryAlgorithmModule();
		evolutionaryAlgorithmModule.setGenerations(opt4JParameters.getIterations());
		evolutionaryAlgorithmModule.setAlpha(opt4JParameters.getAlpha());
		evolutionaryAlgorithmModule.setMu(opt4JParameters.getMu());
		evolutionaryAlgorithmModule.setCrossoverRate(0.3);
		modules.add(evolutionaryAlgorithmModule);

		// SPEA2 specific module
		Spea2Module spea2Module = new Spea2Module();
		spea2Module.setTournament(opt4JParameters.getSpea2Tournament());
		modules.add(spea2Module);
		// Nsga2Module nsga2Module = new Nsga2Module();
		// modules.add(nsga2Module);

		MatingCrossoverMutateRepairModule mating =
				new MatingCrossoverMutateRepairModule(systemModelAdapter);
		modules.add(mating);

		// TODO: Replace {@link MultiEvaluator} with {@link ExplorationGoalPriorityMultiEvaluator}
		// that is
		// not based on a @Priority annotation, but on the priority attribute provided by the goal
		// specifications for the different optimization dimensions.
		// AbstractModule multiEvaluatorModule = new AbstractModule() {
		// @Override
		// protected void configure() {
		// bind(MultiEvaluator.class).to(ExplorationGoalPriorityMultiEvaluator.class);
		//
		// }
		// };
		// modules.add(multiEvaluatorModule);

		archExplorationProblemModule = new ArchitectureExplorationProblemModule(dse,
				systemModelAdapter, requestedSolutions);
		modules.add(archExplorationProblemModule);

		// Configure operators
		// 1. Crossover
		CrossoverAbstractTaskMappingModule crossoverModule =
				new CrossoverAbstractTaskMappingModule(systemModelAdapter, 0.3);
		modules.add(crossoverModule);
		NopCrossoverInstantiatedTaskMappingModule nopCrossoverInstMapping =
				new NopCrossoverInstantiatedTaskMappingModule();
		modules.add(nopCrossoverInstMapping);
		NopCrossoverSFArchModule nopCrossoverSFArch = new NopCrossoverSFArchModule();
		modules.add(nopCrossoverSFArch);

		// =====================================
		// 2. Mutation (in order)
		// =====================================

		// Safety functions (if the safety exploration module is enabled)
		if(pickFirstInstanceOf(SafetyExploration.class, dse.getFeatures()) != null) {
			MutateSafetyFunctionArchitectureModule sfArchModule =
					new MutateSafetyFunctionArchitectureModule(systemModelAdapter);
			modules.add(sfArchModule);
		}

		// Partition Exploration
		modules.add(new PartitionMappingModule(systemModelAdapter));

		// Task instantiation
		MutateTaskInstanceModule mutateInstanceModule =
				new MutateTaskInstanceModule(systemModelAdapter);
		// mutateInstanceModule.setMutationRateType(MutationRateType.ADAPTIVE);
		modules.add(mutateInstanceModule);

		// Task allocation
		MutateTaskMappingModule mutateModule = new MutateTaskMappingModule(systemModelAdapter);
		// mutateModule.setMutationRateType(MutationRateType.ADAPTIVE);
		// mutateModule.setMutationRate(0.125);
		modules.add(mutateModule);

		// 3. Genotype copy
		modules.add(new CopyTaskMappingModule());
		modules.add(new CopySafetyFunctionArchModule());

		// Define a completer module that "controls" the execution of the decode and the evaluation
		// phases. It also allows to parallelize the execution of these phases.
		IndividualCompleterModule comp = new IndividualCompleterModule();
		if(opt4JParameters.isParallelExecution()) {
			comp.setType(IndividualCompleterModule.Type.PARALLEL);
			comp.setThreads(Runtime.getRuntime().availableProcessors());
		}
		modules.add(comp);

		// Create and initialize optimization task
		opt4JTask = new Opt4JTask(false);
		opt4JTask.init(modules);

		if(progressMonitor != null) {
			opt4JTask.addStateListener(new TaskStateListener() {

				@Override
				public void stateChanged(Task task) {
					// stateChanged() is invoked after every generation
					progressMonitor.worked(1);

					if(progressMonitor.isCanceled()) {
						System.out.println("DSE has been canceled.");
						Control control = opt4JTask.getInstance(Control.class);
						control.doStop();
					}
				}
			});
		}
	}

	/**
	 * {@inheritDoc}
	 * 
	 * @throws Exception
	 */
	@Override
	public CompositeExplorationSolution<?> explore() throws Exception {

		System.out.println("Running EA optimization...");
		if(progressMonitor != null) {
			progressMonitor.beginTask("EA optimization", opt4JParameters.getIterations());
		}

		CompositeExplorationSolution<?> rval = null;
		Exception ex = null;
		try {
			// Execute optimization
			opt4JTask.execute();

			// Collect the set objectives
			Collection<Objective> objectiveSet = new HashSet<>();
			for(ExplorationTarget<?> copiedExpTarget : targetAssocMap.values()) {
				if(!(copiedExpTarget instanceof ExplorationConstraint<?>)) {
					Objective constraintObjective =
							archExplorationProblemModule.getObjectiveOf(copiedExpTarget);
					objectiveSet.add(constraintObjective);
				}
			}

			Archive archive = new DefaultArchive();
			Archive completeArchive = opt4JTask.getInstance(Archive.class);

			// Check each individual for constraint violations and only further process those which
			// violate none.
			Collection<Individual> feasibleIndividuals = new ArrayList<>();
			for(Individual curIndividual : completeArchive) {
				boolean satisfiesAllConstraints = true;
				for(ExplorationConstraint<?> target : pickInstanceOf(ExplorationConstraint.class,
						expSpec.getTargets())) {
					Objective objective = archExplorationProblemModule.getObjectiveOf(target);
					Value<?> value = curIndividual.getObjectives().get(objective);
					if(!validateSolution((IExplorationSolution<?>)curIndividual.getPhenotype(),
							target, value, archExplorationProblemModule.getEvaluators())) {
						satisfiesAllConstraints = false;
					}
				}

				if(satisfiesAllConstraints) {
					feasibleIndividuals.add(curIndividual);
				}
			}

			// For the final Pareto-Front calculation, consider only Objectives.
			// Create a new archive with the same individuals that have only objective evaluations.
			// TODO: For robust optimization, we should also consider a distance metric to the
			// constraint's bounds.
			Map<Individual, Objectives> completeObjectives = new HashMap<>();
			for(Individual curIndividual : feasibleIndividuals) {
				Objectives oldObjectives = curIndividual.getObjectives();
				completeObjectives.put(curIndividual, oldObjectives);
				Objectives reducedObjectives = new Objectives();
				for(Entry<Objective, Value<?>> singleObjective : oldObjectives) {
					if(objectiveSet.contains(singleObjective.getKey())) {
						reducedObjectives.add(singleObjective.getKey(), singleObjective.getValue());
					}
				}
				curIndividual.setObjectives(reducedObjectives);
				archive.update(curIndividual);
			}

			// However, readd the complete objective set afterwards such that all evaluations are
			// available (including the constraint evaluations).
			for(Individual curIndividual : feasibleIndividuals) {
				Objectives origObjectives = completeObjectives.get(curIndividual);
				curIndividual.setObjectives(origObjectives);
			}

			rval = new Opt4JExplorationSolutionSet<ArchitectureSolution>(archive, targetAssocMap,
					archExplorationProblemModule);

			rval.setCompletedIterations(opt4JTask.getIteration());
		} catch(Exception e) {
			ex = e;
		} finally {
			opt4JTask.close();
		}

		if(ex != null) {
			throw ex;
		}

		return rval;
	}
}
