/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j;

import static org.fortiss.af3.exploration.util.ExplorationUtils.getReturnTypeOf;

import java.util.Map;
import java.util.Map.Entry;

import org.fortiss.af3.exploration.alg.dse.SolutionQuantification;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem.ProblemModuleBase;
import org.fortiss.af3.exploration.model.ExplorationTarget;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.opt4j.core.Objective;
import org.opt4j.core.Objectives;

/**
 * Exploration result for a single solution computed by Opt4J, i.e. its fitness w.r.t to the
 * investigated optimization dimensions.
 * 
 * @author barner
 */
public class Opt4JSingleExplorationSolution extends SolutionQuantification {

	/**
	 * Constructs the {@SolutionQuantification} representation of the
	 * fitness values for a single solution w.r.t to the investigated optimization dimensions
	 * obtained by an Opt4J run.
	 * 
	 * @throws Exception
	 *             if the evaluation result for an {@link Objective} cannot be identified.
	 */
	Opt4JSingleExplorationSolution(Objectives objectives,
			Map<ExplorationTarget<?>, ExplorationTarget<?>> targetAssocMap,
			ProblemModuleBase<?, ?, ?> problemModule) throws Exception {
		for(Entry<ExplorationTarget<?>, ExplorationTarget<?>> targetAssoc : targetAssocMap
				.entrySet()) {
			// Determine the ExplorationTarget class to identify the correct return value type.
			Objective objective = problemModule.getObjectiveOf(targetAssoc.getValue());
			ExplorationTarget<?> concreteExpTarget = targetAssoc.getKey();
			Class<?> expectedResultType = getReturnTypeOf(concreteExpTarget);
			// FIXME: throw an exception here.
			assert (expectedResultType != null) : "Could not determine the result type of the exploration target " +
					concreteExpTarget.getName() +
					". Possible reasons are a missing deduction from the class ExplorationTarget or a missing parameter substitiution in the concrete ExplorationTarget.";

			// Type check the result vs. the expected type defined by the exploration target.
			// TODO: Possible NPE: Throw a more precise Exception, instead.
			Object result = objectives.get(objective).getValue();

			// FIXME: remove this hack: Instead, the safetycomplianceconstraint should have the
			// right type parameter for the value and interface the repair mechanism
			// differently.
			if(expectedResultType.equals(Boolean.class) ||
					IModelElement.class.isAssignableFrom(expectedResultType)) {
				result = Boolean.valueOf(objectives.get(objective).getDouble() <= 10E-9);
			} else if(result != null && !result.getClass().equals(expectedResultType)) {
				// Note: if the result is null, a serious constraint violation is indicated. Its
				// corresponding solution will be filtered out in a later stage.
				throw new IllegalArgumentException(
						"Incompatible evaluation result type for the exploration target " +
								concreteExpTarget.getName() + ".");
			}
			// "Store" the result if the type of the result is the expected one.
			addSingleExplorationResult(concreteExpTarget, objectives.get(objective));
		}
	}
}
