/*-------------------------------------------------------------------------+
| Copyright 2017 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.partitionmapping;

import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.MappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.genotype.ComposableGenotype;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.ComposablePhenotype;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IExecutionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemModelAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.random.RandomExploration;

import com.google.common.collect.ArrayListMultimap;
import com.google.common.collect.ListMultimap;

/**
 * Encoding that maps partition, which host tasks, to execution units. Typically, these are cores.
 * The encoding is essentially a two-dimensional table.
 * 
 * @author diewald
 */
public class PartitionMappingEncoding extends
		MappingEncoding<PartitionMappingEncoding, Partition, IExecutionUnitAdapter<?>, PartitionMappingEntry>
		implements ComposableGenotype, Cloneable {
	/** The allocation table. */
	private ListMultimap<IResourceAdapter<?>, Partition> allocationTable =
			ArrayListMultimap.create();

	private Collection<? extends IResourceAdapter<?>> targetSet;

	/** Constructor. */
	// TODO: The strategy can be replaced by lambdas for a cleaner design.
	public PartitionMappingEncoding(SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter,
			CREATE_STARTEGY strategy) {
		// Number of tasks to deploy
		// TODO: This does not consider task replicas yet.
		int taskNum = systemModelAdapter.getDeployableComponents().size();
		targetSet = systemModelAdapter.getDeploymentTargets();
		if(strategy == CREATE_STARTEGY.MIN_SET) {
			for(IResourceAdapter<?> target : targetSet) {
				allocationTable.put(target, new Partition());
			}
		} else if(strategy == CREATE_STARTEGY.RANDOM_SET) {
			for(IResourceAdapter<?> target : targetSet) {
				int randMaxPartitions = RandomExploration.getInstance().nextInt(taskNum);
				for(int i = 0; i < randMaxPartitions; ++i) {
					allocationTable.put(target, new Partition());
				}
			}
		} else if(strategy == CREATE_STARTEGY.MAX_SET) {
			for(IResourceAdapter<?> target : targetSet) {
				for(int i = 0; i < taskNum; ++i) {
					allocationTable.put(target, new Partition());
				}
			}
		}
	}

	/** Prototype Constructor. */
	private PartitionMappingEncoding() {
		// Prototype for newInstance().
	}

	public Collection<IResourceAdapter<?>> getResources() {
		return allocationTable.keySet();
	}

	public List<Partition> getRequesters(IResourceAdapter<?> resource) {
		return allocationTable.get(resource);
	}

	/** Creates a {@link Partition} that is allocated to the given {@link IResourceAdapter}. */
	public void createPartition(IResourceAdapter<?> resource) {
		allocationTable.put(resource, new Partition());
	}

	/** Removes a random {@link Partition} allocation to the given {@link IResourceAdapter}. */
	public void removeRandomPartition(IResourceAdapter<?> resource) {
		List<Partition> partitions = allocationTable.get(resource);
		int randIdx = RandomExploration.getInstance().nextInt(partitions.size());
		Partition rmPartition = partitions.get(randIdx);
		removePartition(resource, rmPartition);
	}

	/** Removes the given {@link Partition} allocation to the given {@link IResourceAdapter}. */
	public void removePartition(IResourceAdapter<?> resource, Partition partition) {
		allocationTable.remove(resource, partition);
	}

	/** {@inheritDoc} */
	@Override
	public int size() {
		return allocationTable.keySet().size();
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public PartitionMappingEncoding newInstance() {
		return new PartitionMappingEncoding();
	}

	/** {@inheritDoc} */
	@Override
	public Class<? extends ComposableGenotype> getGTypeToRegister() {
		return getClass();
	}

	/** {@inheritDoc} */
	@Override
	public Class<? extends ComposablePhenotype<?>> getPTypeToRegister() {
		return getClass();
	}

	/** {@inheritDoc} */
	@Override
	public PartitionMappingEncoding clone() {
		PartitionMappingEncoding clonedEncoding = new PartitionMappingEncoding();

		Map<PartitionMappingEntry, PartitionMappingEntry> copyAssocMap = new HashMap<>();
		clonedEncoding.allocationMap = copyAllocationMap(copyAssocMap);

		// TODO: Implement a generic version of the assoc Map copy and call it here.
		// TODO: Implement a generic clone version, that can be called from here.

		return clonedEncoding;
	}
}
