/*-------------------------------------------------------------------------+
| Copyright 2015 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.abstractmapping;

import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.partitionmapping.Partition;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.partitionmapping.PartitionMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskgraph.AbstractTaskGraphEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskgraph.SafeTaskGraphEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.TaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.TaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.genotype.ComposableGenotype;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.ComposablePhenotype;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IDeployableComponentAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemModelAdapter;
import org.fortiss.af3.exploration.alg.exception.ExplorationException;
import org.fortiss.af3.exploration.moea.model.DseSpecification;
import org.opt4j.core.Genotype;

/**
 * {@link Genotype} which encodes an {@link TaskMappingEncoding} that may contain
 * {@link TaskMappingEntry}s which are required to be initialized.
 * 
 * @author diewald
 */
public class AbstractTaskMappingEncoding
		extends TaskMappingEncoding<IDeployableComponentAdapter<?>, AbstractTaskMappingEntry> {

	/**
	 * Constructor that is initialized with a reference to the DSE specification and the input
	 * system model.
	 */
	AbstractTaskMappingEncoding(DseSpecification dse,
			SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter,
			PartitionMappingEncoding partEnc, SafeTaskGraphEncoding<?> tgEncoding) {
		super(dse, systemModelAdapter, partEnc, tgEncoding);
	}

	/**
	 * Constructor that optionally initializes the encoding to a random encoding
	 * 
	 * @throws ExplorationException
	 *             if no valid allocation targets can be identified for one
	 *             {@link IDeployableComponentAdapter}.
	 */
	public AbstractTaskMappingEncoding(DseSpecification dse,
			SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter,
			PartitionMappingEncoding partEnc, SafeTaskGraphEncoding<?> tgEncoding,
			boolean initialize) throws ExplorationException {
		super(dse, systemModelAdapter, partEnc, tgEncoding, initialize);
	}

	/**
	 * Constructor that optionally initializes the encoding to a random encoding
	 * 
	 * @throws ExplorationException
	 *             if no valid allocation targets can be identified for one
	 *             {@link IDeployableComponentAdapter}.
	 */
	public AbstractTaskMappingEncoding(DseSpecification dse,
			SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter, boolean initialize)
			throws ExplorationException {
		super(dse, systemModelAdapter, initialize);
	}

	/**
	 * Constructor for UPDATES.
	 * 
	 * @throws ExplorationException
	 *             if no valid allocation targets can be identified for one
	 *             {@link IDeployableComponentAdapter}.
	 */
	public AbstractTaskMappingEncoding(DseSpecification dse,
			SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter,
			PartitionMappingEncoding partEnc, AbstractTaskGraphEncoding<?, ?> atgEnc,
			boolean initialize) throws ExplorationException {
		super(dse, systemModelAdapter, partEnc, atgEnc, initialize);
	}

	/**
	 * Constructor that generated a new {@link AbstractTaskMappingEncoding} preserving the
	 * allocations from the given {@code partialEncoding}.
	 * 
	 * Constructor for UPDATES.
	 * 
	 * @throws ExplorationException
	 *             if no valid allocation targets can be identified for one
	 *             {@link IDeployableComponentAdapter}.
	 */
	public AbstractTaskMappingEncoding(DseSpecification dse,
			SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter,
			PartitionMappingEncoding partEnc, AbstractTaskGraphEncoding<?, ?> atgEnc,
			AbstractTaskMappingEncoding partialEncoding) throws ExplorationException {
		super(dse, systemModelAdapter, partEnc, atgEnc, partialEncoding);
	}

	/** {@inheritDoc} */
	@Override
	public AbstractTaskMappingEntry createEntry(IDeployableComponentAdapter<?> deployableComponent,
			Partition deploymentTarget) {
		// Request not needed to construct a {@link FailSilentTaskMappingEncoding}
		if(deploymentTarget == null) {
			return null;
		}

		return new AbstractTaskMappingEntry(deployableComponent, deploymentTarget);
	}

	/** {@inheritDoc} */
	@Override
	protected <S extends TaskMappingEntry> AbstractTaskMappingEntry createEntry(S allocEntry) {
		// TODO: Remove the ugly casts.
		return new AbstractTaskMappingEntry((IDeployableComponentAdapter<?>)allocEntry.getSource(),
				(Partition)allocEntry.getTarget());
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public AbstractTaskMappingEncoding newInstance() {
		try {
			return new AbstractTaskMappingEncoding(dse, systemModelAdapter, true);
		} catch(ExplorationException e) {
			// Opt4J does not allow to throw Exceptions here.
			throw new RuntimeException(e);
		}
	}

	/** {@inheritDoc} */
	@Override
	public AbstractTaskMappingEncoding clone() {
		AbstractTaskMappingEncoding clonedEncoding;
		try {
			clonedEncoding = new AbstractTaskMappingEncoding(dse, systemModelAdapter, false);
		} catch(ExplorationException e) {
			// Opt4J does not allow to throw Exceptions here.
			throw new RuntimeException(e);
		}

		Map<AbstractTaskMappingEntry, AbstractTaskMappingEntry> copyAssocMap = new HashMap<>();

		clonedEncoding.allocationMap = copyAllocationMap(copyAssocMap);
		clonedEncoding.constraintMap = copyConstraintMap(copyAssocMap);
		clonedEncoding.getRequesters().forEach(c -> c.updateInstatiationData());
		return clonedEncoding;
	}

	/** {@inheritDoc} */
	@Override
	public Class<? extends ComposableGenotype> getGTypeToRegister() {
		return AbstractTaskMappingEncoding.class;
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public Class<? extends ComposablePhenotype<?>> getPTypeToRegister() {
		return AbstractTaskMappingEncoding.class;
	}

	/** {@inheritDoc} */
	@Override
	public <GT extends Genotype, CP extends ComposablePhenotype<GT>> CP
			getSubPhenotype(Class<CP> subPhenotype) {
		// TODO Auto-generated method stub
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public <GT extends Genotype, CP extends ComposablePhenotype<GT>> void
			putSubPhenotype(Class<CP> subGenotype, CP curPhenotype) {
		// TODO Auto-generated method stub

	}

	/** {@inheritDoc} */
	@Override
	public List<ComposablePhenotype<? extends Genotype>> getSubPhenotypes() {
		// TODO Auto-generated method stub
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public Set<Class<? extends ComposablePhenotype<? extends Genotype>>> keySet() {
		// TODO Auto-generated method stub
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public Iterator<Entry<Class<? extends ComposablePhenotype<? extends Genotype>>, ComposablePhenotype<? extends Genotype>>>
			iterator() {
		// TODO Auto-generated method stub
		return null;
	}
}
