/*-------------------------------------------------------------------------+
| Copyright 2015 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.mating;

import static org.fortiss.af3.exploration.alg.service.ExplorationService.getService;

import java.util.LinkedList;
import java.util.List;
import java.util.Queue;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.genotype.ComposableGenotype;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IDeployableComponentAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemModelAdapter;
import org.fortiss.af3.exploration.alg.exception.ExplorationException;
import org.fortiss.af3.exploration.alg.service.IExplorationRepairService;
import org.fortiss.af3.exploration.alg.service.IRepairModule;
import org.opt4j.core.Genotype;
import org.opt4j.core.Individual;
import org.opt4j.core.IndividualFactory;
import org.opt4j.core.common.random.Rand;
import org.opt4j.core.genotype.CompositeGenotype;
import org.opt4j.operators.copy.Copy;
import org.opt4j.operators.crossover.Crossover;
import org.opt4j.operators.crossover.Pair;
import org.opt4j.operators.mutate.Mutate;
import org.opt4j.operators.mutate.MutationRate;
import org.opt4j.optimizers.ea.Coupler;
import org.opt4j.optimizers.ea.CrossoverRate;
import org.opt4j.optimizers.ea.Mating;
import org.opt4j.optimizers.ea.MatingCrossoverMutate;

import com.google.inject.Inject;
import com.google.inject.name.Named;

/**
 * {@link Mating} module that extends {@link MatingCrossoverMutate} by providing an additional
 * repair mechanism.
 * 
 * @author diewald
 */
public class MatingCrossoverMutateRepair extends MatingCrossoverMutate {

	/** References the input model. */
	@Inject
	@Named("SystemModelAdapter")
	private SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter;

	/**
	 * Constructor for the {@link MatingCrossoverMutateRepair} module which extends
	 * {@link MatingCrossoverMutate}.
	 */
	@Inject
	public MatingCrossoverMutateRepair(Crossover<Genotype> crossover, Mutate<Genotype> mutate,
			Copy<Genotype> copy, Coupler coupler, CrossoverRate crossoverRate,
			MutationRate mutationRate, Rand random, IndividualFactory individualFactory) {
		super(crossover, mutate, copy, coupler, crossoverRate, mutationRate, random,
				individualFactory);
	}

	/** {@inheritDoc} */
	@Override
	protected Pair<Individual> mate(Individual parent1, Individual parent2, boolean doCrossover) {
		Individual repaired1;
		Individual repaired2;
		try {
			repaired1 = repairGenotype(parent1);
			repaired2 = repairGenotype(parent2);
		} catch(ExplorationException e) {
			// Opt4J does not allow to throw Exceptions here.
			throw new RuntimeException(e);
		}

		Pair<Individual> pair;
		if(repaired1 == null && repaired2 == null) {
			pair = super.mate(parent1, parent2, doCrossover);
		} else if(repaired1 == null) {
			pair = super.mate(parent1, repaired2, doCrossover);
		} else if(repaired2 == null) {
			pair = super.mate(repaired1, parent2, doCrossover);
		} else {
			pair = new Pair<Individual>(repaired1, repaired2);
		}

		// TODO #checking: Add a final check here, to capture repair errors.

		return pair;
	}

	/**
	 * Triggers the repair method for {@link Individual}s.
	 * 
	 * @throws ExplorationException
	 *             if no valid allocation targets can be identified for one
	 *             {@link IDeployableComponentAdapter}.
	 */
	private Individual repairGenotype(Individual ind) throws ExplorationException {
		boolean isRepaired = false;
		Genotype genotype = ind.getGenotype();

		for(IRepairModule<?> repairModule : getService(IExplorationRepairService.class)
				.getRepairFunctionsOf(genotype, systemModelAdapter)) {
			if(repairModule.needsRepair(genotype, ind.getObjectives().getValues())) {
				ComposableGenotype copiedGenotype = (ComposableGenotype)copy.copy(genotype);
				repairModule.repairGenotype(copiedGenotype, ind.getObjectives().getValues());
				isRepaired = true;
			}
		}

		if(genotype instanceof CompositeGenotype<?, ?>) {
			@SuppressWarnings("unchecked") Queue<ComposableGenotype> traversalQueue =
					new LinkedList<>(((CompositeGenotype<?, ComposableGenotype>)genotype).values());
			while(!traversalQueue.isEmpty()) {
				ComposableGenotype curGenotype = traversalQueue.poll();
				List<IRepairModule<ComposableGenotype>> repairModules =
						getService(IExplorationRepairService.class).getRepairFunctionsOf(
								curGenotype, systemModelAdapter);
				for(IRepairModule<?> curRepairModule : repairModules) {
					if(curRepairModule.needsRepair(curGenotype, ind.getObjectives().getValues())) {
						ComposableGenotype copiedGenotype =
								(ComposableGenotype)copy.copy(curGenotype);
						curRepairModule.repairGenotype(copiedGenotype, ind.getObjectives()
								.getValues());
						isRepaired = true;
					}
				}

				if(curGenotype instanceof CompositeGenotype) {
					@SuppressWarnings("unchecked") CompositeGenotype<?, ComposableGenotype> castedGenotype =
							(CompositeGenotype<?, ComposableGenotype>)curGenotype;
					traversalQueue.addAll(castedGenotype.values());
				}
			}
		}

		return isRepaired ? individualFactory.create(genotype) : null;
	}
}
