/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;

import org.fortiss.af3.exploration.alg.dse.TaskMappingFactory;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.comm.MessageEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskgraph.InstantiatedAcyclicTaskGraphEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskgraph.InstantiatedTaskGraphEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.InstantiatedTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.InstantiatedTaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.decode.ComposableDecoderBase;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.exception.DecodingException;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.FlatPhenotypeMap;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.solution.StrictTTSchedule;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemModelAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.sched.comm.TDMABusScheduler;
import org.fortiss.af3.exploration.alg.dse.sysmodel.sched.proc.EDFSchedulerBase;
import org.fortiss.af3.exploration.moea.model.DseSpecification;
import org.opt4j.core.Genotype;

import com.google.inject.Inject;

/**
 * The Decoder decodes {@link Genotype}s into phenotype objects.
 * 
 * @author huang
 */
public class StrictTTDecoder<S extends InstantiatedTaskMappingEntry, T extends InstantiatedTaskMappingEncoding<S>>
		extends ComposableDecoderBase<T, StrictTTSchedule<S, T>> implements ArchExpSubDecoder {

	/** Representation of input system model */
	private SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter;

	/** Underlying EDF task scheduler */
	private EDFSchedulerBase<S, T> simpleEDFScheduler;

	/** Underlying bus scheduler */
	private TDMABusScheduler<S, T> simpleBusScheduler;

	/**
	 * Constructs a new decoder that constructs {@link StrictTTSchedule}s for a given exploration
	 * run.
	 */
	@SuppressWarnings("unchecked")
	@Inject
	public StrictTTDecoder(SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter,
			DseSpecification explorationDefinition) {
		this.systemModelAdapter = systemModelAdapter;

		TaskMappingFactory<?, ?> tmFactory = TaskMappingFactory.getInstance();

		simpleEDFScheduler =
				(EDFSchedulerBase<S, T>)tmFactory.createEDFScheduler(systemModelAdapter,
						explorationDefinition.getTargetSpecification());
		simpleBusScheduler = new TDMABusScheduler<S, T>(systemModelAdapter,
				explorationDefinition.getTargetSpecification());
	}

	/** {@inheritDoc} */
	@Override
	public StrictTTSchedule<S, T> decodeComposable(T genotype,
			FlatPhenotypeMap<Phenotype> phenotypeTypeMap) {
		@SuppressWarnings("unchecked") MessageEncoding<S, T> msgEnc =
				phenotypeTypeMap.getEncoding(MessageEncoding.class);
		assert (msgEnc != null) : "The required message encoding could not be found in the map of already decoded phenotypes.";
		@SuppressWarnings("unchecked") InstantiatedAcyclicTaskGraphEncoding<S, ?> acycItgEnc =
				phenotypeTypeMap.getEncoding(InstantiatedAcyclicTaskGraphEncoding.class);
		assert (acycItgEnc != null) : "The required acyclic and instantiated task graph encoding could not be found in the map of already decoded phenotypes.";

		// TODO: Externalize the validation such that the encoding is not required to be passed to
		// the constructor (is there for validation purposes)
		StrictTTSchedule<S, T> schedule = new StrictTTSchedule<S, T>(systemModelAdapter, genotype);
		simpleEDFScheduler.schedule(acycItgEnc, genotype, schedule);

		assert (schedule.validateSchedule(true));

		// TODO: Find out why the message scheduler does not require the task graph. Actually, it
		// should require it...
		simpleBusScheduler.schedule(schedule, msgEnc.getPeriodicMessages(), acycItgEnc);
		assert (schedule.validateSchedule(true));

		return schedule;
	}

	/**
	 * {@inheritDoc}
	 * 
	 * @throws DecodingException
	 *             see {@link ComposableDecoderBase}
	 */
	@Override
	public StrictTTSchedule<S, T> updateComposable(T inGenotype, StrictTTSchedule<S, T> outGenotype,
			FlatPhenotypeMap<Phenotype> decodedPhenotypes) throws DecodingException {
		// Not required: The message encoding is a Phenotype only.
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<Class<?>> getRequiredTypes() {
		Collection<Class<?>> reqTypes = new ArrayList<>();
		reqTypes.add(InstantiatedTaskMappingEncoding.class);
		reqTypes.add(MessageEncoding.class);
		reqTypes.add(InstantiatedAcyclicTaskGraphEncoding.class);
		// FIXME: Remove as soon as the evaluators can declare their required Phenotypes!!! (Needed
		// for the period evaluator)
		reqTypes.add(InstantiatedTaskGraphEncoding.class);

		return reqTypes;
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public Class<StrictTTSchedule<S, T>> getOutputType() {
		Class<? super StrictTTSchedule<S, T>> retType = StrictTTSchedule.class;
		return (Class<StrictTTSchedule<S, T>>)retType;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<Class<?>> getOptionalTypes() {
		return Collections.emptySet();
	}
}
