/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem.comm;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Map;

import org.fortiss.af3.exploration.alg.dse.TaskMappingFactory;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.comm.MessageEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.partitionmapping.Partition;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskgraph.InstantiatedTaskGraphEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.TaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.TaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.InstantiatedTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.decode.ComposableDecoder;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.decode.ComposableDecoderBase;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.exception.DecodingException;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.FlatPhenotypeMap;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IDeployableComponentAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemModelAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.comm.Message;
import org.fortiss.af3.exploration.alg.dse.sysmodel.comm.MessageGenerator;
import org.fortiss.af3.exploration.alg.dse.sysmodel.mapping.comm.IMessageRouter;
import org.jgrapht.DirectedGraph;
import org.jgrapht.graph.DefaultEdge;

import com.google.inject.Inject;

/**
 * {@link ComposableDecoder} that takes an {@link InstantiatedTaskMappingEncoding} and a) an
 * generates the required messages for this {@link InstantiatedTaskMappingEncoding} and b) routes
 * the generated {@link Message} on the target platform.
 * 
 * @author diewald
 */
public class MessageDecoder<S extends TaskMappingEntry<IDeployableComponentAdapter<?>, Partition>, T extends TaskMappingEncoding<IDeployableComponentAdapter<?>, S>>
		extends ComposableDecoderBase<T, MessageEncoding<S, T>> {

	/** References the input system models. */
	protected SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter;

	/**
	 * {@link MessageGenerator} which analyzes an {@link InstantiatedTaskMappingEncoding} and
	 * determines the required {@link Message}s.
	 */
	protected MessageGenerator<S, T> messageGenerator;

	/** References the {@link IMessageRouter} selected in the Constructor. */
	protected IMessageRouter messageRouter;

	/** Default constructor that requires a reference to the input models. */
	@Inject
	public MessageDecoder(SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter) {
		this.systemModelAdapter = systemModelAdapter;

		TaskMappingFactory<?, ?> tmFactory = TaskMappingFactory.getInstance();

		messageGenerator = new MessageGenerator<S, T>(systemModelAdapter);
		messageRouter = tmFactory.createShortestPathRouter(systemModelAdapter);
	}

	/**
	 * {@inheritDoc}
	 * 
	 * @throws DecodingException
	 */
	@Override
	public MessageEncoding<S, T> decodeComposable(T genotype,
			FlatPhenotypeMap<Phenotype> decodedPhenotypes) throws DecodingException {
		Collection<Message> generatedMessages = messageGenerator.generateMessages(genotype);

		Map<Message, DirectedGraph<IResourceAdapter<?>, DefaultEdge>> messageRouteMap =
				messageRouter.calculateMessageRoutes(generatedMessages);
		if(messageRouteMap == null) {
			// At least one *invalid* communication link is required by the given mapping =>
			// infeasible mapping.
			throw new DecodingException(genotype, this,
					"At least one message could not be routed from the sending unit to all its receivers.");
		}

		// Add the calculated routes to the generated Messages.
		for(Message msg : generatedMessages) {
			msg.setRoute(messageRouteMap.get(msg));
		}

		return new MessageEncoding<>(generatedMessages);
	}

	/**
	 * {@inheritDoc}
	 * 
	 * @throws DecodingException
	 *             see {@link ComposableDecoder}.
	 */
	@Override
	public MessageEncoding<S, T> updateComposable(T inGenotype, MessageEncoding<S, T> outGenotype,
			FlatPhenotypeMap<Phenotype> decodedPhenotypes) throws DecodingException {
		// Not required: The message encoding is a Phenotype only.
		// TODO: Throw a decoding exception, instead.
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<Class<?>> getRequiredTypes() {
		Collection<Class<?>> reqTypes = new ArrayList<>();
		reqTypes.add(InstantiatedTaskMappingEncoding.class);
		reqTypes.add(InstantiatedTaskGraphEncoding.class);
		return reqTypes;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<Class<?>> getOptionalTypes() {
		return Collections.emptySet();
	}

	/** {@inheritDoc} */
	@SuppressWarnings("rawtypes")
	@Override
	public Class<MessageEncoding> getOutputType() {
		return MessageEncoding.class;
	}
}
