/*-------------------------------------------------------------------------+
| Copyright 2015 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem.instantiatemapping;

import static org.fortiss.tooling.common.util.LambdaUtils.filterSet;

import java.util.ArrayList;
import java.util.Collection;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.partitionmapping.PartitionMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskgraph.AbstractTaskGraphEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.abstractmapping.AbstractTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.abstractmapping.AbstractTaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.FailSilentTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.FailSilentTaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.decode.ComposableDecoder;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.exception.DecodingException;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.FlatPhenotypeMap;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IDeployableComponentAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemModelAdapter;
import org.fortiss.af3.exploration.moea.model.DseSpecification;

/**
 * Decoder for instantiating an {@link AbstractTaskMappingEncoding} to a
 * {@link FailSilentTaskMappingEncoding}.
 * 
 * @author diewald
 */
public class FailSilentAbstractTaskMappingDecoder extends
		AbstractTaskMappingDecoder<FailSilentTaskMappingEntry, FailSilentTaskMappingEncoding> {

	/** Constructor. */
	public FailSilentAbstractTaskMappingDecoder(
			SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter,
			DseSpecification explorationDefinition) {
		super(systemModelAdapter, explorationDefinition);
	}

	/** {@inheritDoc} */
	@Override
	public FailSilentTaskMappingEncoding decodeComposable(AbstractTaskMappingEncoding genotype,
			FlatPhenotypeMap<Phenotype> phenotypeTypeMap) {
		assert (false);
		AbstractTaskGraphEncoding<?, ?> atgEnc =
				phenotypeTypeMap.getEncoding(AbstractTaskGraphEncoding.class);
		assert (atgEnc != null) : "The required abstract task graph encoding could not be found in the map of already decoded phenotypes.";
		PartitionMappingEncoding partEnc =
				phenotypeTypeMap.getEncoding(PartitionMappingEncoding.class);
		assert (partEnc != null) : "The required partition mapping encoding could not be found in the map of already decoded phenotypes.";
		return new FailSilentTaskMappingEncoding(explorationDefinition, systemModelAdapter, partEnc,
				atgEnc, genotype);
	}

	// TODO: Abstract to the AbstractTaskMappingDecoder.
	/**
	 * {@inheritDoc}
	 * 
	 * @throws DecodingException
	 *             see {@link ComposableDecoder}.
	 */
	@Override
	public FailSilentTaskMappingEncoding updateComposable(AbstractTaskMappingEncoding inGenotype,
			FailSilentTaskMappingEncoding outGenotype,
			FlatPhenotypeMap<Phenotype> decodedPhenotypes) throws DecodingException {
		outGenotype.replaceConstraints(inGenotype.getConstraints());

		for(IDeployableComponentAdapter<?> comp : inGenotype.getRequesters()) {
			Collection<AbstractTaskMappingEntry> abstrEntries =
					inGenotype.getMappingEntriesOf(comp);
			Collection<FailSilentTaskMappingEntry> instEntries;
			if(comp.isInstantiationRequired()) {
				instEntries = filterSet(outGenotype.getMappingEntries(),
						e -> e.getSource().getReplacedComponent() == comp);
			} else {
				instEntries = outGenotype.getMappingEntriesOf(comp);
			}

			// Missing Associations between instantiated and abstract entries indicate the need to
			// create or remove corresp. entries.
			for(AbstractTaskMappingEntry abstrEntry : abstrEntries) {
				if(abstrEntry.getAssociatedElement().isEmpty()) {
					outGenotype.createAndAddEntry(abstrEntry);
				}
			}

			Collection<FailSilentTaskMappingEntry> rmEntries = new ArrayList<>();
			for(FailSilentTaskMappingEntry instEntry : instEntries) {
				if(instEntry.getAssociatedElement().isEmpty()) {
					rmEntries.add(instEntry);
				}
			}
			rmEntries.forEach(e -> outGenotype.removeEntry(e));
		}

		// Remove all tasks that are present in the output encoding but not in the input encoding.
		Collection<IDeployableComponentAdapter<?>> rmTasks = new ArrayList<>();
		for(IDeployableComponentAdapter<?> comp : outGenotype.getRequesters()) {
			IDeployableComponentAdapter<?> rmComponent = comp;
			if(comp.isInstantiation()) {
				comp = comp.getReplacedComponent();
			}
			if(!inGenotype.getRequesters().contains(comp)) {
				rmTasks.add(rmComponent);
			}
		}
		rmTasks.forEach(e -> outGenotype.removeTask(e));

		for(FailSilentTaskMappingEntry entry : outGenotype.getMappingEntries()) {
			entry.resetTarget(
					entry.getAssociatedElement(AbstractTaskMappingEntry.class).getTarget());
		}

		return outGenotype;
	}
}
