/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.evaluator.constraint;

import static org.fortiss.tooling.common.util.LambdaUtils.mapInOut;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.af3.component.model.Component;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.InstantiatedTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.InstantiatedTaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.FlatPhenotypeMap;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IDeployableComponentAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IExecutionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemModelAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.mapping.IMappingEntry;
import org.fortiss.af3.exploration.alg.service.IExplorationTargetEvaluator;
import org.fortiss.af3.exploration.model.ExplorationConstraint;
import org.fortiss.af3.exploration.moea.model.predefined.ComponentMultiAllocationConstraint;
import org.opt4j.core.DoubleValue;
import org.opt4j.core.Value;

import com.google.common.collect.ClassToInstanceMap;

/**
 * Deployment constraint which defines the allowed deployments of
 * {@link IDeployableComponentAdapter}s to {@link IExecutionUnitAdapter}s. All other allocations are
 * forbidden.
 * 
 * @author huang
 */
// TODO: Register this evaluator for AstractTaskMappings.
public class ComponentMultiAllocationConstraintEvaluator<S extends InstantiatedTaskMappingEntry, T extends InstantiatedTaskMappingEncoding<S>>
		extends ConstraintEvaluatorBase<T, Boolean, ComponentMultiAllocationConstraint> {

	/** Constructor. */
	public ComponentMultiAllocationConstraintEvaluator(
			SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter,
			ComponentMultiAllocationConstraint optGoal) throws Exception {
		super(systemModelAdapter, optGoal);
	}

	/** {@inheritDoc} */
	@Override
	public DoubleValue evaluateGoal(T phenotype, FlatPhenotypeMap<Phenotype> phenotypeTypeMap,
			ClassToInstanceMap<EObject> transformedModels) {
		for(InstantiatedTaskMappingEntry mappingEntry : phenotype.getMappingEntries()) {
			if(explorationTarget.getAllRightMultiLocationModelElements()
					.contains(mappingEntry.getSource().getObject())) {
				return new DoubleValue(Double.MAX_VALUE);
			}
		}
		return new DoubleValue(0.0);
	}

	/**
	 * {@inheritDoc}
	 * 
	 * @throws Exception
	 *             if the component or execution unit reference is missing, if the component is not
	 *             part of the deployable component set, or if the referenced execution unit is not
	 *             in the set of deployment targets.
	 */
	@Override
	public boolean validateInputs() throws Exception {
		if(explorationTarget.getComponent() == null || systemModelAdapter
				.getDeployableComponentOf(explorationTarget.getComponent()) == null) {
			throw new Exception("The referenced component of the dislocation constraint " +
					explorationTarget.getName() + "(id: " + explorationTarget.getId() +
					") is not correctly specified." +
					" Either it is missing, or not part of the set of deployable components.");
		}

		if(explorationTarget.getExecutionUnits().isEmpty() ||
				!mapInOut(systemModelAdapter.getDeploymentTargets(), t -> t.getObject(),
						ArrayList::new).containsAll(explorationTarget.getExecutionUnits())) {
			throw new Exception("The referenced execution unit of the dislocation constraint " +
					explorationTarget.getName() + "(id: " + explorationTarget.getId() +
					") is not correctly specified." +
					" Either it is missing, or not part of the set of deployment targets.");
		}

		return true;
	}

	/** {@inheritDoc} */
	@Override
	public boolean validateInternal(ExplorationConstraint<?> expTarget, T solution,
			Value<?> value) {
		Component constrComp = ((ComponentMultiAllocationConstraint)expTarget).getComponent();

		Collection<InstantiatedTaskMappingEntry> entries = new ArrayList<>();
		for(InstantiatedTaskMappingEntry mappingEntry : solution.getMappingEntries()) {
			IDeployableComponentAdapter<?> compAdp = mappingEntry.getSource();
			if(compAdp.getObject() == constrComp) {
				entries.add(mappingEntry);
			} else if(compAdp.isInstantiation() &&
					compAdp.getReplacedComponent().getObject() == constrComp) {
				entries.add(mappingEntry);
			}
		}

		if(entries.isEmpty()) {
			return false;
		}

		for(IMappingEntry entry : entries) {
			if(!((ComponentMultiAllocationConstraint)expTarget).getExecutionUnits()
					.contains(entry.getTarget().getObject())) {
				return false;
			}
		}
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<Class<?>> getRequiredInputTypes() {
		return Collections.emptySet();
	}

	/** {@inheritDoc} */
	@Override
	public Class<?> getPrimaryInputType() {
		return InstantiatedTaskMappingEncoding.class;
	}

	/** {@inheritDoc} */
	@Override
	public IExplorationTargetEvaluator<T, Boolean, ComponentMultiAllocationConstraint>
			createEvaluator(SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter,
					ComponentMultiAllocationConstraint explorationTarget) throws Exception {
		return new ComponentMultiAllocationConstraintEvaluator<>(systemModelAdapter,
				explorationTarget);
	}
}
