/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.evaluator.constraint;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.dse.evaluator.ExplorationTargetEvaluatorBase;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemModelAdapter;
import org.fortiss.af3.exploration.model.ExplorationConstraint;
import org.opt4j.core.Value;
import org.opt4j.core.problem.Evaluator;

/**
 * Base class for {@link Evaluator}s of constraints.
 * 
 * @author diewald
 */
public abstract class ConstraintEvaluatorBase<P extends Phenotype, RT, T extends ExplorationConstraint<RT>>
		extends ExplorationTargetEvaluatorBase<P, RT, T> implements IConstraintEvaluator<P, RT, T> {

	/**
	 * Defines the acceptable numeric imprecision, i.e. the allowed violation of constraints during
	 * the validation phase.
	 */
	protected final static double NUMERIC_IMPRECISION = 10E-9;

	/**
	 * Intended for counting the number of iterations of the DSE. It can be used in the evaluate
	 * method to increase the punishment of constraints with increased progress of the solver.
	 */
	private int iterationCount;

	/** Prototype constructor. */
	protected ConstraintEvaluatorBase() {
	}

	/** Creates an evaluator for a given exploration goal. */
	protected ConstraintEvaluatorBase(SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter,
			T goal) throws Exception {
		super(systemModelAdapter, goal);

		// Init with -1 due to automatic increment during readout via its getter function.
		iterationCount = -1;
	}

	// TODO: remove the line below. A status genotype is needed here.
	/**
	 * Returns the "iteration count". If the iteration count shall be used, it must be called
	 * EXACTLY ONCE per DSE iteration.
	 */
	protected int getIterationCountAndIncrement() {
		return iterationCount++;
	}

	/** {@inheritDoc} */
	@Override
	public <M> boolean validate(ExplorationConstraint<?> expTarget, M solution,
			Class<M> solutionType, Value<?> value) {
		if(!solutionType.isAssignableFrom(solution.getClass())) {
			return false;
		}
		@SuppressWarnings("unchecked") P castSolution = (P)solution;
		return validateInternal(expTarget, castSolution, value);
	}

	/** Actual validation method that must be implemented by each {@link IConstraintEvaluator}. */
	public abstract boolean validateInternal(ExplorationConstraint<?> expTarget, P solution,
			Value<?> value);
}
