/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.evaluator.constraint;

import static java.lang.Math.pow;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.HashSet;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskgraph.InstantiatedAcyclicTaskGraphEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.InstantiatedTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.InstantiatedTaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.FlatPhenotypeMap;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.solution.StrictTTSchedule;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IDeployableComponentAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemModelAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.af3.AF3SystemModelAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph.TaskGraph;
import org.fortiss.af3.exploration.alg.service.IExplorationTargetEvaluator;
import org.fortiss.af3.exploration.model.ExplorationConstraint;
import org.fortiss.af3.exploration.model.ExplorationTarget;
import org.fortiss.af3.exploration.model.time.DeadlineConstraint;
import org.opt4j.core.DoubleValue;
import org.opt4j.core.Value;

import com.google.common.collect.ClassToInstanceMap;

/**
 * Evaluator for deadline objectives.
 * 
 * @author huang
 */
public class DeadlineConstraintEvaluator<S extends InstantiatedTaskMappingEntry, T extends InstantiatedTaskMappingEncoding<S>>
		extends ConstraintEvaluatorBase<StrictTTSchedule<S, T>, Double, DeadlineConstraint> {

	/**
	 * The source component which defines the component associated with the start time of this
	 * deadline.
	 */
	private Collection<IDeployableComponentAdapter<?>> sourceDeployableComponent;

	/**
	 * The target component which defines the component associated with the finish time of this
	 * deadline.
	 */
	private Collection<IDeployableComponentAdapter<?>> targetDeployableComponent;

	/** Stores the deadline for this goal. */
	private double deadline;

	/** Objective to track evaluation of this {@link ExplorationTarget}. */
	public DeadlineConstraintEvaluator(SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter,
			DeadlineConstraint optGoal) throws Exception {
		super(systemModelAdapter, optGoal);
		sourceDeployableComponent = new ArrayList<>();
		targetDeployableComponent = new ArrayList<>();

		sourceDeployableComponent.add(systemModelAdapter.getDeployableComponentOf(optGoal
				.getStartComponent()));
		targetDeployableComponent.add(systemModelAdapter.getDeployableComponentOf(explorationTarget
				.getEndComponent()));

		deadline = explorationTarget.getDeadline();

		// TODO: AD: The cast to the AF3 systemmodeladapter breaks the abstration the goals. It
		// would be rather required to define the source and sink components as
		// IDeployableComponents in the ecore model. The transformation from AF3 components to the
		// IDeployableCompontes should be done when the data from the GUI is written to the model.
		IDeployableComponentAdapter<?> origStartComponent =
				((AF3SystemModelAdapter)systemModelAdapter).getDeployableComponentAdapter(explorationTarget
						.getStartComponent());
		if(origStartComponent.isInstantiation()) {
			sourceDeployableComponent.add(origStartComponent);
		} else {
			sourceDeployableComponent.addAll(origStartComponent.getReplacementComponents());
		}

		IDeployableComponentAdapter<?> origTargetComponent =
				((AF3SystemModelAdapter)systemModelAdapter).getDeployableComponentAdapter(explorationTarget
						.getEndComponent());
		if(origStartComponent.isInstantiation()) {
			targetDeployableComponent.add(origTargetComponent);
		} else {
			targetDeployableComponent.addAll(origTargetComponent.getReplacementComponents());
		}
	}

	/**
	 * evaluate the deadline objective
	 * if the deadline is met, we set the fitness value to -1
	 * if the deadline is violated, we set the fitness value to the gap between the actual latency
	 * and the deadline
	 */
	@Override
	public DoubleValue evaluateGoal(StrictTTSchedule<S, T> schedule,
			FlatPhenotypeMap<Phenotype> phenotypeTypeMap,
			ClassToInstanceMap<EObject> transformedModels) {
		@SuppressWarnings("unchecked") InstantiatedAcyclicTaskGraphEncoding<S, ?> acycItgEnc =
				phenotypeTypeMap.getEncoding(InstantiatedAcyclicTaskGraphEncoding.class);
		assert (acycItgEnc != null) : "The required acyclic and instantiated task graph encoding could not be found in the map of already decoded phenotypes.";

		double retval = 0;

		// One TaskGraph suffices as we know that "replacements" are always located in the same task
		// graph.
		TaskGraph taskGraph =
				acycItgEnc.getTaskGraphOf(sourceDeployableComponent.stream().findFirst().get());
		// we need to check each iteration in the hyper period
		for(int i = 0; i < systemModelAdapter.getHyperPeriod() / taskGraph.getPeriod(); i++) {
			// Required for the lambda operators.
			final int j = i;

			// the start time is the release time of this iteration, which is the beginning
			// of the period
			// double start = i * tg.getPeriod();
			IDeployableComponentAdapter<?> earliestComp =
					sourceDeployableComponent.stream()
							.min(Comparator.comparing(t -> schedule.getFirstStartTime(t, j))).get();
			double start = schedule.getFirstStartTime(earliestComp, i);
			IDeployableComponentAdapter<?> latestComp =
					targetDeployableComponent.stream()
							.min(Comparator.comparing(t -> schedule.getLastEndTime(t, j))).get();
			double end = schedule.getLastEndTime(latestComp, i);
			double latency = end - start;

			// TODO: fix the logic below, it is a stylistic nightmare.
			if(latency <= 0)
				continue;

			// see if the deadline is met
			if(!(latency > deadline)) {
				// deadline ok, keep the retval unchanged
			} else {
				// deadline violated
				if(retval < 0) {
					retval = latency - deadline;
				} else {
					retval += latency - deadline;
				}
			}
		}

		return new DoubleValue(retval > 0 ? pow(retval * getIterationCountAndIncrement(), 2) : 0);
	}

	/** {@inheritDoc} */
	@Override
	public boolean validateInputs() {
		// TODO: AD: is this check really required here? Shouldn't this be done by the
		// inputvalidator?
		if(explorationTarget.getStartComponent() == null || explorationTarget.getEndComponent() == null ||
				explorationTarget.getDeadline() <= 0) {
			return false;
		}

		return true;
	}

	/** {@inheritDoc} */
	@Override
	public boolean validateInternal(ExplorationConstraint<?> expTarget,
			StrictTTSchedule<S, T> solution, Value<?> value) {
		// TODO: Readd the getLimitation support or find another way to express the same.
		Double evalValue = value.getDouble();
		assert (expTarget instanceof DeadlineConstraint);
		if(evalValue > ((DeadlineConstraint)expTarget).getLimitation() - NUMERIC_IMPRECISION) {
			return false;
		}
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<Class<?>> getRequiredInputTypes() {
		Collection<Class<?>> inputTypes = new HashSet<>();
		inputTypes.add(StrictTTSchedule.class);
		inputTypes.add(InstantiatedAcyclicTaskGraphEncoding.class);
		return inputTypes;
	}

	/** {@inheritDoc} */
	@Override
	public Class<?> getPrimaryInputType() {
		return StrictTTSchedule.class;
	}

	/** {@inheritDoc} */
	@Override
	public IExplorationTargetEvaluator<StrictTTSchedule<S, T>, Double, DeadlineConstraint>
			createEvaluator(SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter,
					DeadlineConstraint explorationTarget) throws Exception {
		return new DeadlineConstraintEvaluator<S, T>(systemModelAdapter, explorationTarget);
	}
}
