/*-------------------------------------------------------------------------+
| Copyright 2015 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.evaluator.constraint;

import static java.lang.Math.pow;

import java.util.Collection;
import java.util.HashSet;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskgraph.InstantiatedTaskGraphEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.InstantiatedTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.InstantiatedTaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.FlatPhenotypeMap;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.solution.StrictTTSchedule;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IDeployableComponentAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IExecutionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemModelAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph.TaskGraph;
import org.fortiss.af3.exploration.alg.service.IExplorationTargetEvaluator;
import org.fortiss.af3.exploration.model.ExplorationConstraint;
import org.fortiss.af3.exploration.model.ExplorationTarget;
import org.fortiss.af3.exploration.model.time.PeriodConstraint;
import org.opt4j.core.DoubleValue;
import org.opt4j.core.Value;

import com.google.common.collect.ClassToInstanceMap;

/**
 * This {@link IConstraintEvaluator} evaluates whether all {@link IDeployableComponentAdapter}s are
 * executed within their respective period. Any violation is summed and then squared.
 * 
 * @author diewald
 */
public class PeriodConstraintEvaluator<S extends InstantiatedTaskMappingEntry, T extends InstantiatedTaskMappingEncoding<S>>
		extends ConstraintEvaluatorBase<StrictTTSchedule<S, T>, Double, PeriodConstraint> {

	/** Objective to track evaluation of this {@link ExplorationTarget}. */
	public PeriodConstraintEvaluator(SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter,
			PeriodConstraint optGoal) throws Exception {
		super(systemModelAdapter, optGoal);
	}

	/**
	 * evaluate the deadline objective
	 * if the deadline is met, we set the fitness value to -1
	 * if the deadline is violated, we set the fitness value to the gap between the actual latency
	 * and the deadline
	 */
	@Override
	public DoubleValue evaluateGoal(StrictTTSchedule<S, T> schedule,
			FlatPhenotypeMap<Phenotype> phenotypeTypeMap,
			ClassToInstanceMap<EObject> transformedModels) {
		@SuppressWarnings("unchecked") InstantiatedTaskGraphEncoding<S, ?> itgEnc =
				phenotypeTypeMap.getEncoding(InstantiatedTaskGraphEncoding.class);
		assert (itgEnc != null) : "The required instantiated task graph encoding could not be found in the map of phenotypes.";
		InstantiatedTaskMappingEncoding<?> itmEnc =
				phenotypeTypeMap.getEncoding(InstantiatedTaskMappingEncoding.class);
		assert (itmEnc != null) : "The required instantiated task mapping encoding could not be found in the map of phenotypes.";

		double retval = 0;

		// we evaluate if all tasks are within the period. we want everything to be
		// finished within the period
		for(IDeployableComponentAdapter<?> deployableComponent : itmEnc.getRequesters()) {
			TaskGraph taskGraph = itgEnc.getTaskGraphOf(deployableComponent);
			// we need to check each iteration in the hyper period
			if(taskGraph != null) {
				for(int i = 0; i < systemModelAdapter.getHyperPeriod() /
						taskGraph.getPeriod(); i++) {
					double periodEnd = (i + 1) * taskGraph.getPeriod();
					double end = schedule.getLastEndTime(deployableComponent, i);
					if(end > periodEnd) {
						retval += end - periodEnd;
					}
				}
			}
		}

		return new DoubleValue(retval > 0 ? pow(retval, 2) : 0);
	}

	/**
	 * {@inheritDoc}
	 * 
	 * @throws Exception
	 *             if the period of one component is missing, or one combination of a component and
	 *             execution unit has no WCET specified.
	 */
	@Override
	public boolean validateInputs() throws Exception {
		// Each deployable component has a period.
		for(IDeployableComponentAdapter<?> component : systemModelAdapter
				.getDeployableComponents()) {
			if(component.getTimingRequirementPeriod() == null) {
				throw new Exception("The component " + component.getName() +
						" has no period specified. Each component that shall be considered" +
						" by the DSE must be associated with a period: either defined for" +
						" the component itself or indirectly by its ports.");
			}
		}

		// Check WCETs of each component/execution unit combination.
		for(IDeployableComponentAdapter<?> component : systemModelAdapter
				.getDeployableComponents()) {
			for(IExecutionUnitAdapter<?> execUnit : systemModelAdapter.getDeploymentTargets()) {
				if(component.getWcet(execUnit) == null) {
					throw new Exception("The combination of the component " + component.getName() +
							" and the executuion unit " + execUnit.getName() +
							" does not have a WCET specified. This parameter is required for the DSE.");
				}
			}
		}

		return true;
	}

	/** {@inheritDoc} */
	@Override
	public boolean validateInternal(ExplorationConstraint<?> expTarget,
			StrictTTSchedule<S, T> solution, Value<?> value) {
		Double evalValue = value.getDouble();
		assert (expTarget instanceof PeriodConstraint);
		Double limit = ((PeriodConstraint)expTarget).getLimitation();
		limit = limit == null ? 0 : limit;
		if(evalValue > limit + NUMERIC_IMPRECISION) {
			return false;
		}
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<Class<?>> getRequiredInputTypes() {
		Collection<Class<?>> inputTypes = new HashSet<>();
		inputTypes.add(StrictTTSchedule.class);
		inputTypes.add(InstantiatedTaskGraphEncoding.class);
		inputTypes.add(InstantiatedTaskMappingEncoding.class);
		return inputTypes;
	}

	/** {@inheritDoc} */
	@Override
	public Class<?> getPrimaryInputType() {
		return StrictTTSchedule.class;
	}

	/** {@inheritDoc} */
	@Override
	public IExplorationTargetEvaluator<StrictTTSchedule<S, T>, Double, PeriodConstraint>
			createEvaluator(SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter,
					PeriodConstraint explorationTarget) throws Exception {
		return new PeriodConstraintEvaluator<>(systemModelAdapter, explorationTarget);
	}
}
