/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.evaluator.objective;

import java.util.Collection;
import java.util.HashSet;

import org.apfloat.Apfloat;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.FailSilentTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.FailSilentTaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.solution.StrictTTSchedule;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IDeployableComponentAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemModelAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph.DefaultTaskGraph;
import org.fortiss.af3.exploration.alg.service.IExplorationTargetEvaluator;
import org.fortiss.af3.exploration.moea.model.predefined.FailureMinObjective;
import org.fortiss.af3.exploration.moea.model.predefined.FailureType;

/**
 * Reliability analysis for strict time-triggered execution model (with fail-silent assumption).
 * 
 * @author huang
 */
public class FailSilentReliabilityEvaluator extends
		ReliabilityEvaluatorBase<FailSilentTaskMappingEntry, FailSilentTaskMappingEncoding> {

	/** Creates a fail-silent reliability analysis evaluator for a given exploration goal. */
	public FailSilentReliabilityEvaluator(SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter,
			FailureMinObjective goal) throws Exception {
		super(systemModelAdapter, goal);
	}

	/** {@inheritDoc} */
	@Override
	protected double evaluateInternal(
			StrictTTSchedule<FailSilentTaskMappingEntry, FailSilentTaskMappingEncoding> schedule,
			FailSilentTaskMappingEncoding itmEnc, DefaultTaskGraph taskGraph,
			FailureType failureProbabilityType) {

		// for the StrictTTFaileSilence model, no sdc could occur
		if(failureProbabilityType == FailureType.SDC) {
			return 0;
		}

		Apfloat overallSuccRate = new Apfloat(1, FP_PRECISION);
		Apfloat one = new Apfloat(1, FP_PRECISION);

		for(IDeployableComponentAdapter<?> deployableComponent : itmEnc.getRequesters()) {
			// TODO: Check if this correctly determines the taskFailureRate

			// only software tasks are considered

			// the failure rate of the single request
			Apfloat taskFailureRate = new Apfloat(1, FP_PRECISION);
			// get the set of mappings
			Collection<FailSilentTaskMappingEntry> entries =
					itmEnc.getMappingEntriesOf(deployableComponent);
			for(FailSilentTaskMappingEntry entry : entries) {
				Apfloat fr = getComponentFailureRate(deployableComponent, entry.getTarget());
				if(fr != null) {
					// the task fails only if all of its instances fail
					// ASSUMPTIN: fail-silence
					taskFailureRate = taskFailureRate.multiply(fr);
				}
			}
			// the overall application succeeds only if all of its tasks succeed
			Apfloat taskSuccRate = one.subtract(taskFailureRate);
			overallSuccRate = overallSuccRate.multiply(taskSuccRate);
		}

		// return value is the overall failure probability
		return (one.subtract(overallSuccRate)).doubleValue();
	}

	/** {@inheritDoc} */
	@Override
	public Collection<Class<?>> getRequiredInputTypes() {
		Collection<Class<?>> inputTypes = new HashSet<>();
		inputTypes.add(StrictTTSchedule.class);
		return inputTypes;
	}

	/** {@inheritDoc} */
	@Override
	public Class<?> getPrimaryInputType() {
		return StrictTTSchedule.class;
	}

	/** {@inheritDoc} */
	@Override
	public IExplorationTargetEvaluator<StrictTTSchedule<FailSilentTaskMappingEntry, FailSilentTaskMappingEncoding>, Double, FailureMinObjective>
			createEvaluator(SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter,
					FailureMinObjective explorationTarget) throws InstantiationException {
		try {
			return new FailSilentReliabilityEvaluator(systemModelAdapter, explorationTarget);
		} catch(Exception e) {
			throw new InstantiationException("Cannot instantiate the Evaluator " +
					getClass().getSimpleName() + ": " + e.getMessage());
		}
	}
}
