/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.evaluator.objective;

import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apfloat.Apfloat;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.FaultDetectionVotingTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.FaultDetectionVotingTaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.solution.StrictTTSchedule;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IDeployableComponentAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemModelAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph.DefaultTaskGraph;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph.TaskGraph;
import org.fortiss.af3.exploration.alg.dse.sysmodel.faultmodel.FaultDetector;
import org.fortiss.af3.exploration.alg.service.IExplorationTargetEvaluator;
import org.fortiss.af3.exploration.moea.model.predefined.FailureMinObjective;
import org.fortiss.af3.exploration.moea.model.predefined.FailureType;

/**
 * Reliability analysis for strict time-triggered execution model (with fault-detection and voting).
 * 
 * @author huang
 */
public class FaultDetectionVotingReliabilityEvaluator extends
		ReliabilityEvaluatorBase<FaultDetectionVotingTaskMappingEntry, FaultDetectionVotingTaskMappingEncoding> {

	/** Helper class to save the reliability information */
	private class ReliabilityMeasure {

		/** Success */
		public Apfloat suc;

		/** Detected unrecoverable fault */
		public Apfloat duf;

		/** Silent data corruption */
		public Apfloat sdc;

		/** Constructs a new {@link ReliabilityMeasure} (initialized to all zero). */
		public ReliabilityMeasure() {
			this(new Apfloat(0, FP_PRECISION), new Apfloat(0, FP_PRECISION),
					new Apfloat(0, FP_PRECISION));
		}

		/** Constructs a new {@link ReliabilityMeasure} (initialized to user supplied values) */
		public ReliabilityMeasure(Apfloat suc_in, Apfloat due_in, Apfloat sdc_in) {
			suc = suc_in;
			duf = due_in;
			sdc = sdc_in;
		}

		/** {@inheritDoc} */
		@Override
		public String toString() {
			return "DUE=" + duf.doubleValue() + ",SDC=" + sdc.doubleValue();
		}
	}

	/** Helper class or analyzing reliability of a single task using imperfect fault detectors */
	private class ActorAnalysisNode {

		/** Parent node */
		ActorAnalysisNode parent;

		/** The result: 1: SUC, 0: DUE, -1: SDC */
		int result;

		/** Occurrence probability up to now */
		Apfloat probability;

		/** {@link ActorAnalysisNode}'s level. */
		int level;

		/** Constructs a new {@link ActorAnalysisNode} (initialized to default values). */
		ActorAnalysisNode() {
			this(null, 1, new Apfloat(1, FP_PRECISION), -1);
		}

		/** Constructs a new {@link ActorAnalysisNode} (initialized to user supplied values). */
		public ActorAnalysisNode(ActorAnalysisNode parent, int result, Apfloat probability,
				int level) {
			this.parent = parent;
			this.result = result;
			this.probability = probability;
			this.level = level;
		}
	}

	/** Last schedule */
	private StrictTTSchedule<FaultDetectionVotingTaskMappingEntry, FaultDetectionVotingTaskMappingEncoding> lastSchedule =
			null;

	/** Last task graph */
	private DefaultTaskGraph lastTaskGraph = null;

	/** Last analysis result */
	private ReliabilityMeasure lastResult = null;

	/**
	 * Creates a fault-detection + voting reliability analysis evaluator for a given exploration
	 * goal.
	 */
	public FaultDetectionVotingReliabilityEvaluator(
			SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter, FailureMinObjective target)
			throws Exception {
		super(systemModelAdapter, target);
	}

	/**
	 * Reliability analysis for strict time-triggered execution model (with fault-detection and
	 * voting).
	 */
	@Override
	public double evaluateInternal(
			StrictTTSchedule<FaultDetectionVotingTaskMappingEntry, FaultDetectionVotingTaskMappingEncoding> schedule,
			FaultDetectionVotingTaskMappingEncoding itmEnc, DefaultTaskGraph taskGraph,
			FailureType failureProbabilityType) {

		// TODO: Implement this caching in a safer way, and potentially extend it to other cases.
		// first of all, see if the scenario was just analyzed
		if(schedule == lastSchedule && taskGraph == lastTaskGraph && lastResult != null) {
			// System.out.println("using last result " + lastResult);
			return convertResult(lastResult, failureProbabilityType);
			// remember the last analysis scenario, this is used to avoid duplicated analysis for
			// SDC and DUF
		}

		// not analyzed yet, just do it
		FaultDetectionVotingTaskMappingEncoding mapping = itmEnc;
		// STEP 1: analyze the reliability of each actor
		Map<IDeployableComponentAdapter<?>, ReliabilityMeasure> actorReliabilities =
				new HashMap<IDeployableComponentAdapter<?>, ReliabilityMeasure>();

		for(IDeployableComponentAdapter<?> deployableComponent : taskGraph.getTasks()) {
			Collection<FaultDetectionVotingTaskMappingEntry> mappings =
					mapping.getMappingEntriesOf(deployableComponent);
			// compute the reliability of the task
			ReliabilityMeasure rm = taskAnalysis(deployableComponent, taskGraph, mappings);
			actorReliabilities.put(deployableComponent, rm);

			// System.out.println("task" + ((NamedObject)task.eContainer()).getName() + ":" + rm
			// + "sum="+rm.suc.add(rm.duf).add(rm.sdc));
		}

		// STEP2: analyze the reliability of the application
		ReliabilityMeasure appReliability = appAnalysis(taskGraph, actorReliabilities);
		// System.out.println(solution.getEncoding());
		// System.out.println("appReliability" + taskgraph + appReliability);
		// TODO
		appReliability.duf = probabilityToFit(appReliability.duf, taskGraph.getPeriod());
		appReliability.sdc = probabilityToFit(appReliability.sdc, taskGraph.getPeriod());

		// remember the last analysis
		lastSchedule = schedule;
		lastTaskGraph = taskGraph;
		lastResult = appReliability;

		return convertResult(appReliability, failureProbabilityType);
	}

	/** Converts a failure probability to FIT. */
	private Apfloat probabilityToFit(Apfloat p, double period) {
		// the period is in primary time unit, convert to second
		// TODO: currently, the timeUnit is removed from the model, is set to muesec
		// double putos = Explorer.getInstance().toPrimaryTimeUnit(TimeUnitType.SECOND);
		double putos = 1.0;
		Apfloat numIterationPerHour = new Apfloat(3600.0 * putos / period);
		Apfloat fit =
				p.multiply(numIterationPerHour).multiply(new Apfloat(1000000000, FP_PRECISION));
		return fit;
	}

	/** Entry point for task analysis */
	private ReliabilityMeasure taskAnalysis(IDeployableComponentAdapter<?> deployableComponent,
			DefaultTaskGraph taskgraph, Collection<FaultDetectionVotingTaskMappingEntry> mappings) {

		ReliabilityMeasure rm = new ReliabilityMeasure();
		ActorAnalysisNode start = new ActorAnalysisNode();

		// the iterative analysis routine
		taskAnalysis(deployableComponent, start, taskgraph, rm,
				(List<FaultDetectionVotingTaskMappingEntry>)mappings);
		return rm;
	}

	/** Analyzes the reliability of a task. */
	private void taskAnalysis(IDeployableComponentAdapter<?> deployableComponent,
			ActorAnalysisNode current, DefaultTaskGraph taskgraph, ReliabilityMeasure rm,
			List<FaultDetectionVotingTaskMappingEntry> mappings) {

		// the start node
		if(current.level < 0) {
			ActorAnalysisNode suc = new ActorAnalysisNode(current, 1, new Apfloat(1, FP_PRECISION),
					current.level + 1);
			taskAnalysis(deployableComponent, suc, taskgraph, rm, mappings);
			ActorAnalysisNode due = new ActorAnalysisNode(current, 0, new Apfloat(1, FP_PRECISION),
					current.level + 1);
			taskAnalysis(deployableComponent, due, taskgraph, rm, mappings);
			ActorAnalysisNode sdc = new ActorAnalysisNode(current, -1, new Apfloat(1, FP_PRECISION),
					current.level + 1);
			taskAnalysis(deployableComponent, sdc, taskgraph, rm, mappings);
		} else {
			// compute the local probabilities
			FaultDetectionVotingTaskMappingEntry mp = mappings.get(current.level);
			Apfloat failureRate = getComponentFailureRate(deployableComponent, mp.getTarget());
			if(failureRate == null)
				failureRate = new Apfloat(1, FP_PRECISION);
			FaultDetector fd =
					taskgraph.getFaultDetector(deployableComponent, mp.getFaultDetectorID());
			double coverage = fd.getCoverage();

			Apfloat localPr;
			Apfloat one = new Apfloat(1, FP_PRECISION);
			Apfloat cov = new Apfloat(coverage, FP_PRECISION);
			if(current.result > 0) {
				localPr = one.subtract(failureRate);
			} else if(current.result == 0) {
				localPr = failureRate.multiply(cov);
			} else {
				localPr = failureRate.multiply(one.subtract(cov));
			}

			// System.out.println("level " + current.level+","+current.probability+","+localPr);
			// see the if the next level is already the leaves of the tree, if yes, we compute the
			// system-level probabilities
			// if not, we create the children and continue
			if(current.level >= mappings.size() - 1) {
				// trace the previous results
				ActorAnalysisNode aan = current;
				int result = 0;
				while(aan.parent != null) {
					// sum the results
					result += aan.result;
					aan = aan.parent;
				}

				if(result > 0) {
					// result is suc
					rm.suc = rm.suc.add(current.probability.multiply(localPr));
				} else if(result == 0) {
					rm.duf = rm.duf.add(current.probability.multiply(localPr));
				} else {
					rm.sdc = rm.sdc.add(current.probability.multiply(localPr));
				}
			} else {
				// create the children
				ActorAnalysisNode suc = new ActorAnalysisNode(current, 1,
						current.probability.multiply(localPr), current.level + 1);
				taskAnalysis(deployableComponent, suc, taskgraph, rm, mappings);
				ActorAnalysisNode due = new ActorAnalysisNode(current, 0,
						current.probability.multiply(localPr), current.level + 1);
				taskAnalysis(deployableComponent, due, taskgraph, rm, mappings);
				ActorAnalysisNode sdc = new ActorAnalysisNode(current, -1,
						current.probability.multiply(localPr), current.level + 1);
				taskAnalysis(deployableComponent, sdc, taskgraph, rm, mappings);
			}
		}
	}

	/**
	 * Analyzes an application (based on the reliability of the individual actors).
	 * 
	 * @param tg
	 *            {@link DefaultTaskGraph} representing the application
	 * @param actorReliabilities
	 *            Defines the reliability value assigned to the logical components.
	 * @return {@link ReliabilityMeasure} for the given {@link TaskGraph}.
	 */
	private ReliabilityMeasure appAnalysis(DefaultTaskGraph tg,
			Map<IDeployableComponentAdapter<?>, ReliabilityMeasure> actorReliabilities) {
		ReliabilityMeasure rm = new ReliabilityMeasure();

		// SUC: the task is successful, iff all actors are successful
		Apfloat taskSuc = computeTaskSuc(actorReliabilities);

		// SDC: a task is SDC iff, at least one of its tasks is SDC and non of them are DUE
		Map<IDeployableComponentAdapter<?>, ReliabilityMeasure> actorReliabilitiesCopy =
				new HashMap<IDeployableComponentAdapter<?>, ReliabilityMeasure>(actorReliabilities);
		Apfloat taskSdc = new Apfloat(0, FP_PRECISION);
		taskSdc = computeTaskSdc(actorReliabilitiesCopy);

		// result
		Apfloat one = new Apfloat(1, FP_PRECISION);
		rm.suc = taskSuc;
		rm.sdc = taskSdc;
		rm.duf = one.subtract(rm.suc).subtract(rm.sdc);
		return rm;
	}

	/** Computes the success probability of a task */
	private static Apfloat computeTaskSuc(
			Map<IDeployableComponentAdapter<?>, ReliabilityMeasure> actorReliabilities) {
		// SUC: the task is successful, iff all actors are successful
		Apfloat taskSuc = new Apfloat(1, FP_PRECISION);
		for(IDeployableComponentAdapter<?> deployableComponent : actorReliabilities.keySet()) {
			taskSuc = taskSuc.multiply(actorReliabilities.get(deployableComponent).suc);
		}
		return taskSuc;
	}

	/**
	 * SDC(T)=SDC(t0)(SUC(T\t0)+SDC(T\t0))+SUC(t0)SDC(T\t0)
	 * =SDC(t0)SUC(T\t0) + (SDC(t0)+SUC(t0)) SDC(T\t0)
	 */
	private Apfloat computeTaskSdc(
			Map<IDeployableComponentAdapter<?>, ReliabilityMeasure> actorReliabilities) {

		Apfloat taskSdc;

		// get the first entry (make sure map is not empty)
		IDeployableComponentAdapter<?> deployableComponent =
				(IDeployableComponentAdapter<?>)actorReliabilities.keySet().toArray()[0];
		ReliabilityMeasure rm = actorReliabilities.get(deployableComponent);
		// now we can remove the entry from the map
		actorReliabilities.remove(deployableComponent);

		if(actorReliabilities.isEmpty()) {
			// no more actors
			return rm.sdc;
		}

		// first item
		taskSdc = rm.sdc.multiply(computeTaskSuc(actorReliabilities));
		// second item
		Apfloat t0NonDue = rm.sdc.add(rm.suc);
		taskSdc = taskSdc.add(t0NonDue.multiply(computeTaskSdc(actorReliabilities)));
		return taskSdc;
	}

	/** Converts an analysis result according to the selected {@link FailureType}. */
	private static double convertResult(ReliabilityMeasure measure, FailureType type) {
		if(type == FailureType.DUF) {
			// return measure.duf.doubleValue();
			return measure.duf.floatValue();
		} else if(type == FailureType.SDC) {
			// return measure.sdc.doubleValue();
			return measure.sdc.floatValue();
		} else if(type == FailureType.ALL) {
			// Apfloat one = new Apfloat(1,ReliabilityAnalysis.precision);
			// return one.subtract(measure.suc).doubleValue();
			// return measure.duf.add(measure.sdc).doubleValue();
			return measure.duf.add(measure.sdc).floatValue();
		}

		return 0;
	}

	/** {@inheritDoc} */
	@Override
	public boolean validateInputs() {
		// TODO Auto-generated method stub
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public Class<?> getPrimaryInputType() {
		return StrictTTSchedule.class;
	}

	/** {@inheritDoc} */
	@Override
	public IExplorationTargetEvaluator<StrictTTSchedule<FaultDetectionVotingTaskMappingEntry, FaultDetectionVotingTaskMappingEncoding>, Double, FailureMinObjective>
			createEvaluator(SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter,
					FailureMinObjective explorationTarget) throws InstantiationException {
		try {
			return new FaultDetectionVotingReliabilityEvaluator(systemModelAdapter,
					explorationTarget);
		} catch(Exception e) {
			throw new InstantiationException("Cannot instantiate the Evaluator " +
					getClass().getSimpleName() + ": " + e.getMessage());
		}
	}
}
