/*-------------------------------------------------------------------------+
| Copyright 2015 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.modeltransformation.base;

import static java.util.Collections.sort;
import static org.fortiss.af3.exploration.alg.util.ExplorationAlgUtils.pickInstanceOf;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Comparator;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Queue;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.af3.exploration.alg.dse.modeltransformation.ITransformationModule;
import org.fortiss.af3.exploration.alg.dse.modeltransformation.ITransformationStrategy;
import org.fortiss.af3.exploration.alg.dse.modeltransformation.element.IElementTransformationModule;
import org.fortiss.af3.exploration.alg.exception.TransformationModuleException;
import org.fortiss.af3.exploration.alg.guava.SubClassToInstanceMap;
import org.fortiss.af3.exploration.alg.service.ExplorationEncodingMap;
import org.fortiss.af3.exploration.alg.service.IExplorationEncoding;
import org.fortiss.af3.exploration.alg.service.IExplorationTransformationService.TransformationContext;
import org.fortiss.af3.exploration.alg.service.internal.ExplorationTransformationService;
import org.fortiss.af3.project.model.FileProject;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;
import org.fortiss.tooling.kernel.introspection.IIntrospectionItem;
import org.fortiss.tooling.kernel.model.INamedElement;
import org.fortiss.tooling.kernel.model.IProjectRootElement;
import org.fortiss.tooling.kernel.service.IElementCompositorService;

/**
 * Base class for generating AF3 models programatically.
 * 
 * @author diewald
 */
public abstract class TransformationStrategyBase<T extends EObject> extends
		TransformationModuleBase<T> implements ITransformationStrategy<T> {

	/**
	 * Flag that determines whether {@code this} {@link ITransformationStrategy} shall be launched
	 * in a model context. This is only required for the first {@link ITransformationModule} which
	 * is executed, not for sub modules.
	 */
	protected boolean launchedAsSubModule = false;

	/** The original model from which the {@code transformedModel} is generated. */
	protected T transformedModel;

	/** The name of the top-most model to be generated by this transformation process. */
	protected String transformedModelName = "";

	/** References the original base model (if any). */
	protected T originalModel;

	/** Maintains a reference to the model context in which the transformation is executed. */
	protected ITopLevelElement modelContext;

	/** Caches exceptions that are thrown in the model context. */
	protected TransformationModuleException cachedException;

	/** List of {@link EObject}s that were instantiated in previous generation/transformation steps. */
	protected Collection<EObject> instantiatedElements = new ArrayList<>();

	/**
	 * Map that associates the element types with their corresponding element-wise
	 * {@link ITransformationModule}.
	 */
	protected Map<Class<? extends EObject>, IElementTransformationModule<?, ?>> generationModuleMap =
			new LinkedHashMap<>();

	/**
	 * {@inheritDoc} </p>
	 * This implementation follows the strategy pattern.
	 * </p>
	 */
	@Override
	public <N extends EObject, S extends EObject> T transform(TransformationState state,
			TransformationContext context, SubClassToInstanceMap<S> baseModels,
			ExplorationEncodingMap<IExplorationEncoding> explorationEncodings,
			SubClassToInstanceMap<N> transformedModels, EObject outputModel)
			throws TransformationModuleException {
		if(transformedModels == null) {
			throw new TransformationModuleException(this,
					"The map which will contain all transformed models must be passed.");
		}

		transformedModel =
				transformInternal(state, context, baseModels, explorationEncodings,
						transformedModels);

		if(launchedAsSubModule) {
			doSetModelName();
			return transformedModel;
		}

		if(cachedException != null) {
			TransformationModuleException localException = cachedException;
			localException = cachedException;
			cachedException = null;
			throw localException;
		}

		doSetModelName();

		encEntryModelElemAssoc.clear();
		origToGenAssocMap.clear();
		genToOrigAssocMap.clear();

		return transformedModel;
	}

	/** Sets the name of the transformed model. */
	protected void doSetModelName() {
		if(transformedModel instanceof INamedElement) {
			boolean notificationsEnabled = transformedModel.eDeliver();
			if(notificationsEnabled) {
				transformedModel.eSetDeliver(false);
			}
			((INamedElement)transformedModel).setName(transformedModelName);
			transformedModel.eSetDeliver(true);
		}
	}

	/**
	 * Actual transformation method that launches the registered submodules (i.e., their newly
	 * created instances) in the order that is given by the
	 * {@link org.fortiss.af3.exploration.alg.dse.modeltransformation.ITransformationModule.TransformationState}
	 * s.
	 * 
	 * @param state
	 * 
	 * @param context
	 *            see
	 *            {@link ExplorationTransformationService#getTransformedModel(TransformationContext, ITopLevelElement, ExplorationEncodingMap, SubClassToInstanceMap, Class, String)}
	 *            .
	 * @param baseModels
	 *            see
	 *            {@link ITransformationModule#transform(org.fortiss.af3.exploration.alg.dse.modeltransformation.ITransformationModule.TransformationState, TransformationContext, SubClassToInstanceMap, ExplorationEncodingMap, SubClassToInstanceMap, EObject)}
	 *            .
	 * @param explorationEncodings
	 *            see
	 *            {@link ITransformationModule#transform(org.fortiss.af3.exploration.alg.dse.modeltransformation.ITransformationModule.TransformationState, TransformationContext, SubClassToInstanceMap, ExplorationEncodingMap, SubClassToInstanceMap, EObject)}
	 *            .
	 * @param transformedModels
	 *            see
	 *            {@link ITransformationModule#transform(org.fortiss.af3.exploration.alg.dse.modeltransformation.ITransformationModule.TransformationState, TransformationContext, SubClassToInstanceMap, ExplorationEncodingMap, SubClassToInstanceMap, EObject)}
	 *            .
	 * @return The resulting transformed model.
	 * @throws TransformationModuleException
	 *             if any of the sub modules fails during execution or produces no output.
	 */
	// TODO update doc "state"
	protected <N extends EObject, S extends EObject> T transformInternal(TransformationState state,
			TransformationContext context, SubClassToInstanceMap<S> baseModels,
			ExplorationEncodingMap<IExplorationEncoding> explorationEncodings,
			SubClassToInstanceMap<N> transformedModels) throws TransformationModuleException {
		@SuppressWarnings("unchecked") T castedTransformedBaseModel =
				(T)transformedModels.get(getOutputType());
		this.transformedModel = castedTransformedBaseModel;

		Queue<TransformationState> remainingStates =
				new LinkedList<>(Arrays.asList(TransformationState.values()));

		while(!remainingStates.isEmpty()) {
			TransformationState currentState = remainingStates.poll();

			// Clean the intra-generated element associations if we enter a 'processing' stage.
			if(currentState.isProcessing()) {
				genToGenAssocMap.clear();
			}

			executeSubModules(getSubModules(), context, baseModels, explorationEncodings,
					transformedModels, currentState);
		}

		// Suppressed warnings: Generics are used here to ensure compatibility between the type
		// (class) and the instance.
		@SuppressWarnings("unchecked") Class<? extends N> transformedModelType =
				(Class<N>)getOutputType();
		@SuppressWarnings("unchecked") N outputModel = (N)transformedModel;
		transformedModels.put(transformedModelType, outputModel);
		return this.transformedModel;
	}

	/**
	 * Executes the given set of {@link ITransformationModule}s iff the given
	 * {@link org.fortiss.af3.exploration.alg.dse.modeltransformation.ITransformationModule.TransformationState}
	 * equals the activation state of the sub-module or the sub-module
	 * is an {@link ElementTransformationStrategyBase}.
	 * 
	 * @param subModules
	 *            Collection of {@link ITransformationModule}s to (possibly) execute.
	 * @param context
	 *            Whether or not to export the resulting models into the corresponding project.
	 * @param baseModels
	 *            Input set of original models.
	 * @param explorationEncodings
	 *            DSE internal encoding inputs for the transformation.
	 * @param transformedModels
	 *            Set of models that result from the transformation (output).
	 * @param currentState
	 *            Current state of the parent {@link ITransformationStrategy}.
	 * @throws TransformationModuleException
	 *             If any of the submodules fails during execution or in- or outputs are not as
	 *             expected.
	 */
	protected <N extends EObject, S extends EObject> void executeSubModules(
			Collection<ITransformationModule<? extends EObject>> subModules,
			TransformationContext context, SubClassToInstanceMap<S> baseModels,
			ExplorationEncodingMap<IExplorationEncoding> explorationEncodings,
			SubClassToInstanceMap<N> transformedModels, TransformationState currentState)
			throws TransformationModuleException {

		// TODO: consider using the Dependency graph for a) the order in activation stages (needed)
		// and b) the sorting by modules vs.strategies.
		subModules = new ArrayList<>(subModules);
		sort((List<ITransformationModule<? extends EObject>>)subModules,
				createModuleStrategyOrder());

		for(ITransformationModule<? extends EObject> subModule : subModules) {
			if(subModule.getActivationStage() == currentState ||
					(currentState.ordinal() >= subModule.getActivationStage().ordinal() && subModule instanceof ElementTransformationStrategyBase)) {
				if(subModule.isOptional() &&
						!canExecuteModule(baseModels, transformedModels, explorationEncodings,
								subModule)) {
					continue;
				}

				if(subModule instanceof ITransformationStrategy) {
					((ITransformationStrategy<?>)subModule).setModelContext(modelContext);
					((ITransformationStrategy<?>)subModule).setModelName(transformedModelName);
					((ITransformationStrategy<?>)subModule).setLaunchedAsSubModule(true);
				}
				subModule.loadElementAssociations(encEntryModelElemAssoc, origToGenAssocMap,
						genToOrigAssocMap, genToGenAssocMap);
				@SuppressWarnings("unchecked") T transformedModel =
						(T)subModule.transform(currentState, context, baseModels,
								explorationEncodings, transformedModels, this.transformedModel);
				if(transformedModel == null) {
					throw new TransformationModuleException(subModule,
							"The submodule did not return the transformed model. It is required by the framework.");
				}

				// Impl classes in EMF _always_ have one interface that specifies the methods
				// etc.
				@SuppressWarnings("unchecked") Class<N> transformedModelType =
						(Class<N>)transformedModel.getClass().getInterfaces()[0];
				@SuppressWarnings("unchecked") N tempCastedTransformedModel = (N)transformedModel;
				transformedModels.putInstance(transformedModelType, tempCastedTransformedModel);
				if(getOutputType().isAssignableFrom(transformedModel.getClass())) {
					this.transformedModel = transformedModel;
				}
				applyAssociationUpdates(subModule);
			}
		}
	}

	/**
	 * Determines whether the given {@link ITransformationModule} can be executed with the available
	 * input set.
	 */
	protected <N extends EObject, S extends EObject> boolean canExecuteModule(
			Map<Class<? extends S>, S> baseModels, Map<Class<? extends N>, N> transformedModels,
			ExplorationEncodingMap<IExplorationEncoding> explorationEncodings,
			ITransformationModule<? extends EObject> subModule) {
		for(Class<?> reqInput : subModule.getRequiredTypes()) {
			// This cast is required since the explorationEncodings only accept
			// exploration types -> it is just used for presence checking of a
			// corresponding type. During runtime, the type is erased anyway.
			@SuppressWarnings("unchecked") Class<? extends IExplorationEncoding> explReqInput =
					(Class<? extends IExplorationEncoding>)reqInput;
			if(!(baseModels.keySet().stream().anyMatch(m -> m.isAssignableFrom(reqInput))) &&
					!(transformedModels.keySet().stream().anyMatch(m -> m
							.isAssignableFrom(reqInput))) &&
					!(explorationEncodings.containsEncoding(explReqInput))) {
				return false;
			}
		}
		return true;
	}

	/**
	 * {@inheritDoc}
	 * 
	 * @throws TransformationModuleException
	 */
	@Override
	public void addModelToProject(FileProject fp, ITopLevelElement modelContext)
			throws TransformationModuleException {
		if(modelContext == null) {
			throw new TransformationModuleException(this,
					"Cannot add the transformed model to the AF3 Project without a model context. Please provide one.");
		}
		if(fp == null) {
			throw new TransformationModuleException(
					this,
					"Cannot add the transformed model to the AF3 Project without without a referenced project. Please provide one.");
		}

		for(ITransformationStrategy<?> subModule : pickInstanceOf(ITransformationStrategy.class,
				getSubModules())) {
			subModule.addModelToProject(fp, modelContext);
		}

		if(transformedModel instanceof IProjectRootElement) {
			doAddModelToProject(fp, (IProjectRootElement)transformedModel);
		}
	}

	/** Perform the actual adding operation of the transformed model to the AF3 project. */
	private void doAddModelToProject(FileProject fp, IProjectRootElement transformedModel) {
		Runnable fpAdder = () -> {
			// prepareUniqueID(transformedModel, fp);
				IElementCompositorService.getInstance().compose(fp, transformedModel, null);
				fp.getRootElements().add(transformedModel);
			};
		modelContext.runAsCommand(fpAdder);
	}

	/** {@inheritDoc} */
	@Override
	public void setModelContext(ITopLevelElement modelContext) {
		this.modelContext = modelContext;
	}

	/** {@inheritDoc} */
	@Override
	public void setModelName(String transformedModelName) {
		this.transformedModelName = transformedModelName;
	}

	/** {@inheritDoc} */
	@Override
	public void setLaunchedAsSubModule(boolean isSubModule) {
		this.launchedAsSubModule = isSubModule;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<IIntrospectionItem> getIntrospectionItems() {
		return pickInstanceOf(IIntrospectionItem.class, subModules);
	}

	/**
	 * Comparator that allows sorting a list of {@link ITransformationModule}s such that the
	 * {@link ITransformationStrategy}s are present upfront.
	 */
	private <S extends ITransformationModule<?>> Comparator<S> createModuleStrategyOrder() {
		return new Comparator<S>() {
			@Override
			public int compare(S o1, S o2) {
				if(o1 instanceof ITransformationStrategy<?> &&
						o2 instanceof ITransformationStrategy<?>) {
					return 0;
				} else if(o1 instanceof ITransformationStrategy<?>) {
					return 1;
				} else if(o2 instanceof ITransformationStrategy<?>) {
					return -1;
				}
				return 0;
			}
		};
	}
}
