/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.modeltransformation.comparch;

import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.af3.component.model.Component;
import org.fortiss.af3.component.model.ComponentArchitecture;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.abstractmapping.AbstractTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.abstractmapping.AbstractTaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.modeltransformation.EMFModelConcurrencyHandler;
import org.fortiss.af3.exploration.alg.dse.modeltransformation.ITransformationModule;
import org.fortiss.af3.exploration.alg.dse.modeltransformation.base.TransformationModuleBase;
import org.fortiss.af3.exploration.alg.dse.modeltransformation.element.M2MCopier;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IDeployableComponentAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.mapping.IMappingEntry;
import org.fortiss.af3.exploration.alg.exception.TransformationModuleException;
import org.fortiss.af3.exploration.alg.guava.SubClassToInstanceMap;
import org.fortiss.af3.exploration.alg.service.ExplorationEncodingMap;
import org.fortiss.af3.exploration.alg.service.IExplorationEncoding;
import org.fortiss.af3.exploration.alg.service.IExplorationTransformationService.TransformationContext;
import org.fortiss.af3.exploration.alg.util.ExplorationEcoreUtils;
import org.fortiss.af3.project.model.FileProject;

import com.google.common.collect.HashMultimap;
import com.google.common.collect.Multimap;

/**
 * 
 * @author diewald
 */
public class ComponentArchitectureInitializer
		extends TransformationModuleBase<ComponentArchitecture> {

	/** {@inheritDoc} */
	@Override
	public Class<ComponentArchitecture> getOutputType() {
		return ComponentArchitecture.class;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<Class<?>> getRequiredTypes() {
		Collection<Class<?>> inputTypes = new HashSet<>();
		inputTypes.add(ComponentArchitecture.class);
		inputTypes.add(AbstractTaskMappingEncoding.class);
		return inputTypes;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<Class<?>> getOptionalTypes() {
		return Collections.emptySet();
	}

	/** {@inheritDoc} */
	@Override
	public Collection<Class<? extends ITransformationModule<?>>> getParentModules() {
		Collection<Class<? extends ITransformationModule<?>>> parentModules = new HashSet<>();
		parentModules.add(ComponentArchitectureTransformer.class);
		return parentModules;
	}

	/** {@inheritDoc} */
	@Override
	public boolean isOptional() {
		return false;
	}

	/** {@inheritDoc} */
	@Override
	public TransformationState getActivationStage() {
		return TransformationState.INITIALIZING;
	}

	/**
	 * {@inheritDoc}
	 * 
	 * @throws TransformationModuleException
	 *             see {@link TransformationModuleBase}.
	 */
	@Override
	public <N extends EObject, S extends EObject> ComponentArchitecture transform(
			TransformationState state, TransformationContext context,
			SubClassToInstanceMap<S> baseModels,
			ExplorationEncodingMap<IExplorationEncoding> explorationEncodings,
			SubClassToInstanceMap<N> transformedModels, EObject outputModel)
			throws TransformationModuleException {
		ComponentArchitecture origCA =
				(ComponentArchitecture)baseModels.get(ComponentArchitecture.class);
		if(origCA == null) {
			throw new TransformationModuleException(this,
					"Could not retrieve the original Component Architecture from the set of base models.");
		}

		AbstractTaskMappingEncoding mapping =
				explorationEncodings.getEncoding(AbstractTaskMappingEncoding.class);
		if(mapping == null) {
			throw new TransformationModuleException(this,
					"An IMappingEncoding is required for this module to operate properly.");
		}

		outputModel = setupTransformedComponentArchitecture(origCA, "");

		updateEncodingAssociations(mapping);

		// Temporally disable notifications to avoid useless notifications when creating the
		// component replica and the new IDs.
		EMFModelConcurrencyHandler.getInstance().tryDisableNotificationsSyncOf(outputModel);

		return (ComponentArchitecture)outputModel;
	}

	/**
	 * Replicates the {@link ComponentArchitecture} that was used as an input for the DSE. The
	 * replicated {@link ComponentArchitecture} is added to the {@link FileProject}. The intention
	 * of this method is to provide a {@link ComponentArchitecture} that contains the spatial
	 * replica of {@link Component}s that shall be mapped to different ECUs}
	 */
	private ComponentArchitecture setupTransformedComponentArchitecture(
			ComponentArchitecture af3ComponentArchitecture, String addName) {
		M2MCopier copier = new M2MCopier();
		ComponentArchitecture transformedComponentArchitecture =
				ExplorationEcoreUtils.copy(af3ComponentArchitecture, copier);
		copier.entrySet().stream().forEach(e -> genToOrigAssocMap.put(e.getValue(), e.getKey()));
		copier.entrySet().stream().forEach(e -> origToGenAssocMap.put(e.getKey(), e.getValue()));

		// Set the Architecture name (via the top Component)
		Component topComponent = transformedComponentArchitecture.getTopComponent();
		String caName = topComponent.getName() + "_generated_" + addName;
		topComponent.setName(caName);

		return transformedComponentArchitecture;
	}

	/**
	 * Creates associations of {@link IMappingEntry}s that DSE-internally describe task to execution
	 * unit mappings to generated {@link Component}s that represent the allocated tasks.
	 * This information is used in subsequent steps in the transformation. Hereby, the
	 * transformation process is capable of establishing a unique link between entries and model
	 * elements.
	 */
	private void updateEncodingAssociations(AbstractTaskMappingEncoding mapping) {
		Multimap<Component, Component> processedElements = HashMultimap.create();
		for(AbstractTaskMappingEntry entry : mapping.getMappingEntries()) {
			IDeployableComponentAdapter<?> compAdp = entry.getSource();
			Component origAF3Component = (Component)compAdp.getObject();
			Component genAF3Component = null;

			for(EObject generatedComponent : origToGenAssocMap.get(origAF3Component)) {
				if(!processedElements.get(origAF3Component).contains(generatedComponent)) {
					genAF3Component = (Component)generatedComponent;
					break;
				}
			}

			encEntryModelElemAssoc.put(entry, Component.class, genAF3Component,
					AbstractTaskMappingEntry.class);
		}
	}
}
