/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.modeltransformation.comparch;

import static org.eclipse.emf.ecore.util.EcoreUtil.delete;
import static org.fortiss.af3.exploration.alg.util.AF3Utils.getConnectedChannelsOf;
import static org.fortiss.af3.exploration.alg.util.AF3Utils.getConnectedPortsOf;
import static org.fortiss.af3.exploration.alg.util.AF3Utils.getInputOutputPortsOf;
import static org.fortiss.af3.exploration.alg.util.ExplorationAlgUtils.pickFirstInstanceOf;
import static org.fortiss.af3.exploration.alg.util.TransformationUtils.connectPort;
import static org.fortiss.af3.exploration.alg.util.TransformationUtils.setGenComponentPos;
import static org.fortiss.af3.exploration.alg.util.TransformationUtils.setReplicaAwareInstElementName;
import static org.fortiss.tooling.common.util.LambdaUtils.filterSet;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getChildrenWithType;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.af3.component.model.Component;
import org.fortiss.af3.component.model.ComponentArchitecture;
import org.fortiss.af3.component.model.Port;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.abstractmapping.AbstractTaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.InstantiatedTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.InstantiatedTaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.modeltransformation.ITransformationModule;
import org.fortiss.af3.exploration.alg.dse.modeltransformation.base.TransformationModuleBase;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IDeployableComponentAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.mapping.IMappingEntry;
import org.fortiss.af3.exploration.alg.exception.TransformationModuleException;
import org.fortiss.af3.exploration.alg.guava.SubClassToInstanceMap;
import org.fortiss.af3.exploration.alg.service.ExplorationEncodingMap;
import org.fortiss.af3.exploration.alg.service.IExplorationEncoding;
import org.fortiss.af3.exploration.alg.service.IExplorationTransformationService.TransformationContext;
import org.fortiss.af3.platform.model.ExecutionUnit;
import org.fortiss.tooling.base.model.element.IConnector;
import org.fortiss.tooling.kernel.utils.EcoreUtils;

/**
 * Instantiates abstract {@link Component}s in the generated {@link ComponentArchitecture} based on
 * the instantiation information contained in the input {@link InstantiatedTaskMappingEncoding}.
 * 
 * @author diewald
 */
public class ComponentArchitectureInstantiator
		extends TransformationModuleBase<ComponentArchitecture> {

	/** References the {@link Component} to {@link ExecutionUnit} mapping calculated by the DSE. */
	InstantiatedTaskMappingEncoding<?> mapping;

	/** {@inheritDoc} */
	@Override
	public Class<ComponentArchitecture> getOutputType() {
		return ComponentArchitecture.class;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<Class<?>> getRequiredTypes() {
		Collection<Class<?>> inputTypes = new HashSet<>();
		inputTypes.add(InstantiatedTaskMappingEncoding.class);
		return inputTypes;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<Class<?>> getOptionalTypes() {
		return Collections.emptySet();
	}

	/** {@inheritDoc} */
	@Override
	public Collection<Class<? extends ITransformationModule<?>>> getParentModules() {
		Collection<Class<? extends ITransformationModule<?>>> parentModules = new HashSet<>();
		parentModules.add(ComponentArchitectureTransformer.class);
		return parentModules;
	}

	/** {@inheritDoc} */
	@Override
	public boolean isOptional() {
		return false;
	}

	/** {@inheritDoc} */
	@Override
	public TransformationState getActivationStage() {
		return TransformationState.INSTANCIATING;
	}

	/** {@inheritDoc} */
	@Override
	public <N extends EObject, S extends EObject> ComponentArchitecture transform(
			TransformationState state, TransformationContext context,
			SubClassToInstanceMap<S> baseModels,
			ExplorationEncodingMap<IExplorationEncoding> explorationEncodings,
			SubClassToInstanceMap<N> transformedModels, EObject outputModel)
			throws TransformationModuleException {

		mapping = explorationEncodings.getEncoding(InstantiatedTaskMappingEncoding.class);
		if(mapping == null) {
			throw new TransformationModuleException(this,
					"An IMappingEncoding is required for this module to operate properly.");
		}

		instantiateAbstractAF3Components();

		return (ComponentArchitecture)outputModel;
	}

	/**
	 * Instantiates the referenced AF3 {@link Component} of each {@link IDeployableComponentAdapter}
	 * that demands its instantiation. Therefore, the appropriate {@link Component} from the
	 * Component Pool is copied to the generated {@link ComponentArchitecture}. As a side effect,
	 * this method updates the {@code genToOrigAssocMap} such that instantiated Components can be
	 * tracked.
	 * 
	 * @throws TransformationModuleException
	 *             if the set of deployable {@link Component} in the input
	 *             {@link ComponentArchitecture} cannot be determined.
	 */
	private void instantiateAbstractAF3Components() throws TransformationModuleException {
		Collection<Component> instantiatedComponents = new HashSet<>();
		Collection<EObject> preCleanupElementsCollection = new HashSet<>();

		for(IDeployableComponentAdapter<?> comp : mapping.getRequesters()) {
			if(comp.isInstantiation()) {
				Component origComp = (Component)comp.getReplacedComponent().getObject();
				preCleanupElementsCollection.addAll(getChildrenWithType(origComp, EObject.class));
			}
		}

		preCleanupElementsCollection.forEach(e -> origToGenAssocMap.removeAll(e));

		Collection<Component> abstrCompRemovalList = new ArrayList<>();

		for(InstantiatedTaskMappingEntry entry : mapping.getMappingEntries()) {
			@SuppressWarnings("unchecked") IDeployableComponentAdapter<Component> compAdp =
					(IDeployableComponentAdapter<Component>)entry.getSource();
			if(compAdp.isInstantiation()) {
				// Randomly take one replicated generated Component that belongs to the
				// corresponding adapter.
				// Filter out already instantiated components.
				// Component abstrGenComponent =
				// (Component)origToGenAssocMap
				// .get(compAdp.getReplacedComponent().getObject()).stream()
				// .filter(c -> !instantiatedComponents.contains(c)).findAny().get();
				AbstractTaskMappingEntry abstrEntry =
						entry.getAssociatedElement(AbstractTaskMappingEntry.class);
				Component abstrGenComponent =
						(Component)encEntryModelElemAssoc.get(abstrEntry, Component.class);
				Component poolComponent = compAdp.getObject();
				Component instantiatedComponent =
						instantiateComponent(abstrGenComponent, poolComponent);

				// Remove the associated (possibly replicated) "abstract" component.
				abstrCompRemovalList.add(abstrGenComponent);
				// removeAbstractComponent(abstrGenComponent);
				Component origComponent = compAdp.getReplacedComponent().getObject();

				setReplicaAwareInstElementName(origComponent, abstrGenComponent,
						instantiatedComponent);
				updateAssocMaps(entry, poolComponent, origComponent, abstrGenComponent,
						instantiatedComponent);
				instantiatedComponents.add(instantiatedComponent);
			} else {
				IMappingEntry abstrEntr = pickFirstInstanceOf(AbstractTaskMappingEntry.class,
						entry.getAssociatedElement());
				Component genComp =
						(Component)encEntryModelElemAssoc.get(abstrEntr, Component.class);
				genToGenAssocMap.put(genComp, genComp);
				encEntryModelElemAssoc.put(entry, Component.class, genComp,
						InstantiatedTaskMappingEntry.class);
			}
		}

		abstrCompRemovalList.forEach(c -> removeAbstractComponent(c));
	}

	/**
	 * Replaces the given {#link abstractComponent} with the given {#link replacementComponent}.
	 * This method is intended to replace "abstract" {@link Component} with their instances from the
	 * "Component Pool".
	 */
	private Component instantiateComponent(Component abstrComponent,
			Component replacementComponent) {
		Component instantiatedComponent = EcoreUtils.copy(replacementComponent);
		instantiatedComponent.setContainer(abstrComponent.getContainer());

		Collection<Port> instantiatedConnectors = getInputOutputPortsOf(instantiatedComponent);
		for(Port movingPort : getInputOutputPortsOf(abstrComponent)) {
			// Rewire Channels.
			// TODO: Possible NPE if optional is empty. Throw an exception about non-found port name
			// here.
			Collection<Port> matchingPorts = filterSet(instantiatedConnectors,
					p -> movingPort.getName().equals(p.getName()));
			Port matchInstantatedConnector = matchingPorts.stream().findAny().get();

			genToOrigAssocMap.put(matchInstantatedConnector, movingPort);
			connectPort(matchInstantatedConnector, getConnectedPortsOf(movingPort),
					instantiatedComponent);
		}

		// Add position / layout information
		setGenComponentPos(abstrComponent, instantiatedComponent);

		return instantiatedComponent;
	}

	/**
	 * Removes the "abstract" {@link Component}s from its parent {@link ComponentArchitecture}, i.e.
	 * it removes those {@link Component}s which point to an "implementation" in the
	 * "Component Pool".
	 */
	// TODO: The parameter should actually be a collection of components.
	private void removeAbstractComponent(Component abstrComponent) {
		getConnectedChannelsOf(abstrComponent).stream().forEach(ch -> delete(ch));
		delete(abstrComponent);
	}

	/** Updates the association maps with the modifications introduced by the instantiation. */
	// TODO: generalize this update method to also consider children.
	private void updateAssocMaps(IMappingEntry entry, Component poolComponent,
			Component origComponent, Component abstrGenComponent, Component instantiatedComponent) {
		origToGenAssocMap.remove(origComponent, abstrGenComponent);
		origToGenAssocMap.put(origComponent, instantiatedComponent);
		genToOrigAssocMap.remove(poolComponent);
		genToOrigAssocMap.put(instantiatedComponent, origComponent);
		genToGenAssocMap.put(abstrGenComponent, instantiatedComponent);
		for(IConnector abstrPort : abstrGenComponent.getConnectors()) {
			origToGenAssocMap.removeAll(abstrPort);
			for(IConnector instPort : instantiatedComponent.getConnectors()) {
				// Names MUST be identical (convention).
				// TODO: Identify a method to identify equal Ports without querying the name.
				if(((Port)instPort).getName().equals(((Port)abstrPort).getName())) {
					genToGenAssocMap.put(abstrPort, instPort);
					origToGenAssocMap.put(abstrPort, instPort);
					genToOrigAssocMap.put(instPort, abstrPort);
				}
			}
		}

		// Further we need to update the associations from the ports of the original Component.
		for(IConnector abstrPort : origComponent.getConnectors()) {
			for(IConnector instPort : instantiatedComponent.getConnectors()) {
				// Names MUST be identical (convention).
				// TODO: Identify a method to identify equal Ports without querying the name.
				if(((Port)instPort).getName().equals(((Port)abstrPort).getName())) {
					genToGenAssocMap.put(abstrPort, instPort);
					origToGenAssocMap.put(abstrPort, instPort);
					genToOrigAssocMap.put(instPort, abstrPort);
				}
			}
		}

		encEntryModelElemAssoc.put(entry, Component.class, instantiatedComponent,
				InstantiatedTaskMappingEntry.class);
		// entry.getAssociatedElement().forEach(
		// e -> encEntryModelElemAssoc.put(e, Component.class, instantiatedComponent,
		// e.getClass()));
	}
}
