/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.modeltransformation.comparch;

import static org.fortiss.af3.component.utils.ComponentModelElementFactory.createChannelAndAttach;
import static org.fortiss.af3.exploration.alg.util.AF3Utils.getInputOutputPortsOf;
import static org.fortiss.af3.exploration.alg.util.AF3Utils.isAnyParentComponentDeployable;
import static org.fortiss.af3.exploration.alg.util.ExplorationAlgUtils.getSourcePorts;
import static org.fortiss.af3.exploration.alg.util.TransformationUtils.arrangePorts;
import static org.fortiss.af3.exploration.alg.util.TransformationUtils.connectPort;
import static org.fortiss.af3.exploration.alg.util.TransformationUtils.getDeployableComponentMap;
import static org.fortiss.af3.exploration.alg.util.TransformationUtils.retrieveDeployableComponents;
import static org.fortiss.af3.exploration.alg.util.TransformationUtils.setGenComponentPos;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getChildrenWithType;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.pickInstanceOf;
import static org.fortiss.tooling.kernel.utils.KernelModelElementUtils.findElementById;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;

import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.fortiss.af3.component.model.Channel;
import org.fortiss.af3.component.model.Component;
import org.fortiss.af3.component.model.ComponentArchitecture;
import org.fortiss.af3.component.model.InputPort;
import org.fortiss.af3.component.model.OutputPort;
import org.fortiss.af3.component.model.Port;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.comm.constraint.InternalIsolatedCommunicationSet;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.partitionmapping.Partition;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.TaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.abstractmapping.AbstractTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.abstractmapping.AbstractTaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.modeltransformation.ITransformationModule;
import org.fortiss.af3.exploration.alg.dse.modeltransformation.base.TransformationModuleBase;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IDeployableComponentAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.mapping.IMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.sysmodel.mapping.IMappingEntry;
import org.fortiss.af3.exploration.alg.exception.InvalidTimingModelException;
import org.fortiss.af3.exploration.alg.exception.TransformationModuleException;
import org.fortiss.af3.exploration.alg.guava.SubClassToInstanceMap;
import org.fortiss.af3.exploration.alg.service.ExplorationEncodingMap;
import org.fortiss.af3.exploration.alg.service.IExplorationEncoding;
import org.fortiss.af3.exploration.alg.service.IExplorationTransformationService.TransformationContext;
import org.fortiss.af3.platform.model.ExecutionUnit;
import org.fortiss.af3.platform.model.IPlatformResource;
import org.fortiss.tooling.kernel.model.INamedElement;
import org.fortiss.tooling.kernel.utils.EcoreUtils;

import com.google.common.collect.HashMultimap;
import com.google.common.collect.Multimap;

/**
 * Replicates {@link Component}s in the generated {@link ComponentArchitecture} if
 * {@link IMappingEntry}s of the input {@link AbstractTaskMappingEncoding} point to the same
 * {@link Component}.
 * 
 * @author diewald
 */
public class ComponentArchitectureReplicator
		extends TransformationModuleBase<ComponentArchitecture> {

	/** Collection of {@link Component}s that can be allocated to {@link IPlatformResource}. */
	private Collection<IDeployableComponentAdapter<Component>> deployableComponentCollection =
			new HashSet<>();

	/**
	 * Contains the allocation information between possibly replicated {@link Component}s and their
	 * corresponding {@link IPlatformResource}s.
	 */
	Map<Component, IPlatformResource> deploymentComponents = new HashMap<>();

	/** References the original {@link ComponentArchitecture} that was input to the DSE. */
	ComponentArchitecture originalCA;

	/** {@inheritDoc} */
	@Override
	public Class<ComponentArchitecture> getOutputType() {
		return ComponentArchitecture.class;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<Class<?>> getRequiredTypes() {
		Collection<Class<?>> inputTypes = new HashSet<>();
		inputTypes.add(ComponentArchitecture.class);
		inputTypes.add(AbstractTaskMappingEncoding.class);
		return inputTypes;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<Class<?>> getOptionalTypes() {
		return Collections.emptySet();
	}

	/** {@inheritDoc} */
	@Override
	public Collection<Class<? extends ITransformationModule<?>>> getParentModules() {
		Collection<Class<? extends ITransformationModule<?>>> parentModules = new HashSet<>();
		parentModules.add(ComponentArchitectureTransformer.class);
		return parentModules;
	}

	/** {@inheritDoc} */
	@Override
	public boolean isOptional() {
		return false;
	}

	/** {@inheritDoc} */
	@Override
	public TransformationState getActivationStage() {
		return TransformationState.REPLICATING;
	}

	/** {@inheritDoc} */
	@Override
	public <N extends EObject, S extends EObject> ComponentArchitecture transform(
			TransformationState state, TransformationContext context,
			SubClassToInstanceMap<S> baseModels,
			ExplorationEncodingMap<IExplorationEncoding> explorationEncodings,
			SubClassToInstanceMap<N> transformedModels, EObject outputModel)
			throws TransformationModuleException {
		ComponentArchitecture genCA = (ComponentArchitecture)outputModel;
		if(genCA == null) {
			throw new TransformationModuleException(this,
					"Could not retrieve the generated Component Architecture.");
		}
		originalCA = (ComponentArchitecture)baseModels.get(ComponentArchitecture.class);
		if(originalCA == null) {
			throw new TransformationModuleException(this,
					"Could not retrieve the original Component Architecture from the set of base models.");
		}

		AbstractTaskMappingEncoding mapping =
				explorationEncodings.getEncoding(AbstractTaskMappingEncoding.class);
		if(mapping == null) {
			throw new TransformationModuleException(this,
					"An IMappingEncoding is required for this module to operate properly.");
		}
		deployableComponentCollection = retrieveDeployableComponents(this, mapping);
		Map<IDeployableComponentAdapter<Component>, Component> genCompAssoc =
				getDeployableComponentMap(this, origToGenAssocMap, deployableComponentCollection);
		EList<Port> replicatedPorts = new BasicEList<Port>();

		for(Entry<IDeployableComponentAdapter<Component>, Component> deployedComponentPair : genCompAssoc
				.entrySet()) {
			Component currentDeployableAf3Component = deployedComponentPair.getValue();
			// TODO: This part needs to be updated to match the new Allocation Models.
			Collection<ExecutionUnit> allocatedExecutionUnits =
					getExecutionUnitAllocationsFromMapping(
							mapping.getAllocatedResourcesFor(deployedComponentPair.getKey()));
			Collection<AbstractTaskMappingEntry> currentMappingEntries =
					mapping.getMappingEntriesOf(deployedComponentPair.getKey());
			List<Component> generatedAndReplicatedComponents = new ArrayList<Component>();
			generatedAndReplicatedComponents.add(currentDeployableAf3Component);

			// TODO: Refactor the following code, the re-use of the Iterator is anything but
			// beautiful code.
			// Create an iterator which is used to handle the allocations of the replicated
			// components.
			assert (!allocatedExecutionUnits
					.isEmpty()) : "Each Component must be allocated to at least one ExecutionUnit.";
			Iterator<AbstractTaskMappingEntry> mappingEntryIt = currentMappingEntries.iterator();
			AbstractTaskMappingEntry mappingEntry = mappingEntryIt.next();
			deploymentComponents.put(currentDeployableAf3Component,
					(ExecutionUnit)mappingEntry.getTarget().getObject());
			encEntryModelElemAssoc.put(mappingEntry, Component.class, currentDeployableAf3Component,
					AbstractTaskMappingEntry.class);
			genToGenAssocMap.put(currentDeployableAf3Component, currentDeployableAf3Component);

			if(allocatedExecutionUnits.size() > 1) {
				// Collect the InputPorts and the sources of their connectors to restore the
				// original connections for the replicated components (incoming part).
				Map<InputPort, Collection<Port>> originalComponentInputSources =
						getComponentPortToConnectedPortMap(currentDeployableAf3Component,
								InputPort.class);
				// Collect the OutputPorts and the targets of their connectors to restore the
				// original connections for the replicated components (outgoing part).
				Map<OutputPort, Collection<Port>> originalComponentOutputTargets =
						getComponentPortToConnectedPortMap(currentDeployableAf3Component,
								OutputPort.class);
				Multimap<Port, Port> assocGenWithReplPort = HashMultimap.create();

				// Create replica of the current component based on the number of target resources.
				for(int i = 1; i < allocatedExecutionUnits.size(); i++) {
					Component replicatedComponent =
							replicateComponent(replicatedPorts, currentDeployableAf3Component,
									generatedAndReplicatedComponents, originalComponentInputSources,
									originalComponentOutputTargets, assocGenWithReplPort, i);
					allocateReplicatedComponent(deploymentComponents, mappingEntryIt,
							replicatedComponent);
					genToGenAssocMap.put(currentDeployableAf3Component, replicatedComponent);
				}
			}
			// Remember the association between the replicated component and the "original".
			// TODO: Replace this method with the new assoc map (see todo above).
			createComponentAssocWithReplicaComponents(currentDeployableAf3Component,
					generatedAndReplicatedComponents);
		}

		// After the replicated Components have been added, the wiring of the of the replicated
		// ports and their connected channels must be adjusted through the complete architecture.
		try {
			createReplicaAwareChannelWiring(genCA, replicatedPorts, deploymentComponents.keySet());
		} catch(InvalidTimingModelException e) {
			throw new TransformationModuleException(this,
					"Could not complete the replication transformation step due to an (probably) invalid timing model.",
					e);
		}

		for(IMappingEntry entry : mapping.getMappingEntries()) {
			if(!isPartOfIsolationCommConstraint(
					mapping.getConstraintsOf(InternalIsolatedCommunicationSet.class), entry)) {
				continue;
			}
			Component genComponent = (Component)encEntryModelElemAssoc.get(entry, Component.class);
			Collection<InternalIsolatedCommunicationSet> others =
					getOtherIsolationCommConstraints(mapping, entry);
			for(InternalIsolatedCommunicationSet otherSet : others) {
				for(IMappingEntry conEntry : otherSet.getIsolatedCommunicationSet()) {
					Component otherGenComp =
							(Component)encEntryModelElemAssoc.get(conEntry, Component.class);
					removeAllConnectionsBetween(genComponent, otherGenComp);
				}
			}
		}

		return genCA;
	}

	/** Removes all connections ({@link Channel}s) between the given {@link Component}s. */
	private void removeAllConnectionsBetween(Component firstComp, Component secondComp) {
		Collection<Channel> firstCompChannels = new HashSet<>();
		firstComp.getInputPorts().forEach(p -> firstCompChannels.addAll(p.getIncomingChannels()));
		firstComp.getOutputPorts().forEach(p -> firstCompChannels.addAll(p.getOutgoingChannels()));
		Collection<Channel> rmChannels = new ArrayList<>();
		for(Channel firstCompChannel : firstCompChannels) {
			if(firstCompChannel.getSource().getComponent() == secondComp ||
					firstCompChannel.getTarget().getComponent() == secondComp) {
				rmChannels.add(firstCompChannel);
			}
		}
		rmChannels.forEach(ch -> EcoreUtil.delete(ch));
	}

	/**
	 * Returns a collection of all {@link InternalIsolatedCommunicationSet} constraints that do not
	 * include the given {@link IMappingEntry}.
	 */
	private static Collection<InternalIsolatedCommunicationSet> getOtherIsolationCommConstraints(
			TaskMappingEncoding<?, ?> encoding, IMappingEntry entry) {
		Collection<InternalIsolatedCommunicationSet> otherConstraints = new HashSet<>();
		Collection<InternalIsolatedCommunicationSet> isoComm =
				encoding.getConstraintsOf(InternalIsolatedCommunicationSet.class);
		for(InternalIsolatedCommunicationSet constr : isoComm) {
			if(!constr.getIsolatedCommunicationSet().contains(entry)) {
				otherConstraints.add(constr);
			}
		}

		return otherConstraints;
	}

	/**
	 * Returns whether the given {@link IMappingEntry} is included in any of the given
	 * {@link InternalIsolatedCommunicationSet}s.
	 */
	private static boolean isPartOfIsolationCommConstraint(
			Collection<InternalIsolatedCommunicationSet> isoComm, IMappingEntry entry) {
		for(InternalIsolatedCommunicationSet constr : isoComm) {
			if(constr.contains(entry)) {
				return true;
			}
		}

		return false;
	}

	/**
	 * Transforms a {@link Collection} of {@link IResourceAdapter} that is the result of a mapping
	 * of a single {@link Component} into a {@link Collection} of {@link ExecutionUnit}s.
	 */
	private static Collection<ExecutionUnit> getExecutionUnitAllocationsFromMapping(
			Collection<Partition> mappingEntriesOfRequester) {
		Collection<ExecutionUnit> allocatedExecutionUnits = new ArrayList<ExecutionUnit>();
		for(IResourceAdapter<?> currentAllocationTarget : mappingEntriesOfRequester) {
			assert (currentAllocationTarget
					.getObject() instanceof ExecutionUnit) : "There are currently no other deployment targets than ExecutionUnits allowed.";
			allocatedExecutionUnits.add((ExecutionUnit)currentAllocationTarget.getObject());
		}
		return allocatedExecutionUnits;
	}

	/**
	 * Determines the {@link Port}s to which a {@link Component} is connected to. The result is a
	 * {@link Map} which contains either the input or output ports of the given {@link Component}
	 * and their connected {@link Port}s.
	 */
	@SuppressWarnings("unchecked")
	private static <T extends Port> Map<T, Collection<Port>> getComponentPortToConnectedPortMap(
			Component deployableAF3Component, Class<? extends Port> componentPortType) {
		EList<T> originalComponentInputPorts = new BasicEList<T>();
		if(InputPort.class.isAssignableFrom(componentPortType)) {
			for(InputPort inputPort : deployableAF3Component.getInputPorts()) {
				originalComponentInputPorts.add((T)inputPort);
			}
		} else {
			for(OutputPort outputPort : deployableAF3Component.getOutputPorts()) {
				originalComponentInputPorts.add((T)outputPort);
			}
		}

		Map<T, Collection<Port>> originalComponentInputSources = new HashMap<T, Collection<Port>>();
		for(T currentOriginalPort : originalComponentInputPorts) {
			List<Channel> connectedChannels = null;
			if(InputPort.class.isAssignableFrom(componentPortType)) {
				connectedChannels = currentOriginalPort.getIncomingChannels();
			} else {
				connectedChannels = currentOriginalPort.getOutgoingChannels();
			}
			Collection<Port> connectedPorts = new ArrayList<Port>();
			for(Channel currentChannel : connectedChannels) {
				if(InputPort.class.isAssignableFrom(componentPortType)) {
					connectedPorts.add(currentChannel.getSource());
				} else {
					connectedPorts.add(currentChannel.getTarget());
				}
			}
			originalComponentInputSources.put(currentOriginalPort, connectedPorts);
		}
		return originalComponentInputSources;
	}

	/**
	 * Replicates a single {@link Component} from the input {@link ComponentArchitecture} of the
	 * DSE.
	 */
	// TODO: cleanup this method: It has a too large signature.
	private Component replicateComponent(EList<Port> replicatedPorts,
			Component currentDeployableAf3Component, List<Component> origAndReplicatedComponents,
			Map<InputPort, Collection<Port>> originalComponentInputSources,
			Map<OutputPort, Collection<Port>> originalComponentOutputTargets,
			Multimap<Port, Port> assocGenWithReplPort, int i) {
		Component replicatedComponent = EcoreUtils.copy(currentDeployableAf3Component);

		// TODO: Cleanup, use the new assoc map, the old implementation is buggy.
		List<Component> replSubComponents =
				getChildrenWithType(replicatedComponent, Component.class);
		origAndReplicatedComponents.add(replicatedComponent);
		origAndReplicatedComponents.addAll(replSubComponents);

		replicatedComponent.setContainer(currentDeployableAf3Component.getParentComponent());
		replicatedComponent
				.setName(currentDeployableAf3Component.getName() + "_" + Integer.toString(i));

		setGenComponentPos(currentDeployableAf3Component, replicatedComponent);

		// Connect inputs of the replicated Component.
		connectReplicatedPorts(originalComponentInputSources, replicatedComponent);
		// Connect outputs of the replicated Component.
		connectReplicatedPorts(originalComponentOutputTargets, replicatedComponent);

		Component parent = currentDeployableAf3Component.getParentComponent();
		parent.getContainedElements().add(replicatedComponent);
		replicatedPorts.addAll(getInputOutputPortsOf(replicatedComponent));

		Collection<Port> originalPorts =
				getChildrenWithType(currentDeployableAf3Component, Port.class);
		for(Port currentOrigPort : originalPorts) {
			Port replPort = (Port)findElementById(currentOrigPort.getId(), replicatedComponent);
			assocGenWithReplPort.put(currentOrigPort, replPort);
		}

		// Associate the original Component with its replica.
		// Original (possibly abstract) Component.
		// TODO: Generalize: It should be possible to create the association based on a
		// copy from the pool, not the generated architecture -> Decoupling.
		EObject origComponent = genToOrigAssocMap.get(currentDeployableAf3Component);
		Map<INamedElement, INamedElement> assocMap = updateGenToOrigElementMapById(origComponent,
				currentDeployableAf3Component, replicatedComponent, INamedElement.class);
		genToOrigAssocMap.putAll(assocMap);
		assocMap.entrySet().stream().forEach(e -> origToGenAssocMap.put(e.getValue(), e.getKey()));

		for(Port curPort : pickInstanceOf(Port.class, replicatedComponent.eContents())) {
			Port baseGenPort =
					(Port)findElementById(curPort.getId(), currentDeployableAf3Component);
			genToGenAssocMap.put(baseGenPort, curPort);
		}

		genToOrigAssocMap.put(replicatedComponent, replicatedComponent);

		return replicatedComponent;
	}

	/**
	 * Replicates the {@link Port}s given in the list {#link portReplicationCollection}.
	 * Furthermore, channels between the sources/targets of the original {@link Port}s to the
	 * replicated {@link Port}s are created. Thus, the original data flow is maintained.
	 */
	// TODO: rework documentation.
	private static <T extends Port, S extends Port> void connectReplicatedPorts(
			Map<T, Collection<S>> connectedPortMap, Component replicatedComponent) {
		Collection<T> portReplicationCollection = connectedPortMap.keySet();
		for(T currentPort : portReplicationCollection) {
			@SuppressWarnings("unchecked") T replicatedPort =
					(T)findElementById(currentPort.getId(), replicatedComponent);
			if(replicatedPort != null) {
				connectPort(replicatedPort, connectedPortMap.get(currentPort), replicatedComponent);
			}
		}
	}

	/**
	 * Initializes the {@link Multimap} that associates {@link Component}s in the original
	 * {@link ComponentArchitecture} with their respective {@link Component}s in the generated
	 * {@link ComponentArchitecture}.
	 */
	// TODO: rework doc.
	// TODO: avoid suppresswarning, use common parent class for archs as well.
	@SuppressWarnings("unchecked")
	private <T extends INamedElement> Map<T, T> updateGenToOrigElementMapById(
			EObject originalElement, EObject replElementTemplate, EObject generatedElement,
			Class<T> elementType) {
		Map<T, T> assocOrigToGenElements = new HashMap<>();
		List<T> origElementList = getChildrenWithType(originalElement, elementType);
		origElementList.add((T)originalElement);
		List<T> genElementList = getChildrenWithType(generatedElement, elementType);
		genElementList.add((T)generatedElement);

		for(T origElement : origElementList) {
			Optional<T> equivElem = genElementList.stream()
					.filter(c -> (c.getId() == origElement.getId())).findAny();
			if(equivElem.isPresent()) {
				genToOrigAssocMap.put(equivElem.get(), origElement);
				origToGenAssocMap.put(origElement, equivElem.get());
				genToGenAssocMap.put(replElementTemplate, equivElem.get());
			}
		}

		return assocOrigToGenElements;
	}

	/**
	 * Allocates the given replicated {@link Component} to the corresponding
	 * {@link IPlatformResource} indicated by the given {@link IMappingEncoding}. The result is
	 * stored in the given {@code deployableComponent} Map.
	 */
	private void allocateReplicatedComponent(Map<Component, IPlatformResource> deploymentComponents,
			Iterator<AbstractTaskMappingEntry> mappingEntryIt, Component replicatedComponent) {
		// Allocate the replica of the original component to the ExecutionUnits.
		AbstractTaskMappingEntry mappingEntry = mappingEntryIt.next();
		deploymentComponents.put(replicatedComponent,
				(ExecutionUnit)mappingEntry.getTarget().getObject());
		encEntryModelElemAssoc.put(mappingEntry, Component.class, replicatedComponent,
				AbstractTaskMappingEntry.class);
	}

	/**
	 * Modifies a given map which stores the associations of components with their equivalents in a
	 * {@link ComponentArchitecture} that considers replica. The associations refer each equivalent
	 * {@link Component} of an "original" {@link Component}, be it a replica of the respective
	 * original {@link Component}, or not.
	 */
	private void createComponentAssocWithReplicaComponents(Component currentGeneratedAf3Component,
			List<Component> generatedAndReplicatedComponents) {
		// Component originalComponent =
		// (Component)findElementById(currentDeployableAf3Component.getId(), originalCA);
		Component originalComponent =
				(Component)genToOrigAssocMap.get(currentGeneratedAf3Component);
		origToGenAssocMap.putAll(originalComponent, generatedAndReplicatedComponents);
		genToGenAssocMap.putAll(currentGeneratedAf3Component, generatedAndReplicatedComponents);
		// Additionally, consider contained components.
		for(Component containedComponent : getChildrenWithType(currentGeneratedAf3Component,
				Component.class)) {
			originalComponent = (Component)findElementById(containedComponent.getId(), originalCA);
			origToGenAssocMap.putAll(originalComponent, generatedAndReplicatedComponents);
			genToGenAssocMap.putAll(currentGeneratedAf3Component, generatedAndReplicatedComponents);
		}

		for(Port containedPort : getChildrenWithType(currentGeneratedAf3Component, Port.class)) {
			Port origPort = (Port)findElementById(containedPort.getId(), originalComponent);
			genToGenAssocMap.put(containedPort, origPort);
			origToGenAssocMap.put(origPort, containedPort);
		}
	}

	/**
	 * Adjusts the wiring of a {@link ComponentArchitecture} that represents a mapping with
	 * {@link Component} replica.
	 * First, this method identifies all source {@link Port}s within the replica-aware
	 * {@link ComponentArchitecture} by using information from the original
	 * {@link ComponentArchitecture}. Then, the method traverses each {@link Channel} and replicates
	 * the port and successive {@link Channel}s, if the target {@link Port} belongs to a hull
	 * {@link Component}.
	 * 
	 * @throws InvalidTimingModelException
	 *             if the timing specification of any traversed element is inconsistent.
	 */
	private void createReplicaAwareChannelWiring(
			ComponentArchitecture replicaAwareComponentArchitecture, EList<Port> replicatedPorts,
			Set<Component> deployableComponentsOfReplicatedArch)
			throws InvalidTimingModelException {

		// Create a list of Components that contain deployment targets.
		EList<Component> allComponentsOfReplicatedAf3Architecture =
				getChildrenWithType(replicaAwareComponentArchitecture, Component.class);
		List<Component> deployableComponentHullList = new ArrayList<Component>();
		for(Component currentAf3Component : allComponentsOfReplicatedAf3Architecture) {
			if(!deployableComponentsOfReplicatedArch.contains(currentAf3Component) &&
					!isAnyParentComponentDeployable(currentAf3Component, encEntryModelElemAssoc)) {
				deployableComponentHullList.add(currentAf3Component);
			}
		}

		// Build the list of replicated source ports. Therefore, all source ports from the component
		// architecture containing the replica are determined, and the original ports are removed
		// from the list.
		EList<Port> remainingSourcePortsToConnect =
				getReplicatedSourcePorts(replicaAwareComponentArchitecture, replicatedPorts);

		while(!remainingSourcePortsToConnect.isEmpty()) {
			// Always take the first element form the working list.
			Port currentPortToConnect = remainingSourcePortsToConnect.get(0);
			for(Channel currentOutgoingChannel : currentPortToConnect.getOutgoingChannels()) {
				Port currentTargetPort = currentOutgoingChannel.getTarget();
				if(deployableComponentHullList.contains(currentTargetPort.getOwner())) {
					// For hull Components, the target Port needs to be replicated, since this
					// cannot be considered when replicating the Components according to the
					// mapping.
					// FIXME: remove IOServer hack.
					if(!currentTargetPort.getComponent().getName().contains("IOServer") ||
							currentPortToConnect.getComponent().getName().contains("IOServer")) {
						Port replicatedPort = EcoreUtil.copy(currentTargetPort);
						replicatedPort.setOwner(currentTargetPort.getOwner());
						currentOutgoingChannel.setTarget(replicatedPort);
						String replicatedPortName =
								extractReplicatedPortName(currentPortToConnect, replicatedPort);
						replicatedPort.setName(replicatedPortName);

						// Differentiate the target ports by their property of being a "hull"
						// component
						// or a deployment target, since this results in a different channel wiring.
						for(Channel currentOutgoingTargetChannel : currentTargetPort
								.getOutgoingChannels()) {
							Port tgtPort = currentOutgoingTargetChannel.getTarget();
							createChannelAndAttach(
									currentOutgoingTargetChannel.getParentComponent(), "",
									replicatedPort, tgtPort);
						}
						// Add the "hull Port" to the working list for further processing
						remainingSourcePortsToConnect.add(replicatedPort);
					}
				}
			}
			remainingSourcePortsToConnect.remove(0);
		}

		// Finally, arrange the ports for the graphical representation (avoid overlap).
		for(Component currentHullComponent : deployableComponentHullList) {
			arrangePorts(currentHullComponent.getInputPorts(), currentHullComponent);
			arrangePorts(currentHullComponent.getOutputPorts(), currentHullComponent);
		}
	}

	/**
	 * Returns an {@link EList} of source {@link Port}s that belong to replicated {@link Component}
	 * s.
	 * 
	 * @throws InvalidTimingModelException
	 *             if the timing specification of any traversed element is inconsistent.
	 */
	private static EList<Port> getReplicatedSourcePorts(
			ComponentArchitecture replicaAwareComponentArchitecture, EList<Port> replicatedPorts)
			throws InvalidTimingModelException {
		// Get the source ports from the replicated and the original ComponentArchitecture
		EList<Port> replicatedSourcePorts = new BasicEList<Port>();
		EList<OutputPort> originalAndReplicatedSourcePorts =
				getSourcePorts(replicaAwareComponentArchitecture);
		for(Port currentPort : originalAndReplicatedSourcePorts) {
			if(replicatedPorts.contains(currentPort)) {
				replicatedSourcePorts.add(currentPort);
			}
		}

		return replicatedSourcePorts;
	}

	/**
	 * Generates the name of a replicated Port to obtain a consistent enumeration within the
	 * replicated logical architecture and to have unique port naming of "hull" components
	 * (otherwise the model was invalid).
	 */
	private static String extractReplicatedPortName(Port currentPortToConnect,
			Port replicatedPort) {
		String[] replicatedPortNumber;
		// The port number can be derived from the component of the port to connect, or the port
		// itself (for hull components).
		if(currentPortToConnect.getIncomingChannels().isEmpty()) {
			replicatedPortNumber = currentPortToConnect.getComponent().getName().split("_");
		} else {
			replicatedPortNumber = currentPortToConnect.getName().split("_");
		}
		return replicatedPort.getName() + "_" +
				replicatedPortNumber[replicatedPortNumber.length - 1];
	}
}
