/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.modeltransformation.deployment;

import static org.fortiss.af3.exploration.alg.dse.modeltransformation.ITransformationModule.TransformationState.INITIALIZING;

import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.af3.component.model.ComponentArchitecture;
import org.fortiss.af3.deployment.model.AF3DeploymentFactory;
import org.fortiss.af3.deployment.model.ComponentArchitectureReference;
import org.fortiss.af3.deployment.model.Deployment;
import org.fortiss.af3.deployment.model.PlatformArchitectureReference;
import org.fortiss.af3.exploration.alg.dse.modeltransformation.EMFModelConcurrencyHandler;
import org.fortiss.af3.exploration.alg.dse.modeltransformation.ITransformationModule;
import org.fortiss.af3.exploration.alg.dse.modeltransformation.base.TransformationModuleBase;
import org.fortiss.af3.exploration.alg.exception.TransformationModuleException;
import org.fortiss.af3.exploration.alg.guava.SubClassToInstanceMap;
import org.fortiss.af3.exploration.alg.service.ExplorationEncodingMap;
import org.fortiss.af3.exploration.alg.service.IExplorationEncoding;
import org.fortiss.af3.exploration.alg.service.IExplorationTransformationService.TransformationContext;
import org.fortiss.af3.platform.model.PlatformArchitecture;

/**
 * Initializes a {@link Deployment} by creating a new model and setting its references to the given
 * {@link ComponentArchitecture} and {@link PlatformArchitecture}.
 * 
 * @author diewald
 */
public class DeploymentInitializer extends TransformationModuleBase<Deployment> {

	/** {@inheritDoc} */
	@Override
	public Class<Deployment> getOutputType() {
		return Deployment.class;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<Class<?>> getRequiredTypes() {
		Collection<Class<?>> inputTypes = new HashSet<>();
		inputTypes.add(ComponentArchitecture.class);
		inputTypes.add(PlatformArchitecture.class);
		return inputTypes;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<Class<?>> getOptionalTypes() {
		return Collections.emptySet();
	}

	/** {@inheritDoc} */
	@Override
	public Collection<Class<? extends ITransformationModule<?>>> getParentModules() {
		Collection<Class<? extends ITransformationModule<?>>> parentModules = new HashSet<>();
		parentModules.add(ExplorationSolutionToDeployment.class);
		return parentModules;
	}

	/** {@inheritDoc} */
	@Override
	public boolean isOptional() {
		return false;
	}

	/** {@inheritDoc} */
	@Override
	public TransformationState getActivationStage() {
		return INITIALIZING;
	}

	/** {@inheritDoc} */
	@Override
	public <N extends EObject, S extends EObject> Deployment transform(TransformationState state,
			TransformationContext context, SubClassToInstanceMap<S> baseModels,
			ExplorationEncodingMap<IExplorationEncoding> explorationEncodings,
			SubClassToInstanceMap<N> transformedModels, EObject outputModel)
			throws TransformationModuleException {
		ComponentArchitecture originalCA =
				(ComponentArchitecture)baseModels.get(ComponentArchitecture.class);
		if(originalCA == null) {
			throw new TransformationModuleException(
					this,
					"Could not retrieve the original component architecture from the set of base models. Be sure to add it to the required input types.");
		}

		PlatformArchitecture originalPA =
				(PlatformArchitecture)baseModels.get(PlatformArchitecture.class);
		if(originalPA == null) {
			throw new TransformationModuleException(
					this,
					"Could not retrieve the original platform architecture from the set of base models. Be sure to add it to the required input types.");
		}

		outputModel = createDeployment("Deployment", null, originalPA);

		return (Deployment)outputModel;
	}

	/**
	 * Creates a {@link Deployment} model without requiring a model context (the bi-references do
	 * not notify the referenced model element).
	 */
	private Deployment createDeployment(String name, ComponentArchitecture compArch,
			PlatformArchitecture platArch) {
		Deployment deployment = AF3DeploymentFactory.eINSTANCE.createDeployment();
		deployment.setName(name);

		if(compArch != null) {
			ComponentArchitectureReference caRef =
					AF3DeploymentFactory.eINSTANCE.createComponentArchitectureReference();
			EMFModelConcurrencyHandler.getInstance().setReference(caRef, compArch);
			deployment.setComponentArchitectureReference(caRef);
		}

		if(platArch != null) {
			PlatformArchitectureReference paRef =
					AF3DeploymentFactory.eINSTANCE.createPlatformArchitectureReference();
			EMFModelConcurrencyHandler.getInstance().setReference(paRef, platArch);
			deployment.setPlatformArchitectureReference(paRef);
		}

		return deployment;
	}
}
