/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.modeltransformation.safety;

import static org.fortiss.af3.exploration.alg.util.TransformationUtils.replicateInputPortAndAttach;
import static org.fortiss.tooling.common.util.LambdaUtils.filterSet;

import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.af3.component.model.Channel;
import org.fortiss.af3.component.model.Component;
import org.fortiss.af3.component.model.ComponentArchitecture;
import org.fortiss.af3.component.model.InputPort;
import org.fortiss.af3.component.model.OutputPort;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.safetyfunctionarch.SafetyFunctionArchEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.safetyfunctionarch.SafetyFunctionArchEntry;
import org.fortiss.af3.exploration.alg.dse.modeltransformation.ITransformationModule;
import org.fortiss.af3.exploration.alg.dse.modeltransformation.base.TransformationModuleBase;
import org.fortiss.af3.exploration.alg.dse.modeltransformation.comparch.ComponentArchitectureTransformer;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IDeployableComponentAdapter;
import org.fortiss.af3.exploration.alg.guava.SubClassToInstanceMap;
import org.fortiss.af3.exploration.alg.service.ExplorationEncodingMap;
import org.fortiss.af3.exploration.alg.service.IExplorationEncoding;
import org.fortiss.af3.exploration.alg.service.IExplorationTransformationService.TransformationContext;

/**
 * 
 * @author diewald
 */
public class SafetyChannelPortReplicator extends TransformationModuleBase<ComponentArchitecture> {

	/** {@inheritDoc} */
	@Override
	public boolean isOptional() {
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public ITransformationModule.TransformationState getActivationStage() {
		return TransformationState.INSTANTIATED;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<Class<? extends ITransformationModule<?>>> getParentModules() {
		Collection<Class<? extends ITransformationModule<?>>> parentModules = new HashSet<>();
		parentModules.add(ComponentArchitectureTransformer.class);
		return parentModules;
	}

	/** {@inheritDoc} */
	@Override
	public <N extends EObject, S extends EObject> ComponentArchitecture transform(
			ITransformationModule.TransformationState state, TransformationContext context,
			SubClassToInstanceMap<S> baseModels,
			ExplorationEncodingMap<IExplorationEncoding> explorationEncodings,
			SubClassToInstanceMap<N> transformedModels, EObject outputModel) {
		@SuppressWarnings("unchecked") SafetyFunctionArchEncoding<Component> sfaEnc =
				explorationEncodings.getEncoding(SafetyFunctionArchEncoding.class);

		Collection<Component> sfComponents = new HashSet<>();
		for(SafetyFunctionArchEntry<Component> sfaEntry : sfaEnc.getAllEntries()) {
			Collection<IDeployableComponentAdapter<Component>> entryCompAdps =
					sfaEntry.getSafetyFunctionAdapter().getAllComponents();
			entryCompAdps.forEach(c -> sfComponents.add(c.getObject()));
		}

		Collection<OutputPort> sfOutPorts = new HashSet<>();
		sfComponents.forEach(c -> sfOutPorts.addAll(c.getOutputPorts()));

		Collection<InputPort> connectedInputPorts = new HashSet<>();
		for(OutputPort curOutPort : sfOutPorts) {
			curOutPort.getOutgoingChannels().forEach(
					ch -> connectedInputPorts.add((InputPort)ch.getTarget()));
		}

		Collection<InputPort> nonSfInputPorts =
				filterSet(connectedInputPorts, p -> !sfComponents.contains(p.getComponent()));
		for(InputPort inPort : nonSfInputPorts) {
			Collection<EObject> genPorts = origToGenAssocMap.get(inPort);
			assert (genPorts.size() == 1);
			inPort = (InputPort)genPorts.stream().findAny().get();
			List<Channel> incomingChannels = inPort.getIncomingChannels();
			for(int chIdx = 1; chIdx < incomingChannels.size(); chIdx++) {
				char suffix = (char)(chIdx - 1 + 'a');
				InputPort genPort =
						replicateInputPortAndAttach(inPort, inPort.getComponent(),
								inPort.getName() + "_" + suffix);
				Channel curCh = incomingChannels.get(chIdx);
				curCh.setTarget(genPort);
				InputPort origPort = (InputPort)genToOrigAssocMap.get(inPort);
				genToOrigAssocMap.put(genPort, origPort);
				origToGenAssocMap.put(origPort, genPort);
			}
		}

		return (ComponentArchitecture)outputModel;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<Class<?>> getRequiredTypes() {
		Collection<Class<?>> requiredTypes = new HashSet<>();
		requiredTypes.add(ComponentArchitecture.class);
		requiredTypes.add(SafetyFunctionArchEncoding.class);
		return requiredTypes;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<Class<?>> getOptionalTypes() {
		return Collections.emptySet();
	}

	/** {@inheritDoc} */
	@Override
	public Class<? super ComponentArchitecture> getOutputType() {
		return ComponentArchitecture.class;
	}
}
