/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.modeltransformation.safety;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.af3.component.model.Channel;
import org.fortiss.af3.component.model.Component;
import org.fortiss.af3.component.model.ComponentArchitecture;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.safetyfunctionarch.SafetyFunctionArchEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.safetyfunctionarch.SafetyFunctionArchEntry;
import org.fortiss.af3.exploration.alg.dse.modeltransformation.ITransformationModule;
import org.fortiss.af3.exploration.alg.dse.modeltransformation.base.TransformationModuleBase;
import org.fortiss.af3.exploration.alg.dse.modeltransformation.comparch.ComponentArchitectureTransformer;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IDeployableComponentAdapter;
import org.fortiss.af3.exploration.alg.exception.TransformationModuleException;
import org.fortiss.af3.exploration.alg.guava.SubClassToInstanceMap;
import org.fortiss.af3.exploration.alg.service.ExplorationEncodingMap;
import org.fortiss.af3.exploration.alg.service.IExplorationEncoding;
import org.fortiss.af3.exploration.alg.service.IExplorationTransformationService.TransformationContext;

/**
 * Module that deletes diagnostic {@link Component}s from a generated {@link ComponentArchitecture}
 * if diagnostic units were not selected by the DSE for some safety function.
 * 
 * @author diewald
 */
public class SafetyDiagUnitRemover extends TransformationModuleBase<ComponentArchitecture> {

	/** {@inheritDoc} */
	@Override
	public Class<ComponentArchitecture> getOutputType() {
		return ComponentArchitecture.class;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<Class<?>> getRequiredTypes() {
		Collection<Class<?>> inputModelTypes = new HashSet<>();
		inputModelTypes.add(ComponentArchitecture.class);
		inputModelTypes.add(SafetyFunctionArchEncoding.class);
		return inputModelTypes;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<Class<?>> getOptionalTypes() {
		return Collections.emptySet();
	}

	/** {@inheritDoc} */
	@Override
	public boolean isOptional() {
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public TransformationState getActivationStage() {
		return TransformationState.INITIALIZED;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<Class<? extends ITransformationModule<?>>> getParentModules() {
		Collection<Class<? extends ITransformationModule<?>>> parentModuleTypes = new HashSet<>();
		parentModuleTypes.add(ComponentArchitectureTransformer.class);
		return parentModuleTypes;
	}

	/** {@inheritDoc} */
	@Override
	public <N extends EObject, S extends EObject> ComponentArchitecture transform(
			TransformationState state, TransformationContext context,
			SubClassToInstanceMap<S> baseModels,
			ExplorationEncodingMap<IExplorationEncoding> explorationEncodings,
			SubClassToInstanceMap<N> transformedModels, EObject outputModel)
			throws TransformationModuleException {
		ComponentArchitecture genCA = (ComponentArchitecture)outputModel;
		if(genCA == null) {
			throw new TransformationModuleException(this,
					"Could not retrieve the generated Component Architecture.");
		}

		@SuppressWarnings("unchecked") SafetyFunctionArchEncoding<Component> sfaEnc =
				explorationEncodings.getEncoding(SafetyFunctionArchEncoding.class);
		if(sfaEnc == null) {
			throw new TransformationModuleException(this,
					"A SafetyFunctionArchEncoding is required for this module to operate properly.");
		}

		for(SafetyFunctionArchEntry<Component> sfaEntry : sfaEnc.getAllEntries()) {
			if(!sfaEntry.isDiagnosicUnitPresent()) {
				for(IDeployableComponentAdapter<Component> diagCompAdp : sfaEntry
						.getSafetyFunctionAdapter().getDiagnosticComponents()) {
					for(EObject genDiagComponent : origToGenAssocMap.get(diagCompAdp.getObject())) {
						if(!(genDiagComponent instanceof Component)) {
							throw new TransformationModuleException(this, "The generated element " +
									genDiagComponent +
									" in the generated ComponentArchitectureis not a 'Component' as it is expected.");
						}

						// Get all channels connected to the current diagnostic Component.
						Collection<Channel> genChannels = new ArrayList<>();
						((Component)genDiagComponent).getInputPorts().forEach(
								p -> genChannels.addAll(p.getIncomingChannels()));
						((Component)genDiagComponent).getOutputPorts().forEach(
								p -> genChannels.addAll(p.getOutgoingChannels()));
						// Remove all ports from Components which are connected to the diagnostic
						// Component. Can be included if needed. NOTE: Instantiated components will
						// introduce the "same" Port again.
						// genChannels.stream().map(ch -> ch.getTarget())
						// .filter(p -> p.getOwner() != genDiagComponent)
						// .forEach(c -> removeGenModelElement(c));
						// genChannels.stream().map(ch -> ch.getSource())
						// .filter(p -> p.getOwner() != genDiagComponent)
						// .forEach(c -> removeGenModelElement(c));
						// Remove all Channels connected to the diagnostic Component.
						genChannels.forEach(ch -> removeGenModelElement(ch));

						removeGenModelElement(genDiagComponent);
					}
				}
			}
		}
		return genCA;
	}
}
