/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.sysmodel.arch;

import java.util.Collection;

import org.eclipse.emf.common.util.Enumerator;
import org.fortiss.af3.component.model.Component;

/**
 * Adapter between the deployable components (=tasks) of the system meta-model used in the modeling
 * tool, and the DSE.
 * 
 * @author barner
 */
public interface IDeployableComponentAdapter<T> extends IRequestAdapter<T>, Cloneable {

	/**
	 * Returns the WCET if this {@link IDeployableComponentAdapter} would be executed on the given
	 * {@link IExecutionUnitAdapter}.
	 */
	public Double getWcet(IExecutionUnitAdapter<?> executionUnit);

	/**
	 * Returns the consumed energy if this {@link IDeployableComponentAdapter} would be executed on
	 * the given {@link IExecutionUnitAdapter}.
	 */
	public Double getEnergyConsumption(IExecutionUnitAdapter<?> executionUnit);

	/** Returns the minimum required safety integrity level of the deployable component. */
	public Enumerator getRequiredSafetyIntegrityLevel();

	/** Returns the period of the deployable component. */
	public Double getTimingRequirementPeriod();

	/** Sets the period of the deployable component. */
	public void setTimingRequirementPeriod(double period);

	/**
	 * Predicates whether this {@link IDeployableComponentAdapter} must be instantiated by another
	 * {@link IDeployableComponentAdapter}. This is determined by evaluating whether
	 * {@link IDeployableComponentAdapter}s are given that may be used to instantiate this
	 * {@link IDeployableComponentAdapter}.
	 */
	public boolean isInstantiationRequired();

	/**
	 * Returns those {@link IDeployableComponentAdapter}s that may be used to instantiate this
	 * {@link IDeployableComponentAdapter}.
	 */
	public Collection<IDeployableComponentAdapter<T>> getReplacementComponents();

	/**
	 * Predicates whether this {@link IDeployableComponentAdapter} instantiates an abstract
	 * {@link IDeployableComponentAdapter}.
	 */
	public boolean isInstantiation();

	/**
	 * Returns the {@link IDeployableComponentAdapter} instantiated by this
	 * {@link IDeployableComponentAdapter} (if applicable).
	 */
	public IDeployableComponentAdapter<T> getReplacedComponent();

	/**
	 * Sets the replacedComponent reference of {@code this} {@link IDeployableComponentAdapter} to
	 * allow tracking of the "abstract" {@link IDeployableComponentAdapter}s from the instantiated
	 * {@link IDeployableComponentAdapter}s.
	 */
	// TODO: If possible, find a way to eliminate this suppresswarning.
	public void setReplacedComponent(
			@SuppressWarnings("rawtypes") IDeployableComponentAdapter abstractComponent);

	/**
	 * Whenever an {@link IDeployableComponentAdapter} is copied/cloned its references wrt. the task
	 * instantiation need to be updated. Hence, after cloning, this method should be called for each
	 * copied/cloned {@link IDeployableComponentAdapter}.
	 */
	public void updateInstatiationData();

	/**
	 * Returns the minimum number of replications allowed for this
	 * {@link IDeployableComponentAdapter}.
	 */
	public int getMinReplication();

	/**
	 * Returns the maximum number of replications allowed for this
	 * {@link IDeployableComponentAdapter}.
	 */
	public int getMaxReplication();

	/**
	 * Predicate if this {@link IDeployableComponentAdapter} is strongly causal. This implies that
	 * its outputs to non-strongly causal {@link IDeployableComponentAdapter}s only become available
	 * during the next hyperperiod.
	 */
	public boolean isStronglyCausal();

	/**
	 * Clones the {@link IDeployableComponentAdapter} on which this method is called. If
	 * {@link Component}s are instantiated during the exploration, this is needed. The initially
	 * constructed {@link IDeployableComponentAdapter}s must be replicated for each instantiated
	 * encoding.
	 */
	public IDeployableComponentAdapter<T> clone();
}
