/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.sysmodel.arch;

import java.util.Collection;

import org.fortiss.af3.safety.model.SafetyArchitecture;

/**
 * Adapter to represent safety functions from the input system models in the DSE.
 * 
 * @author diewald
 */
public interface ISafetyFunctionAdapter<T> {

	/**
	 * Initialization routine for externally contributed adapters that are created before the
	 * {@link SystemModelAdapter} of the DSE.
	 */
	void initializeAdapter(SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter);

	/**
	 * Returns the collection of {@link IDeployableComponentAdapter}s that are realizing the safety
	 * function.
	 */
	Collection<IDeployableComponentAdapter<T>> getRealizingComponents();

	/**
	 * Returns the collection of {@link IDeployableComponentAdapter}s that are realizing diagnostic
	 * functions of the actual safety function.
	 */
	Collection<IDeployableComponentAdapter<T>> getDiagnosticComponents();

	/**
	 * Returns the {@link IDeployableComponentAdapter} that generates the output signal of the
	 * safety functions, i.e. a voter or HW interfacing component.
	 */
	// TODO: extend to collection.
	IDeployableComponentAdapter<T> getOutputComponent();

	/**
	 * Returns the collection of all {@link IDeployableComponentAdapter}s that represent
	 * {@code this} safety function.
	 */
	Collection<IDeployableComponentAdapter<T>> getAllComponents();

	/**
	 * Returns the collection of {@link SafetyArchitecture}s that may be used to implement
	 * {@code this} {@link ISafetyFunctionAdapter}.
	 */
	Collection<SafetyArchitecture> getAllowedSafetyArchitectures();

	/**
	 * Returns the allowed number of channels to realize {@code this} {@link ISafetyFunctionAdapter}
	 * .
	 */
	int[] getAllowedChannelNumbers();

	/**
	 * Returns the array of allowed signal numbers that trigger the safety function subject to the
	 * given number of channels. E.g., for the safety function design 1oo2 and 2oo2, one and two
	 * would be the return values of this method.
	 */
	int[] getAllowedTriggerSignalNumbers(int channelNum);

	/**
	 * Returns the collection of trigger signal numbers that may be used to implement the allowed
	 * safety architectures.
	 */
	Collection<Integer> getAllAllowedTriggerSignalNumbers();

	/**
	 * Determines whether a diagnostic unit may be added depending on the number of channels and the
	 * number of signals that trigger the safety function.
	 */
	boolean isDiagUnitAllowed(int channelNum, int triggerSignalNum);

	/**
	 * Returns whether each channel of {@code this} safety function requires a separate output
	 * {@link IDeployableComponentAdapter}.
	 */
	boolean hasSeparateOutPerChannel();
}
