/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.sysmodel.arch;

import static org.fortiss.af3.exploration.util.ExplorationUtils.isDebugVerboseEnabled;
import static org.fortiss.af3.schedule.utils.MathUtils.lcmDoubleCollection;
import static org.fortiss.af3.task.util.TaskAllocationUtils.getAllocatedComponentsTo;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.pickFirstInstanceOf;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;
import java.util.stream.Collectors;

import org.conqat.lib.commons.collections.Pair;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.fortiss.af3.component.model.Component;
import org.fortiss.af3.component.model.ComponentArchitecture;
import org.fortiss.af3.component.model.Port;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskgraph.AbstractTaskGraphEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.abstractmapping.AbstractTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceConnectionAdapter.ConnectionType;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph.AcyclicTaskGraph;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph.ChannelAdapterWeightedEdge;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph.DefaultTaskGraph;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph.TaskGraph;
import org.fortiss.af3.exploration.alg.exception.InvalidTimingModelException;
import org.fortiss.af3.exploration.alg.guava.MutableSubClassToInstanceMap;
import org.fortiss.af3.exploration.alg.guava.SubClassToInstanceMap;
import org.fortiss.af3.exploration.alg.service.IExplorationEncoding;
import org.fortiss.af3.exploration.backend.IDseInputContainer.SignalType;
import org.fortiss.af3.platform.model.ExecutionUnit;
import org.fortiss.af3.platform.model.IPlatformArchitectureElement;
import org.fortiss.af3.platform.model.IPlatformCommunicationResource;
import org.fortiss.af3.platform.model.IPlatformExport;
import org.fortiss.af3.platform.model.IPlatformResource;
import org.fortiss.af3.platform.model.PlatformArchitecture;
import org.fortiss.af3.platform.model.annotation.ResourceLink;
import org.fortiss.af3.platform.utils.PlatformArchitectureUtils;
import org.fortiss.af3.task.model.Task;
import org.fortiss.af3.task.model.TaskArchitecture;
import org.fortiss.tooling.base.model.element.IConnection;
import org.fortiss.tooling.base.model.element.IConnector;
import org.fortiss.tooling.base.model.element.IHierarchicElement;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.jgrapht.Graphs;
import org.jgrapht.alg.ConnectivityInspector;

/**
 * Adapter to the system model used as input for the DSE which is created for a specific DSE run.
 * 
 * @author barner, huang
 */
public class SystemModelAdapter<C, E, S, G, TR, M> {

	/**
	 * Constant defining the allowed tolerance for the calculation of the greatest common divisor
	 * used in the calculation of the effective periods.
	 */
	protected final static int GCD_PRECISION = 3;

	/** Logical architecture of this {@link SystemModelAdapter}. */
	protected ILogicalArchitectureAdapter<?> logicalArchitecture;

	/** Platform architecture of this {@link SystemModelAdapter}. */
	protected IPlatformArchitectureAdapter<?> platformArchitecture;

	/**
	 * Input models that can be retrieved via {@code this} {@link SystemModelAdapter}. It references
	 * the original
	 */
	protected SubClassToInstanceMap<EObject> inputModels = MutableSubClassToInstanceMap.create();

	/** Deployable software components, to be mapped to deployment targets provided by platform. */
	protected Collection<IDeployableComponentAdapter<C>> deployableComponents;

	/** Collection of deployable components from the base model. */
	protected Collection<C> deployableComponentObjects;

	/**
	 * Maps the {@link Port}s of the deployable {@link Component}s to their temporal
	 * {@link SignalType} (aperiodic, sporadic, etc.).
	 */
	protected Map<Port, SignalType> portSignalTypes;

	/**
	 * Look up table to determine the input/output directions of {@link IConnector} types of the
	 * target {@link PlatformArchitecture}.
	 */
	protected Map<Class<? extends IConnector>, ConnectionType> connectionTypeLUT;

	/**
	 * Deployable software components that may be replicated (per default: all @{code
	 * deployableComponent}s.
	 */
	protected Collection<IDeployableComponentAdapter<C>> replicableComponents;

	/** Schedulable software components, to be scheduled by the internal scheduling algorithm. */
	protected Collection<IDeployableComponentAdapter<C>> schedulableComponents = new ArrayList<>();

	/**
	 * Collection of software components that will be actually executed. This includes realizations
	 * of "abstract" software components (-> interfaces) that are contained in the
	 * {@code deployableComponents} collection.
	 */
	protected Collection<IDeployableComponentAdapter<C>> realizableDeployableComponents;

	/**
	 * Deployment targets, i.e. the execution units of the target platform, where deployable
	 * software components will be mapped to.
	 */
	protected Collection<IExecutionUnitAdapter<E>> deploymentTargets;

	/** Channels that define the (allowed) data exchange between logical components. */
	protected Collection<ISignalAdapter<S>> messages;

	/** All {@link ExecutionUnit}s that are present in the target platform. */
	protected Collection<IExecutionUnitAdapter<E>> allPresentExecutionUnits;

	/** {@link Collection} of memories present in the system. */
	protected Collection<IMemoryUnitAdapter<M>> memoryUnits;

	/**
	 * Gateway units that are employed to manage the communication between multiple networks
	 * (connecting the same or different layers).
	 */
	protected Collection<IGatewayUnitAdapter<G>> gatewayUnits;

	// TODO Consider more than one transmission unit
	/** TDMA arbitrated transmission unit */
	protected Collection<ITransmissionUnitAdapter<TR>> transmissionUnits;

	/** Contains all available communication resources of the platform. */
	protected Collection<ICommunicationResourceAdapter<TR>> availableCommunicationResource;

	/** Contains all resources that are available in the given platform. */
	protected Collection<IResourceAdapter<?>> availableResources;

	/**
	 * Contains all deployment target resources, i.e. the deployment targets and their
	 * interconnects.
	 */
	protected Collection<IResourceAdapter<?>> deploymentTargetResources;

	/** Contains the connections between the {@link IResourceAdapter}s of the platform. */
	protected Collection<IResourceConnectionAdapter> deploymentTargetConnectors;

	/**
	 * Collection of the safety functions adapters that represent the safety functions from the
	 * input models.
	 */
	protected Collection<ISafetyFunctionAdapter<C>> safetyFunctionAdapters = new ArrayList<>();

	/** The {@link PlatformCommunicationGraph} that is given by the current platform. */
	protected PlatformCommunicationGraph platformCommunicationGraph;

	/** The set of independent {@link DefaultTaskGraph}s derived from the logical architecture. */
	protected Collection<TaskGraph> taskGraphs;

	/**
	 * The set of independent directed acyclic {@link DefaultTaskGraph}s derived from the logical
	 * architecture.
	 */
	protected Collection<AcyclicTaskGraph> acyclicTaskGraphs;

	/** The "messages" emitted by the sender {@link IDeployableComponentAdapter}. */
	protected Map<IDeployableComponentAdapter<C>, Collection<ISignalAdapter<S>>> emittedMessagesBySender;

	/** Hyper-period of independent tasks graphs */
	protected Double hyperPeriod;

	/**
	 * Map: deployable component (= task) -> corresponding (independent, connected)
	 * {@link DefaultTaskGraph}.
	 */
	protected Map<IDeployableComponentAdapter<C>, TaskGraph> componentInGraph;

	/**
	 * Map: deployable component (= task) -> corresponding (independent, connected)
	 * {@link AcyclicTaskGraph}.
	 */
	protected Map<IDeployableComponentAdapter<C>, AcyclicTaskGraph> componentInAcyclicGraph;

	/**
	 * Mapping virtual resource -> hardware resource onto which virtual resource is (statically)
	 * mapped, e.g., partition -> hardware execution unit / processor core.
	 * 
	 * The virtual resources are considered as deployment targets where {@link IRequestAdapter}s can
	 * be allocated to.
	 * 
	 * However, deployment targets running on the same core share the same time line.
	 * Hence,a map between the deployment targets and the "time line" is maintained which
	 * will be considered during scheduling to separate partitions in time.
	 */
	protected Map<IResourceAdapter<?>, IResourceAdapter<?>> virtualResourceToHardwareResourceMap;

	/**
	 * Identifies the {@link IModelElement} types which establish a connection between two
	 * {@link PlatformArchitecture}s via an {@link ResourceLink}.
	 */
	protected Pair<Class<? extends IModelElement>, Class<? extends IModelElement>> platformInterconnectTypes;

	/** Defines the fault containment region defined for the target {@link PlatformArchitecture}. */
	protected Class<? extends IPlatformResource> faultContainmentRegion;

	/**
	 * Maintains a map of {@link IExplorationEncoding}s that can be provided by {@code this}
	 * {@link SystemModelAdapter} by their corresponding types.
	 */
	private Map<Class<? extends IExplorationEncoding>, IExplorationEncoding> inputEncodings =
			new HashMap<>();

	/** Registers a given input encoding with {@code this} {@link SystemModelAdapter}. */
	protected <T extends IExplorationEncoding> void registerInputEncoding(T encoding) {
		@SuppressWarnings("unchecked") Class<T> encodingType = (Class<T>)encoding.getClass();
		assert (!inputEncodings.containsKey(
				encodingType)) : "You are trying to add an encoding to the input encoding whose type is already registered. This is not supported";
		inputEncodings.put(encodingType, encoding);
	}

	/** Returns the input encoding of the given type. */
	@SuppressWarnings("unchecked")
	public <T extends IExplorationEncoding> T getInputEncoding(Class<T> encodingType) {
		return (T)inputEncodings.get(encodingType);
	}

	/**
	 * Returns the collection of {@link IExplorationEncoding} types that are availably through
	 * {@code this} {@link SystemModelAdapter}.
	 */
	public Set<Class<? extends IExplorationEncoding>> getRegisteredEncodingTypes() {
		return inputEncodings.keySet();
	}

	/** Returns the input models accessible via {@code this} {@link SystemModelAdapter}. */
	public SubClassToInstanceMap<EObject> getInputModels() {
		return inputModels;
	}

	/**
	 * Adds the given model to the input models that can be obtained via {@code this}
	 * {@link SystemModelAdapter}.
	 */
	@SuppressWarnings("unchecked")
	public <T extends EObject> void putInputModel(Class<? extends T> modelType, T inputModel) {
		inputModels.putInstance((Class<T>)modelType, inputModel);
	}

	/**
	 * Creates and registers an {@link AbstractTaskGraphEncoding} with @code this}
	 * {@link SystemModelAdapter}.
	 */
	protected void createAbstractTaskGraphEncoding() {
		AbstractTaskGraphEncoding<AbstractTaskMappingEncoding, C> atgEnc =
				new AbstractTaskGraphEncoding<>(this);
		registerInputEncoding(atgEnc);
	}

	/**
	 * Setups the list of {@link IDeployableComponentAdapter}s that are scheduleable by a
	 * time-triggered scheduler.
	 */
	protected void setupScheduleableComponents() {
		for(IDeployableComponentAdapter<C> comp : deployableComponents) {
			if(comp.getTimingRequirementPeriod() != null) {
				schedulableComponents.add(comp);
			}
		}
	}

	// TODO: Make the task graph setup independent from the AF3-Component Architecture and write a
	// wrapper in the specific AF3 SystemModelAdapter.
	/**
	 * Sets up the {@link DefaultTaskGraph}s for all (independent) connected sub-component
	 * model contained in the logical {@link ComponentArchitecture}. This method
	 * also determines the hyper-period of all {@link DefaultTaskGraph}s.
	 * 
	 * @throws Exception
	 */
	protected void setupTaskGraphs() throws Exception {
		// - Add data dependencies to {@link TaskGraph}s
		// - Print some debugging output
		// - Determine the hyper-period of all {@link TaskGraph}s contained in this {@link
		// SystemModelAdapter}.
		taskGraphs = new ArrayList<TaskGraph>();
		acyclicTaskGraphs = new ArrayList<>();

		// Temporary task graph: add vertices (= software components / tasks that need to be mapped
		// to processing elements, and that potentially require data from other software components)
		DefaultTaskGraph temporaryTaskGraph = new DefaultTaskGraph();
		// createTaskForrest();
		for(IDeployableComponentAdapter<?> deployableComponent : deployableComponents) {
			temporaryTaskGraph.addTask(deployableComponent);
		}
		// Temporary task graph: add edges
		addDataDependenciesToTaskGraph(temporaryTaskGraph);

		// Separate temporary task graph into multiple TaskGraphs that consist of exactly one
		// connected graph component. I.e., there are no data dependencies between the resulting new
		// TaskGraphs.
		// TODO: handle the removal of taskgraphs in another way. Throw an exception / display an
		// error for inconsistent period specifications?

		// Create a generic TaskGraph of the logical architecture (can contain cycles)
		taskGraphs = createDefaultTaskGraphs(temporaryTaskGraph);
		// removeInvalidTaskGraphs(taskGraphs);
		componentInGraph = registerTaskGraphsForComponents(taskGraphs);

		// Construct an Directed Acyclic TaskGraph (e.g., utilized in scheduling).
		acyclicTaskGraphs = taskGraphs.stream().map(tg -> new AcyclicTaskGraph(tg))
				.collect(Collectors.toList());
		removeInvalidTaskGraphs(acyclicTaskGraphs);
		componentInAcyclicGraph = registerTaskGraphsForComponents(acyclicTaskGraphs);

		// Print some debugging output: print the taskgraphs and display them in a dialog.
		if(isDebugVerboseEnabled()) {
			printAndDisplayTaskGraphs(taskGraphs);
			printAndDisplayTaskGraphs(acyclicTaskGraphs);
		}

		Collection<TaskGraph> allTaskGraphs = new ArrayList<TaskGraph>();
		allTaskGraphs.addAll(taskGraphs);
		allTaskGraphs.addAll(acyclicTaskGraphs);
		hyperPeriod = determineHyperPeriod(allTaskGraphs);
	}

	/**
	 * Creates a set of independent {@link TaskGraph}s of the given type ({#link graphClass}). The
	 * subgraphs may have different implementation to construct themselves (via
	 * addDataDependencies).
	 * 
	 * @throws InvalidTimingModelException
	 *             if the periods for a single task graph cannot be set up.
	 */
	private Collection<TaskGraph> createDefaultTaskGraphs(TaskGraph monolithicTaskGraph)
			throws InvalidTimingModelException {
		Collection<TaskGraph> independentTaskGraphs = new ArrayList<>();

		// Compute vertex sets of maximum connected components of temporary task graph
		ConnectivityInspector<IDeployableComponentAdapter<?>, ChannelAdapterWeightedEdge> ci =
				new ConnectivityInspector<IDeployableComponentAdapter<?>, ChannelAdapterWeightedEdge>(
						monolithicTaskGraph.getGraph());
		List<Set<IDeployableComponentAdapter<?>>> connectedSets = ci.connectedSets();
		// Reconstruct TaskGraphs from vertex sets of connected components
		for(Set<IDeployableComponentAdapter<?>> connectedSet : connectedSets) {
			TaskGraph taskGraph = new DefaultTaskGraph();
			Graphs.addAllVertices(taskGraph.getGraph(), connectedSet);

			for(ISignalAdapter<?> message : messages) {
				if(connectedSet.contains(message.getSource()) && !message.getTargets().isEmpty()) {
					taskGraph.addDataDependency(message.getSource(), message.getTargets(), message);
				}
			}

			// Check if tasks within current sub TaskGraph share the same common period
			taskGraph.updatePeriod();
			independentTaskGraphs.add(taskGraph);
		}
		return independentTaskGraphs;
	}

	/**
	 * Adds the given collection of {@link ISafetyFunctionAdapter}s to {@code this}
	 * {@link SystemModelAdapter} that represents the input system models. In contrast to other
	 * input parameters, the {@link ISafetyFunctionAdapter}s are created dynamically if a
	 * safety constraint is present which results in the safety functions to be
	 * extracted <b>after</b> the {@link SystemModelAdapter} has been constructed.
	 * <p>
	 * This method updates the set of replicable components by removing those components that are
	 * referenced by the safety function.
	 */
	// TODO: remove the enclosed cast. The difficulty is that the ISafetyFunctionAdapters are added
	// from within the DSE where the system types are not known any more.
	// It would be more appropriate to shift this addition before launching the DSE.
	@SuppressWarnings({"unchecked", "rawtypes"})
	public void
			addSafetyFunctionAdapters(Collection<ISafetyFunctionAdapter> safetyFunctionAdapters) {
		if(safetyFunctionAdapters == null) {
			return;
		}
		this.safetyFunctionAdapters
				.addAll((Collection<? extends ISafetyFunctionAdapter<C>>)safetyFunctionAdapters);

		for(ISafetyFunctionAdapter<?> sfAdapter : safetyFunctionAdapters) {
			replicableComponents.removeAll(sfAdapter.getDiagnosticComponents());
			replicableComponents.removeAll(sfAdapter.getRealizingComponents());
			replicableComponents.remove(sfAdapter.getOutputComponent());
		}
	}

	/**
	 * Creates a map for the given collection of {@link TaskGraph}s which allows to identify the
	 * {@link TaskGraph} that contains the given {@link IDeployableComponentAdapter} as a vertex.
	 */
	// TODO: Here, we know that the IDeployableCOmponentAdapters are constructed using Components.
	// However, we should try to avoid the "SuppressWarnings".
	@SuppressWarnings("unchecked")
	private <T extends TaskGraph> Map<IDeployableComponentAdapter<C>, T>
			registerTaskGraphsForComponents(Collection<T> graphs) {
		Map<IDeployableComponentAdapter<C>, T> componentToGraphMap =
				new HashMap<IDeployableComponentAdapter<C>, T>();

		for(T taskGraph : graphs) {
			for(@SuppressWarnings("rawtypes") IDeployableComponentAdapter deployableComponent : taskGraph
					.getGraph().vertexSet()) {
				componentToGraphMap.put(deployableComponent, taskGraph);
			}
		}

		return componentToGraphMap;
	}

	/**
	 * For a given collection of {@link TaskGraph}s this method removes all the {@link TaskGraph}s
	 * from the given collection and the contained {@link IDeployableComponentAdapter}s from the set
	 * of deployable components if the corresponding {@link TaskGraph} has a non-strictly positive
	 * hyperperiod.
	 */
	private <T extends TaskGraph> void removeInvalidTaskGraphs(Collection<T> taskGraphs) {
		Set<IDeployableComponentAdapter<?>> tasksToBeRemoved =
				new HashSet<IDeployableComponentAdapter<?>>();
		Collection<T> graphsToBeRemoved = new HashSet<T>();
		// for(T currentGraph : taskGraphs) {
		// if(currentGraph.getPeriod() <= 0) {
		// // No valid period specification: TaskGraph cannot be mapped.
		// tasksToBeRemoved.addAll((currentGraph).getTasks());
		// graphsToBeRemoved.add(currentGraph);
		// System.out.println("No valid period specification: TaskGraph cannot be mapped.");
		// System.out.println(currentGraph);
		// }
		// }

		// TODO: Move these checks to a model validation framework
		// Remove tasks with invalid specification
		if(!tasksToBeRemoved.isEmpty()) {
			System.out.println(
					"The following deployable components are ignored during design exploration due to lack of timing annotations.");
			System.out.println(
					"Each deployable component must be annotated with delay (execution time) and the source commpoent must be annotated with a period!");
			for(IDeployableComponentAdapter<?> deployableComponent : tasksToBeRemoved) {
				System.out.print(deployableComponent.getName() + " , ");
				deployableComponents.remove(deployableComponent);
			}
			for(T currentGraph : graphsToBeRemoved) {
				taskGraphs.remove(currentGraph);
			}
			System.out.print("\n");
		}

	}

	/**
	 * Calculates and returns the hyperperiod of the set of all given {@link TaskGraph}s by
	 * calculating the least common multiple from the hyperperiods of the {@link TaskGraph}s.
	 */
	private Double determineHyperPeriod(Collection<TaskGraph> allTaskGraphs) {
		Set<Double> taskgraphPeriods = new TreeSet<Double>();
		for(TaskGraph currentGraph : allTaskGraphs) {
			Double period = currentGraph.getPeriod();
			if(period != null && period > 0) {
				taskgraphPeriods.add(period);
			}
		}

		if(taskgraphPeriods.isEmpty()) {
			return null;
		}
		return lcmDoubleCollection(taskgraphPeriods, GCD_PRECISION);
	}

	/**
	 * Prints the given collection of {@link TaskGraph}s to the standard output and displays them in
	 * a separate dialog.
	 */
	private <T extends TaskGraph> void printAndDisplayTaskGraphs(Collection<T> taskGraphs) {
		for(T taskGraph : taskGraphs) {
			System.out.println(
					"Application " + taskGraph + ":" + " period = " + taskGraph.getPeriod());
			System.out.println(taskGraph);
			System.out.println("");

			// Open a dialog for the visualization of the TaskGraph
			taskGraph.display();
		}
	}

	/**
	 * Add edges to a {@link TaskGraph} that describe the data dependencies between the software
	 * components of the logical {@link ComponentArchitecture}.
	 * 
	 * @throws Exception
	 */
	private void addDataDependenciesToTaskGraph(DefaultTaskGraph taskGraph) throws Exception {

		// TODO: Currently, top-level components are considered as deployable components (=tasks)
		for(ISignalAdapter<?> channelAdapter : messages) {

			IDeployableComponentAdapter<?> sourceDeployableComponent = channelAdapter.getSource();
			Collection<IDeployableComponentAdapter<?>> targetDeployableComponents =
					channelAdapter.getTargets();

			if(sourceDeployableComponent != null && !(targetDeployableComponents.isEmpty())) {
				taskGraph.addDataDependency(sourceDeployableComponent, targetDeployableComponents,
						channelAdapter);
			}
		}
	}

	/**
	 * Creates the {@link PlatformCommunicationGraph} that is used to determine the routing of
	 * messages between {@link IDeployableComponentAdapter}s and their respective
	 * {@link IExecutionUnitAdapter}s.
	 */
	protected void setupPlatformGraph() {
		platformCommunicationGraph = new PlatformCommunicationGraph(this);
	}

	/**
	 * Setups the {#link emittedMessagesBySender} map which makes it possible to easily identify the
	 * {@link ISignalAdapter}s that are emitted by the given {@link IDeployableComponentAdapter}.
	 */
	protected void setupEmittedMessages() {
		emittedMessagesBySender =
				new HashMap<IDeployableComponentAdapter<C>, Collection<ISignalAdapter<S>>>();

		for(IDeployableComponentAdapter<C> currentSender : deployableComponents) {
			Collection<ISignalAdapter<S>> emittedMessages = new ArrayList<ISignalAdapter<S>>();
			for(ISignalAdapter<S> currentMessage : messages) {
				if(currentSender == currentMessage.getSource()) {
					emittedMessages.add(currentMessage);
				}
			}
			emittedMessagesBySender.put(currentSender, emittedMessages);
		}
	}

	/** Returns the logical architecture that is associated with this {@link SystemModelAdapter}. */
	public ILogicalArchitectureAdapter<?> getLogicalArchitecture() {
		return logicalArchitecture;
	}

	/**
	 * Returns the platform architecture (Middleware/HW) that is associated with this
	 * {@link SystemModelAdapter}.
	 */
	public IPlatformArchitectureAdapter<?> getPlatformArchitecture() {
		return platformArchitecture;
	}

	/**
	 * Returns the deployable (software) component from the original logical architecture. May
	 * contain "abstract" {@link Component}s.
	 */
	public Collection<IDeployableComponentAdapter<C>> getDeployableComponents() {
		return deployableComponents;
	}

	/**
	 * Returns the deployable (software) component from the original logical architecture that my be
	 * replicated. May contain "abstract" {@link Component}s.
	 */
	public Collection<IDeployableComponentAdapter<C>> getReplicableComponents() {
		return replicableComponents;
	}

	/**
	 * Returns the scheduleable (software) component from the original logical architecture. May
	 * contain "abstract" {@link Component}s.
	 */
	public Collection<IDeployableComponentAdapter<C>> getScheduleableComponents() {
		return schedulableComponents;
	}

	/**
	 * Returns the deployable (software) components that can be actually executed on
	 * {@link ExecutionUnit}s, i.e. software components that are backed by an implementation.
	 */
	public Collection<IDeployableComponentAdapter<C>> getRealizableDeployableComponents() {
		return realizableDeployableComponents;
	}

	// TODO: Consider implementing this method as a map, since this operation is run often?
	/**
	 * Returns the {@link IDeployableComponentAdapter} that is constructed form the given object.
	 * 
	 * @param originalEObject
	 *            The object for which the corresponding {@link IDeployableComponentAdapter} shall
	 *            be identified.
	 * @return The asociated {@link IDeployableComponentAdapter}, or {@code null} if none can be
	 *         identified.
	 */
	public synchronized IDeployableComponentAdapter<C>
			getDeployableComponentOf(EObject originalEObject) {
		// FIXME: HACK to work around the missing consideration of the Task Architecture ATM.
		if(originalEObject instanceof Task) {
			TaskArchitecture ta = (TaskArchitecture)((Task)originalEObject).getContainer();
			Collection<Component> cCol = getAllocatedComponentsTo(ta, (Task)originalEObject);
			originalEObject = cCol.iterator().next();
		}

		for(IDeployableComponentAdapter<C> deployableComponentAdapter : deployableComponents) {
			if(deployableComponentAdapter.getObject() == originalEObject) {
				return deployableComponentAdapter;
			}
		}
		return null;
	}

	/**
	 * Returns the {@link IDeployableComponentAdapter} that wraps the given {@link EObject}. Here,
	 * only those {@link IDeployableComponentAdapter} are searched which can be actually executed on
	 * {@link ExecutionUnit}s, i.e. "abstract" components are not considered.
	 */
	public synchronized IDeployableComponentAdapter<C>
			getRealizeableDeployableComponentOf(EObject originalEObject) {
		for(IDeployableComponentAdapter<C> deployableComponentAdapter : realizableDeployableComponents) {
			if(deployableComponentAdapter.getObject() == originalEObject) {
				return deployableComponentAdapter;
			}
		}
		return null;
	}

	/**
	 * Returns the {@link ISignalAdapter} corresponding the sender and receiver {@link Component}.
	 * Returns null if no channel exists between the two given components.
	 */
	public synchronized ISignalAdapter<S> getChannelAdapter(IDeployableComponentAdapter<C> source,
			IDeployableComponentAdapter<C> target) {
		// TODO: Currently, top-level components are considered as deployable components (=tasks)
		for(ISignalAdapter<S> currentChannel : messages) {
			if(currentChannel.getSource() == source &&
					currentChannel.getTargets().contains(target)) {
				return currentChannel;
			}
		}

		return null;
	}

	/**
	 * Returns the deployment targets, i.e. the execution units of the target platform,
	 * where deployable software components will be mapped to.
	 */
	public Collection<IExecutionUnitAdapter<E>> getDeploymentTargets() {
		return deploymentTargets;
	}

	/** Returns the {@link IMemoryUnitAdapter} present in the target platform. */
	public Collection<IMemoryUnitAdapter<M>> getMemoryUnits() {
		return memoryUnits;
	}

	/** Returns the gateway units present in the platform. */
	public Collection<IGatewayUnitAdapter<G>> getGatewayUnits() {
		return gatewayUnits;
	}

	/** Returns the systems TDMA arbitrated transmission unit */
	public Collection<ITransmissionUnitAdapter<TR>> getTransmissionUnits() {
		return transmissionUnits;
	}

	/** Returns all the {@link IResourceAdapter}s present in the platform. */
	public Collection<IResourceAdapter<?>> getAvailableResources() {
		return availableResources;
	}

	/**
	 * Returns the hardware resource (e.g., processor core) onto which a virtual resource
	 * (e.g. partition) is (statically) mapped.
	 */
	public IResourceAdapter<?> getHardwareResource(IResourceAdapter<?> virtualResource) {
		return virtualResourceToHardwareResourceMap.get(virtualResource);
	}

	/**
	 * Returns the {@link ISafetyFunctionAdapter}s constructed from the safety functions that are
	 * present in the input models.
	 */
	public Collection<ISafetyFunctionAdapter<C>> getSafetyFunctionAdapters() {
		if(safetyFunctionAdapters != null) {
			return safetyFunctionAdapters;
		}
		return Collections.emptySet();
	}

	/**
	 * Returns the {@link PlatformCommunicationGraph} required for routing of the messages located
	 * in different parts of the platform.
	 */
	public PlatformCommunicationGraph getPlatformCommunicationGraph() {
		return platformCommunicationGraph;
	}

	/**
	 * Returns the {@link TaskGraph}s of the demanded type that has been constructed by {@code this}
	 * {@link SystemModelAdapter}, if any exists. Otherwise, an empty collection is returned.
	 */
	@SuppressWarnings("unchecked")
	public <T extends TaskGraph> Collection<T> getTaskGraphsByType(Class<T> tgType) {
		if(DefaultTaskGraph.class.isAssignableFrom(tgType)) {
			return (Collection<T>)getTaskGraphs();
		} else if(AcyclicTaskGraph.class.isAssignableFrom(tgType)) {
			return (Collection<T>)getAcyclicTaskGraphs();
		}
		return Collections.emptySet();
	}

	/**
	 * Returns all independent {@link DefaultTaskGraph}s contained in this
	 * {@link SystemModelAdapter}.
	 */
	public Collection<? extends TaskGraph> getTaskGraphs() {
		return taskGraphs;
	}

	/**
	 * Returns all independent {@link DefaultTaskGraph}s contained in this
	 * {@link SystemModelAdapter}.
	 */
	public Collection<? extends TaskGraph> getAcyclicTaskGraphs() {
		return acyclicTaskGraphs;
	}

	/**
	 * Returns the {@link DefaultTaskGraph} in which a given deployable component (=task) is
	 * contained.
	 */
	public TaskGraph getTaskGraph(IDeployableComponentAdapter<?> deployableComponent) {
		return componentInGraph.get(deployableComponent);
	}

	/**
	 * Returns the {@link DefaultTaskGraph} in which a given deployable component (=task) is
	 * contained.
	 */
	public DefaultTaskGraph
			getAcyclicTaskGraph(IDeployableComponentAdapter<?> deployableComponent) {
		return componentInAcyclicGraph.get(deployableComponent);
	}

	/**
	 * Returns the {@link ISignalAdapter}s "emitted" by the given
	 * {@link IDeployableComponentAdapter}. Returns an empty collection if the given
	 * {@link IDeployableComponentAdapter} does not emit anything.
	 */
	public Collection<ISignalAdapter<S>>
			getEmmittedMessages(IDeployableComponentAdapter<?> sender) {
		// TODO: use another representation ro make an access via the reference obsolete
		if(sender.isInstantiation()) {
			sender = sender.getReplacedComponent();
		}
		// TODO: replace by map or similar.
		IDeployableComponentAdapter<?> origSender =
				getDeployableComponentOf((EObject)sender.getObject());
		return emittedMessagesBySender.get(origSender);
	}

	/**
	 * Returns the hyper period of all tasks in the system, i.e. the hyper-period of all independent
	 * {@link DefaultTaskGraph}s.
	 */
	public Double getHyperPeriod() {
		return hyperPeriod;
	}

	/** Returns whether the given {@link Component} is a deployable {@link Component}. */
	public boolean isDeployableComponent(Component component) {
		return deployableComponentObjects.contains(component);
	}

	/**
	 * Returns whether the given {@link Port} is attached to a deployable {@link Component}, i.e.
	 * its temporal behavior must be defined.
	 */
	public boolean isDeployablePort(Port port) {
		return portSignalTypes.get(port) != null;
	}

	/** Returns the input output behavior of the given platform connector. */
	public ConnectionType getConnectionType(Class<? extends IConnector> connector) {
		return connectionTypeLUT.get(connector);
	}

	// ===============================================================
	// To be removed / generalized
	// ===============================================================

	/**
	 * Wrapper for {@link #getEffectiveConnectionTargets(IConnector, Map, Collection)} that disables
	 * the traversal between the physical and the virtual platform.
	 */
	protected Collection<IConnector> getEffectiveConnectionTargets(IConnector sourceConnector,
			Collection<ExecutionUnit> deploymentTargets) throws Exception {
		return getEffectiveConnectionTargets(sourceConnector, null, deploymentTargets);
	}

	/**
	 * <p>
	 * Determines the effective target {@link IConnector}s of a given "source" {@link IConnector}.
	 * Elements/{@link IConnector}s that are irrelevant to communication within the platform are not
	 * added to the target list. Instead, the connectors that are accessible by the source
	 * {@link IConnector} are traversed until relevant targets are identified, like
	 * {@link ExecutionUnit}s selected for the deployment.
	 * </p>
	 * <p>
	 * <b>NOTE:</b> This method may return targets that are not directly involved in the
	 * communication between {@link ExecutionUnit}s. However, 'hull' elements (or similar elements)
	 * that may "disturb" the communication between {@link ExecutionUnit}s will not appear.
	 * </p>
	 * <p>
	 * <b>Implementation Note:</b> The <i>valid</i> targets are determined via path walk following
	 * all connections from the source {@link IConnector} until <i>valid</i> targets are found.
	 * </p>
	 * 
	 * @throws Exception
	 *             if the references between the {@link IModelElement}s are incorrectly specified.
	 */
	protected Collection<IConnector> getEffectiveConnectionTargets(IConnector sourceConnector,
			Map<IModelElement, IModelElement> invElementReferences,
			Collection<ExecutionUnit> af3DeploymentTargets) throws Exception {
		Collection<IConnector> targetConnectors = new HashSet<IConnector>();
		if(sourceConnector == null) {
			return targetConnectors;
		}

		LinkedList<IConnector> possibleTargetConnectors = new LinkedList<IConnector>();
		Collection<IConnector> visitedTargetConnectors = new HashSet<IConnector>();

		possibleTargetConnectors.add(sourceConnector);
		while(!possibleTargetConnectors.isEmpty()) {
			IConnector nextTargetConnector = possibleTargetConnectors.poll();

			// This section handles the traversal between the elements of the same platform
			// architecture.
			targetConnectors.addAll(
					seekIntraPlatformTargetConnectors(sourceConnector, possibleTargetConnectors,
							visitedTargetConnectors, nextTargetConnector, af3DeploymentTargets));

			// This section handles the traversal between a virtual and a hardware platform
			// architecture.
			if(invElementReferences != null) {
				targetConnectors.addAll(seekInterPlatformConnectors(nextTargetConnector,
						invElementReferences, possibleTargetConnectors, af3DeploymentTargets));
			}

			visitedTargetConnectors.add(nextTargetConnector);
		}

		return targetConnectors;
	}

	/**
	 * Identifies those {@link IConnector}s that are actual target {@link IConnector}s of data sent
	 * out by the given {@code sourceConnector}. This methos is needed as in a
	 * {@link PlatformArchitecture}, there exist {@link IConnector}s which are attached to container
	 * elements that are not backed by physical interface ports (These are represented by the
	 * contained elements).
	 * 
	 * @param sourceConnector
	 *            Source {@link IConnector} for which the target {@link IConnector}s need to be
	 *            identified.
	 * @param possibleTargetConnectors
	 *            List to which those {@link IConnector}s are added that are directly connected to
	 *            the {@code sourceConnector} but are no "communication targets" as described above
	 *            an need further traversal.
	 * @param visitedTargetConnectors
	 *            Set of already traversed connectors (to avoid loops).
	 * @param nextTargetConnector
	 *            Current {@link IConnector} whose target {@link IConnector}s need to be searched.
	 * @return Collection of {@link IConnector}s that are actual target {@link IConnector}s of the
	 *         {@code sourceConnector}.
	 */
	protected Collection<IConnector> seekIntraPlatformTargetConnectors(IConnector sourceConnector,
			LinkedList<IConnector> possibleTargetConnectors,
			Collection<IConnector> visitedTargetConnectors, IConnector nextTargetConnector,
			Collection<ExecutionUnit> deploymentTargets) {
		Collection<IConnector> targetConnectors = new ArrayList<IConnector>();

		Collection<IConnection> attachedConnections = new ArrayList<IConnection>();
		attachedConnections.addAll(nextTargetConnector.getOutgoing());
		attachedConnections.addAll(nextTargetConnector.getIncoming());

		for(IConnection connToTarget : attachedConnections) {
			IConnector connectorTgt = null;
			IConnector connectorSrc = null;
			if(connToTarget.getTarget().getOwner() != nextTargetConnector.getOwner()) {
				connectorTgt = connToTarget.getTarget();
				connectorSrc = connToTarget.getSource();
			}
			if(connToTarget.getSource().getOwner() != nextTargetConnector.getOwner()) {
				connectorTgt = connToTarget.getSource();
				connectorSrc = connToTarget.getTarget();
			}
			if(connectorTgt == null || connectorSrc == null) {
				throw new RuntimeException(
						"Error: \"floating\" connection found during transformation for the DSE.");
			}

			IHierarchicElement tgtElement = connectorTgt.getOwner();
			// Do not traverse back (to same connector) & not to the same IHierarchicElement.
			if(!visitedTargetConnectors.contains(connectorTgt) &&
					tgtElement != sourceConnector.getOwner()) {
				if(tgtElement instanceof IPlatformArchitectureElement && isValidConnectionTarget(
						(IPlatformArchitectureElement)tgtElement, deploymentTargets)) {
					targetConnectors.add(connectorTgt);
				} else if(tgtElement == connectorSrc.getOwner().getContainer() &&
						!(nextTargetConnector instanceof IPlatformExport)) {
					// Treatment of hull ports: Do not walk a path that would just traverse a 'hull'
					// port. Hence, just consider the ports to the "outside" (upper level).
					for(IConnection currConn : connectorTgt.getOutgoing()) {
						IConnector currTgtConnector = currConn.getTarget();
						if(currTgtConnector.getOwner().getContainer() != tgtElement) {
							if(tgtElement instanceof IPlatformArchitectureElement &&
									isValidConnectionTarget(
											(IPlatformArchitectureElement)currTgtConnector
													.getOwner(),
											deploymentTargets)) {
								targetConnectors.add(currTgtConnector);
							} else {
								possibleTargetConnectors.add(currTgtConnector);
							}
						}
					}

					for(IConnection currConn : connectorTgt.getIncoming()) {
						IConnector currTgtConnector = currConn.getSource();
						if(currTgtConnector.getOwner().getContainer() != tgtElement) {
							if(tgtElement instanceof IPlatformArchitectureElement &&
									isValidConnectionTarget(
											(IPlatformArchitectureElement)currTgtConnector
													.getOwner(),
											deploymentTargets)) {
								targetConnectors.add(currTgtConnector);
							} else {
								possibleTargetConnectors.add(currTgtConnector);
							}
						}
					}

					visitedTargetConnectors.add(connectorTgt);
				} else {
					possibleTargetConnectors.add(connectorTgt);
				}
			}
		}

		return targetConnectors;
	}

	/**
	 * Determines those {@link IConnector}s (in a different {@link PlatformArchitecture} than the
	 * {@code sourceConnector}) that establish a link between between two elements of different
	 * {@link PlatformArchitecture}s and which are connected by a {@link ResourceLink}.
	 * This method is agnostic of the "direction" of the {@link ResourceLink}, i.e. it does not
	 * matter whether the {@code sourceConnector} is attached to an
	 * {@link IPlatformArchitectureElement} to which the {@link ResourceLink} is annotated or is
	 * referenced by virtual link.
	 * <p>
	 * <b>NOTE:</b> The current implementation assumes that there exist only n:1 links between
	 * {@link IModelElement}s that establish a "communication path" between different platforms
	 * (virtual -> physical).
	 * 
	 * @param sourceConnector
	 *            {@link IConnector} for which the target {@link IConnector}s in another platform
	 *            have to be identified.
	 * @param invElementReferences
	 *            Map representing all 1:1 inverse {@link ResourceLink}s.
	 * @param possibleTargetConnectors
	 *            List of {@link IConnector}s to which those {@link IConnector}s are added that are
	 *            located in referenced {@link PlatformArchitecture} but need to be traversed
	 *            further such that valid "communication targets" are identified.
	 * 
	 * @return List of {@link IConnector}s that a message actually traverses when being sent by the
	 *         {@code sourceConnector}.
	 * @throws Exception
	 *             if the references between the {@link IModelElement}s are incorrectly specified.
	 */
	// TODO: Throw appropriate Exceptions on failure, i.e. wrong types of referenced elements etc.
	protected Collection<IConnector> seekInterPlatformConnectors(IConnector sourceConnector,
			Map<IModelElement, IModelElement> invElementReferences,
			LinkedList<IConnector> possibleTargetConnectors,
			Collection<ExecutionUnit> deploymentTargets) throws Exception {
		Collection<IConnector> identifiedTargets = new ArrayList<IConnector>();

		if(connectsSwHwPlatform(sourceConnector)) {
			// Special treatment for the connection between the virtual
			// platform and the physical platform.
			IPlatformArchitectureElement referencingElement =
					(IPlatformArchitectureElement)sourceConnector.getOwner();
			EList<IPlatformArchitectureElement> referencedElements =
					PlatformArchitectureUtils.getReferencedElementsWithType(referencingElement,
							IPlatformArchitectureElement.class);
			if(referencedElements.size() == 0) {
				throw new Exception(
						"No Network Interface is referenced in the ResourceLink annotation of " +
								referencingElement.getName() +
								". Please specify at least one reference.");
			}

			for(IPlatformArchitectureElement currRefElem : referencedElements) {
				if(sourceConnector.getOwner() instanceof IPlatformArchitectureElement &&
						isValidConnectionTarget(
								(IPlatformArchitectureElement)sourceConnector.getOwner(),
								deploymentTargets)) {
					identifiedTargets.add(pickFirstInstanceOf(IPlatformExport.class,
							currRefElem.getConnectors()));
				} else {
					possibleTargetConnectors.addAll(currRefElem.getPlatformConnectorUnits());
				}
			}
		} else if(invElementReferences.get(sourceConnector.getOwner()) != null) {
			// Special treatment for the connection between the physical platform
			// and the virtual platform.
			IModelElement referencedElement = invElementReferences.get(sourceConnector.getOwner());
			assert (referencedElement instanceof IPlatformArchitectureElement) : "A non-platform model element has been referenced.";
			IPlatformArchitectureElement referencedPlatformElement =
					(IPlatformArchitectureElement)referencedElement;
			IPlatformExport driverExport = pickFirstInstanceOf(IPlatformExport.class,
					referencedPlatformElement.getConnectors());

			if(driverExport.getOwner() instanceof IPlatformArchitectureElement &&
					isValidConnectionTarget((IPlatformArchitectureElement)driverExport.getOwner(),
							deploymentTargets)) {
				identifiedTargets.add(driverExport);
			} else {
				// TODO: dead code path? Is not called for any recent model.
				possibleTargetConnectors
						.addAll(referencedPlatformElement.getPlatformConnectorUnits());
			}
		}

		return identifiedTargets;
	}

	/**
	 * Determines whether the given connector (or its containing element) is used to establish a
	 * connection between separate {@link PlatformArchitecture}s. This method is capable of
	 * identifying connections between different {@link PlatformArchitecture} even if the links are
	 * defined only for the elements of {@link PlatformArchitecture}. This feature requires a map
	 * that tracks the link between the referenced elements and the elements containing the link.
	 * 
	 * @param connector
	 *            {@link IConnection} to be examined for its role of connecting different
	 *            {@link PlatformArchitecture}s.
	 * @param invNetDriverRef
	 *            Map tracking the inverse link of inter-{@link PlatformArchitecture} links.
	 * @return Flag, if the given connector (or its containing {@link IModelElement}) connects model
	 *         elements form different {@link PlatformArchitecture}s.
	 */
	protected boolean connectsSwHwPlatform(IConnector connector,
			Map<IModelElement, IModelElement> invNetDriverRef) {
		if(!connectsSwHwPlatform(connector)) {
			if(connector != null) {
				IHierarchicElement parent = connector.getOwner();
				// Second: physical counter part of inter-platform connection.
				if(platformInterconnectTypes != null &&
						platformInterconnectTypes.getSecond().isAssignableFrom(parent.getClass()) &&
						invNetDriverRef.get(parent) != null) {
					return true;
				}
			}
			return false;
		}
		return true;
	}

	/**
	 * Determines whether the given connector is used to establish a connection between separate
	 * {@link PlatformArchitecture}s.
	 * <p>
	 * <b>NOTE:</b> This method only works correctly if the given {@link IConnector} is attached to
	 * a {@link IModelElement} that has an explicit reference to an {@link IModelElement} of another
	 * {@link PlatformArchitecture}. Otherwise, use {#link connectsSwHwPlatform(IConnector,
	 * Map<NetworkInterface, OnChipNetworkDriver>)}.
	 */
	protected boolean connectsSwHwPlatform(IConnector connector) {
		if(connector instanceof IPlatformExport && platformInterconnectTypes.getFirst()
				.isAssignableFrom(connector.getOwner().getClass())) {
			return true;
		}
		return false;
	}

	/**
	 * Returns whether the given {@link IHierarchicElement} is a valid connection end-point. These
	 * end-points are either {@link ExecutionUnit}s selected as deployment targets for
	 * {@link Component}s or {@link IPlatformCommunicationResource}s whose internal structure is not
	 * modeled.
	 */
	protected boolean isValidConnectionTarget(IPlatformArchitectureElement element,
			Collection<ExecutionUnit> deploymentTargets) {
		if(element instanceof IPlatformResource && deploymentTargets.contains(element)) {
			return true;
		} else if(element instanceof IPlatformCommunicationResource &&
				element.getContainedElements().isEmpty()) {
			return true;
		}
		return false;
	}

	/** Returns the fault containment region specified for the target platform. */
	public Class<? extends IPlatformResource> getFaultContainmentRegion() {
		return faultContainmentRegion;
	}
}
