/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.sysmodel.arch.af3;

import java.util.Collection;
import java.util.Iterator;

import org.fortiss.af3.component.model.Channel;
import org.fortiss.af3.component.model.InputPort;
import org.fortiss.af3.component.model.OutputPort;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IDeployableComponentAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ISignalAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITransmissionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.comm.Message;
import org.fortiss.af3.exploration.backend.IDseInputContainer.SignalType;

/**
 * Adapter between the AF3 {@link Channel} and the DSE.
 * 
 * @author barner
 */
public class AF3SignalAdapter extends AF3RequestAdapter<OutputPort>
		implements ISignalAdapter<OutputPort> {
	/** Conversion constant for the message size. */
	private static final long MBYTES_TO_BYTES = 1000 * 1000;

	/** Message's source component */
	IDeployableComponentAdapter<?> source;

	/** Message's target component adapter */
	Collection<IDeployableComponentAdapter<?>> targets;

	/** Message's target AF3 ports. */
	Collection<InputPort> targetPorts;

	/** Message size annotation */
	long messageSize;

	/**
	 * Contains the {#link SignalType} of this {@link ISignalAdapter}. Defaults to {#link
	 * SignalType.PERIODIC}.
	 */
	SignalType signalType;

	/**
	 * Creates an {@link ISignalAdapter} for an {@link Channel}.
	 * 
	 * @throws Exception
	 */
	public AF3SignalAdapter(OutputPort senderPort, IDeployableComponentAdapter<?> source,
			Collection<InputPort> targetPorts, Collection<IDeployableComponentAdapter<?>> targets,
			Long messageSize) throws Exception {
		super(senderPort);

		this.source = source;
		this.targets = targets;
		this.targetPorts = targetPorts;

		if(messageSize != null) {
			this.messageSize = messageSize;
		} else {
			throw new Exception("Port " + senderPort + " does not have a message size annotation.");
		}

		signalType = SignalType.PERIODIC;
	}

	/** {@inheritDoc} */
	@Override
	public IDeployableComponentAdapter<?> getSource() {
		return source;
	}

	/**
	 * Adds a target to the target list of the channel adapter. Compatibility layer for AF3 /wo
	 * multicast patch.
	 */
	public void addTarget(IDeployableComponentAdapter<?> target) {
		targets.add(target);
	}

	/** {@inheritDoc} */
	@Override
	public Collection<IDeployableComponentAdapter<?>> getTargets() {
		// TODO: use another representation ro make an access via the reference obsolete
		return targets;
	}

	/**
	 * Returns the AF3 {@link OutputPort} that is responsible for emitting the {@link Message} (if
	 * required).
	 */
	public OutputPort getSenderPort() {
		// Known by construction.
		return getObject();
	}

	/**
	 * Returns the AF3 {@link InputPort} that is that receives the {@link Message} (if required).
	 */
	public Collection<InputPort> getTargetPorts() {
		return targetPorts;
	}

	/** {@inheritDoc} */
	@Override
	public long getSignalSize() {
		return messageSize;
	}

	/** {@inheritDoc} */
	@Override
	public double getTransmissionTime(ITransmissionUnitAdapter<?> transmissionUnit) {
		return transmissionUnit.getTransmissionDuration(messageSize / MBYTES_TO_BYTES);
	}

	/** {@inheritDoc} */
	@Override
	public String getName() {
		String targetNameList = "";
		for(Iterator<IDeployableComponentAdapter<?>> targetIterator =
				targets.iterator(); targetIterator.hasNext();) {
			targetNameList += targetIterator.next().getName();
			if(targetIterator.hasNext()) {
				targetNameList += ", ";
			}
		}
		return source.getName() + "->{" + targetNameList + "}";
	}

	/**
	 * Sets the {#link SignalType} of this {@link AF3SignalAdapter}.
	 */
	public void setSignalType(SignalType signalType) {
		this.signalType = signalType;
	}

	/** {@inheritDoc} */
	@Override
	public SignalType getSignalType() {
		return signalType;
	}
}
