/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.sysmodel.arch.af3;

import static org.fortiss.af3.deployment.utils.DeploymentParameterUtils.getDeploymentsWithParametersOf;
import static org.fortiss.af3.exploration.alg.util.AF3Utils.getConnectionTypeDirection;
import static org.fortiss.af3.exploration.alg.util.AF3Utils.getInverseIntraPlatformReferences;
import static org.fortiss.af3.exploration.alg.util.AF3Utils.isAnyChildDeploymentTarget;
import static org.fortiss.af3.exploration.alg.util.AF3Utils.isComponentOrPortsDeployable;
import static org.fortiss.af3.exploration.alg.util.ExplorationAlgUtils.findTargetPorts;
import static org.fortiss.af3.exploration.alg.util.ExplorationAlgUtils.getDeployableOutputPorts;
import static org.fortiss.af3.platform.utils.PlatformArchitectureUtils.getPhysicalPlatformArchitecture;
import static org.fortiss.af3.platform.utils.PlatformArchitectureUtils.getReferencedElementsWithType;
import static org.fortiss.af3.schedule.utils.MathUtils.gcdDoubleCollection;
import static org.fortiss.tooling.base.ui.annotation.view.AnnotationViewPartBase.isUpdateEnabled;
import static org.fortiss.tooling.base.ui.annotation.view.AnnotationViewPartBase.setUpdateEnabled;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getChildrenWithType;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.pickFirstInstanceOf;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.pickInstanceOf;
import static org.fortiss.tooling.kernel.utils.KernelModelElementUtils.getParentElement;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.conqat.lib.commons.collections.Pair;
import org.fortiss.af3.component.model.Channel;
import org.fortiss.af3.component.model.Component;
import org.fortiss.af3.component.model.ComponentArchitecture;
import org.fortiss.af3.component.model.InputPort;
import org.fortiss.af3.component.model.OutputPort;
import org.fortiss.af3.component.model.Port;
import org.fortiss.af3.deployment.model.Deployment;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IDeployableComponentAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IExecutionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IGatewayUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IMemoryUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceConnectionAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceConnectionAdapter.ConnectionType;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ISignalAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITransmissionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.InternalComponentParameters;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemModelAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemParameterContainer;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.af3.helper.AF3DeploymentParameterMapAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph.TaskGraph;
import org.fortiss.af3.exploration.alg.exception.InvalidPlatformModelException;
import org.fortiss.af3.platform.model.ExecutionUnit;
import org.fortiss.af3.platform.model.GatewayUnit;
import org.fortiss.af3.platform.model.IPlatformArchitectureElement;
import org.fortiss.af3.platform.model.IPlatformExport;
import org.fortiss.af3.platform.model.IPlatformResource;
import org.fortiss.af3.platform.model.MemoryUnit;
import org.fortiss.af3.platform.model.PlatformArchitecture;
import org.fortiss.af3.platform.model.TransmissionUnit;
import org.fortiss.tooling.base.model.element.IConnector;
import org.fortiss.tooling.base.model.element.IHierarchicElement;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.kernel.model.IProjectRootElement;
import org.fortiss.tooling.kernel.utils.EcoreUtils;

import com.google.common.collect.HashMultimap;
import com.google.common.collect.Multimap;

/**
 * {@link SystemModelAdapter} which abstracts AutoFOCUS3 models for the DSE.
 * 
 * @author diewald
 */
public class AF3SystemModelAdapter extends
		SystemModelAdapter<Component, ExecutionUnit, OutputPort, GatewayUnit, TransmissionUnit, MemoryUnit> {

	/** References the AF3 {@link PlatformArchitecture} that is adapted. */
	private PlatformArchitecture hardwarePlatformArchitecture;

	/**
	 * References the {@link AF3DeploymentParameterMapAdapter} providing access to the deployment
	 * specific parameters.
	 */
	private AF3DeploymentParameterMapAdapter parameterAdapter;

	/**
	 * Collection of {@link IDeployableComponentAdapter} which can be used to instantiate "abstract"
	 * {@link IDeployableComponentAdapter}s.
	 */
	private Collection<IDeployableComponentAdapter<Component>> componentReplacementAdapters;

	// TODO (2x): package access rights are ok? If so, please add /* package */
	/** Contains informations whether the given AF3 model supports multicast channels. */
	boolean multicastSupport = false;

	/** References the method which returns the multicast targets (if available). */
	Method multicastGetTargets;

	/** Cache the Annotation View's update status. */
	private boolean annotationUpdateEnabled;

	/**
	 * Constructor.
	 * 
	 * @param componentArchitecture
	 *            main logical architecture to deploy.
	 * @param platformArchitecture
	 *            main target platform.
	 * @param systemParameters
	 *            required and optional parameters for the DSE.
	 * @throws Exception
	 *             if the creation of the system model adapter fails due to missing inputs or
	 *             inconsistent specifications.
	 */
	public AF3SystemModelAdapter(ComponentArchitecture componentArchitecture,
			PlatformArchitecture platformArchitecture, SystemParameterContainer systemParameters)
			throws Exception {

		checkMulticastSupport();

		this.logicalArchitecture = new AF3LogicalArchitectureAdapter(componentArchitecture);
		this.platformArchitecture = new AF3PlatformArchitectureAdapter(platformArchitecture);
		this.deployableComponentObjects = systemParameters.getDeployableComponents();
		this.connectionTypeLUT = systemParameters.getConnectionTypeLUT();
		this.portSignalTypes = systemParameters.getPortSignalTypes();
		this.platformInterconnectTypes = systemParameters.getPlatfromInterconnectTypes();
		this.faultContainmentRegion = systemParameters.getFaultContainmentRegion();
		putInputModel(ComponentArchitecture.class, componentArchitecture);
		putInputModel(PlatformArchitecture.class, platformArchitecture);

		// Disable any annotation view update.
		annotationUpdateEnabled = isUpdateEnabled();
		setUpdateEnabled(false);

		// Determine the associated hardware platform given the platform architecture is a a model
		// of a virtual platform.
		hardwarePlatformArchitecture = getPhysicalPlatformArchitecture(platformArchitecture);

		this.deploymentTargets = createDeploymentTargets(systemParameters.getDeploymentTargets(),
				systemParameters.getResourceFailureProbabilities());

		memoryUnits = createMemoryUnits(platformArchitecture,
				systemParameters.getResourceFailureProbabilities());
		gatewayUnits = createGatewayUnits(platformArchitecture,
				systemParameters.getResourceFailureProbabilities(),
				systemParameters.getResourcePowerConsumption());
		transmissionUnits = createTransmissionUnits(platformArchitecture,
				systemParameters.getResourceFailureProbabilities(),
				systemParameters.getResourcePowerConsumption());
		if(platformArchitecture != hardwarePlatformArchitecture) {
			memoryUnits.addAll(createMemoryUnits(hardwarePlatformArchitecture,
					systemParameters.getResourceFailureProbabilities()));
			gatewayUnits.addAll(createGatewayUnits(hardwarePlatformArchitecture,
					systemParameters.getResourceFailureProbabilities(),
					systemParameters.getResourcePowerConsumption()));
			transmissionUnits.addAll(createTransmissionUnits(hardwarePlatformArchitecture,
					systemParameters.getResourceFailureProbabilities(),
					systemParameters.getResourcePowerConsumption()));
		}

		availableResources = createAvailableResources();

		parameterAdapter = createDeploymentParameterMap(componentArchitecture, platformArchitecture,
				systemParameters.getReplacementComponents());
		deploymentTargetConnectors =
				createDeploymentTargetConnectors(systemParameters.getDeploymentTargets());
		Multimap<Component, IDeployableComponentAdapter<Component>> componentToReplacementAdapter =
				createComponentReplacementAdapters(systemParameters.getReplacementComponents());
		deployableComponents = createDeployableComponents(componentToReplacementAdapter,
				systemParameters.getComponentPeriodMap(),
				systemParameters.getComponentReplicationBounds());
		componentReplacementAdapters = componentToReplacementAdapter.values();
		replicableComponents = new HashSet<>(deployableComponents);
		realizableDeployableComponents = createRealizableDeployableComponents(deployableComponents);
		Collection<ISignalAdapter<OutputPort>> unsortedMessages =
				createSignalAdapters(systemParameters.getPortMessageSizes());
		messages = sortMessagesByExpectedOrder(unsortedMessages,
				systemParameters.getComponentPeriodMap());

		virtualResourceToHardwareResourceMap = createVirtualResourceToHardwareResourceMap();

		setupScheduleableComponents();
		setupTaskGraphs();
		setupEmittedMessages();
		setupPlatformGraph();
		createAbstractTaskGraphEncoding();

		// Re-enable annotation view updates.
		setUpdateEnabled(annotationUpdateEnabled);
	}

	/** Determines whether the current AF3 version supports multicast channels. */
	private void checkMulticastSupport() {
		Method[] methodsOfChannel = Channel.class.getMethods();
		for(Method currentChannelMethod : methodsOfChannel) {
			if(currentChannelMethod.getName().equals("getTargets")) {
				multicastSupport = true;
				multicastGetTargets = currentChannelMethod;
			}
		}
	}

	/**
	 * Creates a collection of {@link IDeployableComponentAdapter}s that are available to
	 * instantiate other "abstract" {@link IDeployableComponentAdapter}s. Only those
	 * {@link IDeployableComponentAdapter}s which are referenced by {@link Component}s of the input
	 * {@link ComponentArchitecture} are included in the collection.
	 */
	private Multimap<Component, IDeployableComponentAdapter<Component>>
			createComponentReplacementAdapters(Multimap<Component, Component> replacementComponents)
					throws Exception {
		Multimap<Component, IDeployableComponentAdapter<Component>> rval = HashMultimap.create();
		Collection<Component> allReplacementComponents = replacementComponents.values();
		Map<Component, InternalComponentParameters> internalParameterMap =
				createInternalParameterMaps(allReplacementComponents);

		for(Component curComponent : replacementComponents.keySet()) {
			for(Component curReplacementComponent : replacementComponents.get(curComponent)) {
				rval.put(curComponent, new AF3DeployableComponentAdapter(curReplacementComponent,
						internalParameterMap));
			}
		}

		return rval;
	}

	/**
	 * Determines the deployable components from an AF3 {@link ComponentArchitecture}.
	 * 
	 * @throws Exception
	 */
	private Collection<IDeployableComponentAdapter<Component>> createDeployableComponents(
			Multimap<Component, IDeployableComponentAdapter<Component>> componentReplacementAdapters,
			Map<Component, Double> componentPeriodMap,
			Map<Component, Pair<Integer, Integer>> componentReplicationBounds) throws Exception {
		Collection<IDeployableComponentAdapter<Component>> rval = new HashSet<>();

		Map<Component, InternalComponentParameters> internalParameterMap =
				createInternalParameterMaps(deployableComponentObjects);

		for(Component currentComponent : deployableComponentObjects) {
			Double period = componentPeriodMap.get(currentComponent);
			// TODO: The period was allowed to be null for enabling the VL routing?
			// if (period == null) {
			// throw new Exception("The Period for the deployable component " + currentComponent +
			// " has not been defined.");
			// }
			Pair<Integer, Integer> replicationBounds =
					componentReplicationBounds.get(currentComponent);
			if(replicationBounds == null || replicationBounds.getFirst() == null ||
					replicationBounds.getSecond() == null) {
				throw new Exception("The replication bounds for the deployable component " +
						currentComponent + " have not been defined properly.");
			}

			rval.add(new AF3DeployableComponentAdapter(currentComponent, internalParameterMap,
					replicationBounds.getFirst(), replicationBounds.getSecond(), period,
					componentReplacementAdapters.get(currentComponent)));
		}

		return rval;
	}

	/**
	 * Creates a collection of {@link IDeployableComponentAdapter}s which can be instantiated to run
	 * on execution units.
	 * 
	 * @throws Exception
	 *             If the {@link IDeployableComponentAdapter} cannot be created for one of the
	 *             Components that are concrete implementations of "abstract" {@link Component}s
	 *             defined in the {@link ComponentArchitecture}.
	 */
	private Collection<IDeployableComponentAdapter<Component>> createRealizableDeployableComponents(
			Collection<IDeployableComponentAdapter<Component>> deployableComponents)
			throws Exception {
		Collection<IDeployableComponentAdapter<Component>> rval =
				new ArrayList<IDeployableComponentAdapter<Component>>();

		rval.addAll(componentReplacementAdapters);

		for(IDeployableComponentAdapter<Component> deployableComponent : deployableComponents) {
			if(!deployableComponent.isInstantiationRequired()) {
				rval.add(deployableComponent);
			}
		}

		return rval;
	}

	/**
	 * Collects the deployment-specific parameters (namely the WCET and energy consumption values)
	 * of each given component using the {@link AF3DeploymentParameterMapAdapter} and transforms
	 * them into a {@link InternalComponentParameters} representation.
	 */
	private Map<Component, InternalComponentParameters>
			createInternalParameterMaps(Collection<Component> deployableComponents) {
		if(parameterAdapter == null) {
			return Collections.emptyMap();
		}

		Map<Component, InternalComponentParameters> internalParameterMap =
				new HashMap<Component, InternalComponentParameters>();

		for(Component currentComponent : deployableComponents) {
			Map<IExecutionUnitAdapter<?>, Double> wcetMap =
					new HashMap<IExecutionUnitAdapter<?>, Double>();
			Map<IExecutionUnitAdapter<?>, Double> energyConsumptionMap =
					new HashMap<IExecutionUnitAdapter<?>, Double>();
			for(IExecutionUnitAdapter<?> execUnit : deploymentTargets) {
				wcetMap.put(execUnit, parameterAdapter.getWcet(currentComponent, execUnit));
				// energyConsumptionMap.put(execUnit,
				// parameterAdapter.getEnergyConsumption(currentComponent, execUnit));
			}

			InternalComponentParameters parameterMap = new InternalComponentParameters();
			parameterMap.setWcetMap(wcetMap);
			parameterMap.setEnergyConsumptionMap(energyConsumptionMap);
			internalParameterMap.put(currentComponent, parameterMap);
		}
		return internalParameterMap;
	}

	/**
	 * Propagates the periods defined in a timing model to other deployable components that are
	 * required to maintain the data dependencies of the logical architecture.
	 * This method first propagates the defined periods along the direction of the channels to the
	 * "successor" {@link Component}s. Then, the periods are propagated from those {@link Component}
	 * s for which there exists a period now in the opposite direction, i.e. along the opposite
	 * direction of the channels.
	 * <b>Note:</b> The propagation of periods towards the inputs is dangerous as it "hides"
	 * under-specified models. Instead, a model constraint check should be performed before.
	 * 
	 * @param deployableComponents
	 * @return Map that contains the
	 * @throws Exception
	 *             if the timing specification of any traversed element is inconsistent.
	 */
	public Map<Component, Double> propagatePeriods(Map<Component, Double> deployableComponents)
			throws Exception {
		Map<Component, Double> componentPeriodMap = new HashMap<Component, Double>();
		// TODO: Shouldn't be implemented as a Map containing a collection; Instead a own object
		// should be used (not done here due to method being static).
		Map<Component, Collection<Double>> possibleComponentPeriods =
				new HashMap<Component, Collection<Double>>();
		for(Entry<Component, Double> deployableComponent : deployableComponents.entrySet()) {
			Collection<Double> periodAsList = new ArrayList<Double>();
			periodAsList.add(deployableComponent.getValue());
			possibleComponentPeriods.put(deployableComponent.getKey(), periodAsList);
		}

		// First iterate over all successor components. The period of successor (and predecessor
		// components) is determined by the minimal period that is propagated to components that are
		// not referenced by a periodic constraint.
		propagatePeriodsToSuccessors(deployableComponents, possibleComponentPeriods);
		determineEffectivePeriodOfComponents(componentPeriodMap, possibleComponentPeriods);

		// TODO: remove these methods. As agreed some time ago, propagation of periods to
		// predecessor Components is only needed for under-specified systems, so remove this logic.
		// propagatePeriodsToPredecessors(componentPeriodMap, possibleComponentPeriods);
		// determineEffectivePeriodOfComponents(componentPeriodMap, possibleComponentPeriods);

		componentPeriodMap.putAll(deployableComponents);

		return componentPeriodMap;
	}

	/**
	 * Propagates periods defined in periodic constraints to successor components as defined in the
	 * corresponding {@link ComponentArchitecture}. The successor {@link Component}s are associated
	 * with a list of periods that are propagated to them.
	 * 
	 * @throws Exception
	 *             if the timing specification of any traversed element is inconsistent.
	 */
	private void propagatePeriodsToSuccessors(Map<Component, Double> deployableComponents,
			Map<Component, Collection<Double>> possibleComponentPeriods) throws Exception {
		Set<Component> workingQueue = new HashSet<>();
		Set<Port> traversedInputPorts = new HashSet<>();
		workingQueue.addAll(deployableComponents.keySet());
		while(!workingQueue.isEmpty()) {
			Component currentComponent = (Component)workingQueue.toArray()[0];
			workingQueue.remove(currentComponent);
			for(OutputPort portToSuccessor : currentComponent.getOutputPorts()) {
				for(Port successorPort : findTargetPorts(portToSuccessor,
						deployableComponents.keySet())) {
					if(!traversedInputPorts.contains(successorPort)) {
						traversedInputPorts.add(successorPort);
						Component successorComponent = successorPort.getComponent();
						// Successor is not referenced by a periodic constraint
						if(!deployableComponents.containsKey(successorComponent)) {
							Collection<Double> possiblePeriods =
									possibleComponentPeriods.get(successorComponent);
							// If the predecessor component has not been considered yet, a list for
							// the possible periods must be created.
							if(possiblePeriods == null) {
								possiblePeriods = new ArrayList<Double>();
							}
							possiblePeriods.addAll(possibleComponentPeriods.get(currentComponent));
							possibleComponentPeriods.put(successorComponent, possiblePeriods);
							workingQueue.add(successorComponent);
						}
					}
				}
			}
		}
	}

	/**
	 * Propagates periods defined in periodic constraints to successor components as defined in the
	 * corresponding {@link ComponentArchitecture}. The successor {@link Component}s are associated
	 * with a list of periods that are propagated to them.
	 * 
	 * @throws Exception
	 *             if the timing specification of any traversed element is inconsistent.
	 */
	// private void propagatePeriodsToPredecessors(Map<Component, Double> componentPeriodMap,
	// Map<Component, Collection<Double>> possibleComponentPeriods) throws Exception {
	// Set<Component> workingQueue = new HashSet<Component>();
	// Set<InputPort> traversedInputPorts = new HashSet<InputPort>();
	// workingQueue.addAll(componentPeriodMap.keySet());
	// while(!workingQueue.isEmpty()) {
	// Component currentComponent = (Component)workingQueue.toArray()[0];
	// workingQueue.remove(currentComponent);
	// for(InputPort portToPredeccessor : currentComponent.getInputPorts()) {
	// if(!traversedInputPorts.contains(portToPredeccessor)) {
	// traversedInputPorts.add(portToPredeccessor);
	// for(OutputPort predecessorPort : findSourcePorts(portToPredeccessor)) {
	// Component predecessorComponent = predecessorPort.getComponent();
	// // Predecessor is not referenced by a periodic constraint
	// if(!componentPeriodMap.containsKey(predecessorComponent)) {
	// Collection<Double> possiblePeriods =
	// possibleComponentPeriods.get(predecessorComponent);
	// // If the predecessor component has not been considered yet, a list for
	// // the possible periods must be created.
	// if(possiblePeriods == null) {
	// possiblePeriods = new ArrayList<Double>();
	// }
	// possiblePeriods.addAll(possibleComponentPeriods.get(currentComponent));
	// possibleComponentPeriods.put(predecessorComponent, possiblePeriods);
	// workingQueue.add(predecessorComponent);
	// }
	// }
	// }
	// }
	// }
	// }

	/**
	 * Calculates the effective period of all successor components that are not referenced by a
	 * periodic constraint from the set of plausible periods propagated from other components. This
	 * effective period is calculated by finding the greatest common divisor of all possible periods
	 * of a {@link Component}.
	 * 
	 * @throws Exception
	 *             if the timing specification of any traversed element is inconsistent.
	 */
	private void determineEffectivePeriodOfComponents(Map<Component, Double> componentPeriodMap,
			Map<Component, Collection<Double>> possibleComponentPeriods) throws Exception {

		for(Entry<Component, Collection<Double>> possibleComponentPeriod : possibleComponentPeriods
				.entrySet()) {
			if(isComponentOrPortsDeployable(possibleComponentPeriod.getKey(), this)) {
				componentPeriodMap.put(possibleComponentPeriod.getKey(),
						gcdDoubleCollection(possibleComponentPeriod.getValue(), GCD_PRECISION));
			}
		}
	}

	/**
	 * Determines the allowed data exchange (channels) from an AF3 {@link ComponentArchitecture}.
	 * 
	 * @throws Exception
	 */
	private Collection<ISignalAdapter<OutputPort>>
			createSignalAdapters(Map<OutputPort, Long> messageSizes) throws Exception {
		Collection<ISignalAdapter<OutputPort>> rval = new ArrayList<ISignalAdapter<OutputPort>>();

		Collection<Component> af3DeployableComponents = new ArrayList<>();
		deployableComponents.forEach(d -> af3DeployableComponents.add(d.getObject()));

		for(OutputPort currentSenderPort : getDeployableOutputPorts(af3DeployableComponents)) {
			IDeployableComponentAdapter<?> currentSenderComponentAdapter =
					getDeployableComponentAdapter(currentSenderPort.getComponent());
			if(currentSenderComponentAdapter != null) {
				Collection<IDeployableComponentAdapter<?>> receiverComponentAdapters =
						new ArrayList<IDeployableComponentAdapter<?>>();
				Collection<Port> targetPorts =
						findTargetPorts(currentSenderPort, af3DeployableComponents);
				for(Port currentReceiverPort : targetPorts) {
					IDeployableComponentAdapter<?> receiverComponentAdapter =
							getDeployableComponentAdapter(currentReceiverPort.getComponent());
					receiverComponentAdapters.add(receiverComponentAdapter);
				}
				if(!receiverComponentAdapters.isEmpty()) {
					// Do not consider unconnected Ports
					Collection<InputPort> filteredRecvPorts =
							pickInstanceOf(InputPort.class, new ArrayList<>(targetPorts));
					AF3SignalAdapter sigAdp = new AF3SignalAdapter(currentSenderPort,
							currentSenderComponentAdapter, filteredRecvPorts,
							receiverComponentAdapters, messageSizes.get(currentSenderPort));
					sigAdp.setSignalType(portSignalTypes.get(currentSenderPort));
					rval.add(sigAdp);
				}
			}
		}

		return rval;
	}

	/**
	 * Pre-sort the {@link ISignalAdapter}s such that the {@link IDeployableComponentAdapter}s
	 * referenced by periodic constraints are located at the top of the {@link ISignalAdapter} list.
	 * <p>
	 * <b>NOTE:</b> This method is just for user convenience, do not expect the ordering to be kept
	 * by other algorithms!
	 * 
	 * @throws Exception
	 *             if the period cannot be determined for one component, possibly due to an
	 *             inconsistent timing specification.
	 */
	// TODO: cleanup or get rid of this method.
	private Collection<ISignalAdapter<OutputPort>> sortMessagesByExpectedOrder(
			Collection<ISignalAdapter<OutputPort>> unsortedMessages,
			Map<Component, Double> componentPeriodMap) throws Exception {
		Collection<ISignalAdapter<OutputPort>> rval = new ArrayList<ISignalAdapter<OutputPort>>();
		for(ISignalAdapter<OutputPort> currentMessage : unsortedMessages) {
			Double periodValue = componentPeriodMap.get(currentMessage.getSource().getObject());
			if(periodValue != null && periodValue > 0) {
				rval.add(currentMessage);
			}
		}
		for(ISignalAdapter<OutputPort> currentMessage : unsortedMessages) {
			Double periodValue = componentPeriodMap.get(currentMessage.getSource().getObject());
			if(periodValue == null || periodValue <= 0) {
				rval.add(currentMessage);
			}
		}
		return rval;
	}

	/**
	 * Creates {@link IMemoryUnitAdapter}s from AF3-{@link MemoryUnit}s.
	 * 
	 * @throws Exception
	 */
	private Collection<IMemoryUnitAdapter<MemoryUnit>> createMemoryUnits(
			PlatformArchitecture platformArchitecture,
			Map<IPlatformArchitectureElement, Double> resourceFailureRates) throws Exception {
		Collection<IMemoryUnitAdapter<MemoryUnit>> rval =
				new ArrayList<IMemoryUnitAdapter<MemoryUnit>>();

		for(MemoryUnit currentMemoryUnit : getChildrenWithType(platformArchitecture,
				MemoryUnit.class)) {
			Double failureRate = resourceFailureRates.get(currentMemoryUnit);
			// if(failureRate == null) {
			// throw new Exception("The failure probablity for the " +
			// currentMemoryUnit.getClass().getSimpleName() + " " +
			// currentMemoryUnit.getName() + " has not been defined.");
			// }

			rval.add(new AF3MemoryUnitAdapter(currentMemoryUnit, failureRate));
		}

		return rval;
	}

	/**
	 * Determines the deployment targets from an AF3 {@link PlatformArchitecture}.
	 * 
	 * @throws Exception
	 */
	private Collection<IExecutionUnitAdapter<ExecutionUnit>> createDeploymentTargets(
			Collection<ExecutionUnit> targetExecutionUnits,
			Map<IPlatformArchitectureElement, Double> resourceFailureProbabilies) throws Exception {
		if(targetExecutionUnits.isEmpty()) {
			throw new Exception(
					"No deployment targets have been specified in the target platform architecture(s). Please check whether the correct platform architecture was selected and whether the desired Execution Units were selected as deployment targets (Platform Model --> Annotation View --> Deployment Target)");
		}

		Collection<IExecutionUnitAdapter<ExecutionUnit>> rval =
				new ArrayList<IExecutionUnitAdapter<ExecutionUnit>>();

		for(ExecutionUnit execUnit : targetExecutionUnits) {
			Double failureProb = resourceFailureProbabilies.get(execUnit);
			rval.add(new AF3ExecutionUnitAdapter(execUnit, failureProb));
		}

		return rval;
	}

	// TODO: AD: I assume this mapping can be removed soon. However, there must be a decision
	// whether all possible communication paths shall be stored or if the required paths shall be
	// calculated when needed, i.e. during scheduling.
	/**
	 * Build mapping virtual resource -> hardware resource onto which virtual resource is
	 * statically mapped, e.g., partition -> hardware execution unit / processor core.
	 */
	private Map<IResourceAdapter<?>, IResourceAdapter<?>>
			createVirtualResourceToHardwareResourceMap() {
		Map<IResourceAdapter<?>, IResourceAdapter<?>> rval =
				new HashMap<IResourceAdapter<?>, IResourceAdapter<?>>();

		// TODO: Dummy 1:1 mapping, assuming a flat platform architecture consisting of processors
		// communication units. See {@link #createDeploymentTargets}.
		for(IExecutionUnitAdapter<?> deploymentTarget : getDeploymentTargets()) {
			rval.put(deploymentTarget, deploymentTarget);
		}
		for(ITransmissionUnitAdapter<?> transmissionUnit : getTransmissionUnits()) {
			rval.put(transmissionUnit, transmissionUnit);
		}
		for(IGatewayUnitAdapter<?> gatewayUnit : getGatewayUnits()) {
			rval.put(gatewayUnit, gatewayUnit);
		}

		return rval;
	}

	/**
	 * Collects all {@link GatewayUnit}s from a given AF3 Platform and returns them as a collection
	 * of {@link IGatewayUnitAdapter}s.
	 * 
	 * @throws Exception
	 */
	private Collection<IGatewayUnitAdapter<GatewayUnit>> createGatewayUnits(
			PlatformArchitecture platformArchitecture,
			Map<IPlatformArchitectureElement, Double> resourceFailureRates,
			Map<IPlatformResource, Double> resourcePowerConsumption) throws Exception {
		Collection<IGatewayUnitAdapter<GatewayUnit>> rval =
				new ArrayList<IGatewayUnitAdapter<GatewayUnit>>();

		for(GatewayUnit gatewayUnit : EcoreUtils.getChildrenWithType(platformArchitecture,
				GatewayUnit.class)) {
			Double failureRate = resourceFailureRates.get(gatewayUnit);
			// if(failureRate == null) {
			// throw new Exception("The failure probablity for the " +
			// gatewayUnit.getClass().getSimpleName() + " " + gatewayUnit.getName() +
			// " has not been defined.");
			// }
			Double power = resourcePowerConsumption.get(gatewayUnit);
			// TODO: We do not have a power annotation for gateways yet.
			// if(power == null) {
			// throw new Exception("The failure probability for the " +
			// gatewayUnit.getClass().getSimpleName() + " " + gatewayUnit.getName() +
			// " has not been defined.");
			// }

			rval.add(new AF3GatewayAdapter(gatewayUnit, failureRate, power));
		}

		return rval;
	}

	/**
	 * Determines the communication resource AF3 {@link PlatformArchitecture}.
	 * 
	 * @throws Exception
	 */
	private Collection<ITransmissionUnitAdapter<TransmissionUnit>> createTransmissionUnits(
			PlatformArchitecture platformArchitecture,
			Map<IPlatformArchitectureElement, Double> resourceFailureRates,
			Map<IPlatformResource, Double> resourcePowerConsumption) throws Exception {
		Collection<ITransmissionUnitAdapter<TransmissionUnit>> rval =
				new ArrayList<ITransmissionUnitAdapter<TransmissionUnit>>();

		for(TransmissionUnit transmissionUnit : EcoreUtils.getChildrenWithType(platformArchitecture,
				TransmissionUnit.class)) {
			Double failureRate = resourceFailureRates.get(transmissionUnit);
			// TODO: Virtual TransmissionUnits currently do not have ResourceLinks to their physical
			// counterparts.
			// if(!(transmissionUnit instanceof IVirtualizationPlatformArchitectureElement) &&
			// failureRate == null) {
			// throw new Exception("The failure probablity for the " +
			// transmissionUnit.getClass().getSimpleName() + " " +
			// transmissionUnit.getName() + " has not been defined.");
			// }
			Double power = resourcePowerConsumption.get(transmissionUnit);
			// if(!(transmissionUnit instanceof IVirtualizationPlatformArchitectureElement) &&
			// power == null) {
			// throw new Exception("The failure probablity for the " +
			// transmissionUnit.getClass().getSimpleName() + " " +
			// transmissionUnit.getName() + " has not been defined.");
			// }

			rval.add(new AF3TransmissionUnitAdapter(transmissionUnit, failureRate, power));
		}

		return rval;
	}

	// TODO: Consider removing "available resources", since this might be already implemented in the
	// virtual HW mapping
	// TODO: Rebase createAvailableResources() upon the IPlatformResource, since this the
	// appropriate marker interface. Only create new IResourceAdapters, if no other Resource is
	// present already.
	/**
	 * Creates a {@link Collection} that references all resources present in the given platform.
	 * Currently, those are {@link ExecutionUnit}s, {@link GatewayUnit}s, and
	 * {@link TransmissionUnit}s.
	 */
	private Collection<IResourceAdapter<?>> createAvailableResources() {
		Collection<IResourceAdapter<?>> rval = new ArrayList<IResourceAdapter<?>>();

		// TODO: Do we need to consider more than the deployment target execution units here?
		// Only consider ExecutionUnits whose container are not marked as deployment targets.
		// for(ExecutionUnit execUnit : getChildrenWithType(
		// (PlatformArchitecture)platformArchitecture.getObject(), ExecutionUnit.class)) {
		// if(!isAnyParentDeploymentTarget(execUnit, deploymentTargets)) {
		// rval.add(execUnit);
		// }
		// }
		// // Also for the hardware platform, if the primary platform is a virtual one.
		// if(hardwarePlatformArchitecture != null) {
		// for(ExecutionUnit execUnit : getChildrenWithType(hardwarePlatformArchitecture,
		// ExecutionUnit.class)) {
		// if(!isAnyParentDeploymentTarget(execUnit, deploymentTargets)) {
		// rval.add(execUnit);
		// }
		// }
		// }

		rval.addAll(deploymentTargets);
		rval.addAll(memoryUnits);
		rval.addAll(gatewayUnits);
		rval.addAll(transmissionUnits);

		return rval;
	}

	/**
	 * Creates the {@link IResourceConnectionAdapter}s that describe the connections present in the
	 * {@link PlatformArchitecture}.
	 * 
	 * @throws Exception
	 *             if the end point of a connection cannot be determined.
	 */
	private Collection<IResourceConnectionAdapter> createDeploymentTargetConnectors(
			Collection<ExecutionUnit> af3DeploymentTargets) throws Exception {

		// Create a map that captures the inverse link of from one IModelElement to another
		// IModelelement.
		Map<IModelElement, IModelElement> invElementReferences = getInverseIntraPlatformReferences(
				(PlatformArchitecture)platformArchitecture.getObject(), platformInterconnectTypes);

		// Initial traversal list of IConnectors attached to ExecutionUnits selected as deployment
		// targets.
		LinkedList<IConnector> deployableConnectors = new LinkedList<IConnector>();
		for(IExecutionUnitAdapter<?> execUnit : deploymentTargets) {
			deployableConnectors.addAll(((ExecutionUnit)execUnit.getObject()).getConnectors());
		}

		// TODO: rewrite the code below. Handling of hull elements and virtual targets should be
		// considered in a single util method. Based on the collected target connectors, the
		// resourceconnectionadapters should be generated.
		Collection<IConnector> visitedConnectors = new HashSet<IConnector>();
		Collection<IResourceConnectionAdapter> rval = new HashSet<IResourceConnectionAdapter>();
		while(!deployableConnectors.isEmpty()) {
			IConnector currentConnector = deployableConnectors.poll();
			visitedConnectors.add(currentConnector);
			// NOTE: Here, source element does not imply "source of the connection", but
			// "source of the traversal".
			IHierarchicElement sourceElement = currentConnector.getOwner();
			if(sourceElement instanceof ExecutionUnit || sourceElement instanceof GatewayUnit ||
					sourceElement instanceof TransmissionUnit) {
				// Path walk to determine the *real* connection targets, i.e. ignore hull elements
				// like Tiles whose contained Cores (or partitions) were selected as deployment
				// targets.
				Collection<IConnector> targetConnectors = getEffectiveConnectionTargets(
						currentConnector, invElementReferences, af3DeploymentTargets);

				// Create an IResourceConnectionAdapter between the sourceElement of the traversal
				// and the *real* targets.
				AF3ResourceConnectionAdapter connAdp =
						createResourceConnection(currentConnector, targetConnectors);
				if(connAdp != null) {
					rval.add(connAdp);
				}
				visitedConnectors.addAll(targetConnectors);

				// Collect the IConnectors attached to the target elements of the created
				// IResourceConnectionAdapter for further traversing the PlatformArchitecture.
				// Connections between IModelElements of different PlatformArchitectures which are
				// realized via ResourceLink specifications are automatically considered.
				Collection<IHierarchicElement> targetElements = new HashSet<IHierarchicElement>();
				for(IConnector currentTargetConnector : targetConnectors) {
					targetElements.add(currentTargetConnector.getOwner());
				}
				determineNextConnectionTargets(invElementReferences, deployableConnectors,
						visitedConnectors, targetElements, af3DeploymentTargets);
			}
		}

		return rval;
	}

	/**
	 * Collect the IConnectors attached to the target elements of the created
	 * IResourceConnectionAdapter for further traversing the PlatformArchitecture. Connections
	 * between IModelElements of different PlatformArchitectures which are realized via ResourceLink
	 * specifications are automatically considered.
	 * 
	 * @throws InvalidPlatformModelException
	 *             if the references from the referencing {@link IModelElement} to the target
	 *             {@link IModelElement} are incorrectly specified.
	 */
	// TODO: Refactor this method to use the static methods from the AF3Utils class instead.
	private void determineNextConnectionTargets(
			Map<IModelElement, IModelElement> invElementReferences,
			LinkedList<IConnector> deployableConnectors, Collection<IConnector> visitedConnectors,
			Collection<IHierarchicElement> targetElements,
			Collection<ExecutionUnit> af3DeploymentTargets) throws InvalidPlatformModelException {
		for(IHierarchicElement targetElement : targetElements) {
			if(targetElement instanceof IPlatformResource &&
					!af3DeploymentTargets.contains(targetElement)) {
				for(IConnector connectorAtTarget : targetElement.getConnectors()) {
					// Do not traverse back again.
					if(connectorAtTarget != targetElement) {
						if(!connectsSwHwPlatform(connectorAtTarget, invElementReferences)) {
							if(!visitedConnectors.contains(connectorAtTarget)) {
								deployableConnectors.add(connectorAtTarget);
							}
						} else if(connectsSwHwPlatform(connectorAtTarget)) {
							// Special treatment for the connection between the virtual
							// platform and the physical platform.
							GatewayUnit currentNetworkDriver =
									(GatewayUnit)connectorAtTarget.getOwner();
							List<GatewayUnit> referencedNetworkInterfaces =
									getReferencedElementsWithType(currentNetworkDriver,
											GatewayUnit.class);

							// TODO: consider moving this check to the input model validation.
							if(referencedNetworkInterfaces.size() == 0) {
								throw new InvalidPlatformModelException(currentNetworkDriver,
										"No Network Interface is referenced in the ResourceLink annotation. Please specify at least one reference.");
							}

							if(!visitedConnectors.contains(connectorAtTarget)) {
								for(GatewayUnit refNi : referencedNetworkInterfaces) {
									Collection<IHierarchicElement> targets =
											new ArrayList<IHierarchicElement>();
									targets.add(currentNetworkDriver);
									if(isAnyChildDeploymentTarget(
											(IPlatformResource)refNi.getContainer(),
											af3DeploymentTargets, ExecutionUnit.class)) {
										deployableConnectors.addAll(refNi.getConnectors());
									} else {
										deployableConnectors.addAll(pickInstanceOf(
												IPlatformExport.class, refNi.getConnectors()));
									}
								}
							}
							visitedConnectors.add(connectorAtTarget);
						} else if(invElementReferences.get(connectorAtTarget.getOwner()) != null) {
							// Special treatment for the connection between the physical platform
							// and the virtual platform.
							GatewayUnit currentNetworkDriver = (GatewayUnit)invElementReferences
									.get(connectorAtTarget.getOwner());
							IPlatformExport driverExport = pickFirstInstanceOf(
									IPlatformExport.class, currentNetworkDriver.getConnectors());

							if(!visitedConnectors.contains(driverExport)) {
								Collection<IHierarchicElement> targets =
										new ArrayList<IHierarchicElement>();
								targets.add(driverExport.getOwner());
								deployableConnectors
										.addAll(driverExport.getOwner().getConnectors());

								visitedConnectors.add(driverExport);
							}
						}
					}
				}
			}
		}
	}

	/**
	 * Creates a {@link IResourceConnectionAdapter} between the given source and the target
	 * elements. This method takes {@link IHierarchicElement}s as inputs and identifies the matching
	 * {@link IResourceConnectionAdapter}s of the given elements.
	 * 
	 * @throws InvalidPlatformModelException
	 *             if an invalid combination of of platform ports/interfaces has been detected
	 *             (e.g., Input <-> Input).
	 */
	private AF3ResourceConnectionAdapter createResourceConnection(IConnector sourceConnector,
			Collection<IConnector> targetConnectors) throws InvalidPlatformModelException {
		IResourceAdapter<?> sourceResource = null;
		Collection<IResourceAdapter<?>> targetResources = new HashSet<IResourceAdapter<?>>();
		Map<IResourceAdapter<?>, ConnectionType> targetConnType =
				new HashMap<IResourceAdapter<?>, ConnectionType>();
		Collection<IHierarchicElement> targetElements = new HashSet<IHierarchicElement>();

		// Capture information about the target elements and the IConnectors that are the actual AF3
		// targets of this connection. It is used to determine the connection direction for each
		// target.
		Multimap<IHierarchicElement, IConnector> targetConnsAtElement = HashMultimap.create();
		for(IConnector targetConnector : targetConnectors) {
			targetElements.add(targetConnector.getOwner());
			targetConnsAtElement.put(targetConnector.getOwner(), targetConnector);
		}

		// Identify the corresponding IResourceAdapters of the collection of target
		// IHierarchicalElements and determine the direction of their connection.
		for(IResourceAdapter<?> resource : availableResources) {
			if(resource.getObject() == sourceConnector.getOwner()) {
				sourceResource = resource;
			} else if(targetElements.contains(resource.getObject())) {
				targetResources.add(resource);
				ConnectionType type = ConnectionType.BIDIRECTIONAL;
				boolean firstRoundFlag = true;
				for(IConnector targetConnector : targetConnsAtElement
						.get((IHierarchicElement)resource.getObject())) {
					if(firstRoundFlag) {
						firstRoundFlag = false;
						type = getConnectionTypeDirection(sourceConnector, targetConnector, this);
					} else if(type != getConnectionTypeDirection(sourceConnector, targetConnector,
							this)) {
						type = ConnectionType.BIDIRECTIONAL;
					}
				}
				targetConnType.put(resource, type);
			}
		}
		assert (sourceResource != null);

		return new AF3ResourceConnectionAdapter(sourceResource, targetConnType);
	}

	/**
	 * Returns the (first) {@link IDeployableComponentAdapter} which is associated with the given
	 * AF3 {@link Component}. Returns null if no matching {@link IDeployableComponentAdapter} is
	 * found.
	 */
	public IDeployableComponentAdapter<?> getDeployableComponentAdapter(Component af3Component) {
		if(af3Component == null) {
			throw new IllegalArgumentException(
					"Cannot identify IDeployableComponentAdapters for 'null' Components");
		}

		IDeployableComponentAdapter<?> deployableComponentAdapter = null;
		for(IDeployableComponentAdapter<Component> deployableComponent : deployableComponents) {
			if(deployableComponent.getObject() == af3Component) {
				deployableComponentAdapter = deployableComponent;
				break;
			}
		}

		return deployableComponentAdapter;
	}

	/**
	 * Creates a {@link AF3DeploymentParameterMapAdapter} that provides access to the deployment
	 * specific parameters of an AF3 {@link Deployment}. The corresponding {@link Deployment} is
	 * identified via the referenced {@link ComponentArchitecture} and {@link PlatformArchitecture}.
	 */
	private AF3DeploymentParameterMapAdapter createDeploymentParameterMap(
			ComponentArchitecture componentArchitecture, PlatformArchitecture platformArchitecture,
			Multimap<Component, Component> replacementComponents) {
		Set<Deployment> selectedDeployments = new HashSet<Deployment>();

		addMatchingParameterDeploymentTo(selectedDeployments, componentArchitecture,
				platformArchitecture);

		Set<ComponentArchitecture> diversityArchitectureList =
				getReferencedCompArchs(replacementComponents);
		for(ComponentArchitecture diversityContainingArch : diversityArchitectureList) {
			addMatchingParameterDeploymentTo(selectedDeployments, diversityContainingArch,
					platformArchitecture);
		}

		if(!selectedDeployments.isEmpty()) {
			return new AF3DeploymentParameterMapAdapter(selectedDeployments);
		}
		return null;
	}

	/**
	 * Searches the {@link IProjectRootElement} for {@link Deployment}s of the given
	 * {@link ComponentArchitecture} to the given {@link PlatformArchitecture} and adds the first
	 * matching {@link Deployment} to the given list.
	 */
	private void addMatchingParameterDeploymentTo(Set<Deployment> paramDeployments,
			ComponentArchitecture componentArchitecture,
			PlatformArchitecture platformArchitecture) {
		List<Deployment> matchingDeployments =
				getDeploymentsWithParametersOf(componentArchitecture, platformArchitecture);
		if(matchingDeployments != null && matchingDeployments.size() > 0) {
			// Take the first "template" Deployment until a approach is implemented that allows the
			// user select the desired "template" deployment.
			paramDeployments.add(matchingDeployments.get(0));
		}
	}

	/**
	 * Returns the set of {@link ComponentArchitecture}s that contain the referenced
	 * {@link Component}s given by the {@code replacementComponents} map.
	 */
	private Set<ComponentArchitecture>
			getReferencedCompArchs(Multimap<Component, Component> replacementComponents) {
		Set<ComponentArchitecture> referencedCompArchs = new HashSet<ComponentArchitecture>();
		for(Component diverseComponent : replacementComponents.values()) {
			ComponentArchitecture diversityCompArch =
					getParentElement(diverseComponent, ComponentArchitecture.class, false);
			if(diversityCompArch != null) {
				referencedCompArchs.add(diversityCompArch);
			}
		}
		return referencedCompArchs;
	}

	/**
	 * Returns the (first) {@link IExecutionUnitAdapter} which is associated with the given
	 * AF3 {@link ExecutionUnit}. Returns null if no matching {@link IExecutionUnitAdapter} is
	 * found
	 */
	public IExecutionUnitAdapter<?> getExecutionUnitAdapter(ExecutionUnit af3ExecutionUnit) {
		if(af3ExecutionUnit == null)
			throw new IllegalArgumentException(
					"Cannot identify IExecutionUnitAdapter for 'null' ExecutionUnits");

		IExecutionUnitAdapter<?> executionUnitAdapter = null;
		for(IExecutionUnitAdapter<?> targetExecutionUnit : deploymentTargets) {
			if(targetExecutionUnit.getObject() == af3ExecutionUnit) {
				executionUnitAdapter = targetExecutionUnit;
				break;
			}
		}

		assert (executionUnitAdapter == null) : "No IExecutionUnitAdapter found for the ExecutionUnit " +
				af3ExecutionUnit.getName() + "(ID: " + af3ExecutionUnit.getId() +
				"). Every AF3-ExecutionUnit should have a IExecutionUnitAdapter equivalent.";

		return executionUnitAdapter;
	}

	/** Determine task graph for a given component */
	public TaskGraph getTaskGraph(Component component) {
		// TODO: AD: The cast to the AF3 systemmodeladapter breaks the abstration the goals. It
		// would be rather required to define the source and sink components as
		// IDeployableComponents in the ecore model. The transformation from AF3 components to the
		// IDeployableCompontes should be done when the data from the GUI is written to the model.
		IDeployableComponentAdapter<?> deployableComponent =
				getDeployableComponentAdapter(component);
		return getTaskGraph(deployableComponent);
	}
}
