/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.sysmodel.arch.af3;

import static org.fortiss.tooling.kernel.utils.EcoreUtils.pickFirstInstanceOf;

import java.util.Collection;

import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ISignalAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITransmissionUnitAdapter;
import org.fortiss.af3.platform.model.TransmissionUnit;
import org.fortiss.af3.timing.model.annotation.TransmissionUnitBandwidth;

/**
 * Adapter between the AF3 {@link TransmissionUnit} and the DSE.
 * 
 * @author barner
 */
public class AF3TransmissionUnitAdapter extends AF3ResourceAdapter<TransmissionUnit> implements
		ITransmissionUnitAdapter<TransmissionUnit> {
	/** Conversion constant for the message size. */
	private static final long MBYTES_TO_BYTES = 1000 * 1000;

	/** {@link IResourceAdapter}s that are connected with this {@link AF3TransmissionUnitAdapter}. */
	Collection<IResourceAdapter<?>> connectedResources;

	/** {@link TransmissionUnit} bandwidth annotation */
	TransmissionUnitBandwidth transmissionUnitBandwidth;

	/** {@link TransmissionUnit} power consumptions annotation */
	Double transmissionUnitPower;

	/**
	 * Creates an {@link ISignalAdapter} for an {@link TransmissionUnit}.
	 * 
	 * @param transmissionUnit
	 *            present in the target platform and for which to generate an adapter.
	 * @param failureProb
	 *            failure probability of the {@code transmissionUnit}.
	 * @param power
	 *            Energy required to transfer a single bit.
	 * 
	 * @throws Exception
	 *             if the bandwidth property is invalid: Non-existent or zero.
	 */
	public AF3TransmissionUnitAdapter(TransmissionUnit transmissionUnit, Double failureProb,
			Double power) throws Exception {
		super(transmissionUnit, failureProb);
		this.transmissionUnitPower = power;

		// Extract annotations
		transmissionUnitBandwidth =
				pickFirstInstanceOf(TransmissionUnitBandwidth.class,
						transmissionUnit.getSpecifications());
		if(transmissionUnitBandwidth == null) {
			throw new Exception("Transmission unit " + transmissionUnit +
					" does not have a bandwidth annotation.");
		} else if(transmissionUnitBandwidth.getBandwidth_MBytesPerSecond() == 0.0) {
			throw new Exception("The transmission unit " + transmissionUnit +
					" has a bandwidth of 0 MB/s. This is not a valid attribute.");
		}
	}

	/** {@inheritDoc} */
	@Override
	public double getPower(long messageSizeInBits) {
		return messageSizeInBits /
				(MBYTES_TO_BYTES * transmissionUnitBandwidth.getBandwidth_MBytesPerSecond()) *
				transmissionUnitPower;
	}

	/** {@inheritDoc} */
	@Override
	public double getTransmissionDuration(double messageSizeMB) {
		return messageSizeMB / transmissionUnitBandwidth.getBandwidth_MBytesPerSecond();
	}
}
