/*-------------------------------------------------------------------------+
| Copyright 2015 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.sysmodel.arch.af3.helper;

import static org.fortiss.tooling.kernel.utils.EcoreUtils.getChildrenWithType;

import java.math.BigDecimal;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import org.fortiss.af3.component.model.Component;
import org.fortiss.af3.component.model.ComponentArchitecture;
import org.fortiss.af3.deployment.model.Deployment;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IExecutionUnitAdapter;
import org.fortiss.af3.platform.model.ExecutionUnit;
import org.fortiss.af3.platform.model.annotation.EnergyConsumption;
import org.fortiss.af3.timing.model.annotation.Wcet;

/**
 * AF3 adapter that provides access to the deployment specific parameter map.
 * 
 * @author diewald
 */
public class AF3DeploymentParameterMapAdapter {
	/**
	 * References the map of {@link Deployment}s that contain the parameter map of a
	 * {@link Component}.
	 */
	private Map<Component, Deployment> componentParamDeploymentMap;

	/** Constructor. */
	public AF3DeploymentParameterMapAdapter(Set<Deployment> af3DeploymentList) {
		componentParamDeploymentMap = new HashMap<Component, Deployment>();
		for(Deployment singleParamDeployment : af3DeploymentList) {
			ComponentArchitecture refCompArch = (ComponentArchitecture)singleParamDeployment
					.getComponentArchitectureReference().getReference();
			Collection<Component> containedComponents =
					getChildrenWithType(refCompArch, Component.class);
			for(Component currentComponent : containedComponents) {
				componentParamDeploymentMap.put(currentComponent, singleParamDeployment);
			}
		}
	}

	/** Returns the WCET of a component based on its deployment target. */
	public double getWcet(Component af3Component, IExecutionUnitAdapter<?> deploymentTarget) {
		ExecutionUnit af3ExecutionUnit = getAf3ExecutionUnit(deploymentTarget);
		boolean validAf3Elements = (af3Component != null && af3ExecutionUnit != null);

		Deployment associatedParamDeployment = componentParamDeploymentMap.get(af3Component);
		Object wcetValue = associatedParamDeployment.getDeploymentParameter(af3Component,
				af3ExecutionUnit, Wcet.class, null);
		return (validAf3Elements && wcetValue != null) ? ((BigDecimal)wcetValue).doubleValue()
				: -1.0;
	}

	/** Returns the energy consumption of a component based on its deployment target. */
	public double getEnergyConsumption(Component af3Component,
			IExecutionUnitAdapter<?> deploymentTarget) {
		ExecutionUnit af3ExecutionUnit = getAf3ExecutionUnit(deploymentTarget);
		boolean validAf3Elements = (af3Component != null && af3ExecutionUnit != null);

		Deployment associatedParamDeployment = componentParamDeploymentMap.get(af3Component);
		Object energyValue = associatedParamDeployment.getDeploymentParameter(af3Component,
				af3ExecutionUnit, EnergyConsumption.class, null);
		return (validAf3Elements && energyValue != null) ? (Double)energyValue : -1.0;
	}

	/**
	 * Returns the corresponding {@link ExecutionUnit} of the given {@link IExecutionUnitAdapter}.
	 */
	private ExecutionUnit getAf3ExecutionUnit(IExecutionUnitAdapter<?> deploymentTarget) {
		Object executionUnit = deploymentTarget.getObject();
		return (executionUnit instanceof ExecutionUnit) ? (ExecutionUnit)executionUnit : null;
	}
}
