/*-------------------------------------------------------------------------+
| Copyright 2015 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.SortedSet;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskgraph.TaskGraphEncoding;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IDeployableComponentAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ILogicalArchitectureAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ISignalAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.faultmodel.FaultDetector;
import org.jgrapht.DirectedGraph;
import org.jgrapht.Graphs;
import org.jgrapht.alg.PrimMinimumSpanningTree;
import org.jgrapht.alg.interfaces.MinimumSpanningTree;
import org.jgrapht.graph.ListenableDirectedWeightedGraph;

/**
 * Creates a Directed acyclic task graph of all the {@link IDeployableComponentAdapter}s. <li>Each
 * {@link ISignalAdapter} is examined:</li> <li>If there exists no path between a target of the
 * {@link ISignalAdapter} and it source, add an edge to the {@link DefaultTaskGraph}.</li> <li>
 * Otherwise, a loop is detected. Then, the detected loop is cut either by not adding the edge (if
 * the source vertex is strongly causal) or by removing an edge between the vertices (i.e.
 * {@link IDeployableComponentAdapter}s) that are closest to the current {@link ISignalAdapter} by
 * traversing the found path backwards and adding the current edge.</li>
 * <p>
 * <b>NOTE:</b> This method expects <b>correct</b> {@link ILogicalArchitectureAdapter}s as input,
 * i.e. at least one {@link IDeployableComponentAdapter} of each loop must be 'strongly causal'.
 * 
 * @author diewald
 */
public class AcyclicTaskGraph extends DefaultTaskGraph implements TaskGraph {

	/** Constructor. */
	public AcyclicTaskGraph() {
		super();
	}

	/**
	 * Constructor which creates an acyclic {@link TaskGraph} from a given possibly cyclic
	 * {@link TaskGraph}.
	 */
	public AcyclicTaskGraph(TaskGraph taskGraph) {
		this();
		period = taskGraph.getPeriod();
		createMinSpanAcycTaskGraph(taskGraph);
	}

	/**
	 * Creates an acyclic {@link TaskGraph} from a given general {@link TaskGraph} based on a
	 * calculation of the minimum spanning tree of the transitions between tasks, i.e.
	 * {@link ChannelAdapterWeightedEdge}s.
	 */
	private void createMinSpanAcycTaskGraph(TaskGraph taskGraph) {
		DirectedGraph<IDeployableComponentAdapter<?>, ChannelAdapterWeightedEdge> tg =
				taskGraph.getGraph();
		Graphs.addAllVertices(graph, tg.vertexSet());

		if(taskGraph.getTasks().size() == 1) {
			return;
		}
		MinimumSpanningTree<IDeployableComponentAdapter<?>, ChannelAdapterWeightedEdge> minSpanTree =
				new PrimMinimumSpanningTree<IDeployableComponentAdapter<?>, ChannelAdapterWeightedEdge>(
						taskGraph.getGraph());
		minSpanTree.getMinimumSpanningTreeEdgeSet().stream()
				.forEach(e -> graph.addEdge(tg.getEdgeSource(e), tg.getEdgeTarget(e), e));
	}

	/** {@inheritDoc} */
	@Override
	public AcyclicTaskGraph copy() {
		AcyclicTaskGraph clonedTaskGraph;
		try {
			clonedTaskGraph = (AcyclicTaskGraph)super.clone();
		} catch(CloneNotSupportedException e) {
			throw new AssertionError();
		}
		clonedTaskGraph.graph =
				new ListenableDirectedWeightedGraph<IDeployableComponentAdapter<?>, ChannelAdapterWeightedEdge>(
						ChannelAdapterWeightedEdge.class);
		Graphs.addGraph(clonedTaskGraph.graph, graph);
		clonedTaskGraph.taskFaultDetectorsMap =
				new HashMap<IDeployableComponentAdapter<?>, SortedSet<FaultDetector>>(
						taskFaultDetectorsMap);

		return clonedTaskGraph;
	}

	/**
	 * Transforms all existing {@link TaskGraph}s of {@code this} {@link TaskGraphEncoding} into
	 * {@link AcyclicTaskGraph}s.
	 */
	public static <T extends TaskGraph> Collection<AcyclicTaskGraph> transformToAcyclicGraphs(
			Collection<T> cyclicTGs) {
		Collection<AcyclicTaskGraph> acyclicTGCollection = new ArrayList<>();
		for(TaskGraph tg : cyclicTGs) {
			AcyclicTaskGraph acycTG = new AcyclicTaskGraph(tg);
			acyclicTGCollection.add(acycTG);
		}
		return acyclicTGCollection;
	}
}
