/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph;

import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IDeployableComponentAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ISignalAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemModelAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.faultmodel.FaultDetector;
import org.fortiss.af3.exploration.alg.dse.sysmodel.random.RandomExploration;
import org.fortiss.af3.exploration.alg.exception.InvalidTimingModelException;
import org.fortiss.af3.exploration.alg.graph.display.JGraphTVisualizer;
import org.jgrapht.Graphs;
import org.jgrapht.graph.ListenableDirectedWeightedGraph;
import org.jgrapht.traverse.DepthFirstIterator;

import com.google.common.collect.HashMultimap;
import com.google.common.collect.Multimap;

/**
 * <p>
 * The TaskGraph represents the data dependencies of a single connected component between the
 * deployable components (= tasks) of a systems (Tasks = vertices, channels of logical architecture
 * = edges). Therefore, a system typically consists of multiple {@link DefaultTaskGraph}s that are
 * owned by the corresponding {@link SystemModelAdapter} instance.
 * </p>
 * The TaskGraph also contains
 * <ul>
 * <li>Execution times of tasks.</li>
 * <li>The fault-detectors used by the tasks</li>
 * <li>Task execution times (currently assuming homogeneous cores)</li>
 * </ul>
 * 
 * @author huang
 */

public class DefaultTaskGraph implements TaskGraph, Cloneable {
	/** Underlying graph implementation */
	// TODO: Add a weight w.r.t. transmission time, when deployed
	protected ListenableDirectedWeightedGraph<IDeployableComponentAdapter<?>, ChannelAdapterWeightedEdge> graph;

	// TODO: Currently, all tasks in a {@link TaskGraph} must have the same period
	/** The period time of this {@link DefaultTaskGraph}. */
	protected Double period;

	// TODO: Fault detector library
	/** Fault detector library (map: task -> set of fault detectors) */
	protected Map<IDeployableComponentAdapter<?>, SortedSet<FaultDetector>> taskFaultDetectorsMap;

	/** Random generator */
	private RandomExploration gen = RandomExploration.getInstance();

	/**
	 * Associates "abstract" {@link IDeployableComponentAdapter}s with its instances contained in
	 * this {@link TaskGraph}.
	 */
	private Multimap<IDeployableComponentAdapter<?>, IDeployableComponentAdapter<?>> abstractToInstanceTaskMap;

	/** Creates an empty {@link DefaultTaskGraph}. */
	public DefaultTaskGraph() {

		graph =
				new ListenableDirectedWeightedGraph<IDeployableComponentAdapter<?>, ChannelAdapterWeightedEdge>(
						ChannelAdapterWeightedEdge.class);

		taskFaultDetectorsMap =
				new HashMap<IDeployableComponentAdapter<?>, SortedSet<FaultDetector>>();

		abstractToInstanceTaskMap = HashMultimap.create();
	}

	/** Registers a deployable component (=task) with the {@link DefaultTaskGraph}. */
	@Override
	public boolean addTask(IDeployableComponentAdapter<?> task) {
		if(task.isInstantiation()) {
			abstractToInstanceTaskMap.put(task.getReplacedComponent(), task);
		}
		// FIXME: Hack, remove the map.
		abstractToInstanceTaskMap.put(task, task);
		return graph.addVertex(task);
	}

	/** {@inheritDoc} */
	@Override
	public boolean addTask(IDeployableComponentAdapter<?> task,
			IDeployableComponentAdapter<?> refTask) {
		// FIXME: Hack, remove the map.
		abstractToInstanceTaskMap.put(refTask, task);
		return graph.addVertex(task);
	}

	/** {@inheritDoc} */
	@Override
	public void removeTask(IDeployableComponentAdapter<?> task) {
		Collection<ChannelAdapterWeightedEdge> outgoingEdges = graph.outgoingEdgesOf(task);
		Collection<ChannelAdapterWeightedEdge> incomingEdges = graph.incomingEdgesOf(task);

		graph.removeVertex(task);
		// The outgoing edges must be removed *after* the "abstract" node, since a
		// concurrent modification is performed on the graph resulting in an exception.
		graph.removeAllEdges(outgoingEdges);
		graph.removeAllEdges(incomingEdges);

		// Remove from the cache.
		abstractToInstanceTaskMap.removeAll(task);
	}

	/** {@inheritDoc} */
	@Override
	public boolean addDataDependency(IDeployableComponentAdapter<?> sourceComponent,
			Collection<IDeployableComponentAdapter<?>> destinationComponents,
			ISignalAdapter<?> channelAdapter) {

		boolean addingAllEdgesSuccessful = true;

		// Add edges from the source component to each receiver.
		for(IDeployableComponentAdapter<?> currentDestinationComponent : destinationComponents) {
			// Only add Dependencies which are present in this TaskGraph
			if(graph.containsVertex(sourceComponent) &&
					graph.containsVertex(currentDestinationComponent)) {
				// Generate an edge in the task graph for each receiver. A single failure will cause
				// the whole function to be return 'failed'.
				if(!graph.addEdge(sourceComponent, currentDestinationComponent,
						new ChannelAdapterWeightedEdge(channelAdapter))) {
					addingAllEdgesSuccessful = false;
				}
			}
		}

		return addingAllEdgesSuccessful;
	}

	/** Registers a data dependencies between a source and its destination component. */
	public boolean addDataDependency(IDeployableComponentAdapter<?> sourceComponent,
			IDeployableComponentAdapter<?> destinationComponent, ISignalAdapter<?> channelAdapter) {
		graph.addVertex(sourceComponent);
		graph.addVertex(destinationComponent);
		return graph.addEdge(sourceComponent, destinationComponent, new ChannelAdapterWeightedEdge(
				channelAdapter));
	}

	/** {@inheritDoc} */
	@Override
	public Set<IDeployableComponentAdapter<?>> getTasks() {
		return graph.vertexSet();
	}

	// TODO: Get rid of this method!
	/** {@inheritDoc} */
	@Override
	public
			ListenableDirectedWeightedGraph<IDeployableComponentAdapter<?>, ChannelAdapterWeightedEdge>
			getGraph() {
		return graph;
	}

	/** {@inheritDoc} */
	@Override
	public Double getPeriod() {
		return period;
	}

	/**
	 * {@inheritDoc}
	 * 
	 * @throws InvalidTimingModelException
	 *             see {@link TaskGraph#updatePeriod()}.
	 */
	@Override
	public void updatePeriod() throws InvalidTimingModelException {
		boolean isFirstSet = false;
		// TODO: Move these consistency check to model validation framework
		for(IDeployableComponentAdapter<?> deployableComponent : getTasks()) {
			Double deployableComponentPeriod = deployableComponent.getTimingRequirementPeriod();

			if(!isFirstSet) {
				period = deployableComponentPeriod;
				isFirstSet = true;
			}

			// TODO: readd the period consistency check for task graphs.
			// if(period != null && !period.equals(deployableComponentPeriod)) {
			// throw new InvalidTimingModelException(
			// "The periods of the task graph could not be identified. Currently, only tasks with the same period or non-periodic task graphs are allowed.");
			// }
		}
	}

	/** Returns the fault detector with a given index / ID for a given task. */
	public FaultDetector getFaultDetector(IDeployableComponentAdapter<?> task, int id) {
		if(taskFaultDetectorsMap.containsKey(task)) {
			Object[] faultDetectors = taskFaultDetectorsMap.get(task).toArray();
			if(id < faultDetectors.length) {
				return (FaultDetector)faultDetectors[id];
			}
		}
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public void addFaultDetector(IDeployableComponentAdapter<?> task, FaultDetector faultDetector) {
		if(!taskFaultDetectorsMap.containsKey(task)) {
			SortedSet<FaultDetector> faultDetectors = new TreeSet<FaultDetector>();
			taskFaultDetectorsMap.put(task, faultDetectors);
		}
		taskFaultDetectorsMap.get(task).add(faultDetector);
	}

	/** {@inheritDoc} */
	@Override
	public int getRandomFaultDetectorId(IDeployableComponentAdapter<?> task) {
		if(taskFaultDetectorsMap.get(task) == null)
			return 0;
		return gen.nextInt(taskFaultDetectorsMap.get(task).size());
	}

	/** {@inheritDoc} */
	@Override
	public void display() {
		JGraphTVisualizer<IDeployableComponentAdapter<?>, ChannelAdapterWeightedEdge> taskViz =
				new JGraphTVisualizer<IDeployableComponentAdapter<?>, ChannelAdapterWeightedEdge>(
						this.getGraph());
		taskViz.displayGraph("Task graph");
	}

	/** {@inheritDoc} */
	@Override
	public String toString() {
		String rval = "";

		Iterator<IDeployableComponentAdapter<?>> iter =
				new DepthFirstIterator<IDeployableComponentAdapter<?>, ChannelAdapterWeightedEdge>(
						graph);
		while(iter.hasNext()) {
			IDeployableComponentAdapter<?> deployableComponent = iter.next();
			rval = rval.concat("Task " + deployableComponent.getName() + ": ");
			for(ChannelAdapterWeightedEdge edge : graph.outgoingEdgesOf(deployableComponent)) {
				rval = rval.concat("->" + graph.getEdgeTarget(edge).getName());
			}
			rval = rval.concat("\n");
		}

		return rval;
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public DefaultTaskGraph copy() {
		DefaultTaskGraph clonedTaskGraph;
		try {
			clonedTaskGraph = (DefaultTaskGraph)super.clone();
		} catch(CloneNotSupportedException e) {
			throw new AssertionError();
		}
		clonedTaskGraph.graph =
				new ListenableDirectedWeightedGraph<IDeployableComponentAdapter<?>, ChannelAdapterWeightedEdge>(
						ChannelAdapterWeightedEdge.class);
		Graphs.addGraph(clonedTaskGraph.graph, graph);
		clonedTaskGraph.taskFaultDetectorsMap =
				new HashMap<IDeployableComponentAdapter<?>, SortedSet<FaultDetector>>(
						taskFaultDetectorsMap);

		return clonedTaskGraph;
	}

	/** {@inheritDoc} */
	@Override
	public boolean containsTask(IDeployableComponentAdapter<?> task) {
		if(task.isInstantiationRequired()) {
			// FIXME: Hack to get the DSE working again. To fix it, replace the calls that need the
			// instantiated component from an abstaract component, fix at the coller side.
			return abstractToInstanceTaskMap.containsKey(task) ? true : graph.containsVertex(task);
		}
		return graph.containsVertex(task);
	}

	/** {@inheritDoc} */
	@Override
	public boolean containsAnyTaskOf(Collection<IDeployableComponentAdapter<?>> taskSet) {
		for(IDeployableComponentAdapter<?> task : taskSet) {
			if(containsTask(task)) {
				return true;
			}
		}
		return false;
	}
}
