/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.sysmodel.mapping.comm;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemModelAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.comm.Message;
import org.jgrapht.DirectedGraph;
import org.jgrapht.Graphs;
import org.jgrapht.alg.BellmanFordShortestPath;
import org.jgrapht.alg.ConnectivityInspector;
import org.jgrapht.graph.DefaultDirectedGraph;
import org.jgrapht.graph.DefaultEdge;
import org.jgrapht.graph.DirectedSubgraph;

/**
 * Calculates the shortest path from the sending entity to its receivers on a physical/virtual
 * platform.
 * 
 * @author diewald
 */
public class ShortestPathRouter extends MessageRouterBase {

	/** Buffers the already calculated shortest paths for the sender {@link IResourceAdapter}. */
	private Map<IResourceAdapter<?>, BellmanFordShortestPath<IResourceAdapter<?>, DefaultEdge>> shortestPathForSender;

	/** Constructs a simple shortest path router for a given exploration run */
	public ShortestPathRouter(SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter) {
		super(systemModelAdapter);

		// TODO: Replace with more efficient Dijkstra, which needs an modification to contain the
		// shortest paths to all other vertices in the graph.
		shortestPathForSender =
				new ConcurrentHashMap<IResourceAdapter<?>, BellmanFordShortestPath<IResourceAdapter<?>, DefaultEdge>>();
	}

	/** {@inheritDoc} */
	@Override
	public synchronized DirectedGraph<IResourceAdapter<?>, DefaultEdge> calculateMessageRoute(
			IResourceAdapter<?> source, Collection<IResourceAdapter<?>> receiverResources) {
		// Generates a graph that contains the message's path to each receiver in the following
		// manner:
		// - Find the shortest path from the sender to each receiver (using BellmanFord)
		// - Construct a graph from the edge list of the shortest path to each receiver
		DefaultDirectedGraph<IResourceAdapter<?>, DefaultEdge> messageGraph =
				new DefaultDirectedGraph<IResourceAdapter<?>, DefaultEdge>(DefaultEdge.class);

		// Check, if the shortest paths have already been calculated for the given sender/source. If
		// not start a new shortest path analysis using BellmanFord.
		BellmanFordShortestPath<IResourceAdapter<?>, DefaultEdge> shortestPathForCurrentMessage =
				shortestPathForSender.get(source);
		if(shortestPathForCurrentMessage == null) {
			shortestPathForCurrentMessage =
					new BellmanFordShortestPath<IResourceAdapter<?>, DefaultEdge>(
							platformCommunicationGraph, source);
			shortestPathForSender.put(source, shortestPathForCurrentMessage);
		}

		for(IResourceAdapter<?> currentReceiver : receiverResources) {
			List<DefaultEdge> shortestMessagePath =
					shortestPathForCurrentMessage.getPathEdgeList(currentReceiver);
			// No valid path between source and receiver => invalid mapping;
			if(shortestMessagePath == null) {
				return null;
			}
			// Add all vertices and edges.
			Graphs.addAllEdges(messageGraph, platformCommunicationGraph, shortestMessagePath);
		}

		return messageGraph;
	}

	/** {@inheritDoc} */
	@Override
	public Map<IResourceAdapter<?>, DirectedGraph<IResourceAdapter<?>, DefaultEdge>>
			getRemainingMessageGraphs(Message message, IResourceAdapter<?> communicationResource,
					DirectedGraph<IResourceAdapter<?>, DefaultEdge> messageRoute) {
		// Get the direct successors of the current message sender, which form the senders of the
		// generated messages.
		List<IResourceAdapter<?>> directSuccessorVertices =
				Graphs.successorListOf(messageRoute, communicationResource);

		Set<IResourceAdapter<?>> targetVertices = new HashSet<IResourceAdapter<?>>();
		targetVertices.add(communicationResource);
		targetVertices.addAll(message.getReceiverResources());

		// Generate a working graph, that will be used to identify the correct allocation of
		// receivers to newly created messages. Therefore, the connection to the current sender is
		// cut.
		DirectedGraph<IResourceAdapter<?>, DefaultEdge> workingGraph =
				new DefaultDirectedGraph<IResourceAdapter<?>, DefaultEdge>(DefaultEdge.class);
		Graphs.addGraph(workingGraph, messageRoute);
		for(DefaultEdge edgeFromProcessedResource : workingGraph.edgesOf(communicationResource)) {
			workingGraph.removeEdge(edgeFromProcessedResource);
		}

		// Generate the routes of the successor messages using the connctivity inspector and the
		// workingGraph. For each successor message, a subgraph is determined that contains all
		// reachable receivers.
		ConnectivityInspector<IResourceAdapter<?>, DefaultEdge> messageConnectivity =
				new ConnectivityInspector<IResourceAdapter<?>, DefaultEdge>(workingGraph);
		Map<IResourceAdapter<?>, DirectedGraph<IResourceAdapter<?>, DefaultEdge>> remainingSubGraphs =
				new HashMap<IResourceAdapter<?>, DirectedGraph<IResourceAdapter<?>, DefaultEdge>>();
		for(IResourceAdapter<?> successor : directSuccessorVertices) {
			Set<IResourceAdapter<?>> successorRouteVertices =
					messageConnectivity.connectedSetOf(successor);
			successorRouteVertices.add(communicationResource);
			remainingSubGraphs.put(successor,
					new DirectedSubgraph<IResourceAdapter<?>, DefaultEdge>(messageRoute,
							successorRouteVertices, null));
		}

		return remainingSubGraphs;
	}
}
