/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.sysmodel.sched;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.stream.Collectors;

import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IDeployableComponentAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemModelAdapter;
import org.fortiss.af3.exploration.model.ExplorationSpecification;
import org.fortiss.af3.exploration.model.ExplorationTarget;
import org.fortiss.af3.exploration.model.time.DeadlineConstraint;

/**
 * Base class for schedulers.
 * 
 * @author barner
 */
public class SchedulerBase {
	/** Precision for double */
	protected final static int DOUBLE_PRECISION = 10;

	/** Representation of input system model */
	protected SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter;

	/** Contains the References to the {@link ExplorationTarget}s related to timing. */
	protected Collection<DeadlineConstraint> deadlines;

	/** Contains the reference deployable component -> deadline value. */
	protected Map<IDeployableComponentAdapter<?>, DeadlineConstraint> componentDeadlineSourceMap;

	/** Contains the reference deadline -> target component. */
	protected Map<DeadlineConstraint, IDeployableComponentAdapter<?>> deadlineTargetComponent;

	/** Constructs a scheduler for a given input model. */
	public SchedulerBase(SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter,
			ExplorationSpecification expSpec) {
		this.systemModelAdapter = systemModelAdapter;
		this.deadlines =
				expSpec.getTargets().stream().filter(t -> (t instanceof DeadlineConstraint))
						.map(DeadlineConstraint.class::cast).collect(Collectors.toList());

		componentDeadlineSourceMap =
				new HashMap<IDeployableComponentAdapter<?>, DeadlineConstraint>();
		deadlineTargetComponent = new HashMap<DeadlineConstraint, IDeployableComponentAdapter<?>>();

		createComponentDeadlineMap();
	}

	/**
	 * Creates two maps that <li>relate a {@link IDeployableComponentAdapter} which is defined as
	 * the "source" of a deadline to its {@link DeadlineConstraint} and <li>a map that relates
	 * {@link DeadlineConstraint}s with its "target" {@link IDeployableComponentAdapter}s.
	 */
	private void createComponentDeadlineMap() {
		if(deadlines == null) {
			return;
		}

		for(DeadlineConstraint currentDeadline : deadlines) {
			for(IDeployableComponentAdapter<?> deployableComponent : systemModelAdapter
					.getDeployableComponents()) {
				if(deployableComponent.getObject() == currentDeadline.getStartComponent()) {
					componentDeadlineSourceMap.put(deployableComponent, currentDeadline);
				} else if(deployableComponent.getObject() == currentDeadline.getEndComponent()) {
					deadlineTargetComponent.put(currentDeadline, deployableComponent);
				}
			}
		}
	}
}
