/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.sysmodel.sched.comm;

import static java.lang.Math.max;
import static java.lang.Math.min;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskgraph.InstantiatedAcyclicTaskGraphEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.InstantiatedTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.InstantiatedTaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.solution.StrictTTSchedule;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.solution.StrictTTSchedule.ResourceSchedule;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.solution.TimeSlot;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ICommunicationResourceAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IDeployableComponentAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IExecutionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IGatewayUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITransmissionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemModelAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph.AcyclicTaskGraph;
import org.fortiss.af3.exploration.alg.dse.sysmodel.comm.Message;
import org.fortiss.af3.exploration.alg.dse.sysmodel.mapping.comm.IMessageRouter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.mapping.comm.ShortestPathRouter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.sched.SchedulerBase;
import org.fortiss.af3.exploration.model.ExplorationSpecification;
import org.jgrapht.DirectedGraph;
import org.jgrapht.Graphs;
import org.jgrapht.graph.DefaultEdge;

import com.google.common.collect.HashMultimap;
import com.google.common.collect.Multimap;

/**
 * Simple TDMA bus scheduler that adds the communication time to a pure task schedule
 * ({@link StrictTTSchedule}).
 * 
 * @author huang
 */
public class TDMABusScheduler<S extends InstantiatedTaskMappingEntry, T extends InstantiatedTaskMappingEncoding<S>>
		extends SchedulerBase {

	/**
	 * References the {@link IMessageRouter} used to determine the path of generated {@link Message}
	 * s within this scheduler.
	 */
	private IMessageRouter messageRouter;

	/** Constructs a {@link TDMABusScheduler} for a given input model. */
	public TDMABusScheduler(SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter,
			ExplorationSpecification expSpec) {
		super(systemModelAdapter, expSpec);
		this.messageRouter = new ShortestPathRouter(systemModelAdapter);
	}

	/** Add simple TDMA bus schedule / communication time to a pure task schedule. */
	public void schedule(StrictTTSchedule<S, T> schedule, Collection<Message> generatedMessages,
			InstantiatedAcyclicTaskGraphEncoding<S, ?> acycItgEnc) {
		Multimap<TimeSlot, Message> emitableMessagesPerSlot = HashMultimap.create();
		Map<TimeSlot, Integer> reqRecvMsgsPerSlot = new HashMap<TimeSlot, Integer>();

		// Add the information about the allocated TimeSlots for the senders and the receivers of
		// the messages to the generated messages. This information stems from the schedule of the
		// the execution units
		Map<Message, DirectedGraph<IResourceAdapter<?>, DefaultEdge>> messageRoutes =
				new HashMap<Message, DirectedGraph<IResourceAdapter<?>, DefaultEdge>>();
		for(Message message : generatedMessages) {
			TimeSlot senderSlot = schedule.getTimeSlotOf(message.getSender());
			if(senderSlot == null) {
				// Message is not schedulable (e.g. an RC message) --> Skip scheduling this message.
				continue;
			}
			IDeployableComponentAdapter<?> senderComponent =
					senderSlot.getDeployedRequester().getSource();
			message.addMessageSenderTimeSlot(senderSlot);
			emitableMessagesPerSlot.put(senderSlot, message);

			for(InstantiatedTaskMappingEntry receiverEntry : message.getReceiverMappingEntries()) {
				// If the receiver entry is only executed during the next hyperperiod, do not
				// consider this receiver here.
				AcyclicTaskGraph correspondingTaskGraph =
						acycItgEnc.getTaskGraphOf(senderComponent);
				IDeployableComponentAdapter<?> receiverComponent = receiverEntry.getSource();
				if(correspondingTaskGraph.getGraph().containsEdge(senderComponent,
						receiverComponent)) {
					TimeSlot recvTimeSlot = schedule.getTimeSlotOf(receiverEntry);
					message.addMessageReceiverTimeSlots(recvTimeSlot);
					Integer numReqMsgs = reqRecvMsgsPerSlot.get(recvTimeSlot);
					if(numReqMsgs == null) {
						numReqMsgs = Integer.valueOf(0);
					}
					numReqMsgs++;
					reqRecvMsgsPerSlot.put(recvTimeSlot, numReqMsgs);
				}
			}
			messageRoutes.put(message, messageRouter.calculateMessageRoute(
					message.getSenderTimeSlot().getResource(), message.getReceiverResources()));
		}

		// Construct a message queue whose ordering defines the type of scheduling, i.e. ESF, EDF
		// etc. It is initially filled with the generated messages resulting from the encoding and
		// also captures messages generated by this message scheduler.
		LinkedList<Message> messageReleaseQueue = new LinkedList<Message>();
		LinkedList<Message> messageReadyQueue = new LinkedList<Message>();
		for(Message readyMessage : generatedMessages) {
			TimeSlot senderSlot = readyMessage.getSenderTimeSlot();
			if(senderSlot == null) {
				// Message is not schedulable (e.g. an RC message) --> Skip scheduling this message.
				continue;
			}

			if(!anyPredecessorSlotRequiresMsgs(senderSlot, reqRecvMsgsPerSlot) &&
					reqRecvMsgsPerSlot.get(senderSlot) == null) {
				messageReadyQueue.add(readyMessage);
			} else {
				messageReleaseQueue.add(readyMessage);
			}
		}
		Collections.sort(messageReadyQueue);

		// Process all messages present in the schedule. Note that new messages are generated during
		// message scheduling as a consequence of multi-hop messaging.
		while(!messageReadyQueue.isEmpty()) {
			// Update the currently considered message with the one that has the earliest start
			// time.
			Message currentMessage = messageReadyQueue.poll();

			// First, we need to retrieve the next communication resource along which the current
			// message being routed. This is the second element in the list of participating
			// communication resources, since the first one is the sender.
			DirectedGraph<IResourceAdapter<?>, DefaultEdge> messageGraph =
					messageRoutes.get(currentMessage);

			Set<DefaultEdge> senderResourceEdges =
					messageGraph.edgesOf(currentMessage.getSenderTimeSlot().getResource());
			// Here, we assume that a message may have only one source element. This assumption fits
			// to multi-cast messaging and the principle how new messages are generated.
			assert senderResourceEdges
					.size() == 1 : "Each message must have exactly one source slot!";
			ICommunicationResourceAdapter<?> communicationResource =
					(ICommunicationResourceAdapter<?>)messageGraph
							.getEdgeTarget(senderResourceEdges.iterator().next());

			// Create a time slot for the current message in the schedule of the next
			// communication resource (of the route).
			TimeSlot communicationSlot =
					createCommunicationTimeSlot(schedule, currentMessage, communicationResource);
			TimeSlot senderSlot = currentMessage.getSenderTimeSlot();
			communicationSlot.addPredecessor(currentMessage.getSenderTimeSlot());
			for(TimeSlot successorSlot : currentMessage.getReceiverTimeSlots()) {
				communicationSlot.addSuccessor(successorSlot);
				senderSlot.replaceSuccessor(successorSlot, communicationSlot);
				successorSlot.replacePredecessor(senderSlot, communicationSlot);
			}
			// It is required to shift the receiver tasks by the amount of time
			// that the communicationSlot requires for transmission. This is needed to
			// maintain causality.
			shiftReceiverSlot(schedule, currentMessage, communicationSlot);
			// Determine and generate successor messages from the current one.
			// TODO: remove dependency on the TimeSlot
			Collection<TimeSlot> reachedReceiverSlots = new HashSet<TimeSlot>();
			Collection<Message> succMessages = generateSuccessorMessages(schedule, currentMessage,
					messageGraph, communicationResource, communicationSlot, messageRoutes,
					reqRecvMsgsPerSlot, reachedReceiverSlots);
			messageReadyQueue.addAll(succMessages);
			// The ready and release queues are updated for the next iteration.
			List<TimeSlot> readySenderSlots = findNextSenderSlots(reachedReceiverSlots,
					reqRecvMsgsPerSlot, emitableMessagesPerSlot);
			updateMessageQueues(readySenderSlots, emitableMessagesPerSlot, messageReleaseQueue,
					messageReadyQueue);
		}
		if(!messageReleaseQueue.isEmpty()) {
			schedule = null;
		}
	}

	/**
	 * Determines whether the given {@link TimeSlot} or any of its predecessors must wait for
	 * {@link Message}s to arrive such that they can be executed. Therefore a map is required which
	 * counts the number of remaining {@link Message}s required as input for each {@link TimeSlot}.
	 * Note that {@link TimeSlot}s that do not require {@link Message}s as input may not be present
	 * in the map.
	 * <p>
	 * Implementation note: This method considers the case at which all predecessors are located on
	 * the same {@link IExecutionUnitAdapter} as the given {@link TimeSlot}, i.e. at which no
	 * {@link Message}s are required to pass the output from the predecessors to the input of the
	 * given successor {@link TimeSlot}.
	 * 
	 * @param msgSenderSlot
	 *            {@link TimeSlot} whose predecessors shall be checked for requiring {@link Message}
	 *            s (typically the sender {@link TimeSlot} of a {@link Message}.
	 * @param reqRecvMsgsPerSlot
	 *            Map containing information about the remaining {@link Message}s that are required
	 *            to execute a specific {@link TimeSlot} (-> key).
	 * @return Boolean indicating if any (relevant) predecessor is waiting for a {@link Message} to
	 *         be executed.
	 */
	private boolean anyPredecessorSlotRequiresMsgs(TimeSlot msgSenderSlot,
			Map<TimeSlot, Integer> reqRecvMsgsPerSlot) {
		LinkedList<TimeSlot> predecessorList = new LinkedList<TimeSlot>();
		predecessorList.addAll(msgSenderSlot.getPredecessors());
		while(!predecessorList.isEmpty()) {
			TimeSlot currentPredescessor = predecessorList.poll();
			if(currentPredescessor.getResource() != msgSenderSlot.getResource()) {
				predecessorList.addAll(currentPredescessor.getPredecessors());
				if(reqRecvMsgsPerSlot.get(currentPredescessor) != null &&
						reqRecvMsgsPerSlot.get(currentPredescessor) > 0) {
					return true;
				}
			}
		}

		return false;
	}

	// TODO: split this method into smaller ones, as this is spaghetti code.
	// TODO: another notion for messages is needed: Eliminate the message route map (annotate
	// directly)
	/**
	 * Generates (and registers) new messages for the remaining path of the message. If there are
	 * more than one resource "following" the just considered resource, multiple messages are
	 * generated.
	 * 
	 * @param schedule
	 *            The schedule for which messages shall be generated.
	 * @param currentMessage
	 *            The message for which its successors shall be generated.
	 * @param currentMessageGraph
	 *            The graph containing the path of the message in the platform.
	 * @param communicationResource
	 *            The communication {@link IResourceAdapter} for which the successor messages shall
	 *            be generated. This is typically the (single) successor vertex of the source of the
	 *            current message graph.
	 * @param communicationSlot
	 *            The schedule time slot which identifies the sender of the generated messages.
	 * @param messageRoutes
	 *            Map that associates {@link Message}s with the (remaining) routes/paths they need
	 *            to traverse.
	 * @param reqRecvMsgsPerSlot
	 *            Stores the number of messages a {@link TimeSlot} (of an
	 *            {@link IExecutionUnitAdapter} must receive in order to be executed. Used for
	 *            slot-shifting.
	 * @param reachedReceiverSlots
	 *            Collection of {@link TimeSlot}s that received a message.
	 * @return Collection of {@link Message}s that must be sent from the currently traversed
	 *         Resource such that the signal arrives at the target.
	 */
	private Collection<Message> generateSuccessorMessages(StrictTTSchedule<S, T> schedule,
			Message currentMessage,
			DirectedGraph<IResourceAdapter<?>, DefaultEdge> currentMessageGraph,
			IResourceAdapter<?> communicationResource, TimeSlot communicationSlot,
			Map<Message, DirectedGraph<IResourceAdapter<?>, DefaultEdge>> messageRoutes,
			Map<TimeSlot, Integer> reqRecvMsgsPerSlot, Collection<TimeSlot> reachedReceiverSlots) {
		Collection<Message> generatedMessages = new HashSet<Message>();

		// From the receiver time slots representing the logical receivers of the current message,
		// create a collection of the (platform) resources that onto which these time slots are
		// deployed.
		Collection<TimeSlot> currentMessageRecieverSlots = currentMessage.getReceiverTimeSlots();
		Collection<IResourceAdapter<?>> currentMessageReceiverResources =
				new ArrayList<IResourceAdapter<?>>();
		for(TimeSlot currentReceiverSlot : currentMessageRecieverSlots) {
			currentMessageReceiverResources.add(currentReceiverSlot.getResource());
		}

		// Retrieve the routes of the successor messages
		Map<IResourceAdapter<?>, DirectedGraph<IResourceAdapter<?>, DefaultEdge>> remainingRoutes =
				messageRouter.getRemainingMessageGraphs(currentMessage, communicationResource,
						currentMessageGraph);
		// Generate a new message for each direct successor in the communication graph.
		for(IResourceAdapter<?> nextCommunicationResource : Graphs
				.successorListOf(currentMessageGraph, communicationResource)) {
			// Check, if a new message is required, i.e. the current direct successor is no receiver
			// of the message.
			if(!currentMessageReceiverResources.contains(nextCommunicationResource)) {
				// Identify the route route corresponding to the current successor and generate a
				// list of receiver resources.
				DirectedGraph<IResourceAdapter<?>, DefaultEdge> successorRoute =
						remainingRoutes.get(nextCommunicationResource);
				HashSet<TimeSlot> successorReceiverSlots = new HashSet<TimeSlot>();
				for(TimeSlot receiverSlot : currentMessageRecieverSlots) {
					if(successorRoute.containsVertex(receiverSlot.getResource())) {
						successorReceiverSlots.add(receiverSlot);
					}
				}

				// Finally, generate the message.
				if(!successorReceiverSlots.isEmpty()) {
					Message successorMessage = new Message(currentMessage.getSignalAdapter(),
							communicationSlot, successorReceiverSlots);
					generatedMessages.add(successorMessage);
					messageRoutes.put(successorMessage, successorRoute);
				}
			} else {
				// Book keeping for the messages that have received the target TimeSlots.
				for(TimeSlot currentReceiverSlot : currentMessageRecieverSlots) {
					if(currentReceiverSlot.getResource() == nextCommunicationResource) {
						reachedReceiverSlots.add(currentReceiverSlot);
						Integer remainingMsgCount = reqRecvMsgsPerSlot.get(currentReceiverSlot);
						if(remainingMsgCount != null) {
							remainingMsgCount--;
							reqRecvMsgsPerSlot.put(currentReceiverSlot, remainingMsgCount);
						}
					}
				}
			}
		}

		return generatedMessages;
	}

	/**
	 * From a given collection of {@link TimeSlot}s that have received a {@link Message}, this
	 * method determines those {@link TimeSlot}s that are ready to emit {@link Message}s themselves.
	 * In order to determine these {@link TimeSlot}s a map that counts the missing number of
	 * {@link Message}s that must be received per {@link TimeSlot} and a map referencing the
	 * {@link Message}s that are emitted per {@link TimeSlot} are required.
	 * 
	 * @param reachedReceiverSlots
	 *            {@link TimeSlot}s that have received a {@link Message} in the current iteration of
	 *            the schedule.
	 * @param reqRecvMsgsPerSlot
	 *            Map counting the remaining number of required messages to execute the receiver
	 *            {@link TimeSlot}.
	 * @param emitableMessagesPerSlot
	 *            Map referencing the {@link Message} which are emitted after executing the
	 *            associated {@link TimeSlot} (-> key).
	 * @return List of {@link TimeSlot}s that may emit {@link Message}.
	 */
	private List<TimeSlot> findNextSenderSlots(Collection<TimeSlot> reachedReceiverSlots,
			Map<TimeSlot, Integer> reqRecvMsgsPerSlot,
			Multimap<TimeSlot, Message> emitableMessagesPerSlot) {
		List<TimeSlot> nextReadySenderSlots = new ArrayList<TimeSlot>();

		for(TimeSlot receiverSlot : reachedReceiverSlots) {
			LinkedList<TimeSlot> possiblyReadySuccessors = new LinkedList<TimeSlot>();
			possiblyReadySuccessors.add(receiverSlot);

			while(!possiblyReadySuccessors.isEmpty()) {
				TimeSlot currentSuccessor = possiblyReadySuccessors.poll();
				Integer requiredMsgNumForCurrentSuccessor =
						reqRecvMsgsPerSlot.get(currentSuccessor);
				if(requiredMsgNumForCurrentSuccessor != null &&
						requiredMsgNumForCurrentSuccessor == 0) {
					nextReadySenderSlots.add(currentSuccessor);
				} else if(requiredMsgNumForCurrentSuccessor == null) {
					possiblyReadySuccessors.addAll(currentSuccessor.getSuccessors());
				}
			}
		}

		return nextReadySenderSlots;
	}

	/**
	 * Modifies the given ready and release queues such that the (now) emittable messages are moved
	 * to the ready queue and are eliminated from the release queue.
	 * 
	 * @param readySenderSlots
	 *            List of {@link TimeSlot}s which are ready to emit {@link Message}s, i.e.
	 *            {@link TimeSlot}s which themselves (or their predecessors) have received all
	 *            required {@link Message}s those {@link TimeSlot}s require as input.
	 * @param emittableMessagesPerSlot
	 *            Map containing the {@link Message}s each {@link TimeSlot} emits, once it is ready.
	 * @param messageReleaseQueue
	 *            Queue containing the non-ready {@link Message}s.
	 * @param messageQueueReady
	 *            Queue containing the {@link Message}s that can be scheduled.
	 */
	private void updateMessageQueues(List<TimeSlot> readySenderSlots,
			Multimap<TimeSlot, Message> emittableMessagesPerSlot,
			LinkedList<Message> messageReleaseQueue, LinkedList<Message> messageQueueReady) {

		for(TimeSlot currentSlot : readySenderSlots) {
			Collection<Message> readyMessages = emittableMessagesPerSlot.get(currentSlot);
			messageQueueReady.addAll(readyMessages);
			messageReleaseQueue.removeAll(readyMessages);
		}
		Collections.sort(messageQueueReady);
	}

	/**
	 * Creates the {@link TimeSlot} on a communication resource (like an
	 * {@link ITransmissionUnitAdapter})
	 * for a given Message.
	 * 
	 * @param schedule
	 *            The currently considered schedule.
	 * @param message
	 *            Message, for which the slot has to be created.
	 * @param communicationResource
	 *            An {@link ITransmissionUnitAdapter} or {@link IGatewayUnitAdapter} in whose
	 *            schedule the {@link TimeSlot} is about being registered.
	 * @return The {@link TimeSlot} that has just been added to the schedule.
	 */
	private TimeSlot createCommunicationTimeSlot(StrictTTSchedule<S, T> schedule, Message message,
			ICommunicationResourceAdapter<?> communicationResource) {
		// final double minTransmissionDuration = 0.1;
		// Constant used for bit->MByte transformation.
		final double bitsPerMByte = 8 * 1024 * 1024;
		double messageStartTime = message.getSenderTimeSlot().getEndTime();

		// Determine the time required for the transmission of the message. A minimal transmission
		// time is added to avoid time slots with a duration of zero.
		double resourceWCTT = 0.0;
		if(communicationResource instanceof ITransmissionUnitAdapter) {
			resourceWCTT = ((ITransmissionUnitAdapter<?>)communicationResource)
					.getTransmissionDuration(message.getMessageSize() / bitsPerMByte);
		} else if(communicationResource instanceof IGatewayUnitAdapter) {
			resourceWCTT = ((IGatewayUnitAdapter<?>)communicationResource).getWCFT();
		}
		// resourceWCTT = Math.max(resourceWCTT, minTransmissionDuration);

		// Create a new slot for the message.
		TimeSlot communicationSlot = new TimeSlot(message, communicationResource, messageStartTime,
				messageStartTime + resourceWCTT);

		// Insert the created slot into the schedule of the communication resource.
		if(schedule.getResourceSchedule(communicationResource) != null) {
			shiftSlotToEarliestPossibleTime(schedule.getResourceSchedule(communicationResource),
					communicationSlot);
		}
		schedule.registerTimeSlot(communicationSlot, null, communicationResource);

		return communicationSlot;
	}

	/**
	 * Moves the given {@link TimeSlot} to the earliest 'position' in the given
	 * {@link ResourceSchedule} where it does not overlap with other {@link TimeSlot}s of the target
	 * {@link IResourceAdapter}.
	 * 
	 * @param resourceSchedule
	 *            The {@link ResourceSchedule} containing the {#link insertSlot}.
	 * @param insertSlot
	 *            The {@link TimeSlot} that shall be shifted to a non-overlapping 'position' of the
	 *            given {@link ResourceSchedule}.
	 */
	private void shiftSlotToEarliestPossibleTime(ResourceSchedule resourceSchedule,
			TimeSlot insertSlot) {
		double earliestPossibleStartTime = insertSlot.getStartTime();

		// Determine the slots of the resource schedule that are temporally located 'behind' the
		// start time of the slot to be inserted since these slots may impact the 'position' of the
		// slot to be inserted.
		TreeSet<TimeSlot> orderedResourceSchedule = new TreeSet<TimeSlot>();
		orderedResourceSchedule.addAll(resourceSchedule.keySet());
		TreeSet<TimeSlot> relevantTimeSlots = new TreeSet<TimeSlot>();
		relevantTimeSlots.addAll(orderedResourceSchedule.tailSet(insertSlot));

		// Depending on the existence of a temporally preceding (by the start time) slot, determine
		// the earliest possible start time of the slot to be inserted.
		// - lower(...) returning null implies no preceding slot present
		// - otherwise: the earliest possible start time is that of the inserted slot or the 'end
		// time' of the preceding slot
		if(orderedResourceSchedule.lower(insertSlot) != null) {
			earliestPossibleStartTime =
					Math.max(orderedResourceSchedule.lower(insertSlot).getEndTime(),
							insertSlot.getStartTime());
		}

		// Find the earliest possible start time for the slot to be inserted that is determined by a
		// sufficiently large 'time gap' relative to already present slots in the schedule.
		// NOTE: More "beautiful" solutions are possible, but cause errors if double values are used
		// due to precision errors during variable assignment.
		double timeGap = 0.0;
		for(TimeSlot nextTimeSlot : relevantTimeSlots) {
			timeGap = nextTimeSlot.getStartTime() - earliestPossibleStartTime;
			if(timeGap > insertSlot.getDuration())
				break;
			earliestPossibleStartTime = nextTimeSlot.getEndTime();
		}

		// Perform the shift of the slot to the earliest valid start time.
		double shiftTime = earliestPossibleStartTime - insertSlot.getStartTime();
		insertSlot.shift(shiftTime);
	}

	/**
	 * Shifts {@link TimeSlot}s that are receivers of the given message backwards on their
	 * respective {@link ResourceSchedule}, if required. The amount of time by which the receiving
	 * {@link TimeSlot}s are shifted is determined by the 'end time' of the {#link
	 * communicationSlot}.
	 * Subsequent slots, that are affected by the shift of a receiver, are also shifted until there
	 * exists no more temporal overlap between time slots.
	 * 
	 * @param schedule
	 *            The currently considered schedule.
	 * @param shiftMessage
	 *            The message
	 * @param communicationSlot
	 *            The {@link TimeSlot} allocated at a {@link ICommunicationResourceAdapter} whose
	 *            insertion (possibly) triggers the shift of other {@link TimeSlot}s.
	 */
	private void shiftReceiverSlot(StrictTTSchedule<S, T> schedule, Message shiftMessage,
			TimeSlot communicationSlot) {

		Comparator<TimeSlot> startTimeComparator = new TimeSlot.StartTimeSlotComparator();

		// Perform the shifting for each receiver slot of the message
		for(TimeSlot receiverSlot : shiftMessage.getReceiverTimeSlots()) {
			// Associates a TimeSlot with its direct predecessor TimeSlot.
			Multimap<TimeSlot, TimeSlot> succToPredSlotForShift = HashMultimap.create();
			succToPredSlotForShift.put(receiverSlot, communicationSlot);
			// List of successor TimeSlots that must be considered in the successive shifts.
			LinkedList<TimeSlot> succSlotsToBeShifted = new LinkedList<TimeSlot>();
			succSlotsToBeShifted.add(receiverSlot);
			// Amount of time the affected receivers/predecessors.
			double shiftTime = 0;

			while(!succSlotsToBeShifted.isEmpty()) {
				TimeSlot successorSlot = succSlotsToBeShifted.poll();
				double latestPredFinishTime = 0;
				for(TimeSlot predecessor : succToPredSlotForShift.get(successorSlot)) {
					latestPredFinishTime = max(latestPredFinishTime, predecessor.getEndTime());
				}

				shiftTime = latestPredFinishTime - successorSlot.getStartTime();

				// Is the shifting of the successor required? It may have been become irrelevant due
				// to shifting other TimeSlots on the same resource.
				if(shiftTime > 0) {
					// Shift slots of causality violating slot.
					HashSet<TimeSlot> shiftedSlots = shiftResourceSlots(successorSlot, shiftTime,
							schedule.getResourceSchedule(successorSlot.getResource()));
					assert (schedule.getResourceSchedule(successorSlot.getResource())
							.validateResourceSchedule());

					// Do not consider shifted TimeSlots that are located (temporally) before their
					// successors.
					Collection<TimeSlot> irrelevantSlots = new ArrayList<TimeSlot>();
					for(TimeSlot currentShiftedSlot : shiftedSlots) {
						if(currentShiftedSlot.getEndTime() <= successorSlot.getStartTime()) {
							irrelevantSlots.add(currentShiftedSlot);
						}
					}
					shiftedSlots.removeAll(irrelevantSlots);
					// Add remaining successors with causality violation to the slots to be
					// processed.
					for(TimeSlot currentShiftedSlot : shiftedSlots) {
						Collection<TimeSlot> succSlots = currentShiftedSlot.getSuccessors();
						for(TimeSlot succSlot : succSlots) {
							if(succSlot.getStartTime() < currentShiftedSlot.getEndTime()) {
								if(!succSlotsToBeShifted.contains(succSlot)) {
									succSlotsToBeShifted.add(succSlot);
								}
								succToPredSlotForShift.put(succSlot, currentShiftedSlot);
							}
						}
					}
					Collections.sort(succSlotsToBeShifted, startTimeComparator);
				}
			}
		}
		assert (schedule.validateSchedule(true));
	}

	/**
	 * Shifts the given {@link TimeSlot} in the given {@link ResourceSchedule} by the given amount
	 * of time. The method shifts all {@link TimeSlot}s that are in the time frame of the start time
	 * of the given {@link TimeSlot} and its time shift value. By successive shifting, it is ensured
	 * that no temporal overlaps occur in the given {@link ResourceSchedule}.
	 * 
	 * @param slotToShift
	 *            The slot that shall be shifted at the resource.
	 * @param shiftTime
	 *            The amount of time the given {@link TimeSlot} shall be shifted.
	 * @param receiverResourceSchedule
	 *            The {@link ResourceSchedule} that contains the given {@link TimeSlot}.
	 * @return Set of all shifted slot by this method.
	 */
	private HashSet<TimeSlot> shiftResourceSlots(TimeSlot slotToShift, double shiftTime,
			ResourceSchedule receiverResourceSchedule) {
		HashSet<TimeSlot> allShiftedSlots = new HashSet<TimeSlot>();

		Set<TimeSlot> receiverResourceSlots = new HashSet<TimeSlot>();
		receiverResourceSlots.addAll(receiverResourceSchedule.keySet());

		// Determine the slots that must be 'directly' shifted due to the communication slot.
		// The affected slots are those that are located between the 'end time' of the
		// communication slot and the start time of the receiver slot.
		List<TimeSlot> slotsToBeShifted = new LinkedList<TimeSlot>();
		for(TimeSlot slotAtReceiver : receiverResourceSchedule.keySet()) {
			if(slotAtReceiver.getStartTime() >= slotToShift.getStartTime() &&
					slotAtReceiver.getStartTime() < slotToShift.getStartTime() + shiftTime) {
				slotsToBeShifted.add(slotAtReceiver);
			}
		}

		// Start and end times of the block of shifted slots per round in the loop below.
		double shiftedBlockStartTime;
		double prevShiftedSlotEndTime = slotToShift.getStartTime() + shiftTime;

		// Perform the iterative shifting of the affected time slots.
		HashSet<TimeSlot> shiftedTimeSlots = new HashSet<TimeSlot>();
		while(!slotsToBeShifted.isEmpty()) {
			// Perform the actual shift of the defined slots and mark the start and end times of
			// the shifted slot 'block'.
			shiftedBlockStartTime = Double.MAX_VALUE;
			for(TimeSlot slotAtReceiver : slotsToBeShifted) {
				double blockShiftTime = prevShiftedSlotEndTime - slotAtReceiver.getStartTime();
				assert (blockShiftTime >= 0) : "The TDMAScheduler attempts to shift the TimeSlot " +
						slotAtReceiver + " towards an earlier point in time.";
				slotAtReceiver.shift(blockShiftTime);
				shiftedTimeSlots.add(slotAtReceiver);
				shiftedBlockStartTime = min(shiftedBlockStartTime, slotAtReceiver.getStartTime());
				prevShiftedSlotEndTime = slotAtReceiver.getEndTime();
			}
			slotsToBeShifted.clear();
			allShiftedSlots.addAll(shiftedTimeSlots);

			// Determine the slots that must be shifted due to the shift of the slots at the
			// beginning of the loop.
			for(TimeSlot slotAtReceiver : receiverResourceSlots) {
				if(!containsTimeSlot(shiftedTimeSlots, slotAtReceiver) &&
						slotAtReceiver.getStartTime() >= shiftedBlockStartTime &&
						slotAtReceiver.getStartTime() < prevShiftedSlotEndTime) {
					slotsToBeShifted.add(slotAtReceiver);
				}
			}

			shiftedTimeSlots.clear();
			Collections.sort(slotsToBeShifted);
		}
		return allShiftedSlots;
	}

	/**
	 * Evaluates whether {#link slot} is contained in the given set of {@link TimeSlot}s based on
	 * the {@code hashCode} method.
	 */
	private boolean containsTimeSlot(Set<TimeSlot> timeSlotSet, TimeSlot slot) {
		for(TimeSlot slotOfSet : timeSlotSet) {
			if(slotOfSet.hashCode() == slot.hashCode()) {
				return true;
			}
		}
		return false;
	}
}
