/*-------------------------------------------------------------------------+
| Copyright 2015 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.port.plot;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Shape;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Queue;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.solution.StrictTTSchedule;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.solution.StrictTTSchedule.ResourceSchedule;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.solution.TimeSlot;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ICommunicationResourceAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.comm.Message;
import org.jfree.chart.ChartMouseEvent;
import org.jfree.chart.ChartMouseListener;
import org.jfree.chart.ChartPanel;
import org.jfree.chart.annotations.CategoryLineAnnotation;
import org.jfree.chart.axis.ValueAxis;
import org.jfree.chart.entity.CategoryItemEntity;
import org.jfree.chart.entity.ChartEntity;
import org.jfree.chart.plot.CategoryPlot;

/**
 * Draws lines between the sender {@link TimeSlot}s, the participating {@link TimeSlot}s of
 * communication resource and the receiver {@link TimeSlot}s, if a {@link TimeSlot} of a
 * communication resource is clicked.
 * 
 * @author diewald
 */
public class MessagePathOnClicked implements ChartMouseListener {

	/** Holds the schedule that has been calculated for the selected solution. */
	private StrictTTSchedule<?, ?> schedule;

	/** The plot of the data (includes the data area and the naming etc.). */
	CategoryPlot plot;

	/** The actual "visual" element class showing the dataset. */
	ChartPanel chartPanel;

	/**
	 * Contains the path of all predecessor {@link TimeSlot}s for each {@link TimeSlot} reserved at
	 * {@link ICommunicationResourceAdapter}s.
	 */
	private Map<TimeSlot, List<TimeSlot>> predescessorMsgPath;

	/**
	 * Cache that remembers the {@link CategoryLineAnnotation}s such that they can be deleted again.
	 */
	Collection<CategoryLineAnnotation> paintedMessagePaths;

	/** Constructor. */
	public MessagePathOnClicked(CategoryPlot plot, ChartPanel chartPanel,
			StrictTTSchedule<?, ?> schedule) {
		paintedMessagePaths = new ArrayList<CategoryLineAnnotation>();
		this.plot = plot;
		this.chartPanel = chartPanel;
		this.schedule = schedule;

		predescessorMsgPath = new HashMap<TimeSlot, List<TimeSlot>>();

		List<TimeSlot> communicationSlotDescending = getCommSlotInDescendingOrder(schedule);

		for(TimeSlot currentSlot : communicationSlotDescending) {
			predescessorMsgPath.put(currentSlot, getTimeSlotPredecessorsOf(currentSlot));
		}
	}

	/** {@inheritDoc} */
	@Override
	public void chartMouseClicked(ChartMouseEvent event) {
		ChartEntity entity = event.getEntity();
		if(entity == null)
			return;

		deleteDrawnMessagePath();

		double timeToPixel = getTimeToPixelRatio();
		Double relativeItemXPos = getPositionOfSelectedItem(entity);
		if(relativeItemXPos == null) {
			// No TimeSlot selected.
			return;
		}

		/* Select the resource that corresponds to the given row */
		IResourceAdapter<?> resource = getResourceOf(entity);

		if(resource != null) {
			ResourceSchedule localSchedule = schedule.getResourceSchedule(resource);
			TimeSlot selectedSlot =
					getSelectedTimeSlot(timeToPixel, relativeItemXPos, localSchedule);

			if(selectedSlot != null) {
				// Traverse predecessor TimeSlots.
				paintedMessagePaths.addAll(drawLinesToPredecessors(selectedSlot,
						predescessorMsgPath.get(selectedSlot), plot));

				// Traverse the successors.
				paintedMessagePaths.addAll(drawLinesToSuccessors(selectedSlot, plot));
			}
		}
	}

	/** {@inheritDoc} */
	@Override
	public void chartMouseMoved(ChartMouseEvent event) {
		// Nothing to be done here...
	}

	/**
	 * Returns a list of {@link TimeSlot}s that are predecessors of the given {@link TimeSlot}, if
	 * the the Slot is associated with a {@link Message}, i.e. it is placed on a
	 * {@link ICommunicationResourceAdapter}.
	 * 
	 * @returns List of predecessors or an empty list if the given Slot does not refer to a
	 *          {@link Message}.
	 */
	private List<TimeSlot> getTimeSlotPredecessorsOf(TimeSlot currentSlot) {
		List<TimeSlot> predecessorSlots = new ArrayList<TimeSlot>();
		Message msg = currentSlot.getDeployedMessage();
		while(msg != null) {
			TimeSlot slot = msg.getSenderTimeSlot();
			msg = slot.getDeployedMessage();
			predecessorSlots.add(slot);
		}
		return predecessorSlots;
	}

	/**
	 * For a given {@link StrictTTSchedule}, this method returns all {@link TimeSlot}s of
	 * {@link ICommunicationResourceAdapter}s in descending order w.r.t. the {@link TimeSlot}s start
	 * times.
	 */
	private List<TimeSlot> getCommSlotInDescendingOrder(StrictTTSchedule<?, ?> schedule) {
		LinkedList<TimeSlot> communicationSlotDescending = new LinkedList<TimeSlot>();
		for(IResourceAdapter<?> currentResource : schedule.getResources()) {
			if(currentResource instanceof ICommunicationResourceAdapter) {
				ResourceSchedule resourceSchedule = schedule.getResourceSchedule(currentResource);
				communicationSlotDescending.addAll(resourceSchedule.keySet());
			}
		}
		if(!communicationSlotDescending.isEmpty()) {
			Collections.sort(communicationSlotDescending,
					new TimeSlot.ReverseStartTimeSlotComparator());
		}
		return communicationSlotDescending;
	}

	/**
	 * Determines the ratio of time to the pixels, i.e. which time interval a single pixel
	 * represents.
	 */
	private double getTimeToPixelRatio() {
		// Determine the Ratio between the time and the pixels (to calculate the time of the
		// slot by the position in the chart).
		ValueAxis axis = plot.getRangeAxis(0);
		double rangeAxis = axis.getRange().getUpperBound();
		double range = chartPanel.getChartRenderingInfo().getPlotInfo().getDataArea().getWidth();
		double timeToPixel = rangeAxis / range;
		return timeToPixel;
	}

	/** Returns the resource which is referred to by the given {@link ChartEntity}. */
	private IResourceAdapter<?> getResourceOf(ChartEntity entity) {
		Comparable<?> ckey = ((CategoryItemEntity)entity).getColumnKey();
		IResourceAdapter<?> resource = null;
		for(IResourceAdapter<?> currentResource : schedule.getResources()) {
			if(currentResource.getName().equals(ckey)) {
				resource = currentResource;
				break;
			}
		}
		return resource;
	}

	/** Determines the position of the given {@link ChartEntity} in the data plot. */
	private Double getPositionOfSelectedItem(ChartEntity entity) {
		Double relativeItemXPos = null;
		// Get the offset of the data cart relative to the overall plot (the position of the
		// entity is given in relatively to the overall plot).
		Rectangle2D plotArea = chartPanel.getChartRenderingInfo().getPlotInfo().getPlotArea();
		Rectangle2D dataArea = chartPanel.getChartRenderingInfo().getPlotInfo().getDataArea();
		double dataXOffset = dataArea.getX() - plotArea.getX();

		// Get the item's position and calculate its position in the data chart.
		// Thereby, ensure that the graphical representation of a datapoint has been clicked
		// (instanceof check).
		if(entity instanceof CategoryItemEntity) {
			Shape area = ((CategoryItemEntity)entity).getArea();
			relativeItemXPos = area.getBounds2D().getX() - dataXOffset;
		}
		return relativeItemXPos;
	}

	/**
	 * Determines and returns the selected {@link TimeSlot} utilizing the time to pixel ratio and
	 * the relative x coordinate (in pixel).
	 * 
	 * @returns The selected {@link TimeSlot} or {@code null} if no matching {@link TimeSlot} could
	 *          be identified.
	 */
	private TimeSlot getSelectedTimeSlot(double timeToPixel, double relativeItemXPos,
			ResourceSchedule localSchedule) {
		TimeSlot selectedSlot = null;

		double startXPosOfSlot = relativeItemXPos * timeToPixel;
		for(TimeSlot slot : localSchedule.keySet()) {
			if(slot.getStartTime() <= startXPosOfSlot && slot.getEndTime() >= startXPosOfSlot) {
				selectedSlot = slot;
				break;
			}
		}
		return selectedSlot;
	}

	/**
	 * Draws lines to predecessor {@link TimeSlot}s of the given {@link TimeSlot}. The
	 * {@link CategoryLineAnnotation}s that were drawn are returned by this method.
	 */
	private Collection<CategoryLineAnnotation> drawLinesToPredecessors(TimeSlot selectedSlot,
			List<TimeSlot> predecessorSlots, CategoryPlot plot) {
		Collection<CategoryLineAnnotation> paintedLines = new ArrayList<CategoryLineAnnotation>();
		TimeSlot currentSlot = selectedSlot;
		if(predecessorSlots != null) {
			for(TimeSlot predSlot : predecessorSlots) {
				paintedLines.add(drawTimeSlotLine(plot, predSlot, currentSlot));
				currentSlot = predSlot;
			}
		}
		return paintedLines;
	}

	/**
	 * Draws lines iteratively to successor {@link TimeSlot}s of the given {@link TimeSlot}. The
	 * {@link CategoryLineAnnotation}s that were drawn are returned by this method.
	 */
	private Collection<CategoryLineAnnotation> drawLinesToSuccessors(TimeSlot selectedSlot,
			CategoryPlot plot) {
		Collection<CategoryLineAnnotation> paintedLines = new ArrayList<CategoryLineAnnotation>();

		Queue<TimeSlot> successorSenderQueue = new LinkedList<TimeSlot>();
		successorSenderQueue.add(selectedSlot);
		while(!successorSenderQueue.isEmpty()) {
			TimeSlot senderToSuccessorSlot = successorSenderQueue.poll();
			Collection<TimeSlot> successorSlots = senderToSuccessorSlot.getSuccessors();
			if(successorSlots != null && (senderToSuccessorSlot
					.getResource() instanceof ICommunicationResourceAdapter)) {
				for(TimeSlot succSlot : successorSlots) {
					paintedLines.add(drawTimeSlotLine(plot, senderToSuccessorSlot, succSlot));

					if(!successorSenderQueue.contains(succSlot)) {
						successorSenderQueue.add(succSlot);
					}
				}
			}
		}

		return paintedLines;
	}

	/**
	 * Removes all previously drawn message paths of the selected communication {@link TimeSlot}.
	 */
	private void deleteDrawnMessagePath() {
		for(CategoryLineAnnotation line : paintedMessagePaths) {
			plot.removeAnnotation(line);
		}
		paintedMessagePaths.clear();
	}

	/** Draws a line between the Two given {@link TimeSlot}s. */
	private CategoryLineAnnotation drawTimeSlotLine(CategoryPlot plot, TimeSlot fromSlot,
			TimeSlot toSlot) {
		CategoryLineAnnotation messageLine =
				new CategoryLineAnnotation(fromSlot.getResource().getName(), fromSlot.getEndTime(),
						toSlot.getResource().getName(), toSlot.getStartTime(), Color.green,
						new BasicStroke(1.0f));

		plot.addAnnotation(messageLine);

		return messageLine;
	}
}
