/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.port.plot;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Paint;
import java.awt.geom.Rectangle2D;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.solution.StrictTTSchedule;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.solution.StrictTTSchedule.ResourceSchedule;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.solution.TimeSlot;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITransmissionUnitAdapter;
import org.jfree.chart.axis.CategoryAxis;
import org.jfree.chart.axis.ValueAxis;
import org.jfree.chart.entity.CategoryItemEntity;
import org.jfree.chart.entity.EntityCollection;
import org.jfree.chart.labels.CategoryItemLabelGenerator;
import org.jfree.chart.labels.ItemLabelAnchor;
import org.jfree.chart.labels.ItemLabelPosition;
import org.jfree.chart.plot.CategoryPlot;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.chart.renderer.category.CategoryItemRendererState;
import org.jfree.chart.renderer.category.GanttRenderer;
import org.jfree.data.category.CategoryDataset;
import org.jfree.data.gantt.GanttCategoryDataset;
import org.jfree.ui.RectangleEdge;
import org.jfree.ui.TextAnchor;

/**
 * Renders a Gantt chart for a {@link StrictTTSchedule}.
 * 
 * // TODO: generalize to common schedule class
 * 
 * @author diewald
 */
public class ScheduleGanttRenderer extends GanttRenderer {

	/** Holds the schedule that has been calculated for the selected solution. */
	private StrictTTSchedule<?, ?> schedule;

	// TODO: These variables are state trackers. Remove them.
	/** holds row from previous call */
	private int row;
	/** holds column from previous call */
	private int col;
	/** holds the time slot index */
	private int index;

	/**
	 * Constructor.
	 * 
	 * @param schedule
	 *            The scheduler provided by the solution
	 */
	public ScheduleGanttRenderer(final StrictTTSchedule<?, ?> schedule) {
		super();
		this.schedule = schedule;

		setBaseItemLabelGenerator(new ScheduleLabelGenerator(schedule));
		setSeriesToolTipGenerator(0, new ScheduleToolTipGenerator(schedule));
		setBaseItemLabelsVisible(true);
		setBaseItemLabelPaint(Color.BLACK);
		setDrawBarOutline(true);
		setBasePositiveItemLabelPosition(
				new ItemLabelPosition(ItemLabelAnchor.INSIDE6, TextAnchor.BOTTOM_CENTER));
	}

	/** {@inheritDoc} */
	@Override
	public Paint getItemPaint(int row, int col) {
		CategoryDataset dataset = super.getPlot().getDataset();

		// TODO: These variables are state trackers. Remove them. The index should be set
		// externally.
		if(this.row != row || this.col != col) {
			this.row = row;
			this.col = col;
			index = 0;
		}

		Comparable<?> rkey = dataset.getRowKey(row);
		Comparable<?> ckey = dataset.getColumnKey(col);
		// set default
		Color color = (Color)super.getItemPaint(row, col);

		/*
		 * Paint those time slots of execution units that are actually used. Tasks, voters, and
		 * messages are colored differently.
		 */
		// TODO: Label Gantt chart bars with task and message names
		// TODO: Indicate source and targets of messages (e.g., using dashed arrows)
		// NB: AD: This cannot be done with the GanttRenderer!
		if(rkey.equals("processing")) {
			/* Select the resource that corresponds to the given row */
			IResourceAdapter<?> resource = null;

			for(IResourceAdapter<?> currentResource : schedule.getResources()) {
				if(currentResource.getName().equals(ckey)) {
					resource = currentResource;
					break;
				}
			}

			/* Color tasks and voters differently */
			if(resource != null) {
				// TODO: make the index handling more dynamic
				// The indexing relies on the fact that the diagram is painted from the left to the
				// right (and so is the "call order").
				// HARD CODED HERE
				// we maintain a index ourselves to track the current sub task in the task
				// series.
				// we have to divide the index by 2, because we observe that for the drawing of
				// each item, the getItemPaint function is called twice.

				ResourceSchedule localSchedule = schedule.getResourceSchedule(resource);
				Object[] slots = localSchedule.keySet().toArray();

				if(index / 2 < slots.length) {
					TimeSlot slot = (TimeSlot)slots[index / 2];
					// TODO hard coded color
					if(slot.isVoter()) {
						color = Color.YELLOW;
					} else {
						if(resource instanceof ITransmissionUnitAdapter) {
							color = Color.BLUE;
						} else {
							color = Color.RED;
						}
					}
				}
			}
		}
		// Increase the time slot indexer for the next call.
		index++;
		return color;
	}

	/** {@inheritDoc} */
	@Override
	protected void drawTasks(Graphics2D g2, CategoryItemRendererState state, Rectangle2D dataArea,
			CategoryPlot plot, CategoryAxis domainAxis, ValueAxis rangeAxis,
			GanttCategoryDataset dataset, int row, int column) {

		int count = dataset.getSubIntervalCount(row, column);
		if(count == 0) {
			drawTask(g2, state, dataArea, plot, domainAxis, rangeAxis, dataset, row, column);
		}

		for(int subinterval = 0; subinterval < count; subinterval++) {

			RectangleEdge rangeAxisLocation = plot.getRangeAxisEdge();

			// value 0
			Number value0 = dataset.getStartValue(row, column, subinterval);
			if(value0 == null) {
				return;
			}
			double translatedValue0 =
					rangeAxis.valueToJava2D(value0.doubleValue(), dataArea, rangeAxisLocation);

			// value 1
			Number value1 = dataset.getEndValue(row, column, subinterval);
			if(value1 == null) {
				return;
			}
			double translatedValue1 =
					rangeAxis.valueToJava2D(value1.doubleValue(), dataArea, rangeAxisLocation);

			if(translatedValue1 < translatedValue0) {
				double temp = translatedValue1;
				translatedValue1 = translatedValue0;
				translatedValue0 = temp;
			}

			double rectStart = calculateBarW0(plot, plot.getOrientation(), dataArea, domainAxis,
					state, row, column);
			double rectLength = Math.abs(translatedValue1 - translatedValue0);
			double rectBreadth = state.getBarWidth();

			// DRAW THE BARS...
			Rectangle2D bar = null;

			if(plot.getOrientation() == PlotOrientation.HORIZONTAL) {
				bar = new Rectangle2D.Double(translatedValue0, rectStart, rectLength, rectBreadth);
			} else if(plot.getOrientation() == PlotOrientation.VERTICAL) {
				bar = new Rectangle2D.Double(rectStart, translatedValue0, rectBreadth, rectLength);
			}

			Rectangle2D completeBar = null;
			Rectangle2D incompleteBar = null;
			Number percent = dataset.getPercentComplete(row, column, subinterval);
			double start = getStartPercent();
			double end = getEndPercent();
			if(percent != null) {
				double p = percent.doubleValue();
				if(plot.getOrientation() == PlotOrientation.HORIZONTAL) {
					completeBar = new Rectangle2D.Double(translatedValue0,
							rectStart + start * rectBreadth, rectLength * p,
							rectBreadth * (end - start));
					incompleteBar = new Rectangle2D.Double(translatedValue0 + rectLength * p,
							rectStart + start * rectBreadth, rectLength * (1 - p),
							rectBreadth * (end - start));
				} else if(plot.getOrientation() == PlotOrientation.VERTICAL) {
					completeBar = new Rectangle2D.Double(rectStart + start * rectBreadth,
							translatedValue0 + rectLength * (1 - p), rectBreadth * (end - start),
							rectLength * p);
					incompleteBar = new Rectangle2D.Double(rectStart + start * rectBreadth,
							translatedValue0, rectBreadth * (end - start), rectLength * (1 - p));
				}
			}

			Paint seriesPaint = getItemPaint(row, column);
			g2.setPaint(seriesPaint);
			g2.fill(bar);

			if(completeBar != null) {
				g2.setPaint(getCompletePaint());
				g2.fill(completeBar);
			}
			if(incompleteBar != null) {
				g2.setPaint(getIncompletePaint());
				g2.fill(incompleteBar);
			}
			if(isDrawBarOutline() && state.getBarWidth() > BAR_OUTLINE_WIDTH_THRESHOLD) {
				g2.setStroke(getItemStroke(row, column));
				g2.setPaint(getItemOutlinePaint(row, column));
				g2.draw(bar);
			}

			CategoryItemLabelGenerator generator = getItemLabelGenerator(row, column);
			if(generator != null && isItemLabelVisible(row, column)) {
				drawItemLabel(g2, dataset, row, column, plot, generator, bar, false);
			}

			// collect entity and tool tip information...
			setupTooltip(state, dataset, row, column, bar);
		}
	}

	/** Tooltip when hovering a Pareto-optimal point. */
	private void setupTooltip(CategoryItemRendererState state, GanttCategoryDataset dataset,
			int row, int column, Rectangle2D bar) {
		if(state.getInfo() != null) {
			EntityCollection entities = state.getEntityCollection();
			if(entities != null) {
				String tip = null;
				if(getToolTipGenerator(row, column) != null) {
					tip = getToolTipGenerator(row, column).generateToolTip(dataset, row, column);
				}
				String url = null;
				if(getItemURLGenerator(row, column) != null) {
					url = getItemURLGenerator(row, column).generateURL(dataset, row, column);
				}
				CategoryItemEntity entity = new CategoryItemEntity(bar, tip, url, dataset,
						dataset.getRowKey(row), dataset.getColumnKey(column));
				entities.add(entity);
			}
		}
	}
}
