/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.port.plot;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.solution.StrictTTSchedule;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.solution.StrictTTSchedule.ResourceSchedule;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.solution.TimeSlot;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceAdapter;
import org.jfree.data.gantt.GanttCategoryDataset;
import org.jfree.data.general.AbstractSeriesDataset;

/**
 * This class converts a schedule from a given solution into a format that can be displayed as a
 * Gantt chart. It essentially creates links between the datapoints of the chart and the elements of
 * the schedule.
 * 
 * The rows represent the resources, while the columns are the time slots. The entries in the
 * diagram are then either tasks, messages, or voting elements.
 * 
 * TODO: generalize to common schedule class
 * 
 * @author diewald
 */
public class TaskSeriesCollectionFromSchedule extends AbstractSeriesDataset
		implements GanttCategoryDataset {

	/** Holds the schedule that has been calculated for the selected solution. */

	private StrictTTSchedule<?, ?> schedule;

	// in our drawing, the row keys are the category of scheduling objects, e.g. processing slot or
	// communication slot
	// the column keys are the capability resources, e.g. cores
	// the time slots are considered as intervals to draw in the Gantt chart

	/** Holds the tasks and messages of a row. */
	private List<String> rowkeys;
	/** Holds the resources (busses, execution units) with an associated name */
	private List<String> columnkeys;
	/** Links the names of the resources in {@code columnkeys} with the resource objects. */
	private Map<String, IResourceAdapter<?>> resourceKeyToResourceObject;

	/**
	 * Constructor.
	 * 
	 * @param schedule
	 *            The scheduler provided by the solution
	 */
	public TaskSeriesCollectionFromSchedule(StrictTTSchedule<?, ?> schedule) {
		this.schedule = schedule;
		rowkeys = new ArrayList<String>();
		columnkeys = new ArrayList<String>();
		resourceKeyToResourceObject = new HashMap<String, IResourceAdapter<?>>();

		/* Add a dummy processing row */
		rowkeys.add("processing");

		/*
		 * Add the resources (busses, execution units) in the columnkeys list in an aphabetical
		 * order
		 */
		SortedSet<String> sortedkeys = new TreeSet<String>();
		Set<IResourceAdapter<?>> resources = schedule.getResources();

		for(IResourceAdapter<?> resource : resources) {
			sortedkeys.add(resource.getName());
			resourceKeyToResourceObject.put(resource.getName(), resource);
		}
		columnkeys.addAll(sortedkeys);
	}

	/** {@inheritDoc} */
	@Override
	public Number getEndValue(int row, int column) {
		Comparable<?> rowKey = getRowKey(row);
		Comparable<?> columnKey = getColumnKey(column);
		return getEndValue(rowKey, columnKey);
	}

	/**
	 * {@inheritDoc} This method is intended to derive the overall end value.
	 */
	@Override
	public Number getEndValue(@SuppressWarnings("rawtypes") Comparable rowKey,
			@SuppressWarnings("rawtypes") Comparable columnKey) {
		/*
		 * First, filter for execution units, the last operation is always execution, not messaging.
		 */
		if(rowKey.equals("processing")) {
			if(resourceKeyToResourceObject.get(columnKey) != null) {
				// Retrieve all allocated time allocated for the referenced resource in the
				// schedule.
				ResourceSchedule resourceSchedule =
						schedule.getResourceSchedule(resourceKeyToResourceObject.get(columnKey));

				if(!resourceSchedule.isEmpty()) {
					TimeSlot lastslot = resourceSchedule.lastKey();
					return lastslot.getEndTime();
				}
			}
		}

		return null;
	}

	/** {@inheritDoc} */
	@Override
	public Number getStartValue(int row, int column) {
		Comparable<?> rowKey = getRowKey(row);
		Comparable<?> columnKey = getColumnKey(column);
		return getStartValue(rowKey, columnKey);
	}

	/** {@inheritDoc}. This method is intended to derive the overall start value. */
	@Override
	public Number getStartValue(@SuppressWarnings("rawtypes") Comparable rowKey,
			@SuppressWarnings("rawtypes") Comparable columnKey) {
		// First, filter for execution units, the first operation is always execution, not
		// messaging.
		if(rowKey.equals("processing")) {
			if(resourceKeyToResourceObject.get(columnKey) != null) {
				// Retrieve all allocated time allocated for the referenced resource in the
				// schedule.
				ResourceSchedule resourceSchedule =
						schedule.getResourceSchedule(resourceKeyToResourceObject.get(columnKey));

				if(!resourceSchedule.isEmpty()) {
					TimeSlot firstslot = resourceSchedule.firstKey();
					return firstslot.getStartTime();
				}
			}
		}

		return null;
	}

	/** {@inheritDoc} */
	@Override
	public int getColumnIndex(@SuppressWarnings("rawtypes") Comparable columnKey) {
		if(columnKey == null) {
			throw new IllegalArgumentException("Null 'columnKey' argument.");
		}
		return this.columnkeys.indexOf(columnKey);
	}

	/** {@inheritDoc} */
	@Override
	public Comparable<?> getColumnKey(int index) {
		return columnkeys.get(index);
	}

	/** {@inheritDoc} */
	@Override
	public List<?> getColumnKeys() {
		return columnkeys;
	}

	/** {@inheritDoc} */
	@Override
	public int getRowIndex(@SuppressWarnings("rawtypes") Comparable rowKey) {
		if(rowKey == null) {
			throw new IllegalArgumentException("Null 'rowKey' argument.");
		}
		return this.rowkeys.indexOf(rowKey);
	}

	/** {@inheritDoc} */
	@Override
	public Comparable<?> getRowKey(int index) {
		return rowkeys.get(index);
	}

	/** {@inheritDoc} */
	@Override
	public List<?> getRowKeys() {
		return rowkeys;
	}

	/** {@inheritDoc} */
	@Override
	public Number getValue(@SuppressWarnings("rawtypes") Comparable rowKey,
			@SuppressWarnings("rawtypes") Comparable columnKey) {
		return getStartValue(rowKey, columnKey);
	}

	/** {@inheritDoc} */
	@Override
	public int getColumnCount() {
		return columnkeys.size();
	}

	/** {@inheritDoc} */
	@Override
	public int getRowCount() {
		return rowkeys.size();
	}

	/** {@inheritDoc} */
	@Override
	public Number getValue(int row, int column) {
		return getStartValue(row, column);
	}

	/** {@inheritDoc} */
	@Override
	public Number getEndValue(int row, int column, int subinterval) {
		Comparable<?> rowKey = getRowKey(row);
		Comparable<?> columnKey = getColumnKey(column);
		return getEndValue(rowKey, columnKey, subinterval);
	}

	/** {@inheritDoc} */
	@Override
	public Number getEndValue(@SuppressWarnings("rawtypes") Comparable rowKey,
			@SuppressWarnings("rawtypes") Comparable columnKey, int subinterval) {
		// just return the starting time of a specific slot
		if(rowKey.equals("processing")) {
			if(resourceKeyToResourceObject.get(columnKey) != null) {

				ResourceSchedule resourceSchedule =
						schedule.getResourceSchedule(resourceKeyToResourceObject.get(columnKey));

				if(subinterval < resourceSchedule.size()) {
					Object[] objs = resourceSchedule.keySet().toArray();
					TimeSlot firstslot = (TimeSlot)objs[subinterval];
					// Return the end value of the slot minus some small value for better
					// visualization of separate tasks.
					if(firstslot.getStartTime() - 0.1 <= firstslot.getEndTime()) {
						return firstslot.getEndTime();
					}
					return firstslot.getEndTime() - 0.1;
				}
			}
		}

		return null;
	}

	/** {@inheritDoc} */
	@Override
	public Number getPercentComplete(int arg0, int arg1) {

		return null;
	}

	/** {@inheritDoc} */
	@Override
	public Number getPercentComplete(@SuppressWarnings("rawtypes") Comparable arg0,
			@SuppressWarnings("rawtypes") Comparable arg1) {

		return null;
	}

	/** {@inheritDoc} */
	@Override
	public Number getPercentComplete(int arg0, int arg1, int arg2) {

		return null;
	}

	/** {@inheritDoc} */
	@Override
	public Number getPercentComplete(@SuppressWarnings("rawtypes") Comparable arg0,
			@SuppressWarnings("rawtypes") Comparable arg1, int arg2) {

		return null;
	}

	/** {@inheritDoc} */
	@Override
	public Number getStartValue(int row, int column, int subinterval) {
		Comparable<?> rowKey = getRowKey(row);
		Comparable<?> columnKey = getColumnKey(column);
		return getStartValue(rowKey, columnKey, subinterval);
	}

	/** {@inheritDoc} */
	@Override
	public Number getStartValue(@SuppressWarnings("rawtypes") Comparable rowKey,
			@SuppressWarnings("rawtypes") Comparable columnKey, int subinterval) {
		// just return the starting time of a specific slot
		if(rowKey.equals("processing")) {
			if(resourceKeyToResourceObject.get(columnKey) != null) {

				ResourceSchedule resourceSchedule =
						schedule.getResourceSchedule(resourceKeyToResourceObject.get(columnKey));

				if(subinterval < resourceSchedule.size()) {
					Object[] objs = resourceSchedule.keySet().toArray();
					TimeSlot firstslot = (TimeSlot)objs[subinterval];
					// Return the end value of the slot plus some small value for better
					// visualization of separate tasks.
					if(firstslot.getStartTime() - 0.1 <= firstslot.getEndTime()) {
						return firstslot.getStartTime();
					}
					return firstslot.getStartTime() + 0.1;
				}
			}
		}

		return null;
	}

	/** {@inheritDoc} */
	@Override
	public int getSubIntervalCount(int row, int column) {
		Comparable<?> rowKey = getRowKey(row);
		Comparable<?> columnKey = getColumnKey(column);
		return getSubIntervalCount(rowKey, columnKey);
	}

	/** {@inheritDoc} */
	@Override
	public int getSubIntervalCount(@SuppressWarnings("rawtypes") Comparable rowKey,
			@SuppressWarnings("rawtypes") Comparable columnKey) {
		if(rowKey.equals("processing")) {
			if(resourceKeyToResourceObject.get(columnKey) != null) {

				ResourceSchedule resourceSchedule =
						schedule.getResourceSchedule(resourceKeyToResourceObject.get(columnKey));

				return resourceSchedule.size();
			}
		}
		return 0;
	}

	/** {@inheritDoc} */
	@Override
	public int getSeriesCount() {
		return getRowCount();
	}

	/** {@inheritDoc} */
	@Override
	public Comparable<?> getSeriesKey(int arg0) {
		if(arg0 < getRowCount()) {
			return rowkeys.get(arg0);
		}
		return null;
	}
}
