/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.port.plot;

import static javax.swing.JOptionPane.YES_OPTION;
import static org.fortiss.af3.exploration.alg.ExplorationAlgActivator.PLUGIN_ID;
import static org.fortiss.af3.exploration.alg.service.ExplorationService.getService;
import static org.fortiss.af3.exploration.alg.service.IExplorationTransformationService.TransformationContext.EXPORT;
import static org.fortiss.af3.exploration.util.ExplorationUtils.createExceptionStatus;
import static org.fortiss.af3.project.utils.ProjectUtils.getFileProject;
import static org.fortiss.tooling.kernel.utils.KernelModelElementUtils.getRootElements;

import java.awt.Color;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.geom.Ellipse2D;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.stream.Collectors;

import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.SwingConstants;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.af3.component.model.ComponentArchitecture;
import org.fortiss.af3.deployment.model.Deployment;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.InstantiatedTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.InstantiatedTaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.solution.IScheduleSolution;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.solution.StrictTTSchedule;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemModelAdapter;
import org.fortiss.af3.exploration.alg.guava.MutableSubClassToInstanceMap;
import org.fortiss.af3.exploration.alg.guava.SubClassToInstanceMap;
import org.fortiss.af3.exploration.alg.service.IExplorationTransformationService;
import org.fortiss.af3.exploration.model.ExplorationConstraint;
import org.fortiss.af3.exploration.model.ExplorationSpecification;
import org.fortiss.af3.exploration.model.ExplorationTarget;
import org.fortiss.af3.exploration.model.solutions.ExplorationResult;
import org.fortiss.af3.exploration.model.solutions.ExplorationSolutionSet;
import org.fortiss.af3.exploration.model.solutions.SingleExplorationSolutionMap;
import org.fortiss.af3.platform.model.PlatformArchitecture;
import org.fortiss.af3.project.model.FileProject;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;
import org.fortiss.tooling.kernel.service.IPersistencyService;
import org.jfree.chart.ChartFactory;
import org.jfree.chart.ChartMouseEvent;
import org.jfree.chart.ChartMouseListener;
import org.jfree.chart.ChartPanel;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.entity.ChartEntity;
import org.jfree.chart.entity.XYItemEntity;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.chart.plot.XYPlot;
import org.jfree.chart.renderer.xy.XYLineAndShapeRenderer;
import org.jfree.data.xy.XYDataItem;
import org.jfree.data.xy.XYDataset;
import org.jfree.data.xy.XYSeries;
import org.jfree.data.xy.XYSeriesCollection;
import org.jfree.ui.RectangleEdge;

import com.google.common.collect.ClassToInstanceMap;

/**
 * This class is responsible for creating a dialog that is used to display the identified
 * pareto-optimal solution of the design space exploration. It offers a 2D graph that visualizes the
 * values of two objectives versus each other. The objectives to be displayed can be chosen from
 * Comboboxes.
 * 
 * Whenever an solution is chosen from the graph, its values are displayed in a listing
 * below the graph. There also reside some buttons for further information on the chosen solution,
 * like the schedule visualization.
 * 
 * // TODO: generalize to common schedule class
 * 
 * @author diewald
 */
public class XYPlotter extends JFrame implements ChartMouseListener, ActionListener {

	// TODO remove this class; is only required for identifying the objective gui elements
	/** The object containing a pair of an objective's name and its value. */
	public class ObjectiveNameValueElement {
		/**
		 * Constructor.
		 */
		public ObjectiveNameValueElement(JTextField field, JLabel label) {
			objTextField = field;
			objLabel = label;
		}

		/** text field containing the objective value */
		public JTextField objTextField;
		/** label containing the objective's name */
		public JLabel objLabel;
	}

	/** Class to reference the {@link ExplorationTarget} plus providing a nice UI presentation. */
	private class ExplorationTargetItem {

		/** Target referenced by this item. */
		ExplorationTarget<?> target;

		/** Constructor. */
		ExplorationTargetItem(ExplorationTarget<?> target) {
			this.target = target;
		}

		/** Returns the {@link ExplorationTarget} associated with this target item. */
		public ExplorationTarget<?> getTarget() {
			return target;
		}

		/** {@inheritDoc} */
		@Override
		public String toString() {
			return target.getName();
		}
	}

	/** contains the objective name assigned to the X-axis of the chart */
	private ExplorationTargetItem Xaxis;

	/** contains the objective name assigned to the Y-axis of the chart */
	private ExplorationTargetItem Yaxis;

	/** The ComboBox for assigning an objective to the X-axis */
	private JComboBox<ExplorationTargetItem> XCombo;

	/** The ComboBox for assigning an objective to the Y-axis */
	private JComboBox<ExplorationTargetItem> YCombo;

	/** the chart displaying the selected objective values */
	private JFreeChart currentChart;

	/** holds/buffers the solution that has been selected in the chart */
	private SingleExplorationSolutionMap currentSolution;

	/** the DSE model containing the objective and constraint description */
	private ExplorationSpecification expSpec;

	/** The overall exploration result */
	private ExplorationSolutionSet explorationResult;

	/** The {@link SystemModelAdapter} that references the explored system. */
	private SystemModelAdapter<?, ?, ?, ?, ?, ?> exploredSystemModel;

	/**
	 * maps the names of the objectives to the GUI elements used to display the exact solution
	 * values
	 */
	private Map<ExplorationTarget<?>, ObjectiveNameValueElement> objectiveToGuiItems;

	/**
	 * maps the coordinates from the created points in the chart to a specific solution of the DSE
	 */
	private Map<XYDataItem, SingleExplorationSolutionMap> coordinateToSolution;

	/** buffer for the list of goals that exist in the DSE model */
	private List<ExplorationTarget<?>> goalList;

	/** Associates the {@link ExplorationTarget}s with their UI Elements. */
	private Map<ExplorationTarget<?>, ExplorationTargetItem> targetUIMap;

	/**
	 * Caches a
	 * private Exception transformationException = null;
	 * 
	 * /**
	 * Constructor.
	 * 
	 * @param expSpec
	 *            the exploration/optimization model (required for the identification of goals etc.)
	 * @param explorationResult
	 *            the DSE results which shall be visualized
	 * @param title
	 *            title of the dialog window
	 */
	public XYPlotter(ExplorationSpecification expSpec, ExplorationSolutionSet explorationResult,
			String title) {
		super(title);
		this.expSpec = expSpec;
		this.explorationResult = explorationResult;
		if(explorationResult
				.getInputModelAdapter() instanceof SystemModelAdapter<?, ?, ?, ?, ?, ?>) {
			this.exploredSystemModel =
					(SystemModelAdapter<?, ?, ?, ?, ?, ?>)explorationResult.getInputModelAdapter();
		}
		// TODO: throw an Exception if the SystemModelAdapter is not there.

		Xaxis = null;
		Yaxis = null;
		currentChart = null;
		currentSolution = null;
		coordinateToSolution = new HashMap<>();
		objectiveToGuiItems = new HashMap<>();
		targetUIMap = new HashMap<>();
	}

	/**
	 * Creates the dialog window that contains the chart for displaying the results of the DSE and
	 * some control elements.
	 */
	public void createPlot() {
		// TODO: Do we want to add constraints here?
		goalList = new ArrayList<ExplorationTarget<?>>();
		goalList.addAll(
				expSpec.getTargets().stream()
						.filter(t -> !(t instanceof ExplorationConstraint &&
								((ExplorationConstraint<?>)t).isImplicit()))
						.collect(Collectors.toList()));

		goalList.stream().forEach(t -> targetUIMap.put(t, new ExplorationTargetItem(t)));

		if(goalList.size() > 1) {
			// set x and y to the first two obj by default
			Xaxis = targetUIMap.get(goalList.get(0));
			Yaxis = targetUIMap.get(goalList.get(1));
		}
		// TODO: Nonsense else if? Can't this be handled differently?
		else if(goalList.size() > 0) {
			// get(0) safe due to if.
			Xaxis = targetUIMap.get(goalList.get(0));
			// get(0) safe due to if.
			Yaxis = targetUIMap.get(goalList.get(0));
		} else {
			System.out.println("ERROR: exploration should contain at least one objective!");
			return;
		}
		// create plot
		XYDataset dataset = createDataset();
		JFreeChart chart = createChart(dataset);
		ChartPanel chartPanel = new ChartPanel(chart);
		chartPanel.addChartMouseListener(this);
		currentChart = chart;

		// custom gui items
		// objective selection combo box
		JPanel objPanel = new JPanel();
		GridBagLayout objgridbag = new GridBagLayout();
		GridBagConstraints objcons = new GridBagConstraints();
		objPanel.setLayout(objgridbag);

		createChartControls(objPanel, objcons);

		JPanel editPanel = new JPanel();
		GridBagLayout editgridbag = new GridBagLayout();
		GridBagConstraints editcons = new GridBagConstraints();
		editPanel.setLayout(editgridbag);

		createDetailedGoalList(chartPanel, editPanel, editcons);

		// Button for displaying the schedule of a single solution
		editcons.gridx = 3;
		editcons.gridy = 0;
		editcons.weightx = 0.2;
		JButton showButton = new JButton("Show Schedule");
		editPanel.add(showButton, editcons);
		showButton.setActionCommand("showschedule");
		showButton.addActionListener(this);

		// Button for generating an AF3 deployment of a single solution
		JButton genButton = new JButton("Generate Single Deployment");
		editcons.gridx = 3;
		editcons.gridy = 1;
		editcons.weightx = 0.2;
		editPanel.add(genButton, editcons);
		genButton.setActionCommand("generateDeployment");
		genButton.addActionListener(this);

		// Button for generating an AF3 deployment for all solution
		JButton genAllButton = new JButton("Generate All Deployments");
		editcons.gridx = 3;
		editcons.gridy = 2;
		editcons.weightx = 0.2;
		editPanel.add(genAllButton, editcons);
		genAllButton.setActionCommand("generateAllDeployments");
		genAllButton.addActionListener(this);

		// put everything together
		JPanel content = new JPanel();
		GridBagLayout gridbag = new GridBagLayout();
		content.setLayout(gridbag);
		GridBagConstraints cons = new GridBagConstraints();
		cons.fill = GridBagConstraints.BOTH;
		cons.gridx = 0;
		cons.gridy = 1;
		objcons.weightx = 0.8;
		content.add(chartPanel, cons);
		cons.gridx = 0;
		cons.gridy = 2;
		objcons.weightx = 0.1;
		content.add(objPanel, cons);
		cons.gridx = 0;
		cons.gridy = 3;
		objcons.weightx = 0.1;
		content.add(editPanel, cons);
		// set pane
		setContentPane(content);
	}

	/**
	 * Creates the controls (two comboboxes and a button) that determine which information shall be
	 * illustrated in the diagram.
	 * 
	 * @param objPanel
	 * @param objcons
	 */
	private void createChartControls(JPanel objPanel, GridBagConstraints objcons) {
		XCombo = new JComboBox<ExplorationTargetItem>();
		// add objectives
		for(int i = 0; i < goalList.size(); i++) {
			XCombo.addItem(targetUIMap.get(goalList.get(i)));
		}
		if(Xaxis != null)
			XCombo.setSelectedItem(Xaxis);

		YCombo = new JComboBox<ExplorationTargetItem>();
		// add objectives
		for(int i = 0; i < goalList.size(); i++) {
			YCombo.addItem(targetUIMap.get(goalList.get(i)));
		}
		if(Yaxis != null)
			YCombo.setSelectedItem(Yaxis);

		JButton redrawButton = new JButton("Update Chart");
		redrawButton.setActionCommand("updatechart");
		redrawButton.addActionListener(this);

		objcons.fill = GridBagConstraints.BOTH;
		objcons.ipady = 8;
		objcons.gridx = 0;
		objcons.gridy = 0;
		objcons.weightx = 0.4;
		objPanel.add(XCombo, objcons);
		objcons.gridx = 2;
		objcons.gridy = 0;
		objcons.weightx = 0.4;
		objPanel.add(YCombo, objcons);
		objcons.gridx = 3;
		objcons.gridy = 0;
		objcons.weightx = 0.2;
		objPanel.add(redrawButton, objcons);
	}

	/**
	 * Creates a two column grid for displaying the detailed objective (and constraint) values of
	 * chosen solution.
	 * 
	 * @param chartPanel
	 * @param editPanel
	 * @param editcons
	 */
	private void createDetailedGoalList(ChartPanel chartPanel, JPanel editPanel,
			GridBagConstraints editcons) {
		// TODO: Make the value display non-editable.
		for(int i = 0; i < goalList.size(); i++) {
			ExplorationTarget<?> target = goalList.get(i);
			JTextField goalField = new JTextField(12);
			goalField.setHorizontalAlignment(SwingConstants.LEFT);
			goalField.setActionCommand(target.getName());
			goalField.addActionListener(chartPanel);
			JLabel goalFieldLabel = new JLabel(target.getName() + ": ");
			editcons.fill = GridBagConstraints.BOTH;
			editcons.gridx = 0;
			editcons.gridy = i;
			editcons.weightx = 0.4;
			editPanel.add(goalFieldLabel, editcons);
			editcons.gridx = 1;
			editcons.gridy = i;
			editcons.weightx = 0.4;
			editPanel.add(goalField, editcons);

			objectiveToGuiItems.put(target,
					new ObjectiveNameValueElement(goalField, goalFieldLabel));
		}
	}

	/**
	 * This method creates the chart i.e. the visualization of the dataset that is created from the
	 * solutions of the DSE.
	 * 
	 * @param dataset
	 *            a dataset compatible with the jfreechart framework; contains value pairs
	 * @return the chart
	 */
	public JFreeChart createChart(XYDataset dataset) {
		// create the chart...
		final JFreeChart chart = ChartFactory.createXYLineChart("Design Exploration Result", // chart
																								// title
				Xaxis.toString(), // x axis label
				Yaxis.toString(), // y axis label
				dataset, // data
				PlotOrientation.VERTICAL, true, // include legend
				true, // tooltips
				false // urls
		);

		chart.setBackgroundPaint(Color.white);
		chart.getLegend().setPosition(RectangleEdge.TOP);
		chart.getTitle().visible = false;
		// get a reference to the plot for further customization...
		final XYPlot plot = chart.getXYPlot();
		plot.setBackgroundPaint(Color.lightGray);
		plot.setDomainGridlinePaint(Color.white);
		plot.setRangeGridlinePaint(Color.white);

		final XYLineAndShapeRenderer renderer = new XYLineAndShapeRenderer();
		renderer.setSeriesLinesVisible(0, false);
		// Rectangle(-5,-5,10,10)
		renderer.setSeriesShape(0, new Ellipse2D.Float(-5, -5, 10, 10));

		plot.setRenderer(renderer);

		return chart;
	}

	/**
	 * Triggers an update of the dataset to be displayed. required, when the objectives to be
	 * displayed are changed.
	 * 
	 * @return the updated dataset
	 */
	public XYDataset createDataset() {

		XYSeriesCollection dataset = new XYSeriesCollection();
		updateDataSetFromResult(dataset);

		return dataset;
	}

	/**
	 * This method converts a the goal values of an identified solution to a usable dataset for the
	 * chart.
	 * 
	 * @param dataset
	 */
	public void updateDataSetFromResult(XYSeriesCollection dataset) {
		dataset.removeAllSeries();

		XYSeries series = new XYSeries("Exploration Result");
		// add the result to series
		coordinateToSolution.clear();

		for(SingleExplorationSolutionMap solution : explorationResult.getSolutions()) {
			/* Retrieve the solution values of the selected objectives. */
			Map<ExplorationTarget<?>, ExplorationResult<?>> objValues =
					solution.getSolutionMap().map();

			Object xResultVal = null;
			Object yResultVal = null;
			if(Xaxis != null && Yaxis != null) {
				ExplorationResult<?> xExpResult = objValues.get(Xaxis.getTarget());
				ExplorationResult<?> yExpResult = objValues.get(Yaxis.getTarget());
				if(xExpResult != null && yExpResult != null) {
					xResultVal = xExpResult.getResult();
					yResultVal = yExpResult.getResult();
				}
			}

			/* Convert the values to the data items. */
			if(xResultVal instanceof Double && yResultVal instanceof Double) {
				XYDataItem item = new XYDataItem((Double)xResultVal, (Double)yResultVal);
				series.add(item);
				coordinateToSolution.put(item, solution);
			} else {
				System.out.println("Error drawing chart: wrong x/y axis setting!");
				break;
			}
		}
		dataset.addSeries(series);
	}

	/** {@inheritDoc} */
	@Override
	public void chartMouseClicked(ChartMouseEvent e) {
		ChartEntity entity = e.getEntity();
		if(entity instanceof XYItemEntity) {
			XYItemEntity xyentity = (XYItemEntity)entity;
			XYDataset dataset = xyentity.getDataset();
			double x = dataset.getXValue(xyentity.getSeriesIndex(), xyentity.getItem());
			double y = dataset.getYValue(xyentity.getSeriesIndex(), xyentity.getItem());
			XYDataItem item = new XYDataItem(x, y);
			currentSolution = coordinateToSolution.get(item);

			Optional<SingleExplorationSolutionMap> singleExpSol = explorationResult.getSolutions()
					.stream().filter(r -> r == currentSolution).findAny();
			if(singleExpSol.isPresent()) {
				SingleExplorationSolutionMap currentSolMap = singleExpSol.get();
				Map<ExplorationTarget<?>, ExplorationResult<?>> objectiveValues =
						currentSolMap.getSolutionMap().map();

				// TODO: identifying the elements by their names is not good style, the map is not
				// required.
				/* Update the list of exact objective values */
				if(objectiveValues != null) {
					// display the objective values
					for(int i = 0; i < goalList.size(); i++) {
						ExplorationTarget<?> target = goalList.get(i);
						ExplorationResult<?> expResult = objectiveValues.get(target);
						if(objectiveToGuiItems.get(target) != null &&
								(expResult != null && expResult.getResult() instanceof Double)) {
							objectiveToGuiItems.get(target).objTextField
									.setText(Double.toString((Double)expResult.getResult()));
						}
					}
				}
			}
		}
	}

	/** {@inheritDoc} */
	@Override
	public void chartMouseMoved(ChartMouseEvent arg0) {
		// nothing to be done here
	}

	/** {@inheritDoc} */
	@Override
	public void actionPerformed(ActionEvent e) {
		if("updatechart".equals(e.getActionCommand())) {
			// update x y axis according to the combo box
			Xaxis = (ExplorationTargetItem)XCombo.getSelectedItem();
			Yaxis = (ExplorationTargetItem)YCombo.getSelectedItem();
			XYPlot plot = (XYPlot)currentChart.getPlot();
			plot.getDomainAxis().setLabel(Xaxis.toString());
			plot.getRangeAxis().setLabel(Yaxis.toString());
			// update data set and redraw the chart
			updateDataSetFromResult((XYSeriesCollection)plot.getDataset());
		} else if("generateDeployment".equals(e.getActionCommand())) {
			if(currentSolution != null) {

				String deploymentName = (String)JOptionPane.showInputDialog(this,
						"Generate Deployment", "Enter a name for the Deployment",
						JOptionPane.QUESTION_MESSAGE, null, null, "_Deployment");
				int generateSCSpcec = 0;
				if(deploymentName != null) {
					// if(pickFirstInstanceOf(SafetyComplianceConstraint.class, dse
					// .getTargetSpecification().getTargets()) != null) {
					// generateSCSpcec =
					// JOptionPane.showConfirmDialog(this,
					// "Also export the generated safety compliance model?");
					// }
					try {
						transformSingleDeployment(currentSolution, deploymentName,
								generateSCSpcec == YES_OPTION);
					} catch(Exception ex) {
						createExceptionStatus(ex, PLUGIN_ID, "Deployment Export Failed.");
					}
				}
			} else {
				// message box
				JOptionPane.showMessageDialog(this,
						"Cannot perform model transformation : No Exploration Solution Selected!");
			}
		} else if("generateAllDeployments".equals(e.getActionCommand())) {
			String deploymentName = (String)JOptionPane.showInputDialog(this,
					"Generate All Deployments", "Enter a name prefix for the resulting deployments",
					JOptionPane.QUESTION_MESSAGE, null, null, "GenDeployment_");
			if(deploymentName != null) {
				int generateSCSpcec = 0;
				// if(pickFirstInstanceOf(SafetyComplianceConstraint.class, dse
				// .getTargetSpecification().getTargets()) != null) {
				// generateSCSpcec =
				// JOptionPane.showConfirmDialog(this,
				// "Also export the generated safety compliance model?");
				// }
				// Deployment generation has NOT been cancelled.
				if(deploymentName.isEmpty()) {
					deploymentName = "GenDeployment_";
				}
				try {
					generateAllDeployments(deploymentName, generateSCSpcec == YES_OPTION);
				} catch(Exception ex) {
					createExceptionStatus(ex, PLUGIN_ID, "Deployment Export Failed.");
				}
			}
		} else if("showschedule".equals(e.getActionCommand())) {
			IScheduleSolution<?> scheduleSolution =
					currentSolution.getSolutionModel(IScheduleSolution.class);
			if(scheduleSolution instanceof StrictTTSchedule<?, ?>) {
				SchedulePlotter sp = new SchedulePlotter((StrictTTSchedule<?, ?>)scheduleSolution);
				sp.createPlot();
			} else if(currentSolution == null) {
				// No solution has been selected
				JOptionPane.showMessageDialog(this,
						"Cannot display a schedule : No Exploration Solution Selected!");
			} else {
				// The selected solution does not have a valid schedule type
				JOptionPane.showMessageDialog(this,
						"Cannot display a schedule : The selected solution provides no supported schedule!",
						"Invalid schedule detected", JOptionPane.ERROR_MESSAGE);
			}
		}
	}

	/**
	 * Transform the given solution to an AF3 {@link Deployment} and sets its name to the given
	 * string. After the transformation the deployment is injected to the AF3 project.
	 * 
	 * @throws Exception
	 */
	private Deployment transformSingleDeployment(SingleExplorationSolutionMap solutionMap,
			String deploymentName, boolean generateSCSpec) throws Exception {
		// Deployment generation has NOT been cancelled.
		if(deploymentName.isEmpty()) {
			deploymentName = "_Deployment";
		}

		SubClassToInstanceMap<EObject> baseModels = MutableSubClassToInstanceMap.create();
		// TODO: instead of querying the systemModelAdapter here, we should use the
		// explorationResult -> input models.
		ComponentArchitecture compArch =
				(ComponentArchitecture)exploredSystemModel.getLogicalArchitecture().getObject();
		PlatformArchitecture platArch =
				(PlatformArchitecture)exploredSystemModel.getPlatformArchitecture().getObject();
		baseModels.put(ComponentArchitecture.class, compArch);
		baseModels.put(PlatformArchitecture.class, platArch);
		if(generateSCSpec) {
			// SafetyComplianceConstraint scConstr =
			// pickFirstInstanceOf(SafetyComplianceConstraint.class, dse
			// .getTargetSpecification().getTargets());
			// SafetyComplianceSpecification scSpec = scConstr.getSafetyComplianceSpecification();
			// baseModels.put(SafetyComplianceSpecification.class, scSpec);
		}

		// TODO: The context must be an export here. In a future version, it must be
		// possible to specify, how the transformed model can be injected / exported.
		ITopLevelElement modelContext =
				IPersistencyService.getInstance().getTopLevelElementFor(compArch);

		Collection<Class<? extends EObject>> solutionModelTypes = new HashSet<>();
		solutionModelTypes.add(Deployment.class);
		if(generateSCSpec) {
			// solutionModelTypes.add(SafetyComplianceSpecification.class);
		}

		Deployment deployment = null;
		ClassToInstanceMap<EObject> solutionModels =
				getService(IExplorationTransformationService.class).getTransformedModels(EXPORT,
						modelContext, solutionMap, baseModels, solutionModelTypes, deploymentName);
		deployment = solutionModels.getInstance(Deployment.class);

		// Derive the ComponentType from the Voter Specification.
		// for(VoterType voterSpec : getChildrenWithType(compArch, VoterType.class)) {
		// IModelElement voter = voterSpec.getSpecificationOf();
		// if(voter instanceof Component) {
		// ComponentType cType = getAnnotation(voter, ComponentType.class);
		// modelContext.runAsNonDirtyingCommand(() -> cType
		// .setComponentType(ComponentType_t.VOTER));
		// }
		// }

		// if(generateSCSpec) {
		// exportSCSpec(getFileProject(deployment),
		// solutionModels.getInstance(SafetyComplianceSpecification.class), modelContext,
		// deploymentName);
		// }

		return deployment;
	}

	// /**
	// * Exports a generated {@link SafetyComplianceSpecification} model an stores it on the file
	// * system in the same folder where the {@link FileProject} resides.
	// */
	// private void exportSCSpec(FileProject fp, SafetyComplianceSpecification scSpec,
	// ITopLevelElement modelContext, String transformedModelName) {
	// ResourceSet prjResSet = fp.eResource().getResourceSet();
	//
	// URI prjURI = fp.eResource().getURI().trimSegments(1);
	// Resource resource =
	// prjResSet.createResource(prjURI.appendSegment(transformedModelName +
	// ".safetycompliance"));
	//
	// modelContext.runAsNonDirtyingCommand(() -> resource.getContents().add(scSpec));
	// }

	/**
	 * Transforms each found solution to a {@link Deployment} and injects each into the parent AF3
	 * project.
	 * 
	 * @throws Exception
	 *             if the transformation of any solution to a deployment fails.
	 */
	private <S extends InstantiatedTaskMappingEntry, T extends InstantiatedTaskMappingEncoding<S>>
			void generateAllDeployments(final String deploymentName, boolean generateSCSpec)
					throws Exception {
		Collection<Deployment> transformedDeployments = new ArrayList<>();

		Integer deploymentNumber = 1;
		for(SingleExplorationSolutionMap solution : explorationResult.getSolutions()) {
			transformedDeployments.add(transformSingleDeployment(solution,
					deploymentName + "_" + deploymentNumber, generateSCSpec));
			deploymentNumber++;
		}

		// Defer the injection of deployments such that a failure in the transformation does not
		// inject deployments corresponding to a subset of the solutions to the parent AF3 project.
		for(Deployment curDeployment : transformedDeployments) {
			ITopLevelElement modelContext =
					IPersistencyService.getInstance().getTopLevelElementFor(expSpec);
			modelContext.runAsCommand(() -> addDeploymentToProject(curDeployment));
		}
	}

	/**
	 * Actually generates the an AF3 deployment model from a given solution in the provided model
	 * context and registers the resulting deployment (and a possibly generated
	 * {@link ComponentArchitecture}) in the {@link FileProject} of the DSE.
	 */
	private <S extends InstantiatedTaskMappingEntry, T extends InstantiatedTaskMappingEncoding<S>>
			void addDeploymentToProject(Deployment deployment) {
		ComponentArchitecture deployedArchitecture = (ComponentArchitecture)deployment
				.getComponentArchitectureReference().getReference();
		FileProject fp = getFileProject(expSpec);
		if(!getRootElements(expSpec, ComponentArchitecture.class).contains(deployedArchitecture)) {
			fp.getRootElements().add(deployedArchitecture);
		}
		fp.getRootElements().add(deployment);
	}
}
