/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.service;

import java.util.Collection;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.FlatPhenotypeMap;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.dse.modeltransformation.ITransformationModule;
import org.fortiss.af3.exploration.alg.dse.modeltransformation.ITransformationStrategy;
import org.fortiss.af3.exploration.alg.guava.SubClassToInstanceMap;
import org.fortiss.af3.exploration.alg.service.internal.ExplorationTransformationService;
import org.fortiss.af3.exploration.model.solutions.SingleExplorationSolutionMap;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;

import com.google.common.collect.ClassToInstanceMap;

/**
 * {@link IExplorationService} interface of the {@link IExplorationTransformationService}.
 * It enables users to obtain transformed EMF models of a given type from a provided (candidate)
 * solution calculated by the DSE and the respective base models.
 * A transformed model is returned if:
 * <li>The provided input (base models and candidate solution
 * encodings) match the required ones.</li>
 * <li>All required {@link ITransformationModule}s are
 * present.</li>
 * <li>None of the executed {@link ITransformationModule}s fails.</li>
 * <p>
 * The {@link IExplorationTransformationService} is organized by {@link ITransformationStrategy}s
 * and {@link ITransformationModule}s. The modules encapsulate isolated steps in a transformation
 * process that are defined in the strategy and the {@link ITransformationModule} interface. The
 * strategies link the {@link ITransformationModule}s and are responsible for the correct order of
 * execution.
 * <p>
 * The transformation process (--> strategy) is organized into sub-Modules encapsulating single
 * steps. Each of the sub-Modules may define further submodules that can be also
 * {@link ITransformationStrategy}s (Composite pattern).
 * <p>
 * The {@link IExplorationTransformationService} differs between two context from which the
 * transformation has been triggered and which must be given. If the transformation is triggered by
 * a {@link TransformationContext#DSE_INTERNAL} context, the resulting model will be "floating".
 * Otherwise, if the transformation is launched with a {@link TransformationContext#EXPORT} context,
 * the resulting model will be injected in the corresponding AF3 project.
 * 
 * @author diewald
 */
public interface IExplorationTransformationService
		extends IExplorationService<ITransformationModule<?>> {
	/**
	 * Enum describing the context from which the transformation is requested. It is used to limit
	 * the number of transformations to the required minimum to obtain a better DSE performance.
	 */
	public static enum TransformationContext {
		/** "Full" transformation, including virtual links. */
		EXPORT,
		/** "Minimal" transformation, e.g. for an export towards the SafetyComplianceChecker. */
		DSE_INTERNAL
	}

	/**
	 * Returns whether any registered {@link ITransformationStrategy} can transforms its input to
	 * the given model type.
	 * 
	 * @param transformedModelType
	 *            output model type.
	 * @return if a model of the given type can be produced by the
	 *         {@link IExplorationTransformationService}.
	 */
	boolean canTransformOutputType(Class<?> transformedModelType);

	/**
	 * Generates models of the requested types ({@code transformedModelTypes}) using the set of
	 * solution models contained in the given {@link SingleExplorationSolutionMap} and the
	 * collection of given {@code baseModel}s.
	 * 
	 * @param context
	 *            Defines whether the model transformation is executed DSE-internally or triggered
	 *            by the user to export a solution.
	 * @param modelContext
	 *            Model context in which the transformed models would be stored.
	 * @param solution
	 *            {@link SingleExplorationSolutionMap} containing the solution models.
	 * @param baseModels
	 *            EMF models used as a basis to generate the transformed model.
	 * @param transformedModelTypes
	 *            Type of the transformed models that shall be returned.
	 * @param transformedModelName
	 *            name of the transformed model, used when the model shall be exported.
	 * @return The transformed models which were requested.
	 * @throws Exception
	 *             If no {@link ITransformationModule} has been registered which can produce the
	 *             demanded model, or if some required solution encoding was not passed.
	 */
	<U extends EObject, S extends EObject> ClassToInstanceMap<U> getTransformedModels(
			TransformationContext context, ITopLevelElement modelContext,
			SingleExplorationSolutionMap solution, SubClassToInstanceMap<S> baseModels,
			Collection<Class<? extends EObject>> transformedModelTypes, String transformedModelName)
			throws Exception;

	/**
	 * Generates a model of the requested type ({@code transformedModelType}) using the set of
	 * solution models contained in the given {@link SingleExplorationSolutionMap} and the
	 * collection of given {@code baseModel}s.
	 * 
	 * @param context
	 *            Defines whether the model transformation is executed DSE-internally or triggered
	 *            by the user to export a solution.
	 * @param modelContext
	 *            Model context in which the transformed models would be stored.
	 * @param solution
	 *            {@link SingleExplorationSolutionMap} containing the solution models.
	 * @param baseModels
	 *            EMF models used as a basis to generate the transformed model.
	 * @param transformedModelType
	 *            Type of the transformed model.
	 * @param transformedModelName
	 *            name of the transformed model, used when the model shall be exported.
	 * @return The transformed model.
	 * @throws Exception
	 *             If no {@link ITransformationModule} has been registered which can produce the
	 *             demanded model, or if some required solution encoding was not passed.
	 */
	<U extends EObject, S extends EObject> U getTransformedModel(TransformationContext context,
			ITopLevelElement modelContext, SingleExplorationSolutionMap solution,
			SubClassToInstanceMap<S> baseModels, Class<U> transformedModelType,
			String transformedModelName) throws Exception;

	/**
	 * Generates models of the requested types ({@code transformedModelTypes}) using the set of
	 * solution models contained in the given {@link SingleExplorationSolutionMap} and the
	 * collection of given {@code baseModel}s.
	 * 
	 * @param context
	 *            Defines whether the model transformation is executed DSE-internally or triggered
	 *            by the user to export a solution.
	 * @param modelContext
	 *            Model context in which the transformed models would be stored.
	 * @param encodings
	 *            {@link SingleExplorationSolutionMap} containing the solution models.
	 * @param baseModels
	 *            EMF models used as a basis to generate the transformed model.
	 * @param transformedModelTypes
	 *            Type of the transformed models that shall be returned.
	 * @param transformedModelName
	 *            name of the transformed model, used when the model shall be exported.
	 * @return The transformed models which were requested.
	 * @throws Exception
	 *             If no {@link ITransformationModule} has been registered which can produce the
	 *             demanded model, or if some required solution encoding was not passed.
	 */
	<U extends EObject, S extends EObject> ClassToInstanceMap<U> getTransformedModels(
			TransformationContext context, ITopLevelElement modelContext,
			ExplorationEncodingMap<IExplorationEncoding> encodings,
			SubClassToInstanceMap<S> baseModels,
			Collection<Class<? extends EObject>> transformedModelTypes, String transformedModelName)
			throws Exception;

	/**
	 * Generates a model of the requested type ({@code transformedModelType}) using the set of
	 * {@link Phenotype}s contained in the given {@link FlatPhenotypeMap} and the
	 * collection of given {@code baseModel}s.
	 * 
	 * @param context
	 *            Defines whether the model transformation is executed DSE-internally or triggered
	 *            by the user to export a solution.
	 * @param modelContext
	 *            Model context in which the transformed models would be stored.
	 * @param encodings
	 *            {@link SingleExplorationSolutionMap} containing the solution models.
	 * @param baseModels
	 *            EMF models used as a basis to generate the transformed model.
	 * @param transformedModelType
	 *            Type of the transformed model.
	 * @param transformedModelName
	 *            name of the transformed model, used when the model shall be exported.
	 * @return The transformed model.
	 * @throws Exception
	 *             If no {@link ITransformationModule} has been registered which can produce the
	 *             demanded model, or if some required solution encoding was not passed.
	 */
	<U extends EObject, S extends EObject> U getTransformedModel(TransformationContext context,
			ITopLevelElement modelContext, ExplorationEncodingMap<IExplorationEncoding> encodings,
			SubClassToInstanceMap<S> baseModels, Class<U> transformedModelType,
			String transformedModelName) throws Exception;

	/**
	 * Resets the lists of {@link ITransformationModule}s registered at {@code this}
	 * {@link ExplorationTransformationService}.
	 */
	void resetRegisteredModules();
}
