/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.service;

import java.util.Collection;

import org.fortiss.af3.exploration.alg.exception.ExplorationException;
import org.fortiss.af3.exploration.model.ExplorationSolution;
import org.fortiss.tooling.kernel.service.base.IObjectAware;
import org.opt4j.core.Genotype;
import org.opt4j.core.Value;

/**
 * {@link IRepairModule}s are used to "repair" {@link Genotype}s based on the information found in
 * the {@link Genotype} encoding itself and/or the evaluation results given as an
 * {@link ExplorationSolution}.
 * <p>
 * The ability to "repair" {@link Genotype}s allows to increase the efficiency of genetic algorithms
 * by injecting domain-/problem-specific knowledge into the optimization process. It allows to
 * "correct" parts of the passed {@link Genotype} which cause obvious constraint violations.
 * 
 * @author diewald
 */
public interface IRepairModule<T extends Genotype> extends IObjectAware<T>, IExplorationModule {

	/**
	 * Indicates whether a given {@link Genotype} should be repaired based on an evaluation of the
	 * given {@link Genotype} and the set of passed evaluation {@link Value}s.
	 * 
	 * @param genotype
	 *            {@link Genotype} to be repaired.
	 * @param evaluations
	 *            Set of evaluation results.
	 * @return Whether repairing the {@link Genotype} is indicated.
	 */
	boolean needsRepair(Genotype genotype, Collection<Value<?>> evaluations);

	/**
	 * Performs the repair operation on the given {@link Genotype}. The concrete repair operation
	 * that shall be applied are derived from the {@link Genotype} itself and the passed evaluation
	 * results.
	 * 
	 * @param genotype
	 *            {@link Genotype} to be repaired.
	 * @param evaluations
	 *            Evaluation results.
	 * @throws ExplorationException
	 *             if any repair operation causes a serious error to the overall DSE process.
	 */
	void repairGenotype(Genotype genotype, Collection<Value<?>> evaluations)
			throws ExplorationException;
}
