/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.service.internal;

import static org.fortiss.af3.exploration.alg.util.ExplorationAlgUtils.pickSubTypesOf;
import static org.fortiss.tooling.base.ui.annotation.view.AnnotationViewPartBase.isUpdateEnabled;
import static org.fortiss.tooling.base.ui.annotation.view.AnnotationViewPartBase.setUpdateEnabled;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Queue;

import org.conqat.lib.commons.collections.Pair;
import org.eclipse.emf.ecore.EObject;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.dse.modeltransformation.ITransformationModule;
import org.fortiss.af3.exploration.alg.dse.modeltransformation.ITransformationModule.TransformationState;
import org.fortiss.af3.exploration.alg.dse.modeltransformation.ITransformationStrategy;
import org.fortiss.af3.exploration.alg.guava.MutableSubClassToInstanceMap;
import org.fortiss.af3.exploration.alg.guava.SubClassToInstanceMap;
import org.fortiss.af3.exploration.alg.service.ExplorationEncodingMap;
import org.fortiss.af3.exploration.alg.service.IExplorationEncoding;
import org.fortiss.af3.exploration.alg.service.IExplorationTransformationService;
import org.fortiss.af3.exploration.model.solutions.SingleExplorationSolutionMap;
import org.fortiss.af3.project.model.FileProject;
import org.fortiss.af3.project.utils.ProjectUtils;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;
import org.fortiss.tooling.kernel.model.INamedElement;
import org.fortiss.tooling.kernel.service.ITransformationService;
import org.fortiss.tooling.kernel.service.base.EObjectAwareServiceBase;

import com.google.common.collect.ClassToInstanceMap;
import com.google.common.collect.HashMultimap;
import com.google.common.collect.Multimap;
import com.google.common.collect.MutableClassToInstanceMap;

/**
 * Implementation of the {@link ITransformationService} that manages model transformations from
 * {@link IExplorationEncoding}s to EMF models.
 * 
 * @author diewald
 */
public class ExplorationTransformationService<T extends ITransformationModule<?>>
		extends EObjectAwareServiceBase<T> implements IExplorationTransformationService {

	/** Holds the instance of the transformation service. */
	private volatile static ExplorationTransformationService<?> INSTANCE;

	/** Associates module types that are accessible via the service to their Prototype instance. */
	private Map<Class<? extends ITransformationModule<?>>, ITransformationModule<?>> moduleMap =
			new HashMap<>();

	/**
	 * Associates the types of output models with the corresponding {@link ITransformationModule}.
	 */
	private Map<Class<?>, ITransformationModule<?>> modulesByOutput = new HashMap<>();

	/** Associates the registered {@link ITransformationModule} with their required input types. */
	private Multimap<ITransformationModule<?>, Class<?>> inputTypesOfModules =
			HashMultimap.create();

	/** The {@link ExplorationTransformationService} extension point ID. */
	private static final String EXTENSION_POINT_NAME =
			"org.fortiss.af3.exploration.alg.transformationModule";

	/** The {@link ExplorationTransformationService} configuration element name. */
	private static final String CONFIGURATION_ELEMENT_NAME = "transformationModule";

	/** The {@link ITransformationModule} class attribute name. */
	private static final String HANDLER_CLASS_ATTRIBUTE_NAME = "module";

	/** Cache the Annotation View's update status. */
	private boolean annotationUpdateEnabled;

	/** Constructor. */
	private ExplorationTransformationService() {
		// Prevent direct instantiation.
	}

	/** The instance (singleton) of the ExplorationTransformationService. */
	@SuppressWarnings("rawtypes")
	public static synchronized ExplorationTransformationService<?> getInstance() {
		if(INSTANCE == null) {
			INSTANCE = new ExplorationTransformationService();
		}
		return INSTANCE;
	}

	/** {@inheritDoc} */
	@Override
	public void startService() {
		super.startService();

		for(ITransformationModule<?> module : getAllHandlers()) {
			@SuppressWarnings("unchecked") Class<? extends ITransformationModule<?>> moduleType =
					(Class<? extends ITransformationModule<?>>)module.getClass();
			moduleMap.put(moduleType, module);
		}

		List<Pair<Class<?>, T>> rmAccessibleHandlers = new ArrayList<>();
		for(Entry<Class<?>, List<T>> elementHandlerList : handlerMap.entrySet()) {
			for(T handler : elementHandlerList.getValue()) {

				// Sub-module handling.
				Collection<Class<? extends ITransformationModule<?>>> parentModuleTypes =
						handler.getParentModules();
				for(Class<? extends ITransformationModule<?>> parentType : parentModuleTypes) {
					ITransformationModule<?> parentModule = moduleMap.get(parentType);
					if(parentModule != null) {
						parentModule.getSubModules().add(handler);
					}
				}

				// Mark modules to be removed from the handler map: they shall not be directly
				// accessible.
				// TODO: is a better mechanism than the identification via the strategy needed?
				if(!(handler instanceof ITransformationStrategy)) {
					rmAccessibleHandlers
							.add(new Pair<Class<?>, T>(elementHandlerList.getKey(), handler));
				}
			}
		}

		// Remove modules from the handler map that shall not be directly accessible.
		for(Pair<Class<?>, T> rmElement : rmAccessibleHandlers) {
			allHandlersSet.remove(rmElement.getSecond());
			List<T> handlerList = handlerMap.get(rmElement.getFirst());
			handlerList.remove(rmElement.getSecond());
			if(handlerList.isEmpty()) {
				handlerMap.remove(rmElement.getFirst());
			}
		}

		// Finally, register the remaining modules with the internal data structures.
		for(T module : getAllHandlers()) {
			try {
				registerTransformationModuleInternal(module);
			} catch(Exception e) {
				// TODO: Make the tooling kernel aware of exceptions during startup.
				e.printStackTrace();
			}
		}
	}

	/**
	 * Registers the given {@link ITransformationModule} with the
	 * {@link ExplorationTransformationService}.
	 * 
	 * @param module
	 *            {@link ITransformationModule} to be registered.
	 * @throws Exception
	 *             If an {@link ITransformationModule} that produces the same output model type as
	 *             the given {@link ITransformationModule} has already been registered.
	 */
	protected void registerTransformationModuleInternal(T module) throws Exception {
		// if(modulesByOutput.containsKey(module.getOutputType())) {
		// throw new Exception(
		// "An ITransformationModule has already been registered to produce the target model type "
		// +
		// module.getOutputType() +
		// " (currently registered ITransformationModule: " +
		// module.getClass().getSimpleName() +
		// ". It is only allowed to register one ITransformationModule per target model type.");
		// }

		if(inputTypesOfModules.containsKey(module)) {
			inputTypesOfModules.removeAll(module);
		}
		inputTypesOfModules.putAll(module, module.getRequiredTypes());

		modulesByOutput.put(module.getOutputType(), module);
	}

	/** {@inheritDoc} */
	@Override
	public <U extends EObject, S extends EObject> U getTransformedModel(
			TransformationContext context, ITopLevelElement modelContext,
			SingleExplorationSolutionMap solution, SubClassToInstanceMap<S> baseModels,
			Class<U> transformedModelType, String transformedModelName) throws Exception {
		Collection<Class<? extends EObject>> reqType = new ArrayList<>();
		reqType.add(transformedModelType);
		ClassToInstanceMap<U> retMap = getTransformedModels(context, modelContext, solution,
				baseModels, reqType, transformedModelName);
		return retMap.getInstance(transformedModelType);
	}

	/** {@inheritDoc} */
	@Override
	public boolean canTransformOutputType(Class<?> transformedModelType) {
		return modulesByOutput.get(transformedModelType) != null;
	}

	/** {@inheritDoc} */
	@Override
	public <U extends EObject, S extends EObject> ClassToInstanceMap<U> getTransformedModels(
			TransformationContext context, ITopLevelElement modelContext,
			SingleExplorationSolutionMap solution, SubClassToInstanceMap<S> baseModels,
			Collection<Class<? extends EObject>> transformedModelTypes, String transformedModelName)
			throws Exception {
		// First, collect all transformation modules which would generate the requested output.
		Collection<ITransformationModule<?>> topLvlModules = new HashSet<>();
		for(Class<? extends EObject> tModelType : transformedModelTypes) {
			ITransformationModule<?> module = modulesByOutput.get(tModelType);
			if(module == null) {
				throw new Exception(
						"No ITransformationModule has been registerd which prduces the demanded output model type. Aborting.");
			}
			topLvlModules.add(module);
		}

		// Then, we check whether all required inputs for the corresponding ITransformationModule
		// and its sub-Modules are present.
		// TODO: instead of triggering the transformations here, use a threaded implementation in
		// the ITransformation's base class.
		ExplorationEncodingMap<IExplorationEncoding> solutionEncodings =
				new ExplorationEncodingMap<>(null);
		Queue<ITransformationModule<?>> moduleTraversalList = new LinkedList<>();
		moduleTraversalList.addAll(topLvlModules);
		while(!moduleTraversalList.isEmpty()) {
			ITransformationModule<?> currentModule = moduleTraversalList.poll();
			for(Class<IExplorationEncoding> solutionType : pickSubTypesOf(
					IExplorationEncoding.class, currentModule.getRequiredTypes())) {
				Object solutionModel = solution.getSolutionModel(solutionType);
				if(solutionModel != null) {
					solutionEncodings.putEncoding(solutionType,
							(IExplorationEncoding)solutionModel);
				} else if(!currentModule.isOptional()) {
					throw new Exception("The required solution encoding " +
							solutionType.getSimpleName() +
							" could not be found in the given set of solutions. Please check whether all required solution model have been requested from the DSE.");
				}
			}

			moduleTraversalList.addAll(currentModule.getSubModules());
			currentModule.getSubModules().forEach(m -> topLvlModules.remove(m));
		}

		// TODO: Add a meaningful TransformationException here.
		assert (topLvlModules.size() == 1);

		ITransformationModule<?> rootModule = topLvlModules.stream().findAny().get();
		ITransformationModule<?> moduleInstance = rootModule.createModuleInstance();
		if(moduleInstance instanceof ITransformationStrategy) {
			((ITransformationStrategy<?>)moduleInstance).setModelContext(modelContext);
			((ITransformationStrategy<?>)moduleInstance).setModelName(transformedModelName);
		}

		annotationUpdateEnabled = isUpdateEnabled();
		setUpdateEnabled(false);

		SubClassToInstanceMap<U> transformedModels = MutableSubClassToInstanceMap.create();
		moduleInstance.transform(TransformationState.INITIALIZING, context, baseModels,
				solutionEncodings, transformedModels, null);

		if(context == TransformationContext.EXPORT &&
				moduleInstance instanceof ITransformationStrategy) {
			// Derive the file project from any of the base models
			if(baseModels.isEmpty()) {
				throw new Exception(
						"Cannot add the transformed model to the parent project. No base model has been given by which the project is determined.");
			}

			FileProject fp = null;
			Iterator<S> bMit = baseModels.values().iterator();
			while(bMit.hasNext() && fp == null) {
				fp = ProjectUtils.getFileProject(bMit.next());
			}
			((ITransformationStrategy<?>)moduleInstance).addModelToProject(fp, modelContext);
		}

		// Re-enable Annotation view updates if needed
		setUpdateEnabled(annotationUpdateEnabled);

		ClassToInstanceMap<U> retModels = MutableClassToInstanceMap.create();
		for(Class<? extends EObject> retType : transformedModelTypes) {
			assert (transformedModels.get(retType) != null);
			@SuppressWarnings("unchecked") Class<? extends U> castedRetType =
					(Class<? extends U>)retType;
			retModels.put(castedRetType, transformedModels.get(retType));
		}

		return retModels;
	}

	/** {@inheritDoc} */
	@Override
	public <U extends EObject, S extends EObject> U getTransformedModel(
			TransformationContext context, ITopLevelElement modelContext,
			ExplorationEncodingMap<IExplorationEncoding> phenotypes,
			SubClassToInstanceMap<S> baseModels, Class<U> transformedModelType,
			String transformedModelName) throws Exception {
		Collection<Class<? extends EObject>> reqType = new ArrayList<>();
		reqType.add(transformedModelType);
		ClassToInstanceMap<U> retMap = getTransformedModels(context, modelContext, phenotypes,
				baseModels, reqType, transformedModelName);
		return retMap.getInstance(transformedModelType);
	}

	/** {@inheritDoc} */
	@Override
	public <U extends EObject, S extends EObject> ClassToInstanceMap<U> getTransformedModels(
			TransformationContext context, ITopLevelElement modelContext,
			ExplorationEncodingMap<IExplorationEncoding> phenotypes,
			SubClassToInstanceMap<S> baseModels,
			Collection<Class<? extends EObject>> transformedModelTypes, String transformedModelName)
			throws Exception {
		// First, collect all transformation modules which would generate the requested output.
		Collection<ITransformationModule<?>> topLvlModules = new HashSet<>();
		for(Class<? extends EObject> tModelType : transformedModelTypes) {
			ITransformationModule<?> module = modulesByOutput.get(tModelType);
			if(module == null) {
				throw new Exception(
						"No ITransformationModule has been registerd which produces the demanded output model type. Aborting.");
			}
			topLvlModules.add(module);
		}

		// Then, we check whether all required inputs for the corresponding ITransformationModule
		// and its sub-Modules are present.
		// TODO: instead of triggering the transformations here, use a threaded implementation in
		// the ITransformation's base class.
		ExplorationEncodingMap<IExplorationEncoding> encodings = new ExplorationEncodingMap<>(null);
		Queue<ITransformationModule<?>> moduleTraversalList = new LinkedList<>();
		moduleTraversalList.addAll(topLvlModules);
		while(!moduleTraversalList.isEmpty()) {
			ITransformationModule<?> currentModule = moduleTraversalList.poll();
			for(Class<Phenotype> solutionType : pickSubTypesOf(Phenotype.class,
					currentModule.getRequiredTypes())) {
				IExplorationEncoding phenotype = phenotypes.getEncoding(solutionType);
				if(phenotype != null) {
					encodings.putEncoding(solutionType, phenotype);
				} else if(!currentModule.isOptional()) {
					throw new Exception("The required phenotype encoding " +
							solutionType.getSimpleName() +
							" could not be found in the given set of phenotypes. Please check whether all required phenotypes have been requested from the DSE.");
				}
			}

			moduleTraversalList.addAll(currentModule.getSubModules());
			currentModule.getSubModules().forEach(m -> topLvlModules.remove(m));
		}

		// TODO: Add a meaningful TransformationException here.
		assert (topLvlModules.size() == 1);

		// TODO: here, we will actually need the systemModelAdapter, since this method is called
		// from the DSE process itself.
		ITransformationModule<?> rootModule = topLvlModules.stream().findAny().get();
		ITransformationModule<?> moduleInstance = rootModule.createModuleInstance();
		if(moduleInstance instanceof ITransformationStrategy) {
			((ITransformationStrategy<?>)moduleInstance).setModelContext(modelContext);
		}
		SubClassToInstanceMap<U> transformedModels = MutableSubClassToInstanceMap.create();
		@SuppressWarnings("unchecked") U transformedModel =
				(U)moduleInstance.transform(TransformationState.INITIALIZING, context, baseModels,
						encodings, transformedModels, null);

		if(transformedModel instanceof INamedElement) {
			((INamedElement)transformedModel).setName(transformedModelName);
		}

		ClassToInstanceMap<U> retModels = MutableClassToInstanceMap.create();
		for(Class<? extends EObject> retType : transformedModelTypes) {
			assert (transformedModels.get(retType) != null);
			@SuppressWarnings("unchecked") Class<? extends U> castedRetType =
					(Class<? extends U>)retType;
			retModels.put(castedRetType, transformedModels.get(retType));
		}

		return retModels;
	}

	/** {@inheritDoc} */
	@Override
	public void resetRegisteredModules() {
		modulesByOutput.clear();
		inputTypesOfModules.clear();
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionDescription() {
		return "\n\nThis service manages the transformation of DSE-internal encodings to EMF models." +
				"\nThe resulting EMF models can bei either derived models, i.e., models that are based on some input 'template model', which are" +
				"\nreplicated and enriched with information generated by a DSE run, or 'new' models which are generated using information obtained" +
				"\nfrom a DSE run exclusively." + "\n\nThe service extension point is '" +
				EXTENSION_POINT_NAME + "'.";
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionLabel() {
		return "DSE Transformation Service";
	}

	/** {@inheritDoc} */
	@Override
	protected String getExtensionPointName() {
		return EXTENSION_POINT_NAME;
	}

	/** {@inheritDoc} */
	@Override
	protected String getConfigurationElementName() {
		return CONFIGURATION_ELEMENT_NAME;
	}

	/** {@inheritDoc} */
	@Override
	protected String getHandlerClassAttribute() {
		return HANDLER_CLASS_ATTRIBUTE_NAME;
	}

	/** {@inheritDoc} */
	@Override
	public void initializeService() {
		super.initializeService();
	}
}
