/*-------------------------------------------------------------------------+
| Copyright 2015 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.util;

import static org.fortiss.af3.component.utils.ComponentArchitectureUtils.isAtomicComponent;
import static org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceConnectionAdapter.ConnectionType.BIDIRECTIONAL;
import static org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceConnectionAdapter.ConnectionType.INCOMING;
import static org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceConnectionAdapter.ConnectionType.OUTGOING;
import static org.fortiss.af3.exploration.alg.util.DesignSpaceExplorationModelElementFactory.createComponentMultiAllocationConstraint;
import static org.fortiss.af3.platform.utils.PlatformArchitectureUtils.getReferencedElementsWithType;
import static org.fortiss.af3.platform.utils.PlatformArchitectureUtils.getResourceReferences;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getChildrenWithType;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getParentsWithType;
import static org.fortiss.tooling.kernel.utils.KernelModelElementUtils.getParentElement;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.function.Function;
import java.util.stream.Collectors;

import org.conqat.lib.commons.collections.Pair;
import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.ecore.EObject;
import org.fortiss.af3.component.model.Channel;
import org.fortiss.af3.component.model.Component;
import org.fortiss.af3.component.model.ComponentArchitecture;
import org.fortiss.af3.component.model.InputPort;
import org.fortiss.af3.component.model.OutputPort;
import org.fortiss.af3.component.model.Port;
import org.fortiss.af3.deployment.model.ComponentAllocation;
import org.fortiss.af3.deployment.model.Deployment;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.safetyfunctionarch.SafetyFunctionArchEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.TaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.abstractmapping.AbstractTaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IDeployableComponentAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IModelElementAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceConnectionAdapter.ConnectionType;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemModelAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.mapping.IMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.sysmodel.mapping.IMappingEntry;
import org.fortiss.af3.exploration.alg.exception.InvalidPlatformModelException;
import org.fortiss.af3.exploration.alg.exception.InvalidTimingModelException;
import org.fortiss.af3.exploration.alg.guava.TypedBiMap;
import org.fortiss.af3.exploration.moea.model.predefined.ComponentMultiAllocationConstraint;
import org.fortiss.af3.platform.model.ExecutionUnit;
import org.fortiss.af3.platform.model.IPlatformPort;
import org.fortiss.af3.platform.model.IPlatformResource;
import org.fortiss.af3.platform.model.IVirtualizationPlatformArchitectureElement;
import org.fortiss.af3.platform.model.PlatformArchitecture;
import org.fortiss.af3.platform.model.PlatformConnectorUnit;
import org.fortiss.af3.platform.model.annotation.ResourceLink;
import org.fortiss.af3.safety.model.SafetyArchitecture;
import org.fortiss.tooling.base.model.element.IConnection;
import org.fortiss.tooling.base.model.element.IConnector;
import org.fortiss.tooling.base.model.element.IHierarchicElement;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.kernel.model.IIdLabeled;
import org.jgrapht.DirectedGraph;
import org.jgrapht.graph.DefaultEdge;

/**
 * Utility methods which help to abstract AF3 specifica for the DSE.
 * 
 * @author diewald
 */
public class AF3Utils {

	/** Constructor. Prevent initialization. */
	private AF3Utils() {
	}

	/**
	 * Translates the safety architecture encoded in a {@link SafetyFunctionArchEntry} to a
	 * {@link SafetyArchitecture} enum representation.
	 * 
	 * @param sfaEntry
	 *            {@link SafetyFunctionArchEntry} to be translated.
	 * @return The encoded {@link SafetyArchitecture}, or {@code null} if no matching enum is found.
	 */
	public static SafetyArchitecture getSFArchFromEntry(SafetyFunctionArchEntry<?> sfaEntry) {
		int channelNum = sfaEntry.getNumChannels();
		int triggerNum = sfaEntry.getNumTriggerSignal();
		boolean diagnosisPresent = sfaEntry.isDiagnosicUnitPresent();
		boolean diagnosisConnected = sfaEntry.isDiagnosicUnitConnected();

		if(diagnosisConnected) {
			if(channelNum == 1 && triggerNum == 1) {
				return SafetyArchitecture.ARCH_1OO1_D;
			} else if(channelNum == 2 && triggerNum == 1) {
				return SafetyArchitecture.ARCH_1OO2_D;
			}
		} else {
			if(channelNum == 1 && triggerNum == 1) {
				return diagnosisPresent ? SafetyArchitecture.ARCH_1OO1_DU
						: SafetyArchitecture.ARCH_1OO1;
			} else if(channelNum == 2 && triggerNum == 1) {
				return diagnosisPresent ? SafetyArchitecture.ARCH_1OO2_DU
						: SafetyArchitecture.ARCH_1OO2;
			} else if(channelNum == 3 && triggerNum == 1) {
				return diagnosisPresent ? SafetyArchitecture.ARCH_1OO3_DU
						: SafetyArchitecture.ARCH_1OO3;
			} else if(channelNum == 2 && triggerNum == 2) {
				return diagnosisPresent ? SafetyArchitecture.ARCH_2OO2_DU
						: SafetyArchitecture.ARCH_2OO2;
			} else if(channelNum == 3 && triggerNum == 2) {
				return diagnosisPresent ? SafetyArchitecture.ARCH_2OO3_DU
						: SafetyArchitecture.ARCH_2OO3;
			} else if(channelNum == 4 && triggerNum == 2) {
				return SafetyArchitecture.ARCH_2OO4;
			}
		}

		return null;
	}

	/**
	 * Returns the number of independent safety channels encoded in a {@link SafetyArchitecture}.
	 * 
	 * @param sfArch
	 *            {@link SafetyArchitecture} from which the channel number shall be extracted.
	 * @return Number of channels, or {@code null} if no {@link SafetyArchitecture} is given or the
	 *         architecture is unknown.
	 */
	public static Integer getChannelNumbersSafetyArch(SafetyArchitecture sfArch) {
		if(sfArch == null) {
			return null;
		}

		switch(sfArch) {
			case ARCH_1OO1:
			case ARCH_1OO1_D:
			case ARCH_1OO1_DU:
				return 1;
			case ARCH_1OO2:
			case ARCH_1OO2_D:
			case ARCH_1OO2_DU:
			case ARCH_2OO2:
			case ARCH_2OO2_DU:
				return 2;
			case ARCH_1OO3:
			case ARCH_1OO3_DU:
			case ARCH_2OO3:
			case ARCH_2OO3_DU:
				return 3;
			case ARCH_2OO4:
				return 4;
			default:
				return null;
		}
	}

	/**
	 * Returns the number signals from independent safety channels that are required to trigger the
	 * actual safety function. This information is extracted form the given
	 * {@link SafetyArchitecture}.
	 * 
	 * @param sfArch
	 *            {@link SafetyArchitecture} from which the number of triggers shall be extracted.
	 * @return Number of triggers, or {@code null} if no {@link SafetyArchitecture} is given or the
	 *         architecture is unknown.
	 */
	public static Integer getTriggerNumbersSafetyArch(SafetyArchitecture sfArch) {
		if(sfArch == null) {
			return null;
		}

		switch(sfArch) {
			case ARCH_1OO1:
			case ARCH_1OO1_D:
			case ARCH_1OO1_DU:
			case ARCH_1OO2:
			case ARCH_1OO2_D:
			case ARCH_1OO2_DU:
			case ARCH_1OO3:
			case ARCH_1OO3_DU:
				return 1;
			case ARCH_2OO2:
			case ARCH_2OO2_DU:
			case ARCH_2OO3:
			case ARCH_2OO3_DU:
			case ARCH_2OO4:
				return 2;
			default:
				return null;
		}
	}

	/**
	 * Returns whether the given {@link SafetyArchitecture} encodes the presence of one or more
	 * diagnostic units.
	 * 
	 * @param sfArch
	 *            {@link SafetyArchitecture} encoding the presence of diagnostic unit (s).
	 * @return Presence of diagnostic unit(s), or {@code null} if no {@link SafetyArchitecture} is
	 *         given or the architecture is unknown.
	 */
	public static Optional<Boolean> isDiagPresentSafetyArch(SafetyArchitecture sfArch) {
		switch(sfArch) {
			case ARCH_1OO1_D:
			case ARCH_1OO1_DU:
			case ARCH_1OO2_D:
			case ARCH_2OO2_DU:
			case ARCH_1OO2_DU:
			case ARCH_1OO3_DU:
			case ARCH_2OO3_DU:
				return Optional.of(true);
			case ARCH_1OO1:
			case ARCH_1OO2:
			case ARCH_2OO2:
			case ARCH_1OO3:
			case ARCH_2OO3:
			case ARCH_2OO4:
				return Optional.of(false);
			default:
				return Optional.empty();
		}
	}

	/**
	 * Returns whether the given {@link SafetyArchitecture} encodes the presence of one or more
	 * diagnostic units that directly influence whether the safety function is triggered.
	 * 
	 * @param sfArch
	 *            {@link SafetyArchitecture} encoding the presence of diagnostic unit (s).
	 * @return Presence of diagnostic unit(s) influencing the safety function's output, or
	 *         {@code null} if no {@link SafetyArchitecture} is given or the architecture is
	 *         unknown.
	 */
	public static Optional<Boolean> isDiagConnectedSafetyArch(SafetyArchitecture sfArch) {
		switch(sfArch) {
			case ARCH_1OO1_D:
			case ARCH_1OO2_D:
				return Optional.of(true);
			case ARCH_1OO1:
			case ARCH_1OO1_DU:
			case ARCH_1OO2:
			case ARCH_1OO2_DU:
			case ARCH_2OO2:
			case ARCH_2OO2_DU:
			case ARCH_1OO3:
			case ARCH_1OO3_DU:
			case ARCH_2OO3:
			case ARCH_2OO3_DU:
			case ARCH_2OO4:
				return Optional.of(false);
			default:
				return Optional.empty();
		}
	}

	/**
	 * Determines the {@link ConnectionType} of a connection by examining the class of the given
	 * connector that is the target of the current connection.
	 * 
	 * @param sourceConnector
	 *            "sending" connector.
	 * @param targetConnector
	 *            Target {@link IConnector} of the {@link IConnection} whose direction shall be
	 *            determined.
	 * @param systemModelAdapter
	 *            Abstraction of the input models.
	 * @return {@link ConnectionType} specifying the direction of the {@link IConnection}.
	 * @throws InvalidPlatformModelException
	 *             if an invalid combination of of platform ports/interfaces has been detected
	 *             (e.g., Input <-> Input).
	 */
	// TODO: update doc.
	public static ConnectionType getConnectionTypeDirection(IConnector sourceConnector,
			IConnector targetConnector, SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter)
			throws InvalidPlatformModelException {
		// By the method signature, we know the instance type is a IConnector class.
		@SuppressWarnings("unchecked") ConnectionType connectionTypeSource =
				systemModelAdapter.getConnectionType(
						(Class<? extends IConnector>)sourceConnector.eClass().getInstanceClass());
		@SuppressWarnings("unchecked") ConnectionType connectionTypeTarget =
				systemModelAdapter.getConnectionType(
						(Class<? extends IConnector>)targetConnector.eClass().getInstanceClass());
		// assert (connectionTypeSource != null && connectionTypeTarget != null) :
		// "Could not identify the type of one of the connection targets.";
		if((connectionTypeSource == INCOMING && connectionTypeTarget == INCOMING) ||
				(connectionTypeSource == OUTGOING && connectionTypeTarget == OUTGOING)) {
			throw new InvalidPlatformModelException((PlatformConnectorUnit)sourceConnector,
					"A platform port or interface (see above) for incoming or outgoing traffic is connected to an port or interface of the same type. This is not allowed.");
		}

		if(connectionTypeSource == OUTGOING && connectionTypeTarget == INCOMING) {
			return OUTGOING;
		} else if(connectionTypeSource == INCOMING && connectionTypeTarget == OUTGOING) {
			return INCOMING;
		} else if((connectionTypeSource == BIDIRECTIONAL || connectionTypeSource == null) &&
				(connectionTypeTarget != null && connectionTypeTarget != BIDIRECTIONAL)) {
			// Revert the direction for the target port, since we are examining the connection from
			// its source.
			return connectionTypeTarget == INCOMING ? OUTGOING : INCOMING;
		} else if((connectionTypeTarget == BIDIRECTIONAL || connectionTypeTarget == null) &&
				(connectionTypeSource != null && connectionTypeSource != BIDIRECTIONAL)) {
			return connectionTypeSource;
		}
		return BIDIRECTIONAL;
	}

	/**
	 * Returns the type of communication the given port can handle: incoming, outgoing, or
	 * bidirectional traffic.
	 * 
	 * @param sourceConnector
	 *            {@link IConnector} to be examined.
	 * @param systemModelAdapter
	 *            Abstraction of the input models.
	 * @return The port type (default = bidirectional).
	 */
	public static ConnectionType getPortDirection(IPlatformPort sourceConnector,
			SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter) {
		// By the method signature, we know the instance type is a IConnector class.
		@SuppressWarnings("unchecked") ConnectionType connectionTypeSource =
				systemModelAdapter.getConnectionType(
						(Class<? extends IConnector>)sourceConnector.eClass().getInstanceClass());
		return connectionTypeSource != null ? connectionTypeSource : BIDIRECTIONAL;
	}

	/**
	 * Predicate whether the subject {@link PlatformConnectorUnit} is a connector of the given
	 * {@link IModelElementAdapter}.
	 */
	public static boolean isConnectorOf(PlatformConnectorUnit pConn,
			IModelElementAdapter<?> element) {
		return pConn.getOwner() == element.getObject();
	}

	/**
	 * Predicate whether the subject {@link PlatformConnectorUnit}'s containing element is contained
	 * in the given {@link IModelElementAdapter} and also contained in the given
	 * {@link DirectedGraph}.
	 */
	public static boolean isContainedInResource(PlatformConnectorUnit pConn,
			IModelElementAdapter<?> element) {
		return pConn.getOwner() == ((IHierarchicElement)element.getObject()).getContainer();
	}

	/**
	 * Predicate whether the subject {@link PlatformConnectorUnit} is the predecessor
	 * {@link PlatformConnectorUnit} for the device-internal communication of the given
	 * {@link IModelElementAdapter}.
	 */
	public static boolean isDeviceInternalPredecessor(PlatformConnectorUnit connCandidate,
			IModelElementAdapter<?> element,
			DirectedGraph<PlatformConnectorUnit, DefaultEdge> route) {
		// The following rules apply for device internal communication:
		// - Sender Port is on the same device OR
		// - its reference port is one layer "below", e.g. Gateway -> Router AND
		// - the sender connector must have been traversed (natural for path walking a multicast
		// message.
		if(route.containsVertex(connCandidate) && (isConnectorOf(connCandidate, element) ||
				isContainedInResource(connCandidate, element))) {
			return true;
		}
		return false;
	}

	/**
	 * Returns the collection of {@link InputPort}s and {@link OutputPort}s attached to the given
	 * {@link Component}.
	 * 
	 * @param component
	 *            Containing {@link Component}.
	 * @return Collection of attached {@link InputPort}s and {@link OutputPort}s as {@link Port}s.
	 */
	public static Collection<Port> getInputOutputPortsOf(Component component) {
		Collection<Port> ports = new ArrayList<Port>();
		ports.addAll(component.getInputPorts());
		ports.addAll(component.getOutputPorts());
		return ports;
	}

	/**
	 * Returns the collection of {@link Port}s which are connected via {@link Channel}s with the
	 * given {@link Port}
	 * 
	 * @param port
	 *            {@link Port} whose connected {@link Port}s shall be found.
	 * @return Collection of connected {@link Port}s.
	 */
	public static Collection<Port> getConnectedPortsOf(Port port) {
		Collection<Port> connectedPorts = new ArrayList<Port>();
		port.getIncomingChannels().stream().forEach(conn -> connectedPorts.add(conn.getSource()));
		port.getOutgoingChannels().stream().forEach(conn -> connectedPorts.add(conn.getTarget()));
		return connectedPorts;
	}

	/**
	 * Returns the collection of <i>all</i> directly connected {@link Channel}s of the given
	 * {@link Component}.
	 * 
	 * @param component
	 *            Subject {@link Component}.
	 * @return Collection of directly connected {@link Channel}s.
	 */
	public static Collection<Channel> getConnectedChannelsOf(Component component) {
		Collection<Port> containedPorts = getInputOutputPortsOf(component);
		Collection<Channel> connectedChannels = new ArrayList<>();
		containedPorts.stream().forEach(
				p -> p.getIncomingChannels().stream().forEach(ch -> connectedChannels.add(ch)));
		containedPorts.stream().forEach(
				p -> p.getOutgoingChannels().stream().forEach(ch -> connectedChannels.add(ch)));
		return connectedChannels;
	}

	/**
	 * Checks whether any parent {@link IPlatformResource} of the given {@code element} is a
	 * deployment target defined by the set of deployment targets ({@code deploymentTargets}).
	 * If any parent element has been found that is a deployment target, {@code true} is returned.
	 * The method considers parents in the following order:
	 * <li>Check the direct parent (container).
	 * </li>
	 * <li>Check all referenced resources as defined in the {@link ResourceLink} annotation.</li>
	 * <li>Go to the parent resource and run the same checks.</li>
	 * <li>Run the above checks for all
	 * references Resources as given by the {@link ResourceLink} annotations.</>
	 * 
	 * @param element
	 *            {@link IPlatformResource} whose parents shall be examined if they are deployment
	 *            targets.
	 * @param deploymentTargets
	 *            set of deployment target resources.
	 * @return Whether or not any parent is deployment target by the rules explained above.
	 */
	public static boolean isAnyParentDeploymentTarget(IPlatformResource element,
			Collection<IPlatformResource> deploymentTargets) {
		EObject parent = element.eContainer();
		if(parent == null) {
			return false;
		}

		List<IPlatformResource> resRefs = getResourceReferences((IModelElement)element);

		if(deploymentTargets.contains(parent)) {
			return true;
		} else if(element instanceof IVirtualizationPlatformArchitectureElement &&
				element instanceof IModelElement) {
			if(resRefs.stream().anyMatch(r -> deploymentTargets.contains(r))) {
				return true;
			}
		}

		if(isAnyParentDeploymentTarget(element, deploymentTargets)) {
			return true;
		}

		for(IPlatformResource curRes : resRefs) {
			if(isAnyParentDeploymentTarget(curRes, deploymentTargets)) {
				return true;
			}
		}

		return false;
	}

	/**
	 * Checks whether any child element of the given element is a deplyoment target and conforms to
	 * the given type.
	 */
	public static <T extends IPlatformResource & IHierarchicElement> boolean
			isAnyChildDeploymentTarget(IPlatformResource element, Collection<T> deploymentTargets,
					Class<T> targetType) {
		return getChildrenWithType(element, targetType).stream()
				.anyMatch(child -> deploymentTargets.contains(child));
	}

	/**
	 * Helper method to capture the inverse "links" of {@link ResourceLink} annotations.
	 * 
	 * @param referencingPlatformArchitecture
	 *            {@link PlatformArchitecture} containing the {@link ResourceLink}s.
	 * @param interPlatformConnectingElements
	 *            {@link Pair} of {@link IModelElement} classes identifying those element types that
	 *            reference (first element) another element of another platform (second element) and
	 *            thereby establish a "communication link" between both platforms.
	 * @return Map representing the "inverse" 1:1 references of {@link ResourceLink}s.
	 */
	public static Map<IModelElement, IModelElement> getInverseIntraPlatformReferences(
			PlatformArchitecture referencingPlatformArchitecture,
			Pair<Class<? extends IModelElement>, Class<? extends IModelElement>> interPlatformConnectingElements) {
		if(interPlatformConnectingElements == null) {
			return Collections.emptyMap();
		}

		Map<IModelElement, IModelElement> invElementReferences =
				new HashMap<IModelElement, IModelElement>();

		for(IModelElement referencingElement : getChildrenWithType(referencingPlatformArchitecture,
				interPlatformConnectingElements.getFirst())) {
			for(IModelElement referencedElement : getReferencedElementsWithType(referencingElement,
					interPlatformConnectingElements.getSecond())) {
				invElementReferences.put(referencedElement, referencingElement);
			}
		}

		return invElementReferences;
	}

	/**
	 * Returns a map containing the inverse references represented by {@link ResourceLink}s between
	 * {@link IModelElement}s of {@link PlatformArchitecture}s.
	 * 
	 * @param af3Deployment
	 *            {@link Deployment} for whose linked {@link PlatformArchitecture} the inverse
	 *            {@link ResourceLink} references have to be found.
	 * @param interPlatformConnectingElements
	 *            {@link Pair} of {@link IModelElement} classes identifying those element types that
	 *            reference (first element) another element of another platform (second element) and
	 *            thereby establish a "communication link" between both platforms.
	 * @return The inverse {@link ResourceLink} references, or an empty map if none were identified.
	 */
	public static Map<IModelElement, IModelElement> getInverseIntraPlatformReferences(
			Deployment af3Deployment,
			Pair<Class<? extends IModelElement>, Class<? extends IModelElement>> interPlatformConnectingElements) {
		IModelElement platformReference =
				af3Deployment.getPlatformArchitectureReference().getReference();
		PlatformArchitecture af3PlatformArchitecture = null;
		if(platformReference instanceof PlatformArchitecture) {
			af3PlatformArchitecture = (PlatformArchitecture)platformReference;
			return getInverseIntraPlatformReferences(af3PlatformArchitecture,
					interPlatformConnectingElements);
		}
		return Collections.emptyMap();
	}

	/**
	 * Searches the given {@link ComponentArchitecture} for {@link ComponentDiverseImplRef}s and
	 * returns all referenced {@link Component}s (representing an implementation of the referencing
	 * {@link Component}).
	 * 
	 * @param compArch
	 *            The {@link ComponentArchitecture} to be searched for references.
	 * @return Collection of all {@link Component}s referenced by the
	 *         {@link ComponentDiverseImplRef} {@link IModelElementSpecification} in the given
	 *         {@link ComponentArchitecture} :
	 */
	// public static Collection<Component> getAllReferencedDiverseComponentsOf(
	// ComponentArchitecture compArch) {
	// Collection<Component> referencedComponents = new ArrayList<Component>();
	// List<ComponentDiverseImplRef> diversitySpec =
	// getChildrenWithType(compArch, ComponentDiverseImplRef.class);
	// for(ComponentDiverseImplRef diverseComponentSpec : diversitySpec) {
	// // We assume that the diverse components of one component are located in one and the
	// // same ComponentArchitecture. Hence, it suffices to determine the parent
	// // ComponentArchitecture by the first element in the List.
	// referencedComponents.addAll(diverseComponentSpec.getComponentRef());
	// }
	// return referencedComponents;
	// }

	/**
	 * Creates a set of {@link ComponentMultiAllocationConstraint}s that is constructed from the
	 * {@link ComponentAllocation}s contained in the given {@link Deployment}.
	 * 
	 * @param deployment
	 *            {@link Deployment} containing the {@link ComponentAllocation}s.
	 * @return Set of {@link ComponentMultiAllocationConstraint}s with the same references as the
	 *         {@link ComponentAllocation}s.
	 */
	public static Collection<ComponentMultiAllocationConstraint>
			createConstraintsFromDeployment(Deployment deployment) {
		Collection<ComponentAllocation> allocatedComponents = deployment.getComponentAllocations();
		Function<ComponentAllocation, ComponentMultiAllocationConstraint> mapper =
				ca -> createComponentMultiAllocationConstraint(ca.getComponent(),
						new BasicEList<>(Arrays.asList(ca.getExecutionUnit())));
		return allocatedComponents.stream().map(mapper)
				.map(ComponentMultiAllocationConstraint.class::cast).collect(Collectors.toList());
	}

	/**
	 * Returns the first identified <i>Component Pool</i> that is associated with the given
	 * {@link ComponentArchitecture}. This method uses the set of diverse implementations of
	 * {@link IDeployableComponentAdapter}s from the input {@link ComponentArchitecture} and returns
	 * its containing {@link ComponentArchitecture}. This method assumes that there exists maximally
	 * one <i>Component Pool</i> per "actual" {@link ComponentArchitecture}.
	 * 
	 * @param mapping
	 *            {@link IMappingEncoding} whose {@link IDeployableComponentAdapter}s referenced by
	 *            the {@link IMappingEntry}s are used to identify the <i>Component Pool</i>.
	 * @return The {@link ComponentArchitecture} identified as the <i>Component Pool</i> or
	 *         {@code null} if none is found.
	 */
	public static ComponentArchitecture getFirstComponentPool(
			IMappingEncoding<IDeployableComponentAdapter<?>, ?, IMappingEntry<IDeployableComponentAdapter<?>, ?>> mapping) {
		for(IDeployableComponentAdapter<?> deployableComponent : mapping.getRequesters()) {
			if(deployableComponent.isInstantiation()) {
				// TODO: Avoid casting: Parameterize the IModelElementAdapter to use EObject-derived
				// generics?
				ComponentArchitecture parentCA =
						getParentElement((EObject)deployableComponent.getObject(),
								ComponentArchitecture.class, false);
				if(parentCA != null) {
					return parentCA;
				}
			} else if(deployableComponent.isInstantiationRequired()) {
				for(IDeployableComponentAdapter<?> replacementComponent : deployableComponent
						.getReplacementComponents()) {
					// TODO: Avoid casting: Parameterize the IModelElementAdapter to use
					// EObject-derived generics?
					ComponentArchitecture parentCA =
							getParentElement((EObject)replacementComponent.getObject(),
									ComponentArchitecture.class, false);
					if(parentCA != null) {
						return parentCA;
					}
				}
			}
		}
		return null;
	}

	/**
	 * Resets the IDs of the given {@code root} object and its contained elements to 0.
	 * 
	 * @param root
	 *            {@link EObject} for which and for whose contained elements, the id will be set to
	 *            0.
	 */
	public static void resetIds(IIdLabeled root) {
		for(TreeIterator<EObject> iter = root.eAllContents(); iter.hasNext();) {
			EObject obj = iter.next();
			if(obj instanceof IIdLabeled) {
				((IIdLabeled)obj).setId(0);
			}
		}
		root.setId(0);
	}

	/**
	 * Checks whether any parent {@link Component} of the given {@link EObject} is chosen to be
	 * deployable element.
	 * 
	 * @param startElement
	 *            The element whose parents shall be evaluated w.r.t. the deployable property
	 * @param systemModelAdapter
	 *            Abstraction of the input models.
	 * @return True if any parent {@link Component} is chosen to be deployed, false otherwise.
	 * @throws InvalidTimingModelException
	 *             if the deployable element cannot be identified.
	 */
	public static <T extends EObject> boolean isAnyParentComponentDeployable(EObject startElement,
			SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter)
			throws InvalidTimingModelException {
		EList<Component> parentElements = getParentsWithType(startElement, Component.class);

		for(Component parentComponent : parentElements) {
			if(systemModelAdapter.isDeployableComponent(parentComponent)) {
				return true;
			}
		}

		return false;
	}

	//
	/**
	 * Returns whether any container {@link Component} of the given {@code startElement}
	 * {@link Component} is a deployable {@link Component}. Therefore the
	 * {@code encEntryModelElemAssoc} map is checked for containing any entry that is associated
	 * with one of the containing {@link Component}s.
	 *
	 * @param startElement
	 *            {@link Component} to check.
	 * @param encEntryModelElemAssoc
	 *            associations between {@link TaskMappingEntry}s and {@link Component}s.
	 * @return Whether any containing {@link Component} is deployed.
	 * @throws InvalidTimingModelException
	 */
	public static <T extends EObject> boolean isAnyParentComponentDeployable(EObject startElement,
			TypedBiMap<Object, EObject> encEntryModelElemAssoc) throws InvalidTimingModelException {
		EList<Component> parentElements = getParentsWithType(startElement, Component.class);

		for(Component parentComponent : parentElements) {
			AbstractTaskMappingEntry entry = (AbstractTaskMappingEntry)encEntryModelElemAssoc
					.inverse().get(parentComponent, AbstractTaskMappingEntry.class);
			if(entry != null) {
				return true;
			}
		}

		return false;
	}

	/**
	 * Predicate if a the given {@link Component} is principally deployable, i.e. it is either
	 * defined to be deployable (e.g., by a trigger) or an atomic {@link Component}.
	 * 
	 * @param deployComponent
	 *            The {@link Component} to be checked whether it shall be deployed onto
	 *            {@link ExecutionUnit}s
	 * @param systemModelAdapter
	 *            Abstraction of the input models.
	 * @return boolean, if the given {@link Component} shall be considered deployable.
	 * @throws Exception
	 *             if the timing specification of any traversed element is inconsistent.
	 */
	public static boolean isComponentOrPortsDeployable(Component deployComponent,
			SystemModelAdapter<Component, ?, ?, ?, ?, ?> systemModelAdapter) throws Exception {
		for(Port port : deployComponent.getInputPorts()) {
			if(systemModelAdapter.isDeployablePort(port)) {
				return true;
			}
		}
		for(Port port : deployComponent.getOutputPorts()) {
			if(systemModelAdapter.isDeployablePort(port)) {
				return true;
			}
		}
		if(systemModelAdapter.isDeployableComponent(deployComponent) ||
				(!isAnyParentComponentDeployable(deployComponent, systemModelAdapter) &&
						isAtomicComponent(deployComponent))) {
			return true;
		}
		return false;
	}
}
