/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.util;

import static org.fortiss.af3.exploration.alg.util.ExplorationAlgUtils.pickInstanceOf;

import java.util.Collection;
import java.util.stream.Collectors;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.comm.constraint.InternalIsolatedCommunicationSet;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.partitionmapping.Partition;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.TaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.TaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.abstractmapping.AbstractTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.abstractmapping.AbstractTaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.genotype.ComposableGenotype;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.ComposablePhenotype;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.CompositePhenotype;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IDeployableComponentAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.mapping.IMappingEntry;
import org.opt4j.core.genotype.CompositeGenotype;

/**
 * Collection of methods that are useful for debuggin purposes of the ExplorationAlg plugin, but not
 * during normal operation.
 * 
 * @author diewald
 */
public class ExplorationAlgDebugUtils {

	/**
	 * TBD.
	 */
	public static <E extends TaskMappingEntry<IDeployableComponentAdapter<?>, Partition>> boolean
			checkTaskMappingForSFConsistency(
					TaskMappingEncoding<IDeployableComponentAdapter<?>, E> encoding) {
		// TODO #checking: Implement this method in a generic manner to check the number of channels
		// of safety functions with the entry number in the allocation encodings.
		Collection<E> sfEs = encoding.getMappingEntries().stream()
				.filter(e -> e.getSource().getName().contains("SafetyProtection"))
				.collect(Collectors.toList());
		Collection<E> ioEs = encoding.getMappingEntries().stream()
				.filter(e -> e.getSource().getName().contains("IOServer"))
				.collect(Collectors.toList());

		return sfEs.size() == ioEs.size();
	}

	/**
	 * Returns the set of all {@link ComposablePhenotype}s that are contained in the given
	 * {@link CompositePhenotype}, or any of its sub {@link CompositePhenotype}s.
	 * 
	 * @param compositePhenotype
	 *            {@link CompositePhenotype} whose child {@link ComposablePhenotype}s shall be
	 *            identified.
	 * @return Set of <i>all</i> child {@link ComposablePhenotype}s.
	 */
	public static Collection<ComposablePhenotype<?>>
			getAllSubPhenotypes(CompositePhenotype<?> compositePhenotype) {
		Collection<ComposablePhenotype<?>> allSubPhenotypes = compositePhenotype.getSubPhenotypes();
		pickInstanceOf(CompositePhenotype.class, allSubPhenotypes)
				.forEach(p -> allSubPhenotypes.addAll(getAllSubPhenotypes(p)));
		return allSubPhenotypes;
	}

	/**
	 * Returns the set of all {@link ComposableGenotype}s that are contained in the given
	 * {@link CompositeGenotype}, or any of its sub {@link CompositeGenotype}s.
	 * 
	 * @param compositeGenotype
	 *            {@link CompositeGenotype} whose child {@link ComposableGenotype}s shall be
	 *            identified.
	 * @return Set of <i>all</i> child {@link ComposableGenotype}s.
	 */
	public static Collection<ComposableGenotype>
			getAllSubGenotypes(CompositeGenotype<?, ?> compositeGenotype) {
		@SuppressWarnings("unchecked") Collection<ComposableGenotype> allSubGenotypes =
				(Collection<ComposableGenotype>)compositeGenotype.values();
		pickInstanceOf(CompositeGenotype.class, allSubGenotypes)
				.forEach(p -> allSubGenotypes.addAll(getAllSubGenotypes(p)));
		return allSubGenotypes;
	}

	/**
	 * Checks whether all {@link IMappingEntry}s referenced in
	 * {@link InternalIsolatedCommunicationSet} constraints also exist in the allocation map of the
	 * given {@link AbstractTaskMappingEncoding}.
	 * 
	 * @param encoding
	 *            {@link AbstractTaskMappingEncoding} to examine for consistency.
	 * @return If the mapping is consistent w.r.t. the {@link InternalIsolatedCommunicationSet}.
	 */
	public static boolean
			mappingContainsAllIsolationCommEntries(AbstractTaskMappingEncoding encoding) {
		Collection<InternalIsolatedCommunicationSet> constrs =
				encoding.getConstraintsOf(InternalIsolatedCommunicationSet.class);
		// TODO: Remove the ugly casting when converting the internal constraints into "real" ones.
		for(InternalIsolatedCommunicationSet curConstr : constrs) {
			for(IMappingEntry entry : curConstr.getIsolatedCommunicationSet()) {
				if(!encoding.containsEntry((AbstractTaskMappingEntry)entry)) {
					return false;
				}
			}
		}
		return true;
	}
}
