/*-------------------------------------------------------------------------+
| Copyright 2015 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.util;

import static org.jgrapht.Graphs.predecessorListOf;
import static org.jgrapht.Graphs.successorListOf;

import java.util.ArrayList;
import java.util.Collection;

import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IDeployableComponentAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph.TaskGraph;
import org.jgrapht.DirectedGraph;
import org.jgrapht.Graph;
import org.jgrapht.Graphs;
import org.jgrapht.traverse.BreadthFirstIterator;

/**
 * Utility methods that extend the analysis capabilities for JGrapht graphs.
 * 
 * @author diewald
 */
public class GraphUtils {

	/** Constructor. Prevent initialization. */
	private GraphUtils() {
	}

	/**
	 * Return s the first found {@link Graph} from the given collection that contains the given
	 * vertex.
	 * 
	 * @param vertex
	 *            vertex to search for.
	 * @param graphCollection
	 *            Collection of {@link Graph}s to be examined.
	 * @return First {@link Graph} containing the given vertex.
	 */
	public static <V, E> Graph<V, E> getGraphOf(V vertex,
			Collection<? extends Graph<V, E>> graphCollection) {
		for(Graph<V, E> graph : graphCollection) {
			if(graph.containsVertex(vertex)) {
				return graph;
			}
		}
		return null;
	}

	/**
	 * Returns the first found {@link TaskGraph} from the given collection that contains the given
	 * {@link IDeployableComponentAdapter}.
	 * 
	 * @param vertex
	 *            {@link IDeployableComponentAdapter} to search for.
	 * @param graphCollection
	 *            Collection of {@link TaskGraph}s to be examined.
	 * @return First {@link TaskGraph} containing the given vertex.
	 */
	public static <T extends TaskGraph> T getGraphOf(IDeployableComponentAdapter<?> vertex,
			Collection<T> graphCollection) {
		for(T graph : graphCollection) {
			if(graph.containsTask(vertex)) {
				return graph;
			}
		}
		// Fallback: get graph via the ref object. Must be in the same graph.
		// FIXME: remove this fallback!!!
		return getGraphByObjOf(vertex, graphCollection);
		// return null;
	}

	/**
	 * Returns the first found {@link TaskGraph} from the given collection that contains the given
	 * {@link IDeployableComponentAdapter}. The {@link TaskGraph} is identified by comparing the
	 * referenced objects of the contained tasks, which must be in the same graphs by construction.
	 * 
	 * @param vertex
	 *            {@link IDeployableComponentAdapter} to search for.
	 * @param graphCollection
	 *            Collection of {@link TaskGraph}s to be examined.
	 * @return First {@link TaskGraph} containing the given vertex.
	 */
	public static <T extends TaskGraph> T getGraphByObjOf(IDeployableComponentAdapter<?> vertex,
			Collection<T> graphCollection) {
		for(T graph : graphCollection) {
			for(IDeployableComponentAdapter<?> comp : graph.getTasks()) {
				if(comp.getObject() == vertex.getObject()) {
					return graph;
				}
			}
		}
		return null;
	}

	/**
	 * Determines all predecessors of the given vertex, not just the direct ones as provided by the
	 * {@link Graphs} utility class.
	 * 
	 * @param graph
	 *            The graph containing the vertex {#link vertex}.
	 * @param vertex
	 *            The vertex whose predecessors shall be determined.
	 * @return Collection of predecessor vertices.
	 */
	public static <V, E> Collection<V> getAllPredescesorsOf(DirectedGraph<V, E> graph, V vertex) {
		Collection<V> predescessorList = new ArrayList<V>();
		ArrayList<V> workingQueue = new ArrayList<V>();
		workingQueue.addAll(Graphs.predecessorListOf(graph, vertex));
		while(!workingQueue.isEmpty()) {
			// get(0): Existence ensured by while clause.
			V currentPredescessor = workingQueue.get(0);
			workingQueue.remove(0);

			if(!predescessorList.contains(currentPredescessor)) {
				predescessorList.add(currentPredescessor);
				workingQueue.addAll(predecessorListOf(graph, currentPredescessor));
			}
		}
		return predescessorList;
	}

	/**
	 * Determines all successors of the given vertex, not just the direct ones as provided by the
	 * {@link Graphs} utility class.
	 * 
	 * @param graph
	 *            The graph containing the vertex {#link vertex}.
	 * @param vertex
	 *            The vertex whose successors shall be determined.
	 * @return Collection of successor vertices.
	 */
	public static <V, E> Collection<V> getAllSuccessorsOf(DirectedGraph<V, E> graph, V vertex) {
		Collection<V> successorList = new ArrayList<V>();
		ArrayList<V> workingQueue = new ArrayList<V>();
		workingQueue.addAll(Graphs.successorListOf(graph, vertex));
		while(!workingQueue.isEmpty()) {
			// get(0): Existence ensured by while clause.
			V currentSuccessor = workingQueue.get(0);
			workingQueue.remove(0);

			if(!successorList.contains(currentSuccessor)) {
				successorList.add(currentSuccessor);
				workingQueue.addAll(successorListOf(graph, currentSuccessor));
			}
		}
		return successorList;
	}

	/**
	 * Determines whether a path exists between the given vertices in the given
	 * {@link DirectedGraph}.
	 * 
	 * @param graph
	 *            {@link DirectedGraph} containing the vertices
	 * @param sourceVertex
	 *            Vertex from which the search starts.
	 * @param endVertex
	 *            Vertex to be searched for.
	 * @return boolean indication whether or not the given graph contains a (directed) path between
	 *         source- and the end vertex.
	 */
	public static <V, E> boolean directedPathExisits(DirectedGraph<V, E> graph, V sourceVertex,
			V endVertex) {
		boolean pathExists = false;
		BreadthFirstIterator<V, E> bfsIter = new BreadthFirstIterator<V, E>(graph, sourceVertex);
		while(bfsIter.hasNext() && !pathExists) {
			if(bfsIter.next() == endVertex) {
				pathExists = true;
			}
		}

		return pathExists;
	}

	// public static <V> Class<V> createDummyTypeVertex()
}
